/**
 *  Copyright (C) 2012  plegal  (plegal@appert44.org)
 *  @file         CDallas.h
 *  @brief        Pilote pour les périphériques DALLAS dérivés de CWaveTherm
 *  @version      0.1
 *  @date         17 déc. 2012 21:07:28
 *
 *  @note         Voir la description detaillee explicite dans le fichier
 *                CDallas.cpp
 *                C++ Google Style :
 *                http://google-styleguide.googlecode.com/svn/trunk/cppguide.xml
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef _CDALLAS_H
#define _CDALLAS_H

// Includes system C
//#include <stdint.h>  // definition des types int8_t, int16_t ...

// Includes system C++

// Includes qt

// Includes application
#include "CWaveTherm.h"
#include "CWavePort.h"
#include "CAdrRadio.h"
#include "CDateTime.h"
#include "commun.h"
#include "CWNExceptions.h"
// Constantes
// ex :
// const int kDaysInAWeek = 7;

// Enumerations
// ex :
// enum Couleur
// {
//     kBlue = 0,
//     kWhite,
//     kRed,
// };

// Structures
// ex:
// struct UrlTableProperties
// {
//  string name;
//  int numEntries;
// }

/// Structure de la table de détection des seuils en mode datalogging.
typedef struct {
  /// \arg Numéro du capteur pour lequel le seuil a été atteint.
  Byte NumCapteur;            // le numéro du capteur concerné par la détection
  /// Date de la détection du seuil atteint.
  Byte Date[SIZEDATE];        // tableau des Dates de la détection pour chaque capteur
  /// Durée du dépassement du deuil
  Byte Duree[2];              // tableau des durées des dépassement de seuil pour chaque capteur
  /// Moyenne des mesures relevées durant le dépassement du seuil.
  Byte Moyenne[2];            // tableau des moyennes des mesures pendant le dépassement pour chaque capteur
} DETECTTABLE;

// Declarations de classe avancees
// ex:
// class MaClasse;


/// Espace de noms pour les classes de la librairie WAVENIS
namespace WAVENIS {

/** @brief Pilote pour les périphériques DALLAS dérivés de CWaveTherm
 *  Description detaillee de la classe.
 */
class CDallas : public CWaveTherm
{
public :

	  enum PARAM {
	  PARAMETER_A_SENSOR1 = 0x25,                         // 2
	  PARAMETER_B_SENSOR1 = 0x26,                         // 2
	  PARAMETER_A_SENSOR2 = 0x27,                         // 2
	  PARAMETER_B_SENSOR2 = 0x28,                         // 2
	  MEASUREMENT_PERIOD_THRESHOLD_DETECTION  = 0x23,     // 1      // par défaut 0 en mn
	  HIGHTHRESHOLD_ALARM_SENSOR1 = 0x15,                 // 2      // par défaut 1A0   26°
	  LOWTHRESHOLD_ALARM_SENSOR1  = 0x16,                 // 2      // par défaut 100   16°
	  HIGHTHRESHOLD_EXCESS_TIME_SENSOR1 = 0x17,           // 1      // par défaut 4 en multiple de la période de mesure de détection de seuil
	  LOWTHRESHOLD_EXCESS_TIME_SENSOR1  = 0x18,           // 1      // par défaut 4 en multiple de la période de mesure de détection de seuil
	  HIGHTHRESHOLD_ALARM_SENSOR2 = 0x2B,                 // 2      // par défaut 1A0 26°
	  LOWTHRESHOLD_ALARM_SENSOR2  = 0x2C,                 // 2      // par défaut 100 16°
	  HIGHTHRESHOLD_EXCESS_TIME_SENSOR2 = 0x2D,           // 1      // par défaut 4 en multiple de la période de mesure de détection de seuil
	  LOWTHRESHOLD_EXCESS_TIME_SENSOR2  = 0x2E            // 1      // par défaut 4 en multiple de la période de mesure de détection de seuil
	  };

	  enum SIZEREQPARAM {
	    SIZEREQ_PARAMETER_A_SENSOR1 = 2,
	    SIZEREQ_PARAMETER_A_SENSOR2 = 2,
	    SIZEREQ_PARAMETER_B_SENSOR1 = 2,
	    SIZEREQ_PARAMETER_B_SENSOR2 = 2,
	    SIZEREQ_MEASUREMENT_PERIOD_THRESHOLD_DETECTION = 1,
	    SIZEREQ_HIGHTHRESHOLD_ALARM_SENSOR1 = 2,
	    SIZEREQ_LOWTHRESHOLD_ALARM_SENSOR1 = 2,
	    SIZEREQ_HIGHTHRESHOLD_ALARM_SENSOR2 = 2,
	    SIZEREQ_LOWTHRESHOLD_ALARM_SENSOR2 = 2,
	    SIZEREQ_HIGHTHRESHOLD_EXCESS_TIME_SENSOR1 = 1,
	    SIZEREQ_HIGHTHRESHOLD_EXCESS_TIME_SENSOR2 = 1,
	    SIZEREQ_LOWTHRESHOLD_EXCESS_TIME_SENSOR1 = 1,
	    SIZEREQ_LOWTHRESHOLD_EXCESS_TIME_SENSOR2 = 1,
	  };


    /**
     * Constructeur
     */
	  CDallas( CWavePort* pwp, CAdrRadio* AdresseRadio );
	  CDallas( CWavePort* pwp, RadioAddress radioAddress );

    /**
     * Destructeur
     */
	  virtual ~CDallas();

    // Methodes publiques de la classe
    // ex : ReturnType NomMethode(Type);
	  Byte Init_Sensors();
	    int Value ( );                        // Lire la donnée complète associée au périphérique
	    int TemperatureBrute ( int Voie );    // lire la donnée associée à un capteur
	    double Temperature ( int TempBrute ); // Convertir une mesure brute
	    int TemperatureInt ( float TempReelle ); // Convertir une mesure brute
	    DETECTTABLE* ReadDetectionSeuilsTable( void );    // lecture de la table de détection des seuils en mode datalogging
	    int getHighThreshold_Alarm_Sensor1();
	    int getHighThreshold_Alarm_Sensor2();
	    Byte getHighThreshold_Excess_Time_Sensor1();
	    Byte getHighThreshold_Excess_Time_Sensor2();
	    int getLowThreshold_Alarm_Sensor1();
	    int getLowThreshold_Alarm_Sensor2();
	    Byte getLowThreshold_Excess_Time_Sensor1();
	    Byte getLowThreshold_Excess_Time_Sensor2();
	    Byte getMeasurement_period_threshold_detection();
	    int getParameter_A_sensor1();
	    int getParameter_A_sensor2();
	    int getParameter_B_sensor1();
	    int getParameter_B_sensor2();

	    void _setHighThreshold_Alarm_Sensor1(Byte HighThreshold_Alarm_Sensor1[2]);
	    void setHighThreshold_Alarm_Sensor1(float HighThreshold_Alarm_Sensor1);

	    void _setHighThreshold_Alarm_Sensor2(Byte HighThreshold_Alarm_Sensor2[2]);
	    void setHighThreshold_Alarm_Sensor2(float HighThreshold_Alarm_Sensor2);

	    void setHighThreshold_Excess_Time_Sensor1(Byte HighThreshold_Excess_Time_Sensor1);
	    void setHighThreshold_Excess_Time_Sensor2(Byte HighThreshold_Excess_Time_Sensor2);

	    void _setLowThreshold_Alarm_Sensor1(Byte LowThreshold_Alarm_Sensor1[2]);
	    void setLowThreshold_Alarm_Sensor1(float LowThreshold_Alarm_Sensor1);

	    void _setLowThreshold_Alarm_Sensor2(Byte LowThreshold_Alarm_Sensor2[2]);
	    void setLowThreshold_Alarm_Sensor2(float LowThreshold_Alarm_Sensor2);

	    void setLowThreshold_Excess_Time_Sensor1(Byte LowThreshold_Excess_Time_Sensor1);
	    void setLowThreshold_Excess_Time_Sensor2(Byte LowThreshold_Excess_Time_Sensor2);
	    void setMeasurement_period_threshold_detection(Byte Measurement_period_threshold_detection);
	    void setParameter_A_sensor1(Byte Parameter_A_sensor1[2]);
	    void setParameter_A_sensor2(Byte Parameter_A_sensor2[2]);
	    void setParameter_B_sensor1(Byte Parameter_B_sensor1[2]);
	    void setParameter_B_sensor2(Byte Parameter_B_sensor2[2]);

	    /// La table de détection des seuils en mode datalogging
	    DETECTTABLE DetectionSeuilsTable[5];
    // Pour les associations :
    // Methodes publiques setter/getter (mutateurs/accesseurs) des attributs prives
    // ex :
    // void setNomAttribut(Type nomAttribut) { nomAttribut_ = nomAttribut; }
    // Type getNomAttribut(void) const { return nomAttribut_; }

protected :
    // Attributs proteges

    // Methode protegees

private :
    // Attributs prives
    // ex :
    // Type nomAttribut_;
	    /// 2 entiers 16 bits : 0xAAAA mesure sur la voie 1 0xBBBB mesure sur la voie 2
	    /// \par Lorsqu'une voie n'existe pas la valeur brute est 0x4FFF
	    int m_TemperatureBrute[2];
	    /// Tableau des identificateurs des deux capteurs d'un composant DALLAS
	    /// \arg octet1 : code famille
	    /// -# DS18S20 0x10
	    /// -# DS18B20 0x28
	    /// \arg octet2 à octet7 : numéro de série du capteur
	    /// \arg octet8 : Code CRC 8 bits
	    Byte m_IDCapteur[2][8];            // pour stocker les identificateurs des capteurs

	    Byte m_Measurement_period_threshold_detection;
	    Byte m_Parameter_A_sensor1[2];
	    Byte m_Parameter_A_sensor2[2];
	    Byte m_HighThreshold_Alarm_Sensor1[2];
	    Byte m_LowThreshold_Alarm_Sensor1[2];
	    Byte m_HighThreshold_Excess_Time_Sensor1;
	    Byte m_LowThreshold_Excess_Time_Sensor1;
	    Byte m_Parameter_B_sensor1[2];
	    Byte m_Parameter_B_sensor2[2];
	    Byte m_HighThreshold_Alarm_Sensor2[2];
	    Byte m_LowThreshold_Alarm_Sensor2[2];
	    Byte m_HighThreshold_Excess_Time_Sensor2;
	    Byte m_LowThreshold_Excess_Time_Sensor2;
    // Methodes privees
};

// Methodes publiques inline
// ex :
// inline void CDallas::maMethodeInline(Type valeur)
// {
//   
// }
// inline Type CDallas::monAutreMethode_inline_(void)
// {
//   return 0;
// }


} // namespace WAVENIS

#endif  // _CDALLAS_H

