/**
 *  Copyright (C) 2012  plegal  (plegal@appert44.org)
 *  @file         CWaveTherm.h
 *  @brief        Pilote pour les éléments communs des périphériques capteurs thermiques
 *  @version      0.1
 *  @date         17 déc. 2012 21:08:44
 *
 *  @note         Voir la description detaillee explicite dans le fichier
 *                CWaveTherm.cpp
 *                C++ Google Style :
 *                http://google-styleguide.googlecode.com/svn/trunk/cppguide.xml
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef _CWAVETHERM_H
#define _CWAVETHERM_H

// Includes system C
//#include <stdint.h>  // definition des types int8_t, int16_t ...

// Includes system C++

// Includes qt

// Includes application
#include "CModuleFinal.h"
#include "CWavePort.h"
#include "CAdrRadio.h"
#include "CDateTime.h"
#include "commun.h"

// Constantes
// ex :
// const int kDaysInAWeek = 7;
// La taille de la table de mesures en mode datalogging est fixe et représente 96 octets soit :
// 48 mesures de 2 octets pour un seul capteur dans un composant DALLAS
// 24 mesures de 2 octets par capteur pour 2 capteurs dans un composant DALLAS
// 24 mesures de 4 octets pour un seul capteur dans un composant PT100 ou PT1000
// 12 mesures de 4 octets par capteur pour 2 capteurs dans un composant PT100 ou PT1000
#define   SIZEDLTABLE   2*24


// Enumerations
// ex :
// enum Couleur
// {
//     kBlue = 0,
//     kWhite,
//     kRed,
// };

// Structures
// ex:
// struct UrlTableProperties
// {
//  string name;
//  int numEntries;
// }
/// Structure décrivant la table des données enregistrées dans le mode datalogging du composant.
/// \arg DataLogged[2][SIZEDLTABLE] : 2 tableaux de 24 entiers 16 bits. Un tableau par capteur
/// \arg LastLogDate[SIZEDATE] : date de la dernière valeur stockée au format wavenis d'une date soit :
/// -# octet 1 : le jour
/// -# octet 2 : le mois
/// -# octet 3 : l'année dont la valeur complète est 2000+octet3
/// -# octet 4 : le jour de la semaine de 0 pour Dimanche à 6 pour samedi
/// -# octet 5 : l'heure
/// -# octet 6 : la minute
/// \arg Measurement_Period : Période de mesure en mode "Time steps" seulement
typedef struct {
  /// 2 tableaux de 24 entiers 16 bits. Un tableau par capteur
  Byte DataLogged[2][SIZEDLTABLE];  // 2 tableaux de 24 entiers 16 bits. Un tableau par capteur
  /// Date de la dernière valeur stockée au format wavenis d'une date soit :
  Byte LastLogDate[SIZEDATE];       // au format d'une date
  /// Measurement_Period : Période de mesure en mode "Time steps" seulement
  Byte Measurement_Period;          // Période de mesure en mode "Time steps" seulement
} DATALOGGINGTABLE;

/// Type pointeur vers une table de données loggées
typedef DATALOGGINGTABLE* LPDATALOGGINGTABLE;


// Declarations de classe avancees
// ex:
// class MaClasse;

/// Espace de noms pour les classes de la librairie WAVENIS
namespace WAVENIS {
class CModuleFinal;

/// Cette classe encapsule les propriétés communes des modules distants de type WAVETHERM, c'est à dire des capteurs thermiques.
/// Elle dérive de la classe CModuleFinal. On y trouve :
/// \arg Gestion des alarmes : OUI
/// \arg Gestion du stockage périodique des données ( datalogging) : OUI
/// \arg Gestion du stockage étendu périodique des données ( advanced datalogging) : NON implique la réception multiframe non géré dans CWavePort.
/// \arg Gestion de la durée de vie de la pile : OUI
/** @brief Pilote pour les éléments communs des périphériques capteurs thermiques
 *  Description detaillee de la classe.
 */
class CWaveTherm : public CModuleFinal
{
public :
#define SIZEUSRDATA   32

/// la table des données stockées en mode datalogging
/// Les mesures sont stockées MSB LSB
/// Le premier élément du tableau est la dernière mesure enregistrée.
/// La date du dernier enregistrement est stockée au format :
/// \arg octet 1 : le jour
/// \arg octet 2 : le mois
/// \arg octet 3 : l'année dont la valeur complète est 2000+octet3
/// \arg octet 4 : le jour de la semaine de 0 pour Dimanche à 6 pour samedi
/// \arg octet 5 : l'heure
/// \arg octet 6 : la minute
/// \see DATALOGGINGTABLE
DATALOGGINGTABLE DataLoggingTable;

#define  _MR_WT_CMDWRITEPARAM(a,b,c) \
  try { \
    this->CmdWriteParam((enum CModuleFinal::PARAM)a,(enum CModuleFinal::SIZEREQPARAM)b,c); \
  } \
  catch ( CWPCWPException& e ) { \
    _MR_MF_THROW(e);\
  } \
  catch ( CWPCNSException& e ) { \
    _MR_MF_THROW(e);\
  } \
  catch ( CWPCRIException& e ) { \
    _MR_MF_THROW(e);\
  }\
  catch (CWPCPCException& e) {\
    _MR_MF_THROW(e);\
  }

// les deux capteurs internes possibles
enum VOIE {
  VOIE1,
  VOIE2
};

/// Les modes de fonctionnement possibles pour le module.
/// \par L'octet de mode est une somme de ces constantes.
/// \arg exemple :
/// -# pas de détection de seuil
/// -# datalogging time steps en permanent loop
/// l'octet de mode vaut :
/// \par LOTHRSHDETECT_OFF + HITHRSHDETECT_OFF + DATALOGMODE_TS + STOPDATALOGMODE_PL = xx00 0100 = 0x04
/// \see setDay_week_month
/// \see setTime_Measurement
enum OPMODES {
  /// mode successif pour la détection de seuil
  THRSHDTMODE_SUC = 0x0,
  /// mode cumulatif pour la détection de seuil
  THRSHDTMODE_CUM = 0x40,
  /// Détection d'un seuil bas inactive
  LOTHRSHDETECT_OFF = 0x0,
  /// Détection d'un seuil bas active
  LOTHRSHDETECT_ON = 0x20,
  /// Détection d'un seuil haut inactive
  HITHRSHDETECT_OFF = 0x0,
  /// Détection d'un seuil haut active
  HITHRSHDETECT_ON = 0x10,
  /// pas de datalogging
  DATALOGMODE_OFF = 0x0,
  /// Datalogging en mode "time steps"
  DATALOGMODE_TS = 0x04,
  /// Datalogging en mode "once a week"
  DATALOGMODE_OW = 0x08,
  /// Datalogging en mode "once a month"
  DATALOGMODE_OM = 0x0C,
  /// Mode Permanent Loop : lorsque la taille des buffer est atteinte, les données en tête sont écrasées.
  STOPDATALOGMODE_PL = 0x0,
  /// Mode Memory Full: lorsque la taille des buffer est atteinte l'enregistrement de données s'arrête
  STOPDATALOGMODE_MF = 0x2
};

/// les différents modes d'alarme disponibles.
/// L'octet de configuration des alarmes est une combinaison OU de ces constantes.
/// alarme sur défaut d'un capteur
/// alarme sur fin de vie de la batterie
/// alarme sur détection d'un seuil bas en mode datalogging
/// alarme sur détection d'un seuil haut en mode datalogging
enum MODESALARM {
  LOWTHRESHDETECT = 1,
  HIGHTHRESHDETECT = 2,
  ENDBATTLIFEDETECT = 4,
  PROBEFAULTDETECT = 8
};

enum CMDAPP {
  REQ_CURRENT_TEMP  = 0x01,
  REQ_READ_DATALOGGING_TABLE = 0x03,
  REQ_INIT_SENSORS = 0x04,
  REQ_READ_THRESHOLD_DETECTION_TABLE = 0X05,
  REQ_READ_USER_DATA = 0X16,
  REQ_WRITE_USER_DATA = 0X17
};

enum CMDAPPACK {
  RES_CURRENT_TEMP  = 0x81,
  RES_READ_DATALOGGING_TABLE = 0x83,
  RES_INIT_SENSORS = 0x84,
  RES_READ_THRESHOLD_DETECTION_TABLE = 0X85,
  RES_READ_USER_DATA = 0X96,
  RES_WRITE_USER_DATA = 0X97
};

enum PARAM {
  NUMBER_DATALOGGING_LOOP = 0x0A,
  NUMBER_VALUES_STORED  = 0x0B,
  NUMBER_RETRIES_ALARM  = 0x19,
  DELAY_TRANSMISSION_ALARM  = 0x1A,
  ALARM_CONFIGURATION = 0x22,
  MEASUREMENT_PERIOD  = 0x80,
  START_HOUR_DATALOGGING  = 0x81,
  DAY_WEEK_MONTH  = 0x82,
  TIME_MEASUREMENT  = 0x83,
  BATTERYLIFE_DURATION_COUNTER = 0xA2
};

enum SIZEREQPARAM {
  SIZEREQ_NUMBER_DATALOGGING_LOOP = 1,
  SIZEREQ_NUMBER_VALUES_STORED  = 2,
  SIZEREQ_NUMBER_RETRIES_ALARM  = 1,
  SIZEREQ_DELAY_TRANSMISSION_ALARM  = 1,
  SIZEREQ_ALARM_CONFIGURATION = 1,
  SIZEREQ_MEASUREMENT_PERIOD  = 1,
  SIZEREQ_START_HOUR_DATALOGGING  = 1,
  SIZEREQ_DAY_WEEK_MONTH  = 1,
  SIZEREQ_TIME_MEASUREMENT  = 1,
  SIZEREQ_BATTERYLIFE_DURATION_COUNTER = 2
};


    /**
     * Constructeur
     */
	CWaveTherm(CWavePort *pwp, CAdrRadio* AdresseRadio);
	CWaveTherm( CWavePort* pwp, RadioAddress radioAddress );
    /**
     * Destructeur
     */
	virtual ~CWaveTherm();

    // Methodes publiques de la classe
    // ex : ReturnType NomMethode(Type);
    Byte Init_Sensors( void );
    void setApplication_Status(Byte m_Application_Status);
    Byte getApplication_Status();
    LPDATALOGGINGTABLE ReadDataLoggingTable ( void );
    /// Fixer le mode d'alarme du module.
    void setAlarm(Byte AlarmConfig);
    Byte getNumber_Datalogging_Loop();
    int getBatterylife_duration_counter();
    Byte getDay_week_month();
    Byte getDelay_transmission_alarm();
    Byte getMeasurement_Period();
    Byte getNumber_retries_alarm();
    int getNumber_values_stored();
    Byte getStart_hour_datalogging();
    Byte getTime_measurement();
    void setAlarm_Configuration(Byte m_Alarm_Configuration);
    void setDay_week_month(Byte m_Day_week_month);
    void setDelay_transmission_alarm(Byte m_Delay_transmission_alarm);
    void setMeasurement_Period(Byte m_Measurement_Period);
    void setNumber_retries_alarm(Byte m_Number_retries_alarm);
    void setNumber_values_stored( int m_Number_values_stored );
    void setStart_hour_datalogging(Byte m_Start_hour_datalogging);
    void setTime_measurement(Byte m_Time_measurement);
    Byte* getUser_data( int FirstData, int NbData );
    void setUser_data( int FirstData, int NbData, Byte Data[] );
    // Pour les associations :
    // Methodes publiques setter/getter (mutateurs/accesseurs) des attributs prives
    // ex :
    // void setNomAttribut(Type nomAttribut) { nomAttribut_ = nomAttribut; }
    // Type getNomAttribut(void) const { return nomAttribut_; }

protected :
    // Attributs proteges
    int m_NCapteurs;      // le nombre de capteurs présents
    int NTures;           // nombre de températures stockées en mode datalogging standard
    int NTuresCapt;       // Nombre de températures stockées par capteur
    // Methode protegees
    void ReadDetectionSeuilsTable(void);

private :
    // Attributs prives
    // ex :
    // Type nomAttribut_;
    // paramètres de contrôle accédé par CMDAPP spécifique
    Byte m_Alarm_Configuration;
    Byte m_Number_Datalogging_Loop;
    // R
    int m_Number_values_stored;
    // R LSB first
    Byte m_Number_retries_alarm;
    Byte m_Delay_transmission_alarm;
    Byte m_Measurement_Period;
    Byte m_Start_hour_datalogging;
    Byte m_Day_week_month;
    Byte m_Time_measurement;
    int m_Batterylife_duration_counter;
    Byte m_User_data[SIZEUSRDATA];
    // Methodes privees
};

// Methodes publiques inline
// ex :
// inline void CWaveTherm::maMethodeInline(Type valeur)
// {
//   
// }
// inline Type CWaveTherm::monAutreMethode_inline_(void)
// {
//   return 0;
// }

} // namespace WAVENIS

#endif  // _CWAVETHERM_H

