/*******************************************************************************
 * Copyright (c) 2012 Coronis
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Contributors:
 *     Coronis - initial API and implementation
 *******************************************************************************/
#ifdef _JAVA_WRAPPER
#include <serial_driver_event.h>
#include <loader_logger.h>
#include <JNI/domain_jni.h>

#include "com_coronis_sdk_protocol_waveport_Waveport.h"
#include "protocol_waveport_enum_converssion.h"

#include "../protocol.h"
#include "protocol_types.h"
#include "protocol_types_utils.h"
#include "serial_driver_types_utils.h"
#include "jni_utils.h"

#include <loader_serial_driver.h>
#include <loader_protocol.h>

static jclass g_classProtocolEvent = NULL;
static jclass g_classRadioExchangeCommand = NULL;
static jclass g_classSerialFrame = NULL;
static jclass g_classPayload = NULL;



/**
 * Method used to send a application/service message to a device
 * without waiting for a device answer.
 * @param env - JNI Environment instance
 * @param jniThis - so named "this" of jobject type that indicate pointer to object wich method a called.
 * @param command - contains the information of the message to send
 * @param commandId - A valid id of application/service message command.
 * @param responseId - A valid id of application/service message response.
 * @return CommandResponse - the sending status of the command
 * @throws ProtocolException if unexpected error occurs during sending process
 */
jobject sendMessage(JNIEnv *env, jobject jniThis, jobject command, byte commandId, byte responseId );

/**
 * Method used to send a application/service request to a device
 * without waiting for a device answer.
 * @param env - JNI Environment instance
 * @param jniThis - so named "this" of jobject type that indicate pointer to object wich method a called.
 * @param command - contains the information of the message to send
 * @param commandId - A valid id of application/service request command.
 * @param responseId - A valid id of application/service request response.
 * @return CommandResponse - the sending status of the command
 * @throws ProtocolException if unexpected error occurs during sending process
 */
jobject sendRequest(JNIEnv *env, jobject jniThis, jobject command, byte commandId, byte responseId );

/**
 * Function used to notify sent/receive spies about sent/received frames.
 * @param waveProtocol - a waveport protocol instance.
 * @param command - a radio exchange frame about witch spies will be notified.
 * @param sendEvent - a flag that manage witch of send or receive spies to be notified.
 * @return error code obtained from frame.
 */
void jni_notify_spies_command( const ProtocolInfo* waveProtocol, jobject command, _bool sendEvent );

/**
 * Function used to notify sent/receive spies about sent/received frames.
 * @param waveProtocol - a waveport protocol instance.
 * @param frame - a radio frame about witch spies will be notified.
 * @param sendEvent - a flag that manage witch of send or receive spies to be notified.
 * @return error code obtained from frame.
 */
void jni_notify_spies_hex( const ProtocolInfo* waveProtocol, const char* frame, _bool sendEvent );


ProtocolInfo* get_protocol_from_java( JNIEnv *env, jobject jniThis ) {
	jclass classID = (*env)->GetObjectClass( env, jniThis);
	if ( classID != NULL ) {
		jfieldID fieldID = (*env)->GetFieldID( env, classID, "waveporthandle", "I");
		(*env)->DeleteLocalRef( env, classID);
		if ( fieldID != NULL )
			return get_protocol( (void*)(*env)->GetIntField( env, jniThis, fieldID ) );
	}
	return 0;
}


_bool is_java_serial_driver_loaded( JNIEnv *env, ProtocolInfo* waveProtocol ) {
	return  ( waveProtocol->jniSerialDriver != NULL && ( JNI_FALSE == (*env)->IsSameObject( env, (jobject)waveProtocol->jniSerialDriver, NULL) ) );
}


_bool get_serial_frame_from_object( JNIEnv *env, jobject frame, SerialFrame* serialFrame ) {
	_bool bRet = _false;
	jclass classSerialDriverID = (*env)->GetObjectClass( env, frame);
	if ( classSerialDriverID != NULL ) {
		jmethodID methodID = (*env)->GetMethodID( env, classSerialDriverID, "toString", "()Ljava/lang/String;");
		if ( methodID != NULL ) {
			jobject jstr = (*env)->CallObjectMethod( env, frame, methodID );
			if ( jstr != NULL ) {
				char* str = (char*)(*env)->GetStringUTFChars( env, jstr, NULL);
				if ( str != NULL && is_hex(str)) {
					char *stopstring;
					size_t index = 0;
					size_t i = 0;
					size_t len = strlen(str);


					for ( ; i < len ; i += 2 ) {
						char buff = str[i+2];
						str[i+2] = 0;
						serialFrame->frame[index++] = (byte)strtol( str + i, &stopstring, 16 );
						str[i+2] = buff;
					}
					serialFrame->len = index;
					bRet = _true;
					LOG_DEBUG( g_Logger, "get_serial_frame_from_object: %s", str);
					(*env)->ReleaseStringUTFChars( env, jstr, str);
				}
				(*env)->DeleteLocalRef( env, jstr);
			}
		}
		(*env)->DeleteLocalRef( env, classSerialDriverID);
	}
	return bRet;
}

_bool get_payload_from_object( JNIEnv *env, jobject jniPayload, Payload* payload ) {
	_bool bRet = _false;
	jclass classSerialDriverID = (*env)->GetObjectClass( env, jniPayload);
	if ( classSerialDriverID != NULL ) {
		jmethodID methodID = (*env)->GetMethodID( env, classSerialDriverID, "toString", "()Ljava/lang/String;");
		if ( methodID != NULL ) {
			jobject jstr = (*env)->CallObjectMethod( env, jniPayload, methodID );
			if ( jstr != NULL ) {
				char* str = (char*)(*env)->GetStringUTFChars( env, jstr, NULL);
				if ( str != NULL && is_hex(str)) {
					char *stopstring;
					size_t index = 0;
					size_t i = 0;
					size_t len = strlen(str);
					for ( ; i < len ; i += 2 ) {
						char buff = str[i+2];
						str[i+2] = 0;
						payload->frame[index++] = (byte)strtol( str + i, &stopstring, 16 );
						str[i+2] = buff;
					}

					payload->len = index;
					bRet = _true;
					LOG_DEBUG( g_Logger, "get_payload_from_object: %s", str);
					(*env)->ReleaseStringUTFChars( env, jstr, str);
				}
				(*env)->DeleteLocalRef( env, jstr);
			}
		}
		(*env)->DeleteLocalRef( env, classSerialDriverID);
	}
	return bRet;
}

/**
 * Function used to send command frames
 * @param waveProtocol - a valid handle to an open protocol.
 * @param env - a valid handle to JNIEnv.
 * @param serialDriver - JNI Serial Driver instance.
 * @param sendMethodID - JNI if for "sent" method of Serial Driver class
 * @param frameString - hex string that represents itself serial frame that contain command, requested parameter id
 * @param wprCode - response id for read command
 * @param response - response code and message
 * @param responseFrame - frame that contain requested value if any
 * @return OK_PROTOCOL_SUCCESS if success,
 * KO_PROTOCOL_PROCESS_FAILURE if send failed;
 */
int jni_send_frame_from_hex( ProtocolInfo* waveProtocol,
		JNIEnv *env,
		jobject serialDriver,
		jmethodID sendMethodID,
		const char* frameString,
		WavePortResponses wprCode,
		CommandResponse* response,
		SerialFrame* responseFrame ) {

	response->status = KO_SERIALDRIVER_SENDING_FAILURE;
	jstring jstr = (*env)->NewStringUTF( env, frameString);
	if ( jstr ) {
		jclass serialFrameClassID = (*env)->FindClass( env, "com/coronis/sdk/domain/SerialFrame" );

		LOG_DEBUG(g_Logger, "jni_send_frame_from_hex: hex frame is %s", frameString);
		if ( serialFrameClassID != NULL ) {
			jmethodID constructorMethodID = (*env)->GetMethodID( env, serialFrameClassID, "<init>", "(Ljava/lang/String;)V");

			LOG_DEBUG(g_Logger, "jni_send_frame_from_hex: 2", 0);
			if ((*env)->ExceptionCheck(env)) {
					LOG_ERROR(g_Logger, "jni_send_frame_from_hex: Java exception occurred A1", 0);
			}

			if ( constructorMethodID != NULL ) {

				if ((*env)->ExceptionCheck(env)) {
						LOG_ERROR(g_Logger, "jni_send_frame_from_hex: Java exception occurred A2", 0);
				}

				jobject frame = (*env)->NewObject( env, serialFrameClassID, constructorMethodID, jstr);
				LOG_DEBUG(g_Logger, "jni_send_frame_from_hex: 3", 0);
				if ((*env)->ExceptionCheck(env)) {
						LOG_ERROR(g_Logger, "jni_send_frame_from_hex: Java exception occurred A3", 0);
				}


				if ( frame != NULL ) {
					int iRepeatIfError =  waveProtocol->params.configurationRetryNumber;
					response->status = KO_PROTOCOL_PROCESS_FAILURE;
					do {
						memset( &waveProtocol->receivedFrame, 0, sizeof(waveProtocol->receivedFrame) );
						LOG_DEBUG(g_Logger, "jni_send_frame_from_hex: 4", 0);
						jobject sendingStatus = (*env)->CallObjectMethod( env, serialDriver, sendMethodID, frame );
						LOG_DEBUG(g_Logger, "jni_send_frame_from_hex: 5", 0);
						if ((*env)->ExceptionOccurred(env)) {
							LOG_DEBUG( g_Logger, "jni_send_frame_from_hex: Send method exited with exception", 0);
						}
						(*env)->ExceptionClear( env );

						LOG_DEBUG(g_Logger, "jni_send_frame_from_hex: 6", 0);
						if ( sendingStatus ) {
							jclass classSendingStatusID = (*env)->GetObjectClass( env, sendingStatus);
							LOG_DEBUG(g_Logger, "jni_send_frame_from_hex: 7", 0);
							if ( classSendingStatusID != NULL ) {
								jmethodID methodID = (*env)->GetMethodID( env, classSendingStatusID, "ordinal", "()I");
								if ( methodID != NULL ){
									response->status = (*env)->CallIntMethod( env, sendingStatus, methodID );
								}
								(*env)->DeleteLocalRef( env, classSendingStatusID);
							}
							(*env)->DeleteLocalRef( env, sendingStatus);
						}
						LOG_DEBUG(g_Logger, "jni_send_frame_from_hex: 8", 0);

						if ((*env)->ExceptionCheck(env)) {
								LOG_ERROR(g_Logger, "jni_send_frame_from_hex: Java exception occurred 1", 0);
						}

						LOG_DEBUG( g_Logger, "jni_send_frame_from_hex status :%s for frame %s", get_sending_status_name(response->status), frameString );

						LOG_DEBUG(g_Logger, "jni_send_frame_from_hex: 9", 0);
						if ( OK_SERIALDRIVER_SUCCESS == (int)response->status ) {
							LOG_DEBUG( g_Logger, "jni_send_frame_from_hex: Wait for response ...", 0);
							wait_for_response( waveProtocol, response, wprCode);
							LOG_DEBUG( g_Logger, "jni_send_frame_from_hex: Response got with status %d ", response->status);
							iRepeatIfError = -1;
						} else {

							LOG_ERROR( g_Logger, "jni_send_frame_from_hex : An error occurred [%d], retrying ...", response->status);
							--iRepeatIfError;
							SLEEP( 20 );

							sprintf( response->statusDescription, "fails to send command on serial driver error code:%s", get_sending_status_name( response->status ) );
							response->status = KO_PROTOCOL_PROCESS_FAILURE;
						}
					}while ( ((int)response->status != OK_SERIALDRIVER_SUCCESS) && (iRepeatIfError >= 0) );

					if ((*env)->ExceptionCheck(env)) {
							LOG_ERROR(g_Logger, "jni_send_frame_from_hex: Java exception occurred 2", 0);
					}

					if ( response->status == OK_PROTOCOL_SUCCESS ) {
						if ( responseFrame && (waveProtocol->receivedFrame.len-2) > 0 ) {
							responseFrame->len = waveProtocol->receivedFrame.len-2;
							memcpy( responseFrame->frame, waveProtocol->receivedFrame.frame+2, responseFrame->len );
						}
					}
					(*env)->DeleteLocalRef( env, frame);
				}
			}
			(*env)->DeleteLocalRef( env, serialFrameClassID);
		}
		(*env)->DeleteLocalRef( env, jstr);
	}

	LOG_DEBUG(g_Logger, "jni_send_frame_from_hex: 10", 0);
	return response->status;
}


/**
 * Function used to send command frames
 * @param waveProtocol - a valid handle to an open protocol.
 * @param env - a valid handle to JNIEnv.
 * @param serialDriver - JNI Serial Driver instance.
 * @param sendMethodID - JNI if for "sent" method of Serial Driver class
 * @param frameString - hex string that represents itself serial frame that contain command, requested parameter id
 * @param wprCode - response id for read command
 * @param response - response code and message
 * @param responsePayload - payload that contain requested value if any
 * @return OK_PROTOCOL_SUCCESS if success,
 * KO_PROTOCOL_PROCESS_FAILURE if send failed;
 */
int jni_send_payload_from_hex( ProtocolInfo* waveProtocol,
		JNIEnv *env,
		jobject serialDriver,
		jmethodID sendMethodID,
		const char* frameString,
		WavePortResponses wprCode,
		CommandResponse* response,
		Payload* responsePayload) {

	response->status = KO_SERIALDRIVER_SENDING_FAILURE;
	jstring jstr = (*env)->NewStringUTF( env, frameString);
	LOG_DEBUG( g_Logger, "TCD jni_send_payload_from_hex %s: ", frameString);
	if ( jstr ) {
		jclass serialFrameClassID = (*env)->FindClass( env, "com/coronis/sdk/domain/SerialFrame" );
		if ( serialFrameClassID != NULL ) {
			jmethodID constructorMethodID = (*env)->GetMethodID( env, serialFrameClassID, "<init>", "(Ljava/lang/String;)V");
			if ( constructorMethodID != NULL ) {
				jobject frame = (*env)->NewObject( env, serialFrameClassID, constructorMethodID, jstr);
				if ( frame != NULL ) {
					int iRepeatIfError =  waveProtocol->params.configurationRetryNumber;
					response->status = KO_PROTOCOL_PROCESS_FAILURE;
					do {
						memset( &waveProtocol->receivedFrame, 0, sizeof(waveProtocol->receivedFrame) );

						jobject sendingStatus = (*env)->CallObjectMethod( env, serialDriver, sendMethodID, frame );

						if ((*env)->ExceptionOccurred(env)) {
							LOG_DEBUG( g_Logger, "jni_send_frame_from_hex: Send method exited with exception", 0);
						}
						(*env)->ExceptionClear( env );
						if ( sendingStatus ) {
							jclass classSendingStatusID = (*env)->GetObjectClass( env, sendingStatus);
							if ( classSendingStatusID != NULL ) {
								jmethodID methodID = (*env)->GetMethodID( env, classSendingStatusID, "ordinal", "()I");
								if ( methodID != NULL ){
									response->status = (*env)->CallIntMethod( env, sendingStatus, methodID );
								}
								(*env)->DeleteLocalRef( env, classSendingStatusID);
							}
							(*env)->DeleteLocalRef( env, sendingStatus);
						}

						LOG_DEBUG( g_Logger, "jni_send_frame_from_hex status :%s for frame %s", get_sending_status_name(response->status), frameString );
						if ( OK_SERIALDRIVER_SUCCESS == (int)response->status ) {
							LOG_DEBUG( g_Logger, "jni_send_frame_from_hex: Wait for response ...", 0);
							wait_for_response( waveProtocol, response, wprCode);
							LOG_DEBUG( g_Logger, "jni_send_frame_from_hex: Response got with status %d ", response->status);
							iRepeatIfError = -1;
						} else {

							LOG_ERROR( g_Logger, "jni_send_frame_from_hex : An error occurred [%d], retrying ...", response->status);
							--iRepeatIfError;
							SLEEP( 20 );

							sprintf( response->statusDescription, "fails to send command on serial driver error code:%s", get_sending_status_name( response->status ) );
							response->status = KO_PROTOCOL_PROCESS_FAILURE;
						}
					}while ( ((int)response->status != OK_SERIALDRIVER_SUCCESS) && (iRepeatIfError >= 0) );

					if ( response->status == OK_PROTOCOL_SUCCESS ) {

						if ( responsePayload ) {
							// if no specific response waited, return all received frame
							if (wprCode == -1) {
								responsePayload->len = waveProtocol->receivedFrame.len;
								memcpy( responsePayload->frame, waveProtocol->receivedFrame.frame, responsePayload->len );
							// else extract the 2 first byte (command id and status)
							} else if ((waveProtocol->receivedFrame.len-2) > 0 ) {
								responsePayload->len = waveProtocol->receivedFrame.len-2;
								memcpy( responsePayload->frame, waveProtocol->receivedFrame.frame+2, responsePayload->len );
							}
						}
					}
					(*env)->DeleteLocalRef( env, frame);
				}
			}
			(*env)->DeleteLocalRef( env, serialFrameClassID);
		}
		(*env)->DeleteLocalRef( env, jstr);
	}
	return response->status;
}


/*
 *    JNI implementation of com.coronis.sdk.protocol.waveport.Waveport java class
 */
jint JNICALL JNIFUNCTION(getProtocolVersion)(JNIEnv *env, jobject jniThis) {
	return (jint)protocol_get_version();
}

jint JNICALL JNIFUNCTION(create)(JNIEnv *env, jobject jniThis) {
	return (jint)protocol_create();
}

void JNICALL JNIFUNCTION(initNative)(JNIEnv *env, jobject jniThis) {
	JNIFUNCTION(init)( env,  jniThis, NULL );
}

void JNICALL JNIFUNCTION(init)(JNIEnv *env, jobject jniThis, jobject urlFile) {
	ProtocolInfo* waveProtocol = get_protocol_from_java( env, jniThis );

	//char buff[1024] = "";
	//int fd = creat( "d:\\logfile.jni_init.txt", _O_RDWR);
	//sprintf(buff, "Init entered ...\n");
	//write (fd, buff, strlen(buff));
	// close(fd);

	LOG_DEBUG( g_Logger, "init: Entered ... " , 0);


	if ( !waveProtocol ) {
		return;
	}
	// 1. Protocol is already open ?
	if ( JNI_TRUE == JNIFUNCTION(isOpen)( env, jniThis ) ) {
		// 1. Protocol is already open ? -> Yes! we should throw an exception
		throw_exception_id(env, KO_PROTOCOL_ALREADY_OPEN);
		return;
	}

	//--------------------------------------------------------------------------
	// load driver in C way
	if ( urlFile ) {

		jclass classID = (*env)->GetObjectClass( env, urlFile );
		if ( classID != NULL ) {
			jmethodID methodID = (*env)->GetMethodID( env, classID, "getPath", "()Ljava/lang/String;");
			if ( methodID != NULL ) {
				jstring jstr = (jobject)(*env)->CallObjectMethod( env, urlFile, methodID );
				if ( jstr != NULL ) {
					const char* str = (*env)->GetStringUTFChars( env, jstr, NULL);
					if ( str != NULL) {
						const char* propFileName = str;
#ifdef _WINDOWS
						propFileName = str[0] == '/' ? str+1 : str;
#endif

						int retCode = protocol_init_from_file( waveProtocol, propFileName);

						if (retCode != OK_PROTOCOL_SUCCESS) {
							retCode = KO_PROTOCOL_SETTING_ERROR;
						}

						LOG_DEBUG( g_Logger, "url: %s", str);
						(*env)->ReleaseStringUTFChars( env, jstr, str);
					}
					(*env)->DeleteLocalRef( env, jstr);
				}
			}
			(*env)->DeleteLocalRef( env, classID);
		}
	}
	else {
		// init protocol
		// The exception is not thrown here, only the serialDriverHandle
		// is reset, in this case, the driver will be loaded
		protocol_init( waveProtocol );
	}

	LOG_DEBUG( g_Logger, "init: Check if serial driver handler is not loaded yet ..." , 0);
	//--------------------------------------------------------------------------
	// fail to load Serial Driver in "C" way
	if ( waveProtocol->serialDriverHandler.serialDriverHandle == 0 ) {

		LOG_DEBUG( g_Logger, "init: serialDriverHandle not loaded, load Java Serial Driver ...", 0);
		// load Serial Driver in Java way
		if ( waveProtocol->params.isSet & FLAG_SERIAL_DRIVER_ID ) {
			if ( !is_java_serial_driver_loaded( env, waveProtocol ) ) {
				char serialDriveJNIClassPath[MAX_PATH] = "";
				size_t i = 0;
				size_t len = strlen(waveProtocol->params.serialDriverId) + 1;

				LOG_DEBUG( g_Logger, "init: Will use serial driver ID %s", waveProtocol->params.serialDriverId);

				for ( ; i < len; ++i ) {
					if ( waveProtocol->params.serialDriverId[i] == '.' )
						serialDriveJNIClassPath[i] = '/';
					else
						serialDriveJNIClassPath[i] = waveProtocol->params.serialDriverId[i];
				}

				LOG_DEBUG( g_Logger, "Driver not loaded yet, trying path %s", serialDriveJNIClassPath);

				jclass serialDriverClass = NULL;
				if ( strlen(serialDriveJNIClassPath) > 0 ) {
					LOG_DEBUG( g_Logger, "Search class at determined path ...", 0);
					serialDriverClass = (*env)->FindClass( env, serialDriveJNIClassPath);
				}

				if ( serialDriverClass != NULL ) {
					LOG_DEBUG( g_Logger, "Class found, get class constructor ... ", 0);
					jmethodID methodID = (*env)->GetMethodID( env, serialDriverClass, "<init>", "()V");// get constructor
					if ( methodID != NULL ) {
						LOG_DEBUG( g_Logger, "Constructor loaded, create JNI Java object... ", 0);

						jobject serialDriver = (*env)->NewObject( env, serialDriverClass, methodID);
						JNIFUNCTION(setSerialDriverInstance)( env,  jniThis, serialDriver);
						if ( serialDriver )
							(*env)->DeleteLocalRef( env, serialDriver);
					}
					(*env)->DeleteLocalRef( env, serialDriverClass);
				} else
					throw_exception_id(env, KO_SERIALDRIVER_ID_NOT_EXIST);
			}
		} else
			throw_exception_id(env, KO_SERIALDRIVER_ID_NOT_EXIST);
	}

	LOG_DEBUG( g_Logger, "init: Exiting" , 0);
}

jboolean JNICALL JNIFUNCTION(isOpen)(JNIEnv *env, jobject jniThis) {

	LOG_DEBUG( g_Logger, "\"Java\" isOpen: Entered", 0);
	ProtocolInfo* waveProtocol = get_protocol_from_java( env, jniThis );
	LOG_DEBUG( g_Logger, "\"Java\" isOpen: Got protocol", 0);

	if ( waveProtocol ) {
		LOG_DEBUG( g_Logger, "\"Java\" isOpen: Check C protocol", 0);

		if ( protocol_is_open( waveProtocol ) ) {
			LOG_DEBUG( g_Logger, "\"C\" isOpen: true", 0);
			return JNI_TRUE; // protocol is open in "C" way
		} else if ( is_java_serial_driver_loaded( env, waveProtocol ) ) {
			jboolean jbRet = JNI_FALSE;
			jobject serialDriver = waveProtocol->jniSerialDriver;
			jclass serialDriverClass = (*env)->GetObjectClass( env, serialDriver);
			if ( serialDriverClass != NULL ) {
				jmethodID methodID = (*env)->GetMethodID( env, serialDriverClass, "isOpen", "()Z");
				if ( methodID != NULL ) {
					jbRet = (*env)->CallBooleanMethod( env, serialDriver, methodID );
					LOG_DEBUG( g_Logger, "\"Java\" isOpen: %d", jbRet);
				}
				(*env)->DeleteLocalRef( env, serialDriverClass);
			}
			return jbRet; // protocol is open in Java way
		}
	}
	return JNI_FALSE;
}

void JNICALL JNIFUNCTION(open)(JNIEnv *env, jobject jniThis) {

	ProtocolInfo* waveProtocol = get_protocol_from_java( env, jniThis );
	if ( !waveProtocol ) {
		return;
	}

	CommandResponse rsp;
	memset( &rsp, 0, sizeof(rsp) );
	rsp.status = KO_SERIALDRIVER_NOT_AVAILABLE;
	int iRet = OK_PROTOCOL_SUCCESS;

	LOG_DEBUG( g_Logger, "Open JNI: Entered function", 0);
	//--------------------------------------------------------------------------
	if ( !protocol_is_open( waveProtocol ) ) {

		LOG_DEBUG( g_Logger, "Open JNI: protocol is not opened", 0);

		if ( waveProtocol->serialDriverHandler.serialDriverHandle ) {
			LOG_DEBUG( g_Logger, "Open JNI: serial handler is defined %X", waveProtocol);


			iRet = waveProtocol->serialDriverHandler.add_listener(waveProtocol->serialDriver,
					receive_frame, receive_event, (void*)waveProtocol );
			LOG_DEBUG( g_Logger, "Open JNI: listener added  with ret code %d", iRet);

			// LOG_DEBUG( g_Logger, "Open JNI: SerialDriver Try to open SP %s", waveProtocol->serialDriver.);
			// port id is not set here
			iRet = waveProtocol->serialDriverHandler.open( waveProtocol->serialDriver );
			LOG_DEBUG( g_Logger, "Open JNI: SerialDrover openend with ret code %d", iRet);

			if ( iRet == OK_PROTOCOL_SUCCESS )
				rsp.status = OK_PROTOCOL_SUCCESS;
			else
				LOG_DEBUG( g_Logger, "Open JNI: %d", iRet);

		} else // waveProtocol->serialDriver.serialDriverHandle is 0! 	serial driver library is not initialized
		{
			rsp.status = KO_SERIALDRIVER_ID_NOT_EXIST;
			LOG_DEBUG( g_Logger, "Open JNI: KO_SERIALDRIVER_ID_NOT_EXIST error, try another", 0);
		}
	} else {// "C" serial driver is opened
		rsp.status = OK_PROTOCOL_SUCCESS;
		LOG_DEBUG( g_Logger, "Open JNI: protocol is already opened, use it", 0);
	}

	//--------------------------------------------------------------------------
	// Try to write params 0x06 and 0x0E
	LOG_DEBUG( g_Logger, "Open JNI: Start writing of params", 0);
	if ( rsp.status == OK_PROTOCOL_SUCCESS  ) {
		RadioSettingsCommand cmd;
		memset( &cmd, 0, sizeof(cmd) );
		cmd.dynamicStandardParams.groupId = -1;
		strcpy( ((RadioAddress*)&cmd.dynamicStandardParams.nodeAddress)->addressValue, "000000000000");
		if ( waveProtocol->params.writeDefaultRadioSettings ) {
			// write all default radio configuration params
			memcpy( &cmd.dynamicExpertParams, &waveProtocol->params.dynamicExpertParams, sizeof(cmd.dynamicExpertParams));
			memcpy( &cmd.staticStandardParams, &waveProtocol->params.staticStandardParams, sizeof(cmd.staticStandardParams));
			memcpy( &cmd.staticExpertParams, &waveProtocol->params.staticExpertParams, sizeof(cmd.staticExpertParams));
		}

		LOG_DEBUG( g_Logger, "Open JNI: Write exchange status", 0);
		((Payload*)&cmd.command.payload)->frame[0] = REQ_WRITE_RADIO_PARAM;
		((Payload*)&cmd.command.payload)->frame[1] = EXCHANGE_STATUS;
		((Payload*)&cmd.command.payload)->frame[2] = 0x03; // both status and error frames activated
		((Payload*)&cmd.command.payload)->len = 3;
		protocol_write_radio_params( waveProtocol, &cmd, &rsp );

		if (rsp.status == OK_PROTOCOL_SUCCESS) {
			LOG_DEBUG( g_Logger, "Open JNI: Write relay route status", 0);
			((Payload*)&cmd.command.payload)->frame[1] = RELAY_ROUTE_STATUS;
			((Payload*)&cmd.command.payload)->frame[2] = 0x01; // activate relay mode
			protocol_write_radio_params( waveProtocol, &cmd, &rsp );
		} else {
			rsp.status = KO_PROTOCOL_SETTING_ERROR;
			throw_exception( env, &rsp );
		}
	}else
	{
		LOG_DEBUG( g_Logger, "Open JNI: Will not write params", 0);
	}

	//--------------------------------------------------------------------------
	if ( rsp.status != OK_PROTOCOL_SUCCESS ) { // try to open in Java way

		LOG_DEBUG( g_Logger, "Open JNI: Not opened in C, try to open in JAVA", 0);

		if ( is_java_serial_driver_loaded( env, waveProtocol ) ) {
			LOG_DEBUG( g_Logger, "Open JNI: Serial JAVA driver is loaded", 0);
			rsp.status = KO_PROTOCOL_NOT_OPEN;

			if ( JNI_FALSE == JNIFUNCTION(isOpen)( env,  jniThis) ) {
				LOG_DEBUG( g_Logger, "Open JNI: Serial Driver is not open, so open it!", 0);
				jobject serialDriver = waveProtocol->jniSerialDriver;
				jclass serialDriverClass = (*env)->GetObjectClass( env, serialDriver);
				if ( serialDriverClass != NULL ) {
					jmethodID methodID = (*env)->GetMethodID( env, serialDriverClass, "open", "()V");
					if ( methodID != NULL ) {
						LOG_DEBUG( g_Logger, "Open JNI: Call Java Open function ...", 0);
						(*env)->CallVoidMethod( env, serialDriver, methodID );
						LOG_DEBUG( g_Logger, "Open JNI: Check for java env exceptions", 0);

						// if it will be required to throw exception from Protocol,
						// the pass a second parameter which will mean to ovveride
						// the occurred exception.
						if (catch_and_throw_exception( env, KO_SERIALDRIVER_NOT_AVAILABLE)) {
							(*env)->DeleteLocalRef( env, serialDriverClass);
							return;
						}
					}

					methodID = (*env)->GetMethodID( env, serialDriverClass, "addSerialDriverListener", "(Lcom/coronis/sdk/serialdriver/api/ISerialDriverListener;)V");
					if ( methodID != NULL ) {
						LOG_DEBUG( g_Logger, "Open JNI: Call java method, addSerialDriverListener", 0);
						(*env)->CallVoidMethod( env, serialDriver, methodID, jniThis );
					}
					(*env)->DeleteLocalRef( env, serialDriverClass);
				}
			}

			LOG_DEBUG( g_Logger, "Open JNI: Check if Java has been opened, to set params", 0);
			if ( JNIFUNCTION(isOpen)( env,  jniThis) ) {
				LOG_DEBUG( g_Logger, "Open JNI: Ok, driver has been opened in Java", 0);
				jobject dynamicExpertParams = NULL;
				jobject staticStandardParams = NULL;
				jobject staticExpertParams = NULL;

				if ( waveProtocol->params.writeDefaultRadioSettings ) {

					dynamicExpertParams = create_dynamic_expert_params( env, &waveProtocol->params.dynamicExpertParams );
					staticStandardParams = create_static_standard_params( env, &waveProtocol->params.staticStandardParams );
					staticExpertParams = create_static_expert_params( env, &waveProtocol->params.staticExpertParams );
				}

				char value[3*2+1] = "";	// to be adjusted if number of bytes will change
				LOG_DEBUG( g_Logger, "Open JNI: Try to write EXCHANGE_STATUS", 0);
				sprintf( value, "%02X%02X%02X", REQ_WRITE_RADIO_PARAM, EXCHANGE_STATUS, 0x03);

				if ((*env)->ExceptionCheck(env)) {
					LOG_ERROR(g_Logger, "jni open: Java exception occurred 1", 0);
				}

				jobject cmd = create_radio_settings_command( env, g_classPayload, NULL, dynamicExpertParams, staticStandardParams, staticExpertParams, value);
				if ((*env)->ExceptionCheck(env)) {
					LOG_ERROR(g_Logger, "jni open: Java exception occurred 2", 0);
				}

				if ( cmd != NULL ) {
					LOG_DEBUG( g_Logger, "Open JNI: Call Java writeRadioParams with EXCHANGE_STATUS .. ", 0);
					jobject commandResponse = JNIFUNCTION(writeRadioParams)( env, jniThis, cmd);
					if ( commandResponse != NULL ) {
						LOG_DEBUG( g_Logger, "Open JNI: Ok", 0);
						rsp.status = get_response_status( env, commandResponse );
						(*env)->DeleteLocalRef( env, commandResponse );
					} else
						LOG_ERROR( g_Logger, "Open JNI: Failed to call writeRadioParams", 0);
					(*env)->DeleteLocalRef( env, cmd );

					if (rsp.status ==  KO_PROTOCOL_NO_VALID_PARAMETER) {
						LOG_DEBUG (g_Logger, "Open JNI: There was an error writing parameters, but it is OK when no parameter are specified on port open", 0);
						rsp.status = OK_PROTOCOL_SUCCESS;

					}
				}

				// write parameter 0x06 (enable relay mode)
				if (rsp.status == OK_PROTOCOL_SUCCESS) {

					LOG_DEBUG( g_Logger, "Open JNI: Writing java enable relay mode", 0);
					sprintf( value, "%02X%02X%02X", REQ_WRITE_RADIO_PARAM, RELAY_ROUTE_STATUS, 0x01);
					cmd = create_radio_settings_command( env, g_classPayload, NULL, dynamicExpertParams, staticStandardParams, staticExpertParams, value);
					if ( cmd != NULL ) {
						LOG_DEBUG( g_Logger, "Open JNI: Java Call writeRadioParams to enable relay mode", 0);
						jobject commandResponse = JNIFUNCTION(writeRadioParams)( env, jniThis, cmd);
						if ( commandResponse != NULL ) {
							rsp.status = get_response_status( env, commandResponse );
							(*env)->DeleteLocalRef( env, commandResponse );
						}
						(*env)->DeleteLocalRef( env, cmd );
					}
				}

				if ( dynamicExpertParams != NULL )
					(*env)->DeleteLocalRef( env, dynamicExpertParams );
				if ( staticStandardParams != NULL )
					(*env)->DeleteLocalRef( env, staticStandardParams );
				if ( staticExpertParams != NULL )
					(*env)->DeleteLocalRef( env, staticExpertParams );
			}

			if (rsp.status ==  KO_PROTOCOL_NO_VALID_PARAMETER) {
				LOG_DEBUG (g_Logger, "Open JNI: There was an error writing parameters, "
						"but it is OK when no parameter are specified on port open - [1]", 0);
				rsp.status = OK_PROTOCOL_SUCCESS;

			}

			if ( rsp.status != OK_PROTOCOL_SUCCESS ) {

				LOG_DEBUG( g_Logger, "Open JNI: There was an error, close java protocol", 0);
				JNIFUNCTION(close)( env, jniThis );
				rsp.status = KO_PROTOCOL_SETTING_ERROR;
				LOG_DEBUG( g_Logger, "Open JNI: Prepare to throw KO_PROTOCOL_SETTING_ERROR exception", 0);
				throw_exception( env, &rsp );
				LOG_DEBUG( g_Logger, "Open JNI: After exception throw", 0);

			}
		} else  {// Java instance has not been loaded, handle "C" error
			LOG_DEBUG( g_Logger, "Open JNI: There was an error, close C protocol", 0);
			protocol_close(waveProtocol);
			if (rsp.status != KO_SERIALDRIVER_NOT_AVAILABLE)
				rsp.status = KO_PROTOCOL_SETTING_ERROR;
			throw_exception( env, &rsp );
		}
	}

	LOG_DEBUG( g_Logger, "Open JNI: Quit from 'open' function", 0);
}

void JNICALL JNIFUNCTION(close)(JNIEnv *env, jobject jniThis) {
	ProtocolInfo* waveProtocol = get_protocol_from_java( env, jniThis );
	if ( waveProtocol ) {
		if ( is_java_serial_driver_loaded( env, waveProtocol ) ) {
			jobject serialDriver = waveProtocol->jniSerialDriver;
			jclass serialDriverClass = (*env)->GetObjectClass( env, serialDriver);
			if ( serialDriverClass != NULL ) {
				jmethodID methodID = (*env)->GetMethodID( env, serialDriverClass, "close", "()V");
				if ( methodID != NULL ) {
					(*env)->CallVoidMethod( env, serialDriver, methodID );
					catch_and_throw_exception( env, -1 );
				}
				(*env)->DeleteLocalRef( env, serialDriverClass);
			}
		}
		protocol_close( waveProtocol );
	}
}

void JNICALL JNIFUNCTION(destroy)(JNIEnv *env, jobject jniThis) {
	ProtocolInfo* waveProtocol = get_protocol_from_java( env, jniThis );
	if ( waveProtocol ) {
		JNIFUNCTION(removeMessageListener)( env, jniThis );
		JNIFUNCTION(removeSpyMessageListener)( env, jniThis, NULL );
		if ( is_java_serial_driver_loaded( env, waveProtocol ) ) {
			jobject serialDriver = waveProtocol->jniSerialDriver;
			jclass serialDriverClass = (*env)->GetObjectClass( env, serialDriver);
			if ( serialDriverClass != NULL ) {
				jmethodID methodID = (*env)->GetMethodID( env, serialDriverClass, "destroy", "()V");
				if ( methodID != NULL ) {
					(*env)->CallVoidMethod( env, serialDriver, methodID );
					catch_and_throw_exception( env, -1);
				}
				(*env)->DeleteLocalRef( env, serialDriverClass);
			}
			(*env)->DeleteGlobalRef( env, (jobject)waveProtocol->jniSerialDriver);
			waveProtocol->jniSerialDriver = NULL;
		}
		protocol_destroy( waveProtocol );
	}
}

void JNICALL JNIFUNCTION(setSerialDriverId)(JNIEnv *env, jobject jniThis, jstring serialDriverId) {

	LOG_DEBUG(g_Logger, "setSerialDriverId: Entered!", 0);

	ProtocolInfo* waveProtocol = get_protocol_from_java( env, jniThis );
	if (!waveProtocol){

		LOG_DEBUG(g_Logger, "setSerialDriverId: Cannot get protocol!", 0);
		throw_exception_id( env, KO_PROTOCOL_NOT_INITIALIZED);
		return;
	}

	// get string
	const char* strDriverId = (*env)->GetStringUTFChars( env, serialDriverId, NULL);
	if (!strDriverId){
		throw_exception_id( env, KO_PROTOCOL_NOT_INITIALIZED);
		return;
	}

	//  try C way
	LOG_DEBUG(g_Logger, "setSerialDriverId: try to set serial DriverID using C: %s", strDriverId);
	int iRet = protocol_set_serial_driver_id( (const void* )waveProtocol, strDriverId) ;
	if ( iRet == OK_PROTOCOL_SUCCESS){
		LOG_DEBUG(g_Logger, "setSerialDriverId: Ok, was set using C", 0);
		return;
	}

	// failed in C, try using java
	// 1. check if protocol is opened
	LOG_DEBUG(g_Logger, "setSerialDriverId: Check if port is opened", 0);
	if ( JNIFUNCTION(isOpen)( env,  jniThis) == JNI_TRUE ) {
		// we have other instance of Serial Driver opened!
		LOG_ERROR(g_Logger, "setSerialDriverId: Already opened", 0);
		throw_exception_id( env, KO_SERIALDRIVER_ALREADY_OPENED);
		return;
	}

	// 2. Low level, serial driver. Close serial driver if it is opened
	LOG_DEBUG(g_Logger, "setSerialDriverId: Check if serial driver is opened", 0);
	if ( waveProtocol->serialDriverHandler.serialDriverHandle != NULL ) {
		jobject localSerialDriver = waveProtocol->jniSerialDriver;
		LOG_DEBUG(g_Logger, "setSerialDriverId: Check jni serial driver ... ", 0);
		if ( localSerialDriver != NULL ) {
			LOG_DEBUG(g_Logger, "setSerialDriverId: Ok, call destroy method", 0);

			jclass serialDriverClass = (*env)->GetObjectClass( env, localSerialDriver);
			if ( serialDriverClass != NULL ) {
				jmethodID methodID = (*env)->GetMethodID( env, serialDriverClass, "destroy", "()V");
				if ( methodID != NULL ) {
					(*env)->CallVoidMethod( env, localSerialDriver, methodID );
					catch_and_throw_exception( env, -1 );
				}
				(*env)->DeleteLocalRef( env, serialDriverClass);
			}
			(*env)->DeleteGlobalRef( env, localSerialDriver );
			waveProtocol->jniSerialDriver = NULL;
			waveProtocol->params.isSet ^= FLAG_SERIAL_DRIVER_ID;
			waveProtocol->params.serialDriverId[0] = 0;
		}
	}

	//-------------------------------------------------------------------------
	// 3. load new driver (by ID)
	strncpy(waveProtocol->params.serialDriverId, strDriverId, MAX_PATH - 1);

	LOG_DEBUG(g_Logger, "-Id: Set serialDrvierId JNI", 0);
	char serialDriveJNIClassPath[MAX_PATH] = "";
	size_t i = 0;
	size_t len = strlen(waveProtocol->params.serialDriverId) + 1;
	for ( ; i < len; ++i ) {
		if ( waveProtocol->params.serialDriverId[i] == '.' )
			serialDriveJNIClassPath[i] = '/';
		else
			serialDriveJNIClassPath[i] = waveProtocol->params.serialDriverId[i];
	}

	LOG_DEBUG( g_Logger, "Driver not loaded yet, trying path %s", serialDriveJNIClassPath);

	jclass serialDriverClass = NULL;
	if ( strlen(serialDriveJNIClassPath) > 0 ) {
		LOG_DEBUG( g_Logger, "Search class at determined path ...", 0);
		serialDriverClass = (*env)->FindClass( env, serialDriveJNIClassPath);
	}

	if ( serialDriverClass != NULL ) {
		LOG_DEBUG( g_Logger, "Call found, get class constructor ... ", 0);
		jmethodID methodID = (*env)->GetMethodID( env, serialDriverClass, "<init>", "()V");// get constructor
		if ( methodID != NULL ) {
			LOG_DEBUG( g_Logger, "Constructor loaded, create JNI Java object... ", 0);

			jobject serialDriver = (*env)->NewObject( env, serialDriverClass, methodID);
			JNIFUNCTION(setSerialDriverInstance)( env,  jniThis, serialDriver);
			if ( serialDriver )
				(*env)->DeleteLocalRef( env, serialDriver);
		}
		(*env)->DeleteLocalRef( env, serialDriverClass);
	} else
		throw_exception_id(env, KO_SERIALDRIVER_ID_NOT_EXIST);
}

jstring JNICALL JNIFUNCTION(getSerialDriverId)(JNIEnv *env, jobject jniThis) {
	ProtocolInfo* waveProtocol = get_protocol_from_java( env, jniThis );
	if (!waveProtocol) return NULL;

	jstring jstr = (*env)->NewStringUTF( env, waveProtocol->params.serialDriverId );
	return jstr;
}

// a serial DriverObject is received
void JNICALL JNIFUNCTION(setSerialDriverInstance)(JNIEnv *env, jobject jniThis, jobject serialDriver) {
	ProtocolInfo* waveProtocol = get_protocol_from_java( env, jniThis );

	if ( !waveProtocol ) return;

	if ( JNI_TRUE == JNIFUNCTION(isOpen)( env,  jniThis) ) {
		throw_exception_id(env, KO_PROTOCOL_ALREADY_OPEN);
		return;
	}

	if ( waveProtocol->serialDriverHandler.serialDriverHandle != NULL ) {
		// throw exception Native serial driver has been loaded
		throw_exception_id( env, KO_SERIALDRIVER_ALREADY_LOADED);
		return;
	}

	// check if the serial driver object is already set, remove it
	jobject localSerialDriver = waveProtocol->jniSerialDriver;
	if ( localSerialDriver != NULL ) {

		jclass serialDriverClass = (*env)->GetObjectClass( env, localSerialDriver);
		if ( serialDriverClass != NULL ) {

			jmethodID methodID = (*env)->GetMethodID( env, serialDriverClass, "destroy", "()V");
			if ( methodID != NULL ) {
				(*env)->CallVoidMethod( env, localSerialDriver, methodID );
				catch_and_throw_exception( env, -1 );
			}
			(*env)->DeleteLocalRef( env, serialDriverClass);
		}
		(*env)->DeleteGlobalRef( env, localSerialDriver );
		waveProtocol->jniSerialDriver = NULL;
		waveProtocol->params.isSet ^= FLAG_SERIAL_DRIVER_ID;
		waveProtocol->params.serialDriverId[0] = 0;
	}

	if ( serialDriver == NULL ) {
		throw_exception_id(env, KO_SERIALDRIVER_ALREADY_LOADED);
	}

	jclass serialDriverClass = (*env)->GetObjectClass( env, serialDriver);
	if ( serialDriverClass != NULL ) {
		jmethodID methodID = (*env)->GetMethodID( env, serialDriverClass, "init", "()V");
		if ( methodID != NULL ) {

			(*env)->CallVoidMethod( env, serialDriver, methodID );
			waveProtocol->jniSerialDriver = (*env)->NewGlobalRef( env, serialDriver);
			methodID = (*env)->GetMethodID( env, serialDriverClass, "getClass", "()Ljava/lang/Class;");
			if ( methodID != NULL ) {
				jobject clazz = (*env)->CallObjectMethod( env, serialDriver, methodID );
				if ( clazz != NULL ) {
					jclass classClass = (*env)->GetObjectClass( env, clazz);
					if ( classClass != NULL ) {
						methodID = (*env)->GetMethodID( env, classClass, "getName", "()Ljava/lang/String;");
						if ( methodID != NULL ) {
							jobject name = (*env)->CallObjectMethod( env, clazz, methodID );
							if ( name != NULL ) {
								const char* str = (*env)->GetStringUTFChars( env, name, NULL);
								if ( str != NULL) {

									strcpy( waveProtocol->params.serialDriverId, str );
									LOG_DEBUG( g_Logger, "setSerialDriverInstance: %s", str);
									waveProtocol->params.isSet |= FLAG_SERIAL_DRIVER_ID;
									(*env)->ReleaseStringUTFChars( env, name, str);
								}
								(*env)->DeleteLocalRef( env, name);
							}
						}
						(*env)->DeleteLocalRef( env, classClass);
					}
					(*env)->DeleteLocalRef( env, clazz);
				}
			}
		}
		(*env)->DeleteLocalRef( env, serialDriverClass);
	}
}

void JNICALL JNIFUNCTION(addMessageListener)(JNIEnv *env, jobject jniThis, jobject listener) {
	ProtocolInfo* waveProtocol = get_protocol_from_java( env, jniThis );
	if ( waveProtocol ) {
		if ( !waveProtocol->paramProtocolListener && listener )
			waveProtocol->paramProtocolListener = (*env)->NewGlobalRef( env, (jobject)listener );
	}
}

void JNICALL JNIFUNCTION(removeMessageListener)(JNIEnv *env, jobject jniThis) {
	ProtocolInfo* waveProtocol = get_protocol_from_java( env, jniThis );
	if ( waveProtocol ) {
		if ( waveProtocol->paramProtocolListener ) {
			(*env)->DeleteGlobalRef( env, (jobject)waveProtocol->paramProtocolListener );
			waveProtocol->paramProtocolListener = NULL;
		}
	}
}

void JNICALL JNIFUNCTION(addSpyMessageListener)(JNIEnv *env, jobject jniThis, jobject listener) {
	ProtocolInfo* waveProtocol = get_protocol_from_java( env, jniThis );
	if ( waveProtocol ) {
		int i = 0;
		if ( listener ) {
			jboolean hasSameListener = JNI_FALSE;
			for ( ; i < MAX_SPY_NUM && hasSameListener == JNI_FALSE; ++i )
				hasSameListener = (*env)->IsSameObject( env, listener, (jobject)waveProtocol->paramSpies[i]);
			if ( hasSameListener == JNI_FALSE )
				for ( i = 0; i < MAX_SPY_NUM; ++i )
					if ( waveProtocol->paramSpies[i] == NULL &&
							waveProtocol->spiesRecv[i] == NULL &&
							waveProtocol->spiesSend[i] == NULL	) {
						waveProtocol->paramSpies[i] = (*env)->NewGlobalRef( env, (jobject)listener );
						break;
					}
		}
	}
}

void JNICALL JNIFUNCTION(removeSpyMessageListener)(JNIEnv *env, jobject jniThis, jobject listener) {
	ProtocolInfo* waveProtocol = get_protocol_from_java( env, jniThis );
	if ( waveProtocol ) {
		int i = 0;
		if ( listener ) {
			for ( ; i < MAX_SPY_NUM ; ++i )
				if ( JNI_TRUE == (*env)->IsSameObject( env, listener, (jobject)waveProtocol->paramSpies[i]) ) {
					(*env)->DeleteGlobalRef( env, (jobject)waveProtocol->paramSpies[i] );
					waveProtocol->paramSpies[i] = NULL;
					break;
				}
		} else {
			for ( ; i < MAX_SPY_NUM ; ++i ) {
				if ( waveProtocol->paramSpies[i] )
					(*env)->DeleteGlobalRef( env, (jobject)waveProtocol->paramSpies[i] );
				waveProtocol->paramSpies[i] = NULL;
			}
		}
	}
}

// callback method that receive event from Serial Driver instance
void JNICALL JNIFUNCTION(receiveEvent)(JNIEnv *env, jobject jniThis, jobject event) {
	ProtocolInfo* waveProtocol = get_protocol_from_java( env, jniThis );
	if ( waveProtocol ) {
		jint eventId = -1;
		jclass classID = (*env)->GetObjectClass( env, event);
		if ( classID != NULL ) {
			jmethodID methodID = (*env)->GetMethodID( env, classID, "ordinal", "()I");
			if ( methodID != NULL )
				eventId = (*env)->CallIntMethod( env, event, methodID );
			(*env)->DeleteLocalRef( env, classID);

			jni_notify_event_native( waveProtocol, eventId);
		}
	}
}

// callback method that receives frame from Serial Driver Instance
void JNICALL JNIFUNCTION(receiveFrame)(JNIEnv *env, jobject jniThis, jobject frame) {
	ProtocolInfo* waveProtocol = get_protocol_from_java( env, jniThis );

	LOG_DEBUG(g_Logger, "JAVA receiveFrame", 0);

	if ( waveProtocol ) {
		get_serial_frame_from_object( env, frame, &waveProtocol->receivedFrame );
		receiveInternalFrame(env, jniThis);
	}
}

// callback method that receives frame from internal process
// internal frame must be store into &waveProtocol->receivedFrame
void receiveInternalFrame(JNIEnv *env, jobject jniThis) {
	ProtocolInfo* waveProtocol = get_protocol_from_java( env, jniThis );
	LOG_DEBUG(g_Logger, "JAVA receiveInternalFrame", 0);

	if ( waveProtocol ) {

		waveProtocol->isResponseSessionActive = 1;

		if ( waveProtocol->status == PS_IDLE ) {

			RadioExchangeCommand cmd;
			WavePortResponses responseId = -1;
			int ret = parse_frame(waveProtocol, (SerialFrame*)&waveProtocol->receivedFrame, &cmd, NULL, NULL, NULL, NULL );
			LOG_DEBUG(g_Logger,"create_radio_exchange_command_object: response->commandsLen is : Called from receive_frame", 0);


			if(ret == OK_PROTOCOL_SUCCESS && ((SerialFrame*)&waveProtocol->receivedFrame)->len > 0) {
				responseId = (WavePortResponses) ((SerialFrame*)&waveProtocol->receivedFrame)->frame[0];
			}

			char frameString[MAX_SERIALFRAME_LEN*2+1] = "";
			LOG_DEBUG(g_Logger,	"Received frame is : %s with response ID 0x%02X",
					get_hex_serialframe(frameString, sizeof(frameString),
							(SerialFrame*)&waveProtocol->receivedFrame), responseId);

			jobject reCommand = create_radio_exchange_command_object( env, g_classPayload, g_classRadioExchangeCommand, &cmd, 0, 0, responseId);
			if ( reCommand ) {
				if ( waveProtocol->paramProtocolListener ) {
					jclass classID = (*env)->GetObjectClass( env, (jobject)waveProtocol->paramProtocolListener);
					if ( classID != NULL ) {
						jmethodID methodID = (*env)->GetMethodID( env, classID, "receiveRadioExchangeCommand", "(Lcom/coronis/sdk/protocol/api/RadioExchangeCommand;)V");
						if ( methodID != NULL && reCommand != NULL ) {
							(*env)->CallVoidMethod( env, (jobject)waveProtocol->paramProtocolListener, methodID, reCommand );
						}
						(*env)->DeleteLocalRef( env, classID);
					}
				}

				jni_notify_spies_command( waveProtocol, reCommand, _false);
				(*env)->DeleteLocalRef( env, reCommand);
			}
		}

		protocol_close_response_session(waveProtocol);
	}
}

jobject JNICALL JNIFUNCTION(requestLocalNodeService)(JNIEnv *env, jobject jniThis, jobject command) {

	jobject localNodeServiceCommand = NULL;
	LocalNodeServiceCommand ldsCommand;
	LocalNodeServiceResponse rsp;
	memset( &ldsCommand, 0, sizeof(ldsCommand) );
	memset( &rsp, 0, sizeof(rsp) );

	rsp.response.status = KO_PROTOCOL_NOT_OPEN;
	ProtocolInfo* waveProtocol = get_protocol_from_java( env, jniThis );

	if (! waveProtocol ) {
		return create_local_node_service_response( env, rsp.response.status, NULL);
	}

	if ( JNIFUNCTION(isOpen)( env,  jniThis) == JNI_FALSE ) {
		return create_local_node_service_response( env, rsp.response.status, NULL);
	}

	// 1. A command session is already in progress?
	if ( waveProtocol->status == PS_PROCESS_IN_PROGRESS ) {
		rsp.response.status = KO_PROTOCOL_ALREADY_IN_USE;
		throw_exception_id( env, rsp.response.status);
		return create_local_node_service_response( env, rsp.response.status, NULL);
	}

	// put protocol status in busy mode
	waveProtocol->status = PS_PROCESS_IN_PROGRESS;

	ldsCommand.serviceId = NODE_SERVICE_ID_UNDEFINED;
	jclass classID = (*env)->GetObjectClass( env, command);
	if ( classID != NULL ) {
		LOG_DEBUG(g_Logger, "jni requestLocalNodeService: getting nodeServiceId ... ", 0);
		ldsCommand.serviceId = get_node_service_id( env, classID, command);
		LOG_DEBUG(g_Logger, "jni requestLocalNodeService: getting payload ... ", 0);
		rsp.response.status = KO_PAYLOAD_REQUIRED;
		jfieldID fieldID = (*env)->GetFieldID( env, classID, "payload", "Lcom/coronis/sdk/protocol/api/IPayload;");
		if ( fieldID != NULL ) {
			jobject payload = (*env)->GetObjectField( env, command, fieldID );
			if ( payload != NULL ) {
				get_payload_from_object( env, payload, (Payload*)&ldsCommand.command.payload );
			}
			(*env)->DeleteLocalRef( env, payload);

			//			if (((Payload*)&ldsCommand.command.payload)->len <= 0) {
			//				rsp.response.status = KO_PAYLOAD_REQUIRED;
			//				break;
			//			}
			//
			//			if (((Payload*)&ldsCommand.command.payload)->len > MAX_REQ_SEND_FRAME_BYTES_LEN) {
			//				rsp.response.status = KO_PAYLOAD_NOT_VALID;
			//				break;
			//			}
		}

		(*env)->DeleteLocalRef( env, classID);
	}

	// put protocol status in idle mode
	waveProtocol->status = PS_IDLE;
	// read dynamic expert params in "C" way
	protocol_request_local_node_service( waveProtocol, &ldsCommand, &rsp);
	// put protocol status in busy mode
	waveProtocol->status = PS_PROCESS_IN_PROGRESS;


	if ( rsp.response.status == KO_PROTOCOL_NOT_OPEN ) {

		// fail to read it in "C" way, so then try Java way
		jmethodID methodID = NULL;
		jobject serialDriver = waveProtocol->jniSerialDriver;
		jclass serialDriverClassID = (*env)->GetObjectClass( env, serialDriver);

		if ( serialDriverClassID != NULL ) {
			methodID = (*env)->GetMethodID( env, serialDriverClassID, "send",
					"(Lcom/coronis/sdk/serialdriver/api/ISerialFrame;)Lcom/coronis/sdk/serialdriver/api/SendingStatus;");
			(*env)->DeleteLocalRef( env, serialDriverClassID);
		}

		if ( methodID != NULL  ) {

			char data[1024] = "";
			char payloadString[MAX_SERIALFRAME_LEN*2+1] = "";
			_bool isSupported = _false;
			byte responseId = -1;

			rsp.command.serviceId = ldsCommand.serviceId;

			switch( ldsCommand.serviceId ) {
			case SDP_SERVICE:
				rsp.response.status = KO_PROTOCOL_NODE_SERVICE_NOT_SUPPORTED;
				break;
			case FIRMWARE_UPGRADE_SERVICE:
				rsp.response.status = KO_PROTOCOL_NODE_SERVICE_NOT_SUPPORTED;
				break;
			case STACK_SERVICE:
				sprintf( data, "%s", get_hex_payload(payloadString, sizeof(payloadString), (Payload*)ldsCommand.command.payload ));
				isSupported = _true;
				break;
			case NODE_SERVICE_ID_UNDEFINED:
				LOG_DEBUG( g_Logger, "TCD requestLocalNodeService CASE NODE_SERVICE_ID_UNDEFINED %d", ldsCommand.serviceId );
			default:
				// 3. wrong/unknown serviceId!
				rsp.response.status = KO_PROTOCOL_NODE_SERVICE_ID_REQUIRED;
				break;
			}

			if ( isSupported ) {
				if (jni_send_payload_from_hex( waveProtocol, env,
						serialDriver, methodID, data,
						responseId, &rsp.response,
						(Payload*)&rsp.command.command.payload ) !=
								OK_PROTOCOL_SUCCESS) {
					rsp.response.status = KO_PROTOCOL_PROCESS_FAILURE;
				}
			}
		}
	}

	if ( rsp.response.status == OK_PROTOCOL_SUCCESS ) {

		localNodeServiceCommand = create_local_node_service_command_from_payload_object( env,
				g_classPayload, rsp.command.serviceId,
				(const Payload* )&rsp.command.command.payload );
	}

	// put protocol status in idle mode
	waveProtocol->status = PS_IDLE;
	LOG_DEBUG( g_Logger, "requestLocalNodeService status %d %X", rsp.response.status, localNodeServiceCommand );
	return create_local_node_service_response( env, rsp.response.status, localNodeServiceCommand);
}

jobject JNICALL JNIFUNCTION(sendApplicationBroadcastMessage)(JNIEnv *env, jobject jniThis, jobject command) {
	return 0;
}

jobject JNICALL JNIFUNCTION(sendApplicationBroadcastRequest)(JNIEnv *env, jobject jniThis, jobject command) {
	jobject jDynamicExpertParams = NULL;
	RadioExchangeCommand reCommand;
	MulticastRequestResponse rsp;
	memset( &reCommand, 0, sizeof(reCommand) );
	memset( &rsp, 0, sizeof(rsp) );
	rsp.response.status = KO_PROTOCOL_NOT_OPEN;
	ProtocolInfo* waveProtocol = get_protocol_from_java( env, jniThis );

	if ( !waveProtocol ) {
		rsp.response.status = KO_PROTOCOL_NOT_OPEN;
		return create_multicast_request_response( env,
				g_classPayload, g_classRadioExchangeCommand, waveProtocol, &rsp, jDynamicExpertParams);
	}

	if (! JNIFUNCTION(isOpen)( env,  jniThis)) {
		rsp.response.status = KO_PROTOCOL_NOT_OPEN;
		return create_multicast_request_response( env, g_classPayload, g_classRadioExchangeCommand, waveProtocol, &rsp, jDynamicExpertParams);
	}

	// 1. A command session is already in progress?
	if ( waveProtocol->status == PS_PROCESS_IN_PROGRESS ) {
		rsp.response.status = KO_PROTOCOL_ALREADY_IN_USE;
		throw_exception_id( env, rsp.response.status);
		return create_multicast_request_response( env, g_classPayload, g_classRadioExchangeCommand, waveProtocol, &rsp, jDynamicExpertParams);
	}

	LOG_DEBUG(g_Logger, "sendApplicationBroadcastRequest: Sending broadcast request ...", 0);
	// put protocol status in busy mode
	waveProtocol->status = PS_PROCESS_IN_PROGRESS;

	rsp.response.status = KO_PROTOCOL_NO_VALID_PARAMETER;
	jclass classID = (*env)->GetObjectClass( env, command);
	if ( classID != NULL ) {
		rsp.response.status = KO_PAYLOAD_REQUIRED;
		jfieldID fieldID = (*env)->GetFieldID( env, classID, "payload", "Lcom/coronis/sdk/protocol/api/IPayload;");
		if ( fieldID != NULL ) {
			LOG_DEBUG(g_Logger, "sendApplicationBroadcastRequest: Try to get payload ... ", 0);
			jobject serialFrame = (*env)->GetObjectField( env, command, fieldID );
			if ( serialFrame != NULL ) {
				if ( get_payload_from_object( env, serialFrame, (Payload*)&reCommand.command.payload ) )
					rsp.response.status = OK_PROTOCOL_SUCCESS;
				(*env)->DeleteLocalRef( env, serialFrame);
			}

			if (((Payload*)&reCommand.command.payload)->len <= 0)
				rsp.response.status = KO_PAYLOAD_REQUIRED;

			if (((Payload*)&reCommand.command.payload)->len > MAX_REQ_SEND_FRAME_BYTES_LEN)
				rsp.response.status = KO_PAYLOAD_NOT_VALID;

			LOG_DEBUG(g_Logger, "sendApplicationBroadcastRequest: Got payload "
					"with status code %d of len %d", rsp.response.status,
					((Payload*)&reCommand.command.payload)->len);
		}

		LOG_DEBUG(g_Logger, "sendApplicationBroadcastRequest: Try to get dynamic expert parameters ... ", 0);
		if (rsp.response.status == OK_PROTOCOL_SUCCESS) {

			rsp.response.status = get_dynamic_expert_params( env, classID, command, &reCommand.dynamicExpertParams );

			if (rsp.response.status == OK_PROTOCOL_SUCCESS) {
				LOG_DEBUG(g_Logger, "sendApplicationBroadcastRequest: Get field ID", 0);
				fieldID = (*env)->GetFieldID( env, classID, "dynamicExpertParams", "Lcom/coronis/sdk/protocol/api/DynamicExpertParams;");
				if ( fieldID != NULL ) {
					LOG_DEBUG(g_Logger, "sendApplicationBroadcastRequest: Getting jDynamicExpertParams", 0);
					jDynamicExpertParams = (*env)->GetObjectField( env, command, fieldID );

					if (jDynamicExpertParams == NULL)
						LOG_ERROR(g_Logger, "sendApplicationBroadcastRequest: Failed to get jDynamicExpertParams", 0);

					if (jDynamicExpertParams != NULL)
						LOG_ERROR(g_Logger, "sendApplicationBroadcastRequest: GOT jDynamicExpertParams", 0);

				}
			} else {
				LOG_ERROR(g_Logger, "sendApplicationBroadcastRequest: Failed to get dynamic expert parameters !", 0);
			}

		} else {
			LOG_ERROR(g_Logger, "sendApplicationBroadcastRequest: Failed to get dynamic expert parameters !", 0);
		}

		(*env)->DeleteLocalRef( env, classID);
	}

	if ( rsp.response.status == OK_PROTOCOL_SUCCESS ) {
		// put protocol status in idle mode
		waveProtocol->status = PS_IDLE;
		// send application broadcast request in "C" way
		LOG_DEBUG(g_Logger, "sendApplicationBroadcastRequest: Sending broadcast request using C ", 0);
		protocol_send_application_broadcast_request( waveProtocol, &reCommand, &rsp);
		// put protocol status in busy mode
		waveProtocol->status = PS_PROCESS_IN_PROGRESS;
		if ( rsp.response.status == KO_PROTOCOL_NOT_OPEN ) {
			// fail to read it in "C" way, so then try Java way
			LOG_DEBUG(g_Logger, "sendApplicationBroadcastRequest: Sending broadcast request using Java ", 0);

			jmethodID methodID = NULL;
			jobject serialDriver = waveProtocol->jniSerialDriver;
			jclass serialDriverClassID = (*env)->GetObjectClass( env, serialDriver);
			if ( serialDriverClassID != NULL ) {
				methodID = (*env)->GetMethodID( env, serialDriverClassID, "send", "(Lcom/coronis/sdk/serialdriver/api/ISerialFrame;)Lcom/coronis/sdk/serialdriver/api/SendingStatus;");
				(*env)->DeleteLocalRef( env, serialDriverClassID);
			}

			if ( methodID != NULL  ) {
				// 3. Payload filled in?
				if ( ((Payload*)&reCommand.command.payload)->len > 0 &&
						((Payload*)&reCommand.command.payload)->len <= MAX_REQ_SEND_FRAME_BYTES_LEN ) {

					rsp.response.status = KO_PARAM_WRITING_FAILURE;
					jobject cmd = create_radio_settings_command( env, g_classPayload, NULL, jDynamicExpertParams, NULL, NULL, NULL);
					if ( cmd != NULL ) {
						// put protocol status in idle mode
						waveProtocol->status = PS_IDLE;

						jobject commandResponse = JNIFUNCTION(writeRadioParams)( env, jniThis, cmd);
						// put protocol status in busy mode
						waveProtocol->status = PS_PROCESS_IN_PROGRESS;
						if ( commandResponse != NULL ) {
							rsp.response.status = get_response_status( env, commandResponse );

							LOG_DEBUG( g_Logger, "sendApplicationBroadcastRequest -> writeRadioParams: %d", rsp.response.status);
							(*env)->DeleteLocalRef( env, commandResponse );

							if (rsp.response.status ==  KO_PROTOCOL_NO_VALID_PARAMETER) {
								LOG_DEBUG (g_Logger, "JNI sendApplicationBroadcastRequest: There was an error writing parameters, but it is OK when no parameter are specified on port open", 0);
								rsp.response.status = OK_PROTOCOL_SUCCESS;
							}

						}
						(*env)->DeleteLocalRef( env, cmd );
					}

					if ( rsp.response.status == OK_PROTOCOL_SUCCESS )  {

						char payloadString[MAX_SERIALFRAME_LEN*2+1] = "";
						char data[MAX_SERIALFRAME_LEN] = "";
						snprintf( data, sizeof(data)-1, "%02XFFFFFFFFFFFF", REQ_SEND_BROADCAST );
						sprintf( data+strlen(data), "%s", get_hex_payload(payloadString, sizeof(payloadString), (Payload*)&reCommand.command.payload) );
						LOG_DEBUG(g_Logger, "sendApplicationBroadcastRequest: Sending broadcast request jni_send_frame_from_hex", 0);
						jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, RES_SEND_FRAME, &rsp.response, NULL  );
						// prepare receive buffer for requested data
						waveProtocol->receivedFrame.len = 0;

						LOG_DEBUG(g_Logger, "sendApplicationBroadcastRequest: Broadcast request sent, check it's status...", 0);

						if ( rsp.response.status == OK_PROTOCOL_SUCCESS ) {
							LOG_DEBUG(g_Logger, "sendApplicationBroadcastRequest: Broadcast request succesfully sent", 0);

							jni_notify_spies_hex( waveProtocol, data, _true );
							_bool allFramesReceived = _false;
							int totalNumberOfFramesReceived = 0;
							int index = 0;
							int i = 0;
							char frameString[MAX_SERIALFRAME_LEN*2+1] = "";

							do {
								LOG_DEBUG(g_Logger, "sendApplicationBroadcastRequest: Wait for response", 0);

								WavePortResponses wprCode = -1;
								// prepare receive buffer for requested data
								waveProtocol->receivedFrame.len = 0;
								DynamicExpertParams* dynamicExpertParams = &waveProtocol->params.dynamicExpertParams;
								dynamicExpertParams = &reCommand.dynamicExpertParams;

								LOG_DEBUG(g_Logger, "sendApplicationBroadcastRequest: Wait for frame ...", 0);

								wait_for_frame( waveProtocol, dynamicExpertParams,_true, _true, &rsp.response, &wprCode);
								LOG_DEBUG(g_Logger, "sendApplicationBroadcastRequest: Wait for frame done", 0);
								if ( rsp.response.status == OK_PROTOCOL_SUCCESS ) {
									LOG_DEBUG(g_Logger, "sendApplicationBroadcastRequest: Response status is Ok with code %d", wprCode);
									switch( wprCode ) {
									case RECEIVED_BROADCAST_RESPONSE:
										LOG_DEBUG(g_Logger, "sendApplicationBroadcastRequest: Received broadcast response", 0);
										totalNumberOfFramesReceived = waveProtocol->receivedFrame.frame[2];
										if ( totalNumberOfFramesReceived == 0 ) {
											LOG_DEBUG(g_Logger, "sendApplicationBroadcastRequest: No broadcast frames received", 0);
											allFramesReceived = _true;
											break;
										}
										// verify if we receive last frame ( last frame will have index 1 )
										if ( waveProtocol->receivedFrame.frame[3] == 1 )
											allFramesReceived = _true;

										LOG_DEBUG(g_Logger, "sendApplicationBroadcastRequest: copy address RadioExchangeCommands", 0);

										// copy address RadioExchangeCommands
										for ( i = 4; i < ((MAX_RADIOADDRESS_LEN-1)/2 + 4); ++i )
											sprintf( ((RadioAddress*)&rsp.commands[index].dynamicStandardParams.nodeAddress)->addressValue +
													strlen(((RadioAddress*)&rsp.commands[index].dynamicStandardParams.nodeAddress)->addressValue), "%02X",
													waveProtocol->receivedFrame.frame[i] );
										// set length
										((Payload*)&rsp.commands[index].command.payload)->len = waveProtocol->receivedFrame.len - ((MAX_RADIOADDRESS_LEN-1)/2 + 4);

										LOG_DEBUG(g_Logger, "sendApplicationBroadcastRequest:  copy payload", 0);
										// copy payload
										memcpy( ((Payload*)&rsp.commands[index].command.payload)->frame,
												waveProtocol->receivedFrame.frame + (waveProtocol->receivedFrame.len - ((Payload*)&rsp.commands[index].command.payload)->len),
												((Payload*)&rsp.commands[index].command.payload)->len );

										LOG_DEBUG(g_Logger, "sendApplicationBroadcastRequest:  Setup reception mode", 0);
										// set reception mode to broadcast for any received message
										rsp.commands[index].dynamicExpertParams.receptionMode = RECEPTION_MODE_BROADCAST;
										rsp.commands[index].dynamicExpertParams.isSet |= FLAG_RECEPTION_MODE;

										// set the number of received broadcast frames
										rsp.commandsLen = ++index;

										LOG_DEBUG( g_Logger, "TN: %02d index: %02d address: %s ",
												totalNumberOfFramesReceived,
												waveProtocol->receivedFrame.frame[3],
												((RadioAddress*)&rsp.commands[index-1].dynamicStandardParams.nodeAddress)->addressValue );
										break;
									default:
										sprintf( rsp.response.statusDescription, "unexpected received frame during waiting for waveport END_MESSAGE_EXCHANGE %s",
												get_hex_serialframe(frameString, sizeof(frameString), &waveProtocol->receivedFrame) );
										rsp.response.status = KO_RADIO_TRANSMISSION_FAILURE;
										break;
									}
								} else
								{
									LOG_DEBUG(g_Logger, "sendApplicationBroadcastRequest: Response status is NOk: %d", rsp.response.status);
								}
							} while ( rsp.response.status == OK_PROTOCOL_SUCCESS && !allFramesReceived && rsp.commandsLen < MAX_MULTICAST_REQUEST_COMMANDS_LEN );
						} else {
							rsp.response.status = KO_PROTOCOL_PROCESS_FAILURE;
							LOG_DEBUG(g_Logger, "sendApplicationBroadcastRequest: Broadcast request was NOT sent!", 0);
						}
					} else {
						sprintf( rsp.response.statusDescription, "signals param in error %s", get_response_status_name( rsp.response.status ) );
					}
				} else {// 3. Payload filled in? -> NO

					if (((Payload*)&reCommand.command.payload)->len > 0)
						rsp.response.status = KO_PAYLOAD_REQUIRED;
					else
						rsp.response.status = KO_PAYLOAD_NOT_VALID;
				}
			}
		}
	}

	// put protocol status in idle mode
	waveProtocol->status = PS_IDLE;

	LOG_DEBUG(g_Logger, "sendApplicationBroadcastRequest: Call  create_multicast_request_response from sendApplicationBroadcastRequest ... ", 0);
	if (jDynamicExpertParams != NULL)
		LOG_DEBUG(g_Logger, "sendApplicationBroadcastRequest: jDynamicExpertParams is NOT null", 0);

	if (jDynamicExpertParams == NULL)
		LOG_ERROR(g_Logger, "sendApplicationBroadcastRequest: jDynamicExpertParams is NULL", 0);

	LOG_DEBUG(g_Logger, "sendApplicationBroadcastRequest: Exiting with code %d", rsp.response.status);
	return create_multicast_request_response( env, g_classPayload, g_classRadioExchangeCommand, waveProtocol, &rsp, jDynamicExpertParams);
}

jobject JNICALL JNIFUNCTION(sendApplicationMessage)(JNIEnv *env, jobject jniThis, jobject command) {
	return sendMessage( env, jniThis, command, REQ_SEND_MESSAGE, RES_SEND_FRAME);
}

jobject JNICALL JNIFUNCTION(sendServiceMessage)(JNIEnv *env, jobject jniThis, jobject nodeServiceId, jobject command) {
	return sendMessage( env, jniThis, command, REQ_SEND_SERVICE_MESSAGE, RES_SEND_SERVICE);
}

jobject JNICALL JNIFUNCTION(sendApplicationRequest)(JNIEnv *env, jobject jniThis, jobject command) {
	return sendRequest( env, jniThis, command, REQ_SEND_FRAME, RES_SEND_FRAME);
}

JNIEXPORT jobject JNICALL JNIFUNCTION(sendServiceRequest)(JNIEnv *env, jobject jniThis, jobject nodeServiceId, jobject command) {
	return sendRequest( env, jniThis, command, REQ_SEND_SERVICE, RES_SEND_SERVICE);
}


// payload is of type SerialFrame
// Invoked from JAVA, the java environment is passed in env.
jobject JNICALL JNIFUNCTION(sendApplicationResponse)(JNIEnv *env, jobject jniThis, jobject payload) {

	RadioExchangeCommand	reCommand;
	CommandResponse 		rsp;
	ProtocolInfo			*waveProtocol;

	memset( &reCommand, 0, sizeof(reCommand) );
	memset( &rsp, 0, sizeof(rsp) );

	//--------------------------------------------------------------------------
	// 0. check if protocol is opened
	LOG_DEBUG( g_Logger, "Check if protocol is opened", 0);
	waveProtocol = get_protocol_from_java( env, jniThis );
	if (!waveProtocol || !JNIFUNCTION(isOpen)( env,  jniThis)) {
		rsp.status = KO_PROTOCOL_NOT_OPEN;
		return create_command_response( env, g_classSerialFrame, &rsp );
	}
	LOG_DEBUG( g_Logger, " Response session status is : %d", waveProtocol->isResponseSessionActive);

	//--------------------------------------------------------------------------
	// 3. check if a command session is not already in progress
	LOG_DEBUG( g_Logger, "Check if a command session is not already in progress", 0);
	if (waveProtocol->status == PS_PROCESS_IN_PROGRESS ){
		rsp.status = KO_PROTOCOL_ALREADY_IN_USE;
		throw_exception_id(env, rsp.status);
		return create_command_response( env, g_classSerialFrame, &rsp );
	}

	LOG_DEBUG( g_Logger, " Response session status is : %d", waveProtocol->isResponseSessionActive);

	waveProtocol->status = PS_PROCESS_IN_PROGRESS;

	//--------------------------------------------------------------------------
	// 1. Check if pay-load is filled in
	LOG_DEBUG( g_Logger, " Check if pay-load is filled in", 0);
	if ( payload == NULL ) {
		waveProtocol->status = PS_IDLE;
		rsp.status = KO_PAYLOAD_REQUIRED;
		return create_command_response( env, g_classSerialFrame, &rsp );
	}

	//--------------------------------------------------------------------------
	// 2. Get payload data
	LOG_DEBUG( g_Logger, " Get JNI payload", 0);
	Payload cPayload;

	if ((get_payload_from_object(env, payload, &cPayload ) == _false) ||
		(cPayload.len > MAX_REQ_SEND_FRAME_BYTES_LEN)) {

		LOG_DEBUG( g_Logger, " Could not get pay-load from JNI or it is invalid", 0);
		waveProtocol->status = PS_IDLE;
		strcpy( rsp.statusDescription, "Could not get pay-load from JNI" );
		rsp.status = KO_PAYLOAD_NOT_VALID;
		return create_command_response( env, g_classSerialFrame, &rsp );
	}

	//==========================================================================
	// A - send application response in C way
	LOG_DEBUG( g_Logger, " Response session status is : %d", waveProtocol->isResponseSessionActive);
	waveProtocol->status = PS_IDLE;
	Command appCommand;
	set_payload_value(appCommand.payload, &cPayload);

	protocol_send_application_response( waveProtocol, &appCommand, &rsp);
	if ( rsp.status != KO_PROTOCOL_NOT_OPEN) {
		return create_command_response( env, g_classSerialFrame, &rsp );
	}

	LOG_DEBUG( g_Logger, " Response session status is : %d", waveProtocol->isResponseSessionActive);

	//==========================================================================
	LOG_DEBUG( g_Logger, "Data was not sent in C way, trying to send in JAVA way", 0);
	waveProtocol->status = PS_PROCESS_IN_PROGRESS;

	LOG_DEBUG( g_Logger, " Response session status is : %d", waveProtocol->isResponseSessionActive);
	//--------------------------------------------------------------------------
	// 4. Check if a response session is already in progress
	LOG_DEBUG( g_Logger, " Check if a response session is already in progress", 0);
	if (waveProtocol->isResponseSessionActive == 0 ||
		(strlen(((RadioAddress*)&waveProtocol->senderNodeAddress)->addressValue) == 0)	) {
		waveProtocol->status = PS_IDLE;
		LOG_DEBUG( g_Logger, " This is not a response session, exiting.", 0);
		strcpy( rsp.statusDescription, "This is not a response session" );
		rsp.status = KO_PROTOCOL_NO_RESPONSE_SESSION;
		return create_command_response( env, g_classSerialFrame, &rsp );
	}

	// process data in Java way
	waveProtocol->status = PS_PROCESS_IN_PROGRESS;

	LOG_DEBUG( g_Logger, " Get java serial class ID", 0);
	jobject serialDriver = waveProtocol->jniSerialDriver;
	if(!is_java_serial_driver_loaded(env, waveProtocol)) {
		LOG_DEBUG( g_Logger, " JNI Serial driver, from protocol is null!", 0);
		waveProtocol->status = PS_IDLE;
		strcpy( rsp.statusDescription, "JNI Serial driver, from protocol is null" );
		rsp.status = KO_UNEXPECTED_ERROR;
		return create_command_response( env, g_classSerialFrame, &rsp );
	}

	LOG_DEBUG( g_Logger, " Get serialDriverClassID", 0);
	jclass serialDriverClassID = (*env)->GetObjectClass( env, serialDriver);
	LOG_DEBUG( g_Logger, " Get serialDriverClassID - check results", 0);
	if ( serialDriverClassID == NULL ) {
		LOG_DEBUG( g_Logger, " Could not get java serial class ID", 0);
		waveProtocol->status = PS_IDLE;
		strcpy( rsp.statusDescription, " Could not get java serial class ID" );
		rsp.status = KO_UNEXPECTED_ERROR;
		return create_command_response( env, g_classSerialFrame, &rsp );
	}

	LOG_DEBUG( g_Logger, " Get java sending method ID", 0);
	jmethodID methodID = NULL;
	methodID = (*env)->GetMethodID( env, serialDriverClassID, "send", "(Lcom/coronis/sdk/serialdriver/api/ISerialFrame;)Lcom/coronis/sdk/serialdriver/api/SendingStatus;");
	(*env)->DeleteLocalRef( env, serialDriverClassID);

	if (methodID == NULL) {
		LOG_DEBUG( g_Logger, " Could not get java sending method ID", 0);
		waveProtocol->status = PS_IDLE;
		strcpy( rsp.statusDescription, "Could not get java sending method ID" );
		rsp.status = KO_UNEXPECTED_ERROR;
		return create_command_response( env, g_classSerialFrame, &rsp );
	}

	//--------------------------------------------------------------------------
	// REQ_SEND_RESPONSE + <NodeAddress> +  <Payload>
	char data[1024] = "";
	char payloadString[MAX_SERIALFRAME_LEN*2+1] = "";

	LOG_DEBUG(g_Logger, "sendApplicationResponse: Compose response having flag reception mode %s and value %d",
			((waveProtocol->params.dynamicExpertParams.isSet & FLAG_RECEPTION_MODE) ? "true" : "false"),
			waveProtocol->params.dynamicExpertParams.receptionMode);

	_bool isBroadcastResponse = (waveProtocol->receptionSessionMode == RECEPTION_MODE_BROADCAST) ||
			((waveProtocol->params.dynamicExpertParams.isSet & FLAG_RECEPTION_MODE) &&
			(waveProtocol->params.dynamicExpertParams.receptionMode == RECEPTION_MODE_BROADCAST));

	if (isBroadcastResponse) {

		LOG_DEBUG(g_Logger, "sendApplicationResponse: Sending BROADCAST response", 0);
		sprintf( data, "%02X%s%s", REQ_SEND_BROADCAST_RESPONSE,
				((RadioAddress*)&waveProtocol->senderNodeAddress)->addressValue,
				get_hex_payload (payloadString, sizeof(payloadString), &cPayload));
	}
	else {
		LOG_DEBUG(g_Logger, "sendApplicationResponse: Sending NORMAL response", 0);
		sprintf( data, "%02X%s%s", REQ_SEND_RESPONSE,
				((RadioAddress*)&waveProtocol->senderNodeAddress)->addressValue,
				get_hex_payload (payloadString, sizeof(payloadString), &cPayload));
	}

	// send request and wait for confirmation 0x21
	if (jni_send_frame_from_hex( waveProtocol, env, serialDriver,
			methodID, data, RES_SEND_FRAME, &rsp, NULL) != OK_PROTOCOL_SUCCESS) {
		LOG_ERROR( g_Logger, "sendApplicationResponse: Failed to send frame - jni_send_frame_from_hex", 0);

		waveProtocol->isResponseSessionActive = 0;
		rsp.status = KO_PROTOCOL_PROCESS_FAILURE;
		waveProtocol->status = PS_IDLE;
		return create_command_response( env, g_classSerialFrame, &rsp );
	}

	LOG_ERROR( g_Logger, "sendApplicationResponse: Hex frame sent, and ACK from local modem has been received", 0);
	//--------------------------------------------------------------------------
	// consume old frame, and wait for new one
	waveProtocol->receivedFrame.len = 0;

	// close response session
	waveProtocol->isResponseSessionActive = 0;

	//--------------------------------------------------------------------------
	// wait for END_MESSAGE_EXCHANGE
	WavePortResponses responseId;
	wait_for_frame(waveProtocol, &waveProtocol->params.dynamicExpertParams,
			_false, isBroadcastResponse, &rsp, &responseId);

	if ((rsp.status != OK_PROTOCOL_SUCCESS) || (responseId != END_MESSAGE_EXCHANGE)) {
		waveProtocol->status =  PS_IDLE;
		rsp.status = KO_RADIO_TRANSMISSION_FAILURE;
		waveProtocol->status = PS_IDLE;
		return create_command_response( env, g_classSerialFrame, &rsp );
	}

	waveProtocol->status = PS_IDLE;
	return create_command_response( env, g_classSerialFrame, &rsp );
}

jobject JNICALL JNIFUNCTION(sendTransparentMessage)(JNIEnv *env, jobject jniThis, jobject serialFrame) {
	//Command reCommand;
	SerialExchangeData reSerialData;
	CommandResponse rsp;
	//memset( &reCommand, 0, sizeof(reCommand) );
	memset( &reSerialData, 0, sizeof(reSerialData) );
	memset( &rsp, 0, sizeof(rsp) );
	rsp.status = KO_PROTOCOL_NOT_OPEN;
	ProtocolInfo* waveProtocol = get_protocol_from_java( env, jniThis );
	if ( waveProtocol ) {
		if ( JNIFUNCTION(isOpen)( env,  jniThis) ) {
			rsp.status = KO_PROTOCOL_ALREADY_IN_USE;
			// 1. A command session is already in progress?
			if ( waveProtocol->status != PS_PROCESS_IN_PROGRESS ) {
				// put protocol status in busy mode
				waveProtocol->status = PS_PROCESS_IN_PROGRESS;

				if ((*env)->ExceptionCheck(env)) {
					LOG_ERROR(g_Logger, "sendTransparentMessage: Java exception occurred 1", 0);
				}

				rsp.status = KO_PROTOCOL_NO_VALID_PARAMETER;


				/*jclass classID = (*env)->GetObjectClass( env, command);
				if ( classID != NULL ) {
					rsp.status = KO_PAYLOAD_REQUIRED;
					jfieldID fieldID = (*env)->GetFieldID( env, classID, "serialFrame", "Lcom/coronis/sdk/serialdriver/api/ISerialFrame;");
					if ( fieldID != NULL ) {
						jobject serialFrame = (*env)->GetObjectField( env, command, fieldID );
						if ( serialFrame != NULL ) {
							if ( get_serial_frame_from_object( env, serialFrame,
									(SerialFrame*)&reSerialData.serialFrame ) )
								rsp.status = OK_PROTOCOL_SUCCESS;
							(*env)->DeleteLocalRef( env, serialFrame);
						}
					}
					(*env)->DeleteLocalRef( env, classID);
				} */

				if ( get_serial_frame_from_object( env, serialFrame,
						(SerialFrame*)&reSerialData.serialFrame ) )
					rsp.status = OK_PROTOCOL_SUCCESS;

				if ((*env)->ExceptionCheck(env)) {
					LOG_ERROR(g_Logger, "sendTransparentMessage: Java exception occurred 2", 0);
				}

				if ( rsp.status == OK_PROTOCOL_SUCCESS ) {

					// put protocol status in idle mode
					waveProtocol->status = PS_IDLE;
					// read dynamic expert params in "C" way

					//int protocol_send_transparent_message( const void* protocol, const pSerialExchangeFrame frame, CommandResponse* response ) {
					protocol_send_transparent_message( waveProtocol, &reSerialData, &rsp);
					// put protocol status in busy mode
					waveProtocol->status = PS_PROCESS_IN_PROGRESS;
					if ( rsp.status == KO_PROTOCOL_NOT_OPEN ) {
						// fail to read it in "C" way, so then try Java way

						if ((*env)->ExceptionCheck(env)) {
							LOG_ERROR(g_Logger, "sendTransparentMessage: Java exception occurred 3", 0);
						}

						jmethodID methodID = NULL;
						jobject serialDriver = waveProtocol->jniSerialDriver;
						jclass serialDriverClassID = (*env)->GetObjectClass( env, serialDriver);
						if ( serialDriverClassID != NULL ) {
							methodID = (*env)->GetMethodID( env, serialDriverClassID, "send", "(Lcom/coronis/sdk/serialdriver/api/ISerialFrame;)Lcom/coronis/sdk/serialdriver/api/SendingStatus;");
							(*env)->DeleteLocalRef( env, serialDriverClassID);
						}
						if ((*env)->ExceptionCheck(env)) {
							LOG_ERROR(g_Logger, "sendTransparentMessage: Java exception occurred 4", 0);
						}

						if ( methodID != NULL  ) {
							// 3. Payload filled in?
							if ( ((SerialFrame*)&reSerialData.serialFrame)->len > 0 &&
									((SerialFrame*)&reSerialData.serialFrame)->len <= MAX_REQ_SEND_FRAME_BYTES_LEN ) {

								char data[MAX_SERIALFRAME_LEN] = "";
								char frameString[MAX_SERIALFRAME_LEN*2+1] = "";
								sprintf( data, "%s",
										get_hex_serialframe(frameString, sizeof(frameString), (SerialFrame*) &reSerialData.serialFrame) );

								SerialFrame responseFrame;
								jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, -1, &rsp, &responseFrame );
								if ( responseFrame.len ) {
									set_serial_frame_value(reSerialData.serialFrame, &responseFrame);
									rsp.status = OK_PROTOCOL_SUCCESS;
								}
							} else // 3. Payload filled in? -> NO
								rsp.status = KO_PAYLOAD_REQUIRED;
						}
					}

					if ((*env)->ExceptionCheck(env)) {
						LOG_ERROR(g_Logger, "sendTransparentMessage: Java exception occurred 5", 0);
					}

					// place serial frame into response command
					/*
					jobject jSerialFrame = create_serial_frame_object( env, g_classSerialFrame,
							get_hex_serialframe( (SerialFrame*)reSerialData.serialFrame) );
					if ( jSerialFrame ) {
						jclass classID = (*env)->GetObjectClass( env, command);
						if ( classID ) {
							jmethodID methodID = (*env)->GetMethodID( env, classID, "setSerialFrame", "(Lcom/coronis/sdk/serialdriver/api/ISerialFrame;)V");
							if ( methodID != NULL )
								(*env)->CallVoidMethod( env, command, methodID, jSerialFrame );
							(*env)->DeleteLocalRef( env, classID );
						}
						(*env)->DeleteLocalRef( env, jSerialFrame );
					}
					*/
				}
				if ((*env)->ExceptionCheck(env)) {
					LOG_ERROR(g_Logger, "sendTransparentMessage: Java exception occurred 6", 0);
				}

				// put protocol status in idle mode
				waveProtocol->status = PS_IDLE;
			} else // 1. A command session is already in progress? -> NO
				throw_exception_id( env, KO_PROTOCOL_ALREADY_IN_USE);
		}
	}

	return create_command_response( env, g_classSerialFrame, &rsp );
}

jobject JNICALL JNIFUNCTION(writeRadioParams)(JNIEnv *env, jobject jniThis, jobject command) {
	CommandResponse rsp;
	ProtocolInfo* waveProtocol = get_protocol_from_java( env, jniThis );

	memset(&rsp, 0, sizeof(CommandResponse));

	if ((*env)->ExceptionCheck(env)) {
		LOG_ERROR(g_Logger, "writeRadioParams: Java exception occurred 15", 0);
	}

	//--------------------------------------------------------------------------
	if ( !waveProtocol ) {
		rsp.status = KO_PROTOCOL_SETTING_ERROR;
		throw_exception_id( env, rsp.status);
		return create_command_response( env, g_classSerialFrame, &rsp );
	}

	if ( JNIFUNCTION(isOpen)( env,  jniThis) ==JNI_FALSE ) {
		rsp.status = KO_PROTOCOL_NOT_OPEN;
		return create_command_response( env, g_classSerialFrame, &rsp );
	}
	if ((*env)->ExceptionCheck(env)) {
		LOG_ERROR(g_Logger, "writeRadioParams: Java exception occurred 16", 0);
	}

	// 1. A command session is already in progress?
	if ( waveProtocol->status == PS_PROCESS_IN_PROGRESS ) {
		rsp.status = KO_PROTOCOL_ALREADY_IN_USE;
		throw_exception(env, &rsp);
		return create_command_response( env, g_classSerialFrame, &rsp );
	}
	if ((*env)->ExceptionCheck(env)) {
		LOG_ERROR(g_Logger, "writeRadioParams: Java exception occurred 17", 0);
	}


	//--------------------------------------------------------------------------
	waveProtocol->status = PS_PROCESS_IN_PROGRESS;
	rsp.status = KO_PROTOCOL_NO_VALID_PARAMETER;
	RadioSettingsCommand cmd;
	memset( &cmd, 0, sizeof(cmd) );
	cmd.dynamicStandardParams.groupId = -1;

	jclass classID = (*env)->GetObjectClass( env, command);
	if ( classID == NULL ) {
		waveProtocol->status = PS_IDLE;
		return create_command_response( env, g_classSerialFrame, &rsp );
	}

	if ((*env)->ExceptionCheck(env)) {
		LOG_ERROR(g_Logger, "writeRadioParams: Java exception occurred 18", 0);
	}

	/*+++++++++++++++++++++++++++++++++++++++ START handling of dynamicStandardParams +++++++++++++++++++++++++++++++++++++++*/
	// [A]
	// 1. Get parameters from command
	LOG_DEBUG(g_Logger, "writeRadioParams: call of get_dynamic_standard_params", 0);

	rsp.status = get_dynamic_standard_params( env, classID, command, &cmd.dynamicStandardParams );
	LOG_DEBUG(g_Logger, "writeRadioParams: Exited from get_dynamic_standard_params with status %d", rsp.status);

	_bool areDynamicStandardParamsSet  = rsp.status == KO_PROTOCOL_NO_VALID_PARAMETER ? _false : _true;
	if (rsp.status == KO_PROTOCOL_NO_VALID_PARAMETER ) {
		LOG_DEBUG(g_Logger, "writeRadioParams: Dynamic standard params are not set, take this into account, and reset error code ", 0);
		rsp.status = OK_PROTOCOL_SUCCESS;
	}

	if ((*env)->ExceptionCheck(env)) {
		LOG_ERROR(g_Logger, "writeRadioParams: Java exception occurred 12", 0);
	}

	/*+++++++++++++++++++++++++++++++++++++++ START handling of dynamicExpertParams +++++++++++++++++++++++++++++++++++++++*/
	// [B].
	if ( rsp.status == OK_PROTOCOL_SUCCESS )
		rsp.status = get_dynamic_expert_params( env, classID, command, &cmd.dynamicExpertParams );

	/*+++++++++++++++++++++++++++++++++++++++ START handling of staticStandardParams +++++++++++++++++++++++++++++++++++++++*/
	// [C]
	LOG_DEBUG(g_Logger, "writeRadioParams: start of reading staticStandardParams", 0);

	jfieldID fieldID = (*env)->GetFieldID( env, classID, "staticStandardParams", "Lcom/coronis/sdk/protocol/api/StaticStandardParams;");
	if ( ( rsp.status == OK_PROTOCOL_SUCCESS ) && ( fieldID != NULL ) ) {
		LOG_DEBUG(g_Logger, "writeRadioParams: Got fieldID of staticStandardParams", 0);
		jobject staticStandardParams = (*env)->GetObjectField( env, command, fieldID );
		if ( staticStandardParams != NULL ) {
			LOG_DEBUG(g_Logger, "writeRadioParams: Got staticStandardParams ", 0);
			jclass classStaticStandardParamsID = (*env)->GetObjectClass( env, staticStandardParams);
			if ( classStaticStandardParamsID != NULL ) {
				LOG_DEBUG(g_Logger, "writeRadioParams: Got classStaticStandardParamsID ", 0);
				/************************* START handling of staticStandardParams->firstScanActivityPeriod *************************/
				jint firstScanActivityPeriod = 0;
				if ( ( rsp.status == OK_PROTOCOL_SUCCESS ) && get_int_field_from_object( env, staticStandardParams, classStaticStandardParamsID, "firstScanActivityPeriod", &firstScanActivityPeriod) ) {
					if ( firstScanActivityPeriod < ACTIVITY_PERIOD_MIN ||
							firstScanActivityPeriod > ACTIVITY_PERIOD_MAX ||
							((firstScanActivityPeriod % 100) != 0) )
						rsp.status = KO_PARAMETER_FIRST_SCAN_ACTIVITY_PERIOD_NOT_VALID;
					else {
						cmd.staticStandardParams.firstScanActivityPeriod = firstScanActivityPeriod;
						cmd.staticStandardParams.isSet |= FLAG_ACTIVITY_PERIOD_ON_FIRST_SCAN;
					}
				}
				/************************* END handling of staticStandardParams->firstScanActivityPeriod *************************/
				if ((*env)->ExceptionCheck(env)) {
					LOG_ERROR(g_Logger, "create_command_response: Java exception occurred 13", 0);
				}

				/************************* START handling of staticStandardParams->switchingModeOnRadioAddressActivation *************************/
				jboolean switchingModeOnRadioAddressActivation = JNI_FALSE;
				if ( ( rsp.status == OK_PROTOCOL_SUCCESS ) && get_boolean_field_from_object( env, staticStandardParams, classStaticStandardParamsID, "switchingModeOnRadioAddressActivation", &switchingModeOnRadioAddressActivation) ) {
					if ( switchingModeOnRadioAddressActivation != JNI_FALSE &&
							switchingModeOnRadioAddressActivation != JNI_TRUE )
						rsp.status = KO_PARAMETER_SWITCHING_MODE_ON_RADIO_ADDRESS_ACTIVATION_NOT_VALID;
					else {
						cmd.staticStandardParams.switchingModeOnRadioAddressActivation = (switchingModeOnRadioAddressActivation == JNI_TRUE) ? _true: _false;
						cmd.staticStandardParams.isSet |= FLAG_SWITCHING_MODE_ON_RADIO_ADDRESS_ACTIVATION;
					}
				}
				/************************* END handling of staticStandardParams->switchingModeOnRadioAddressActivation *************************/

				/************************* START handling of staticStandardParams->radioAckActivation *************************/
				jboolean radioAckActivation = JNI_FALSE;
				LOG_DEBUG(g_Logger, "writeRadioParams: Try to write radioAckActivation ... ", 0);
				if ( ( rsp.status == OK_PROTOCOL_SUCCESS ) && get_boolean_field_from_object( env, staticStandardParams, classStaticStandardParamsID, "radioAckActivation", &radioAckActivation) ) {

					LOG_DEBUG(g_Logger, "writeRadioParams: Got radioAckActivation from object", 0);
					if ( radioAckActivation != JNI_FALSE &&
							radioAckActivation != JNI_TRUE ) {
						rsp.status = KO_PARAMETER_RADIO_ACK_ACTIVATION_NOT_VALID;
						LOG_DEBUG(g_Logger, "writeRadioParams: radioAckActivation not valid", 0);
					}
					else {
						// LOG_WARN( g_Logger, "Radio acknowledge activation is not accessible in Waveport!", 0 );
						cmd.staticStandardParams.radioAckActivation = (radioAckActivation == JNI_TRUE) ? _true: _false;
						cmd.staticStandardParams.isSet |= FLAG_RADIO_ACK_ACTIVATION;
						LOG_DEBUG(g_Logger, "writeRadioParams: radioAckActivation is set to value %s", radioAckActivation == JNI_TRUE  ?"true" : "false");
					}
				} else {
					LOG_DEBUG(g_Logger, "writeRadioParams: Could not process radioAckActivation status code: %d", rsp.status );
				}
				/************************* END handling of staticStandardParams->radioAckActivation *************************/

				/************************* START handling of staticStandardParams->receptionGroupId *************************/
				jint receptionGroupId = 0;
				if ( ( rsp.status == OK_PROTOCOL_SUCCESS ) && get_int_field_from_object( env, staticStandardParams, classStaticStandardParamsID, "receptionGroupId", &receptionGroupId) ) {
					if ( receptionGroupId < 0 ||
							receptionGroupId > 0xFF )
						rsp.status = KO_PARAMETER_RECEPTION_GROUP_ID_NOT_VALID;
					else {
						cmd.staticStandardParams.receptionGroupId = receptionGroupId;
						cmd.staticStandardParams.isSet |= FLAG_RECEPTION_GROUPID;
					}
				}
				/************************* END handling of staticStandardParams->firstScanActivityPeriod *************************/
				(*env)->DeleteLocalRef( env, classStaticStandardParamsID);
			}
			(*env)->DeleteLocalRef( env, staticStandardParams);
		}
	}
	/*+++++++++++++++++++++++++++++++++++++++ END handling of dynamicStandardParams +++++++++++++++++++++++++++++++++++++++*/

	/*+++++++++++++++++++++++++++++++++++++++ START handling of staticExpertParams +++++++++++++++++++++++++++++++++++++++*/
	unsigned short secondScanPhysicalConfigPayload = 0;
	unsigned short firstScanPhysicalConfigPayload = 0;

	LOG_DEBUG(g_Logger, "writeRadioParams: Handling of staticExpertParams ... ", 0);
	fieldID = (*env)->GetFieldID(env, classID, "staticExpertParams", "Lcom/coronis/sdk/protocol/api/StaticExpertParams;");
	if ( fieldID != NULL ) {

		LOG_DEBUG(g_Logger, "writeRadioParams: Got field ID of staticExpertParams ... ", 0);
		jobject staticExpertParams = (*env)->GetObjectField( env, command, fieldID );
		if ( staticExpertParams != NULL ) {
			LOG_DEBUG(g_Logger, "writeRadioParams: Got object field of staticExpertParams ... ", 0);
			jclass classStaticExpertParamsID = (*env)->GetObjectClass( env, staticExpertParams);
			if ( classStaticExpertParamsID != NULL ) {

				LOG_DEBUG(g_Logger, "writeRadioParams: Got object Class of staticExpertParams ... ", 0);
				/************************* START handling of staticExpertParams->secondScanEnableActivation *************************/
				jboolean secondScanEnableActivation = JNI_FALSE;
				if ( ( rsp.status == OK_PROTOCOL_SUCCESS ) && get_boolean_field_from_object( env, staticExpertParams, classStaticExpertParamsID, "secondScanEnableActivation", &secondScanEnableActivation) ) {
					if ( secondScanEnableActivation != JNI_FALSE &&
							secondScanEnableActivation != JNI_TRUE )
						rsp.status = KO_PARAMETER_SECOND_SCAN_ENABLE_ACTIVATION_NOT_VALID;
					else {
						cmd.staticExpertParams.secondScanEnableActivation = (secondScanEnableActivation == JNI_TRUE) ? _true: _false;
						cmd.staticExpertParams.isSet |= FLAG_SECOND_SCAN_ENABLE_ACTIVATION;
					}
				}
				/************************* END handling of staticExpertParams->secondScanEnableActivation *************************/

				/************************* START handling of staticExpertParams->radioCRCActivation *************************/
				jboolean radioCRCActivation = JNI_FALSE;
				if ( ( rsp.status == OK_PROTOCOL_SUCCESS ) && get_boolean_field_from_object( env, staticExpertParams, classStaticExpertParamsID, "radioCRCActivation", &radioCRCActivation) ) {
					if ( radioCRCActivation != JNI_FALSE &&
							radioCRCActivation != JNI_TRUE )
						rsp.status = KO_PARAMETER_RADIO_CRC_ACTIVATION_NOT_VALID;
					else {
						cmd.staticExpertParams.radioCRCActivation = (radioCRCActivation == JNI_TRUE) ? _true: _false;
						cmd.staticExpertParams.isSet |= FLAG_RADIO_CRC_ACTIVATION;
					}
				}
				/************************* END handling of staticExpertParams->radioCRCActivation *************************/

				/************************* START handling of staticExpertParams->firstScanPhysicalConfig *************************/
				PhysicalLayer firstScanPhysicalConfig = 0;
				if ( ( rsp.status == OK_PROTOCOL_SUCCESS ) &&
						get_physical_config_field_from_object( env, staticExpertParams, classStaticExpertParamsID,
								"firstScanPhysicalConfig", &firstScanPhysicalConfig,(int*) &rsp.status,
								KO_PARAMETER_FIRST_SCAN_PHYSICAL_CONFIG_NOT_VALID) ) {

					if (!physicalLayerToPayload (firstScanPhysicalConfig, &firstScanPhysicalConfigPayload))
						rsp.status = KO_PARAMETER_FIRST_SCAN_PHYSICAL_CONFIG_NOT_VALID;
					else {
						cmd.staticExpertParams.firstScanPhysicalConfig = firstScanPhysicalConfig;
						cmd.staticExpertParams.isSet |= FLAG_PHYSICAL_CONFIG_ON_FIRST_SCAN;
					}
				}
				/************************* END handling of staticExpertParams->firstScanPhysicalConfig *************************/

				/************************* START handling of staticExpertParams->firstScanPhysicalConfig *************************/
				PhysicalLayer secondScanPhysicalConfig = 0;

				if ( ( rsp.status == OK_PROTOCOL_SUCCESS ) &&
					   get_physical_config_field_from_object( env, staticExpertParams, classStaticExpertParamsID,
								"secondScanPhysicalConfig", &secondScanPhysicalConfig,
								(int*)&rsp.status, KO_PARAMETER_SECOND_SCAN_PHYSICAL_CONFIG_NOT_VALID) ) {

					if (!physicalLayerToPayload (secondScanPhysicalConfig, &secondScanPhysicalConfigPayload))
						rsp.status = KO_PARAMETER_SECOND_SCAN_PHYSICAL_CONFIG_NOT_VALID;
					else {
						cmd.staticExpertParams.firstScanPhysicalConfig = secondScanPhysicalConfig;
						cmd.staticExpertParams.isSet |= FLAG_PHYSICAL_CONFIG_ON_SECOND_SCAN;
					}

				}
				/************************* END handling of staticExpertParams->firstScanPhysicalConfig *************************/

				(*env)->DeleteLocalRef( env, classStaticExpertParamsID);
			}

			(*env)->DeleteLocalRef( env, staticExpertParams);
		}
	}

	LOG_DEBUG(g_Logger, "writeRadioParams: End handling of staticExpertParams", 0);
	/*+++++++++++++++++++++++++++++++++++++++ END handling of staticExpertParams +++++++++++++++++++++++++++++++++++++++*/

	/*+++++++++++++++++++++++++++++++++++++++ START handling of additional parameter (store in payload) +++++++++++++++++++++++++++++++++++++++*/
	LOG_DEBUG(g_Logger, "writeRadioParams: Try to get serial frame from jObject (with status %d)", rsp.status);
	if ( (rsp.status == OK_PROTOCOL_SUCCESS)) {
		jfieldID fieldID = (*env)->GetFieldID( env, classID, "payload", "Lcom/coronis/sdk/protocol/api/IPayload;");
		if ( fieldID != NULL ) {
			LOG_DEBUG(g_Logger, "writeRadioParams: Got class Id of payload", 0);
			jobject jPayload = (*env)->GetObjectField( env, command, fieldID );
			if ( jPayload != NULL ) {
				LOG_DEBUG(g_Logger, "writeRadioParams: Got payload", 0);
				get_payload_from_object( env, jPayload,
						(Payload*)&cmd.command.payload );
				(*env)->DeleteLocalRef( env, jPayload);
			}
		}
	}

	if ((*env)->ExceptionCheck(env)) {
		LOG_ERROR(g_Logger, "writeRadioParams: Java exception occurred A.3", 0);
	}

	/*+++++++++++++++++++++++++++++++++++++++ END handling of serialFrame +++++++++++++++++++++++++++++++++++++++*/

	if (cmd.command.payload != NULL)
	LOG_ERROR(g_Logger, "writeRadioParams:  Check if at least one parameter is set - payload %s,%i", cmd.command.payload, ((Payload*)&cmd.command.payload)->len);
	/*+++++++++++++++++++++++++++++++++++++++ START Check if at least one parameter is set +++++++++++++++++++++++++++++++++++*/
	if((rsp.status == OK_PROTOCOL_SUCCESS) 	    &&
		(areDynamicStandardParamsSet != _true) &&
	    (cmd.staticStandardParams.isSet == 0)   &&
	    (cmd.dynamicExpertParams.isSet == 0)    &&
	    (cmd.staticExpertParams.isSet  == 0) &&
	    (cmd.command.payload == NULL || ((Payload*)&cmd.command.payload)->len)==0) {

		LOG_DEBUG(g_Logger, "writeRadioParams: There are no parameters set ", 0);
		rsp.status = KO_PROTOCOL_NO_VALID_PARAMETER;
	}
	/*+++++++++++++++++++++++++++++++++++++++ END Check if at least one parameter is set +++++++++++++++++++++++++++++++++++*/

	if ((*env)->ExceptionCheck(env)) {
		LOG_ERROR(g_Logger, "writeRadioParams: Java exception occurred A.4", 0);
	}



	// 2. write parameters
	LOG_DEBUG(g_Logger, "writeRadioParams: Try to write parameters (status now is %d)", rsp.status);
	if ((rsp.status == OK_PROTOCOL_SUCCESS)) {
		waveProtocol->status = PS_IDLE;


		if ((*env)->ExceptionCheck(env)) {
			LOG_ERROR(g_Logger, "writeRadioParams: Java exception occurred A.2", 0);
		}
		protocol_write_radio_params( waveProtocol, &cmd, &rsp ); // try to write radio params in "C" way
		if ((*env)->ExceptionCheck(env)) {
			LOG_ERROR(g_Logger, "writeRadioParams: Java exception occurred A.1", 0);
		}

		waveProtocol->status = PS_PROCESS_IN_PROGRESS;

		LOG_DEBUG(g_Logger, "writeRadioParams: Try to write params in Java way (after code: %d) ", rsp.status);
		if ( rsp.status == KO_PROTOCOL_NOT_OPEN ) {// "C" protocol has not opened, so Java should be!
			LOG_DEBUG(g_Logger, "writeRadioParams: Write params in Java, because C protoclo was not written )", 0);

			jobject serialDriver = waveProtocol->jniSerialDriver;
			jclass serialDriverClassID = (*env)->GetObjectClass( env, serialDriver);
			if ( serialDriverClassID != NULL ) {
				jmethodID methodID = (*env)->GetMethodID( env, serialDriverClassID, "send", "(Lcom/coronis/sdk/serialdriver/api/ISerialFrame;)Lcom/coronis/sdk/serialdriver/api/SendingStatus;");
				(*env)->DeleteLocalRef( env, serialDriverClassID);

				if ( methodID != NULL ) {
					char data[1024] = "";
					rsp.status = OK_PROTOCOL_SUCCESS;
					/************************* START writing of dynamicStandardParams->groupId *************************/
					if ( cmd.dynamicStandardParams.groupId > 0 ) {
						sprintf( data, "%02X%02X%02X", REQ_WRITE_RADIO_PARAM, GROUP_NUMBER, cmd.dynamicStandardParams.groupId);
						if ( OK_PROTOCOL_SUCCESS != jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, RES_WRITE_RADIO_PARAM, &rsp, NULL  ) ) {
							rsp.status = KO_PARAMETER_GROUP_NUMBER_VALUE_NOT_SUPPORTED;
							//rsp.status =  KO_PROTOCOL_PROCESS_FAILURE;
							strcpy( rsp.statusDescription, "GROUP_NUMBER" );
						}
					}
					/************************* END writing of dynamicStandardParams->groupId *************************/
					if ((*env)->ExceptionCheck(env)) {
						LOG_ERROR(g_Logger, "writeRadioParams: Java exception occurred 11", 0);
					}

					waveProtocol->relayNb = 0;
					/************************* START writing of dynamicStandardParams->routerPath *************************/
					if ( ( rsp.status == OK_PROTOCOL_SUCCESS ) && strlen(((RadioAddress*)&cmd.dynamicStandardParams.routerPath[0])->addressValue) > 0 ) {
						size_t i = 0;
						sprintf( data, "%02X%02X%02X", REQ_WRITE_RADIO_PARAM, RELAY_ROUTE, 0);
						for ( ; i < ARRAYSIZE(cmd.dynamicStandardParams.routerPath); ++i ) {
							if ( strlen(((RadioAddress*)&cmd.dynamicStandardParams.routerPath[i])->addressValue) > 0 ) {
								++waveProtocol->relayNb;
								strcat( data, ((RadioAddress*)&cmd.dynamicStandardParams.routerPath[i])->addressValue);
								data[5] = '0' + (i+1);
							}
						}
						if ( OK_PROTOCOL_SUCCESS != jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, RES_WRITE_RADIO_PARAM, &rsp, NULL  ) ) {
							rsp.status = KO_PARAMETER_RELAY_ROUTE_VALUE_NOT_SUPPORTED;
							//rsp.status = KO_PROTOCOL_PROCESS_FAILURE;
							strcpy( rsp.statusDescription, "RELAY_ROUTE" );
						}
					}
					/************************* END writing of dynamicStandardParams->routerPath *************************/

					/************************* START writing of dynamicExpertParams->wakeupLength *************************/
					if ( ( rsp.status == OK_PROTOCOL_SUCCESS ) && cmd.dynamicExpertParams.isSet & FLAG_WAKEUP_LENGTH ) {

						sprintf( data, "%02X%02X%02X%02X", REQ_WRITE_RADIO_PARAM, WAKEUP_LENGTH,
								(byte)(cmd.dynamicExpertParams.wakeupLength & 0x00FF),
								(byte)(cmd.dynamicExpertParams.wakeupLength >> 8));

						LOG_DEBUG(g_Logger, "Writting wakeup length of value %d and payload %s", cmd.dynamicExpertParams.wakeupLength, data);
						if ( OK_PROTOCOL_SUCCESS != jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, RES_WRITE_RADIO_PARAM, &rsp, NULL  ) ) {
							rsp.status = KO_PARAMETER_WAKEUP_LENGTH_VALUE_NOT_SUPPORTED;
							//rsp.status = KO_PROTOCOL_PROCESS_FAILURE;
							strcpy( rsp.statusDescription, "WAKEUP_LENGTH" );
						}
					}
					/************************* END writing of dynamicExpertParams->wakeupLength *************************/

					/************************* START writing of dynamicExpertParams->responseTimeout *************************/
					if ( ( rsp.status == OK_PROTOCOL_SUCCESS ) && cmd.dynamicExpertParams.isSet & FLAG_RESPONSE_TIMEOUT ) {
						sprintf( data, "%02X%02X%02X", REQ_WRITE_RADIO_PARAM, RADIO_USER_TIMEOUT, (byte)(cmd.dynamicExpertParams.responseTimeout / 100));
						if ( OK_PROTOCOL_SUCCESS != jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, RES_WRITE_RADIO_PARAM, &rsp, NULL  ) ) {
							rsp.status = KO_PARAMETER_RESPONSE_TIMEOUT_VALUE_NOT_SUPPORTED;
							//rsp.status = KO_PROTOCOL_PROCESS_FAILURE;
							strcpy( rsp.statusDescription, "RADIO_USER_TIMEOUT" );
						}
					}
					/************************* END writing of dynamicExpertParams->responseTimeout *************************/
					if ((*env)->ExceptionCheck(env)) {
						LOG_ERROR(g_Logger, "writeRadioParams: Java exception occurred 10", 0);
					}

					/************************* START writing of dynamicExpertParams->broadcastSessionTimeout *************************/
					if ( ( rsp.status == OK_PROTOCOL_SUCCESS ) && cmd.dynamicExpertParams.isSet & FLAG_BROADCAST_SESSION_TIMEOUT ) {
						sprintf( data, "%02X%02X%02X", REQ_WRITE_RADIO_PARAM, BCST_RECEPTION_TIMEOUT, (byte)(cmd.dynamicExpertParams.broadcastSessionTimeout / 100));
						if ( OK_PROTOCOL_SUCCESS != jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, RES_WRITE_RADIO_PARAM, &rsp, NULL  ) ) {
							rsp.status = KO_PARAMETER_BROADCAST_SESSION_TIMEOUT_VALUE_NOT_SUPPORTED;
							//rsp.status = KO_PROTOCOL_PROCESS_FAILURE;
							strcpy( rsp.statusDescription, "BCST_RECEPTION_TIMEOUT" );
						}
					}
					/************************* END writing of dynamicExpertParams->broadcastSessionTimeout *************************/

					/************************* START writing of dynamicExpertParams->broadcastSessionMaxResponsesNb *************************/
					if ( ( rsp.status == OK_PROTOCOL_SUCCESS ) && cmd.dynamicExpertParams.isSet & FLAG_BROADCAST_SESSION_MAX_RESPONSES_NB ) {
						sprintf( data, "%02X%02X%02X", REQ_WRITE_RADIO_PARAM, BCST_MAX_RESPONSES_NB, cmd.dynamicExpertParams.broadcastSessionMaxResponsesNb);
						if ( OK_PROTOCOL_SUCCESS != jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, RES_WRITE_RADIO_PARAM, &rsp, NULL  ) ) {
							LOG_DEBUG( g_Logger, "BCST_MAX_RESPONSES_NB", 0);
							rsp.status = KO_PARAMETER_BROADCAST_SESSION_MAX_RESPONSES_NB_VALUE_NOT_SUPPORTED;
							//rsp.status = KO_PROTOCOL_PROCESS_FAILURE;
							strcpy( rsp.statusDescription, "BCST_MAX_RESPONSES_NB" );
						}
					}
					/************************* END writing of dynamicExpertParams->broadcastSessionMaxResponsesNb *************************/

					/************************* START writing of dynamicExpertParams->transmissionRadioChannel *************************/
					if ( ( rsp.status == OK_PROTOCOL_SUCCESS ) && cmd.dynamicExpertParams.isSet & FLAG_TRANSMISSION_RADIO_CHANNEL ) {
						sprintf( data, "%02X%02X", REQ_SELECT_CHANNEL, cmd.dynamicExpertParams.transmissionRadioChannel);
						if ( OK_PROTOCOL_SUCCESS != jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, RES_SELECT_CHANNEL, &rsp, NULL  ) ) {
							//rsp.status = KO_PROTOCOL_PROCESS_FAILURE;
							rsp.status = KO_PARAMETER_TRANSMISSION_RADIO_CHANNEL_VALUE_NOT_SUPPORTED;
							strcpy( rsp.statusDescription, "REQ_SELECT_CHANNEL" );
						}
					}
					/************************* END writing of dynamicExpertParams->transmissionRadioChannel *************************/
					if ((*env)->ExceptionCheck(env)) {
						LOG_ERROR(g_Logger, "writeRadioParams: Java exception occurred 9", 0);
					}

					/************************* START writing of dynamicExpertParams->transmissionRadioChannel *************************/
					if ( ( rsp.status == OK_PROTOCOL_SUCCESS ) && cmd.dynamicExpertParams.isSet & FLAG_TRANSMISSION_POWER_LEVEL ) {

						// Perform conversion from float to payload
						byte transmissionPowerLevelPayload = 0;
						if ( transmissionPowerLevelToPayload (cmd.dynamicExpertParams.transmissionPowerLevel,
								&transmissionPowerLevelPayload)) {

							sprintf( data, "%02X%02X", REQ_CHANGE_TX_POWER, transmissionPowerLevelPayload);
							if ( OK_PROTOCOL_SUCCESS != jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, RES_CHANGE_TX_POWER, &rsp, NULL  ) ) {

								rsp.status = KO_PARAMETER_TRANSMISSION_POWER_LEVEL_VALUE_NOT_SUPPORTED;
								//rsp.status = KO_PROTOCOL_PROCESS_FAILURE;
								strcpy( rsp.statusDescription, "REQ_CHANGE_TX_POWER" );
							}
						} else {
							rsp.status = KO_PARAMETER_TRANSMISSION_POWER_LEVEL_VALUE_NOT_SUPPORTED;
						}
					}
					/************************* END writing of dynamicExpertParams->transmissionRadioChannel *************************/

					/************************* START writing of staticStandardParams->firstScanActivityPeriod *************************/
					if ( ( rsp.status == OK_PROTOCOL_SUCCESS ) && cmd.staticStandardParams.isSet & FLAG_ACTIVITY_PERIOD_ON_FIRST_SCAN ) {
						sprintf( data, "%02X%02X%02X", REQ_WRITE_RADIO_PARAM, AWAKENING_PERIOD, cmd.staticStandardParams.firstScanActivityPeriod / 100 );
						if ( OK_PROTOCOL_SUCCESS != jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, RES_WRITE_RADIO_PARAM, &rsp, NULL  ) ) {
							rsp.status = KO_PARAMETER_FIRST_SCAN_ACTIVITY_PERIOD_VALUE_NOT_SUPPORTED;
							//rsp.status = KO_PROTOCOL_PROCESS_FAILURE;
							strcpy( rsp.statusDescription, "AWAKENING_PERIOD" );
						}
					}
					/************************* END writing of staticStandardParams->firstScanActivityPeriod *************************/

					/************************* START writing of staticStandardParams->switchingModeOnRadioAddressActivation *************************/
					if ( ( rsp.status == OK_PROTOCOL_SUCCESS ) && cmd.staticStandardParams.isSet & FLAG_SWITCHING_MODE_ON_RADIO_ADDRESS_ACTIVATION ) {
						sprintf( data, "%02X%02X%02X", REQ_WRITE_RADIO_PARAM, SWITCH_MODE_STATUS, ( cmd.staticStandardParams.switchingModeOnRadioAddressActivation == _true ) ? 1 : 0 );
						if ( OK_PROTOCOL_SUCCESS != jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, RES_WRITE_RADIO_PARAM, &rsp, NULL  ) ) {
							rsp.status = KO_PARAMETER_SWITCHING_MODE_ON_RADIO_ADDRESS_ACTIVATION_VALUE_NOT_SUPPORTED;
							//rsp.status = KO_PROTOCOL_PROCESS_FAILURE;
							strcpy( rsp.statusDescription, "SWITCH_MODE_STATUS" );
						}
					}
					/************************* END writing of staticStandardParams->switchingModeOnRadioAddressActivation *************************/
					if ((*env)->ExceptionCheck(env)) {
						LOG_ERROR(g_Logger, "writeRadioParams: Java exception occurred 8", 0);
					}

					/************************* START writing of staticStandardParams->receptionGroupId *************************/
					if ( ( rsp.status == OK_PROTOCOL_SUCCESS ) && cmd.staticStandardParams.isSet & FLAG_RECEPTION_GROUPID ) {
						sprintf( data, "%02X%02X%02X", REQ_WRITE_RADIO_PARAM, WAVECARD_MULTICAST_GROUP, cmd.staticStandardParams.receptionGroupId );
						if ( OK_PROTOCOL_SUCCESS != jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, RES_WRITE_RADIO_PARAM, &rsp, NULL  ) ) {
							rsp.status = KO_PARAMETER_RECEPTION_GROUP_ID_VALUE_NOT_SUPPORTED;
							//rsp.status = KO_PROTOCOL_PROCESS_FAILURE;
							strcpy( rsp.statusDescription, "WAVECARD_MULTICAST_GROUP" );
						}
					}
					/************************* END writing of staticStandardParams->receptionGroupId *************************/

					/************************* START writing of staticStandardParams->radioAckActivation *************************/
					LOG_DEBUG(g_Logger, "writeRadioParams: RADIO_ACKNOWLEDGE : Try to write this param", 0);
					if ( ( rsp.status == OK_PROTOCOL_SUCCESS ) && (cmd.staticStandardParams.isSet & FLAG_RADIO_ACK_ACTIVATION)) {
						sprintf( data, "%02X%02X%02X", REQ_WRITE_RADIO_PARAM, RADIO_ACKNOWLEDGE, cmd.staticStandardParams.radioAckActivation == _true ? 1 : 0 );
						if ( OK_PROTOCOL_SUCCESS != jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, RES_WRITE_RADIO_PARAM, &rsp, NULL  ) ) {
							rsp.status = KO_PARAMETER_RADIO_ACK_ACTIVATION_VALUE_NOT_SUPPORTED;
							//rsp.status = KO_PROTOCOL_PROCESS_FAILURE;
							strcpy( rsp.statusDescription, "RADIO_ACKNOWLEDGE" );
						}
					}else {
						LOG_DEBUG(g_Logger, "writeRadioParams: RADIO_ACKNOWLEDGE : "
								"Did not wrote status %d, flag %s", rsp.status,
								cmd.staticStandardParams.isSet & FLAG_RADIO_ACK_ACTIVATION ? "true" : "false");

					}
					/************************* END writing of staticStandardParams->receptionGroupId *************************/
					if ((*env)->ExceptionCheck(env)) {
						LOG_ERROR(g_Logger, "writeRadioParams: Java exception occurred 7", 0);
					}

					/************************* START writing of staticExpertParams->secondScanEnableActivation *************************/
					if ( ( rsp.status == OK_PROTOCOL_SUCCESS ) && cmd.staticExpertParams.isSet & FLAG_SECOND_SCAN_ENABLE_ACTIVATION ) {
						sprintf( data, "%02X%02X%02X", REQ_WRITE_RADIO_PARAM, SECOND_SCAN_ENABLE_ACTIVATION, ( cmd.staticExpertParams.secondScanEnableActivation == _true ) ? 1 : 0 );
						if ( OK_PROTOCOL_SUCCESS != jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, RES_WRITE_RADIO_PARAM, &rsp, NULL  ) ) {
							// rsp.status = KO_PROTOCOL_PROCESS_FAILURE;
							rsp.status = KO_PARAMETER_SECOND_SCAN_ENABLE_ACTIVATION_VALUE_NOT_SUPPORTED;
							strcpy( rsp.statusDescription, "SECOND_SCAN_ENABLE_ACTIVATION" );
						}
					}
					/************************* END writing of staticExpertParams->secondScanEnableActivation *************************/

					/************************* START writing of staticExpertParams->radioCRCActivation *************************/
					if ( ( rsp.status == OK_PROTOCOL_SUCCESS ) && cmd.staticExpertParams.isSet & FLAG_RADIO_CRC_ACTIVATION ) {
						sprintf( data, "%02X%02X%02X", REQ_WRITE_RADIO_PARAM, RADIO_CRC_ACTIVATION, ( cmd.staticExpertParams.radioCRCActivation == _true ) ? 1 : 0 );
						if ( OK_PROTOCOL_SUCCESS != jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, RES_WRITE_RADIO_PARAM, &rsp, NULL  ) ) {
							//rsp.status = KO_PROTOCOL_PROCESS_FAILURE;
							rsp.status = KO_PARAMETER_RADIO_CRC_ACTIVATION_VALUE_NOT_SUPPORTED;
							strcpy( rsp.statusDescription, "RADIO_CRC_ACTIVATION" );
						}
					}
					/************************* END writing of staticExpertParams->radioCRCActivation *************************/
					if ((*env)->ExceptionCheck(env)) {
						LOG_ERROR(g_Logger, "writeRadioParams: Java exception occurred 6", 0);
					}

					/************************* START writing of staticExpertParams->secondScanPhysicalConfig *************************/
					if ( ( rsp.status == OK_PROTOCOL_SUCCESS ) && cmd.staticExpertParams.isSet & FLAG_PHYSICAL_CONFIG_ON_SECOND_SCAN ) {
						sprintf( data, "%02X%04X", REQ_SELECT_SECOND_PHYCONFIG, secondScanPhysicalConfigPayload );
						if ( OK_PROTOCOL_SUCCESS != jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, RES_SELECT_SECOND_PHYCONFIG, &rsp, NULL  ) ) {
							//rsp.status = KO_PROTOCOL_PROCESS_FAILURE;
							rsp.status = KO_PARAMETER_SECOND_SCAN_PHYSICAL_CONFIG_VALUE_NOT_SUPPORTED;
							strcpy( rsp.statusDescription, "REQ_SELECT_SECOND_PHYCONFIG" );
						}
					}
					/************************* END writing of staticExpertParams->secondScanPhysicalConfig *************************/
					if ((*env)->ExceptionCheck(env)) {
						LOG_ERROR(g_Logger, "writeRadioParams: Java exception occurred 5", 0);
					}

					/************************* START writing of staticExpertParams->firstScanPhysicalConfig *************************/
					if ( ( rsp.status == OK_PROTOCOL_SUCCESS ) && cmd.staticExpertParams.isSet & FLAG_PHYSICAL_CONFIG_ON_FIRST_SCAN ) {
						sprintf( data, "%02X%04X", REQ_SELECT_PHYCONFIG, firstScanPhysicalConfigPayload );
						if ( OK_PROTOCOL_SUCCESS != jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, RES_SELECT_PHYCONFIG, &rsp, NULL  ) ) {
							// rsp.status = KO_PROTOCOL_PROCESS_FAILURE;
							rsp.status = KO_PARAMETER_FIRST_SCAN_PHYSICAL_CONFIG_VALUE_NOT_SUPPORTED;
							strcpy( rsp.statusDescription, "REQ_SELECT_PHYCONFIG" );
						}
					}
					/************************* END writing of staticExpertParams->firstScanPhysicalConfig *************************/
					if ((*env)->ExceptionCheck(env)) {
						LOG_ERROR(g_Logger, "writeRadioParams: Java exception occurred 4", 0);
					}

					/************************* START writing of serialFrame *************************/
					LOG_DEBUG(g_Logger, "writeRadioParams: Try to write additional Frame "
							"having status : %d and frame[0] = 0x%02x", rsp.status,
							((Payload*)&cmd.command.payload)->frame[0]);
					if ( rsp.status == OK_PROTOCOL_SUCCESS && ((Payload*)&cmd.command.payload)->frame[0] != 0 ) {

						char payloadString[MAX_SERIALFRAME_LEN*2+1] = "";
						WavePortResponses wprCode = ((Payload*)&cmd.command.payload)->frame[0] + 1;
						sprintf( data, "%s", get_hex_payload(payloadString, sizeof(payloadString), (Payload*)&cmd.command.payload) );
						LOG_DEBUG(g_Logger, "writeRadioParams: Write additional SerialFrame : %s", data);

						if ((*env)->ExceptionCheck(env)) {
							LOG_ERROR(g_Logger, "writeRadioParams: Java exception occurred 4.1", 0);
						}

						if ( OK_PROTOCOL_SUCCESS != jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, wprCode, &rsp, NULL  ) )
							rsp.status = KO_PROTOCOL_SETTING_ERROR;

						if ((*env)->ExceptionCheck(env)) {
							LOG_ERROR(g_Logger, "writeRadioParams: Java exception occurred 4.2", 0);
						}

					}
					/************************* END writing of serialFrame *************************/
				}
			}
		}
	}

	if ((*env)->ExceptionCheck(env)) {
		LOG_ERROR(g_Logger, "writeRadioParams: Java exception occurred 3", 0);
	}

	(*env)->DeleteLocalRef( env, classID);


	// do not filter unsupported parameter error messages
	if (rsp.status != OK_PROTOCOL_SUCCESS
		&& rsp.status != KO_PARAMETER_WAKEUP_LENGTH_NOT_VALID
		&& rsp.status != KO_PROTOCOL_NO_VALID_PARAMETER
		&& rsp.status != KO_PARAMETER_WAKEUP_LENGTH_VALUE_NOT_SUPPORTED
		&& rsp.status != KO_PARAMETER_RESPONSE_TIMEOUT_VALUE_NOT_SUPPORTED
		&& rsp.status != KO_PARAMETER_BROADCAST_SESSION_TIMEOUT_VALUE_NOT_SUPPORTED
		&& rsp.status != KO_PARAMETER_BROADCAST_SESSION_MAX_RESPONSES_NB_VALUE_NOT_SUPPORTED
		&& rsp.status != KO_PARAMETER_TRANSMISSION_RADIO_CHANNEL_VALUE_NOT_SUPPORTED
		&& rsp.status != KO_PARAMETER_TRANSMISSION_POWER_LEVEL_VALUE_NOT_SUPPORTED ) {

		rsp.status = KO_PROTOCOL_PROCESS_FAILURE;
	}


	waveProtocol->status = PS_IDLE;

	if ( rsp.status == KO_PROTOCOL_SETTING_ERROR ) {
		LOG_DEBUG(g_Logger, "writeRadioParams: throwing an exception", 0);
		throw_exception( env, &rsp);
	}

	LOG_DEBUG(g_Logger, "writeRadioParams: Create command response for status %d", rsp.status);
	jobject commandResponse = create_command_response( env, g_classSerialFrame, &rsp );

	LOG_DEBUG(g_Logger, "writeRadioParams: Command response created with status %d", get_response_status( env, commandResponse ));
	return commandResponse;
}

jobject JNICALL JNIFUNCTION(readDynamicExpertParams)(JNIEnv *env, jobject jniThis) {
	jobject radioSettingsCommand = NULL;
	RadioSettingsResponse rsp;
	memset( &rsp, 0, sizeof(rsp) );
	ResponseStatus rs = KO_PROTOCOL_NOT_OPEN;
	rsp.response.status = rs;

	ProtocolInfo* waveProtocol = get_protocol_from_java( env, jniThis );
	if ( !waveProtocol ) {
		rsp.response.status = KO_PROTOCOL_NOT_OPEN;
		return create_radio_settings_response( env, rsp.response.status, NULL);
	}

	if ( JNIFUNCTION(isOpen)( env,  jniThis) ==JNI_FALSE ) {
		rsp.response.status = KO_PROTOCOL_NOT_OPEN;
		return create_radio_settings_response( env, rsp.response.status, NULL);
	}

	rsp.response.status = KO_PROTOCOL_ALREADY_IN_USE;
	// 1. A command session is already in progress?
	if ( waveProtocol->status == PS_PROCESS_IN_PROGRESS ) {
		rsp.response.status = KO_PROTOCOL_ALREADY_IN_USE;
		throw_exception_id( env, rsp.response.status);
		return create_radio_settings_response( env, rsp.response.status, NULL);
	}

	// read dynamic expert params in "C" way
	protocol_read_dynamic_expert_params( waveProtocol, &rsp);

	// put protocol status in busy mode
	waveProtocol->status = PS_PROCESS_IN_PROGRESS;
	if ( rsp.response.status == KO_PROTOCOL_NOT_OPEN ) {

		LOG_DEBUG(g_Logger, "readDynamicExpertParams: Failed to read in C way, read using java", 0);

		// fail to read it in "C" way, so then try Java way
		jmethodID methodID = NULL;
		jobject serialDriver = waveProtocol->jniSerialDriver;
		jclass serialDriverClassID = (*env)->GetObjectClass( env, serialDriver);
		if ( serialDriverClassID != NULL ) {
			methodID = (*env)->GetMethodID( env, serialDriverClassID, "send", "(Lcom/coronis/sdk/serialdriver/api/ISerialFrame;)Lcom/coronis/sdk/serialdriver/api/SendingStatus;");
			(*env)->DeleteLocalRef( env, serialDriverClassID);
		}

		if ( methodID != NULL  ) {
			SerialFrame responseFrame;
			char data[2*2+1] = ""; // 2 bytes converted to hex + 1 for ending zero

			sprintf( data, "%02X%02X", REQ_READ_RADIO_PARAM, WAKEUP_LENGTH);
			rs = jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, RES_READ_RADIO_PARAM, &rsp.response, &responseFrame  );
			if ( rs == OK_PROTOCOL_SUCCESS && ( responseFrame.len == 2 ) ) {
				rsp.command.dynamicExpertParams.wakeupLength = responseFrame.frame[1];
				rsp.command.dynamicExpertParams.wakeupLength <<= 8;
				rsp.command.dynamicExpertParams.wakeupLength += responseFrame.frame[0];
				rsp.command.dynamicExpertParams.isSet		 |= FLAG_WAKEUP_LENGTH;
			}

			sprintf( data, "%02X%02X", REQ_READ_RADIO_PARAM, RADIO_USER_TIMEOUT);
			rs = jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, RES_READ_RADIO_PARAM, &rsp.response, &responseFrame  );
			if ( rs == OK_PROTOCOL_SUCCESS ) {
				rsp.command.dynamicExpertParams.responseTimeout = ((unsigned int)responseFrame.frame[0])*100;
				rsp.command.dynamicExpertParams.isSet |= FLAG_RESPONSE_TIMEOUT;
			}

			sprintf( data, "%02X%02X", REQ_READ_RADIO_PARAM, BCST_RECEPTION_TIMEOUT);
			rs = jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, RES_READ_RADIO_PARAM, &rsp.response, &responseFrame  );
			if ( rs == OK_PROTOCOL_SUCCESS ) {
				rsp.command.dynamicExpertParams.broadcastSessionTimeout = ((unsigned int)responseFrame.frame[0])*100;
				rsp.command.dynamicExpertParams.isSet |= FLAG_BROADCAST_SESSION_TIMEOUT;
			}

			sprintf( data, "%02X%02X", REQ_READ_RADIO_PARAM, BCST_MAX_RESPONSES_NB);
			rs = jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, RES_READ_RADIO_PARAM, &rsp.response, &responseFrame  );
			if ( rs == OK_PROTOCOL_SUCCESS ) {
				rsp.command.dynamicExpertParams.broadcastSessionMaxResponsesNb = responseFrame.frame[0];
				rsp.command.dynamicExpertParams.isSet |= FLAG_BROADCAST_SESSION_MAX_RESPONSES_NB;
			}

			sprintf( data, "%02X", REQ_READ_CHANNEL);
			rs = jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, RES_READ_CHANNEL, &rsp.response, &responseFrame  );
			if ( rs == OK_PROTOCOL_SUCCESS ) {
				rsp.command.dynamicExpertParams.transmissionRadioChannel = responseFrame.frame[0];
				rsp.command.dynamicExpertParams.isSet |= FLAG_TRANSMISSION_RADIO_CHANNEL;
				LOG_DEBUG(g_Logger, "readDynamicExpertParams : Channel read %d", responseFrame.frame[0]);
			} else {
				LOG_DEBUG(g_Logger, "readDynamicExpertParams : Failed to read Channel", 0);

			}

			LOG_DEBUG(g_Logger, "readDynamicExpertParams : Reading of TX Power level", 0);
			sprintf( data, "%02X", REQ_READ_TX_POWER);
			rs = jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, RES_READ_TX_POWER, &rsp.response, &responseFrame  );
			if ( rs == OK_PROTOCOL_SUCCESS ) {

				float transmissionPowerLevelRefValue = 0;

				if (payloadToTransmissionPowerLevel (responseFrame.frame[0],
						&transmissionPowerLevelRefValue)) {
					rsp.command.dynamicExpertParams.transmissionPowerLevel = transmissionPowerLevelRefValue;
					rsp.command.dynamicExpertParams.isSet |= FLAG_TRANSMISSION_POWER_LEVEL;
					LOG_DEBUG(g_Logger, "readDynamicExpertParams : TX Power level %d", transmissionPowerLevelRefValue);
				} else {
					LOG_DEBUG(g_Logger, "readDynamicExpertParams : Failed to convert TX Power level %d", responseFrame.frame[0]);
				}
			} else {
				LOG_DEBUG(g_Logger, "readDynamicExpertParams : Failed to read TX Power level ",0);

			}

			rsp.response.status = OK_PROTOCOL_SUCCESS;
		}
	}

	if ( rsp.response.status == OK_PROTOCOL_SUCCESS ) {
		jobject dynamicExpertParams = create_dynamic_expert_params( env, &rsp.command.dynamicExpertParams );
		radioSettingsCommand = create_radio_settings_command( env, g_classPayload, NULL, dynamicExpertParams, NULL, NULL, NULL);
		if ( dynamicExpertParams != NULL )
			(*env)->DeleteLocalRef( env, dynamicExpertParams );
	}

	// put protocol status in idle mode
	waveProtocol->status = PS_IDLE;
	return create_radio_settings_response( env, rsp.response.status, radioSettingsCommand);
}

jobject JNICALL JNIFUNCTION(readStaticExpertParams)(JNIEnv *env, jobject jniThis) {
	jobject radioSettingsCommand = NULL;
	RadioSettingsResponse rsp;
	memset( &rsp, 0, sizeof(rsp) );
	ResponseStatus rs = KO_PROTOCOL_NOT_OPEN;
	rsp.response.status = rs;

	ProtocolInfo* waveProtocol = get_protocol_from_java( env, jniThis );
	if ( !waveProtocol ) {
		rsp.response.status = KO_PROTOCOL_NOT_OPEN;
		return create_radio_settings_response( env, rsp.response.status, NULL);
	}

	if ( JNIFUNCTION(isOpen)( env, jniThis) == JNI_FALSE) {
		rsp.response.status = KO_PROTOCOL_NOT_OPEN;
		return create_radio_settings_response( env, rsp.response.status, NULL);
	}

	// 1. A command session is already in progress?
	if ( waveProtocol->status == PS_PROCESS_IN_PROGRESS ) {

		throw_exception_id( env, KO_PROTOCOL_ALREADY_IN_USE);
		rsp.response.status = KO_PROTOCOL_ALREADY_IN_USE;
		return create_radio_settings_response( env, rsp.response.status, NULL);
	}

	// read dynamic expert params in "C" way
	protocol_read_static_expert_params( waveProtocol, &rsp);

	LOG_DEBUG(g_Logger, "readStaticExpertParams: Ended of read in C way of static expert params, wit ret code %d", rsp.response.status);

	// put protocol status in busy mode
	waveProtocol->status = PS_PROCESS_IN_PROGRESS;
	if ( rsp.response.status == KO_PROTOCOL_NOT_OPEN ) {
		// fail to read it in "C" way, so then try Java way

		jmethodID methodID = NULL;
		jobject serialDriver = waveProtocol->jniSerialDriver;
		jclass serialDriverClassID = (*env)->GetObjectClass( env, serialDriver);
		if ( serialDriverClassID != NULL ) {
			methodID = (*env)->GetMethodID( env, serialDriverClassID, "send", "(Lcom/coronis/sdk/serialdriver/api/ISerialFrame;)Lcom/coronis/sdk/serialdriver/api/SendingStatus;");
			(*env)->DeleteLocalRef( env, serialDriverClassID);
		}

		if ( methodID != NULL  ) {
			SerialFrame responseFrame;
			char data[2*2+1] = ""; // 2 bytes converted to hex + 1 for ending zero

			sprintf( data, "%02X%02X", REQ_READ_RADIO_PARAM, SECOND_SCAN_ENABLE_ACTIVATION);
			rs = jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, RES_READ_RADIO_PARAM, &rsp.response, &responseFrame  );
			if ( rs == OK_PROTOCOL_SUCCESS ) {
				rsp.command.staticExpertParams.secondScanEnableActivation = responseFrame.frame[0];
				rsp.command.staticExpertParams.isSet |= FLAG_SECOND_SCAN_ENABLE_ACTIVATION;
			}

			sprintf( data, "%02X%02X", REQ_READ_RADIO_PARAM, RADIO_CRC_ACTIVATION);
			rs = jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, RES_READ_RADIO_PARAM, &rsp.response, &responseFrame  );
			if ( rs == OK_PROTOCOL_SUCCESS ) {
				rsp.command.staticExpertParams.radioCRCActivation = responseFrame.frame[0];
				rsp.command.staticExpertParams.isSet |= FLAG_RADIO_CRC_ACTIVATION;
			}

			sprintf( data, "%02X%02X", REQ_READ_RADIO_PARAM, AWAKENING_PERIOD);
			rs = jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, RES_READ_RADIO_PARAM, &rsp.response, &responseFrame  );
			if ( rs == OK_PROTOCOL_SUCCESS ) {
				rsp.command.staticExpertParams.secondScanActivityPeriod = ((unsigned int)responseFrame.frame[0])*100;
				rsp.command.staticExpertParams.isSet |= FLAG_ACTIVITY_PERIOD_ON_SECOND_SCAN;
			}

			sprintf( data, "%02X", REQ_READ_CHANNEL);
			rs = jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, RES_READ_CHANNEL, &rsp.response, &responseFrame  );
			if ( rs == OK_PROTOCOL_SUCCESS ) {
				rsp.command.staticExpertParams.firstScanRadioChannel = responseFrame.frame[0];
				rsp.command.staticExpertParams.isSet |= FLAG_RADIO_CHANNEL_ON_FIRST_SCAN;
			}

			sprintf( data, "%02X", REQ_READ_PHYCONFIG);
			rs = jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, RES_READ_PHYCONFIG, &rsp.response, &responseFrame  );
			if ( rs == OK_PROTOCOL_SUCCESS ) {
				unsigned short firstScanPhysicalConfigPayload = responseFrame.frame[0];
				firstScanPhysicalConfigPayload <<= 8;
				firstScanPhysicalConfigPayload += responseFrame.frame[1];

				if (payloadToPhysicalLayer(firstScanPhysicalConfigPayload, &rsp.command.staticExpertParams.firstScanPhysicalConfig))
					rsp.command.staticExpertParams.isSet |= FLAG_PHYSICAL_CONFIG_ON_FIRST_SCAN;
			}

			sprintf( data, "%02X", REQ_READ_SECOND_PHYCONFIG);
			rs = jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, RES_READ_SECOND_PHYCONFIG, &rsp.response, &responseFrame  );
			if ( rs == OK_PROTOCOL_SUCCESS ) {
				unsigned short secondScanPhysicalConfigPayload = responseFrame.frame[0];
				secondScanPhysicalConfigPayload <<= 8;
				secondScanPhysicalConfigPayload += responseFrame.frame[1];

				if (payloadToPhysicalLayer(secondScanPhysicalConfigPayload, &rsp.command.staticExpertParams.secondScanPhysicalConfig))
					rsp.command.staticExpertParams.isSet |= FLAG_PHYSICAL_CONFIG_ON_SECOND_SCAN;
			}

			rsp.command.staticExpertParams.complexScanEnableActivation = _false;
			rsp.command.staticExpertParams.isSet |= FLAG_COMPLEX_SCAN_ENABLE_ACTIVATION;
			rsp.command.staticExpertParams.firstScanSensitivityReductionFactor = 0;
			rsp.command.staticExpertParams.isSet |= FLAG_SENSITIVITY_REDUCTION_FACTOR_ON_FIRST_SCAN;
			rsp.command.staticExpertParams.secondScanRadioChannel = 0;
			rsp.command.staticExpertParams.isSet |= FLAG_RADIO_CHANNEL_ON_SECOND_SCAN;
			rsp.command.staticExpertParams.secondScanSensitivityReductionFactor = 0;
			rsp.command.staticExpertParams.isSet |= FLAG_SENSITIVITY_REDUCTION_FACTOR_ON_SECOND_SCAN;

			// Note: following parameters do not exist on waveport:
			// rsp.command.staticExpertParams.cwupTransmissionActivation;
			// rsp.command.staticExpertParams.cwupAutoSelectionActivation;
			// rsp.command.staticExpertParams.cwupRequestOnlyActivation;
			// rsp.command.staticExpertParams.cwupOnlyActivation;
			// rsp.command.staticExpertParams.nbScanOn;
			// rsp.command.staticExpertParams.nbScanOff;
			// rsp.command.staticExpertParams.frequencyHoppingChannelTable;

			rsp.response.status = OK_PROTOCOL_SUCCESS;
		}
	}

	LOG_DEBUG(g_Logger, "readStaticExpertParams: Finished reading of expert param, returning status code %d", rsp.response.status);

	if ( rsp.response.status == OK_PROTOCOL_SUCCESS ) {
		LOG_DEBUG(g_Logger, "readStaticExpertParams: Start creation of parameteres", 0);
		jobject staticExpertParams = create_static_expert_params( env, &rsp.command.staticExpertParams );
		if (staticExpertParams == NULL) {
			LOG_ERROR(g_Logger, "readStaticExpertParams: Could not create java object staticExpertParams", 0);
		}
		radioSettingsCommand = create_radio_settings_command( env, g_classPayload, NULL, NULL, NULL, staticExpertParams, NULL);
		if ( staticExpertParams != NULL )
			(*env)->DeleteLocalRef( env, staticExpertParams );
	}

	// put protocol status in idle mode
	waveProtocol->status = PS_IDLE;

	LOG_DEBUG(g_Logger, "readStaticExpertParams: Calling create_radio_settings_response with status %d", rsp.response.status);
	return create_radio_settings_response( env, rsp.response.status, radioSettingsCommand);
}

jobject JNICALL JNIFUNCTION(readStaticStandardParams)(JNIEnv *env, jobject jniThis) {
	jobject radioSettingsCommand = NULL;
	RadioSettingsResponse rsp;
	memset( &rsp, 0, sizeof(rsp) );
	ResponseStatus rs = KO_PROTOCOL_NOT_OPEN;
	rsp.response.status = rs;

	ProtocolInfo* waveProtocol = get_protocol_from_java( env, jniThis );
	if (! waveProtocol ) {
		rsp.response.status = KO_PROTOCOL_NOT_OPEN;
		return create_radio_settings_response( env, rsp.response.status, NULL);
	}

	if ( JNIFUNCTION(isOpen)( env,  jniThis) == JNI_FALSE) {
		rsp.response.status = KO_PROTOCOL_NOT_OPEN;
		return create_radio_settings_response( env, rsp.response.status, NULL);
	}

	// 1. A command session is already in progress?
	if ( waveProtocol->status == PS_PROCESS_IN_PROGRESS ) {
		rsp.response.status = KO_PROTOCOL_ALREADY_IN_USE;
		throw_exception_id( env, rsp.response.status);
		return create_radio_settings_response( env, rsp.response.status, NULL);
	}

	// read dynamic expert params in "C" way
	protocol_read_static_standard_params( waveProtocol, &rsp);
	// put protocol status in busy mode
	waveProtocol->status = PS_PROCESS_IN_PROGRESS;
	if ( rsp.response.status == KO_PROTOCOL_NOT_OPEN ) {
		// fail to read it in "C" way, so then try Java way

		jmethodID methodID = NULL;
		jobject serialDriver = waveProtocol->jniSerialDriver;
		jclass serialDriverClassID = (*env)->GetObjectClass( env, serialDriver);
		if ( serialDriverClassID != NULL ) {
			methodID = (*env)->GetMethodID( env, serialDriverClassID, "send", "(Lcom/coronis/sdk/serialdriver/api/ISerialFrame;)Lcom/coronis/sdk/serialdriver/api/SendingStatus;");
			(*env)->DeleteLocalRef( env, serialDriverClassID);
		}

		if ( methodID != NULL  ) {
			SerialFrame responseFrame;
			char data[2*2+1] = ""; // 2 bytes converted to hex + 1 for ending zero

			sprintf( data, "%02X%02X", REQ_READ_RADIO_PARAM, RADIO_ACKNOWLEDGE);
			rs = jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, RES_READ_RADIO_PARAM, &rsp.response, &responseFrame  );
			if ( rs == OK_PROTOCOL_SUCCESS ) {
				rsp.command.staticStandardParams.radioAckActivation = responseFrame.frame[0];
				rsp.command.staticStandardParams.isSet |= FLAG_RADIO_ACK_ACTIVATION;
			}

			sprintf( data, "%02X%02X", REQ_READ_RADIO_PARAM, AWAKENING_PERIOD);
			rs = jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, RES_READ_RADIO_PARAM, &rsp.response, &responseFrame  );
			if ( rs == OK_PROTOCOL_SUCCESS ) {
				rsp.command.staticStandardParams.firstScanActivityPeriod = ((unsigned int)responseFrame.frame[0])*100;
				rsp.command.staticStandardParams.isSet |= FLAG_ACTIVITY_PERIOD_ON_FIRST_SCAN;
			}

			sprintf( data, "%02X%02X", REQ_READ_RADIO_PARAM, SWITCH_MODE_STATUS);
			rs = jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, RES_READ_RADIO_PARAM, &rsp.response, &responseFrame  );
			if ( rs == OK_PROTOCOL_SUCCESS ) {
				rsp.command.staticStandardParams.switchingModeOnRadioAddressActivation = responseFrame.frame[0];
				rsp.command.staticStandardParams.isSet |= FLAG_SWITCHING_MODE_ON_RADIO_ADDRESS_ACTIVATION;
			}

			sprintf( data, "%02X%02X", REQ_READ_RADIO_PARAM, WAVECARD_MULTICAST_GROUP);
			rs = jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, RES_READ_RADIO_PARAM, &rsp.response, &responseFrame  );
			if ( rs == OK_PROTOCOL_SUCCESS ) {
				rsp.command.staticStandardParams.receptionGroupId = responseFrame.frame[0];
				rsp.command.staticStandardParams.isSet |= FLAG_RECEPTION_GROUPID;
			}

			sprintf( data, "%02X%02X", REQ_READ_RADIO_PARAM, RADIO_ACKNOWLEDGE);
			rs = jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, RES_READ_RADIO_PARAM, &rsp.response, &responseFrame  );
			if ( rs == OK_PROTOCOL_SUCCESS ) {
				rsp.command.staticStandardParams.radioAckActivation = responseFrame.frame[0];
				rsp.command.staticStandardParams.isSet |= FLAG_RADIO_ACK_ACTIVATION;
			}

			rsp.command.staticStandardParams.firstScanEnableActivation = _true;
			rsp.command.staticStandardParams.isSet |= FLAG_FIRST_SCAN_ENABLE_ACTIVATION;

			rsp.response.status = OK_PROTOCOL_SUCCESS;
		}
	}

	if ( rsp.response.status == OK_PROTOCOL_SUCCESS ) {
		jobject staticStandardParams = create_static_standard_params( env, &rsp.command.staticStandardParams );
		radioSettingsCommand = create_radio_settings_command( env, g_classPayload, NULL, NULL, staticStandardParams, NULL, NULL);
		if ( staticStandardParams != NULL )
			(*env)->DeleteLocalRef( env, staticStandardParams );
	}

	// put protocol status in idle mode
	waveProtocol->status = PS_IDLE;
	return create_radio_settings_response( env, rsp.response.status, radioSettingsCommand);
}

/**
 * Function used to notify sent/receive spies about sent/received frames.
 * @param waveProtocol - a waveport protocol instance.
 * @param command - a radio exchange frame about witch spies will be notified.
 * @param sendEvent - a flag that manage witch of send or receive spies to be notified.
 * @return error code obtained from frame.
 */
void jni_notify_spies_command( const ProtocolInfo* waveProtocol, jobject command, _bool sendEvent ) {
	JNIEnv* env = get_jni_env();
	LOG_DEBUG(g_Logger, "jni_notify_spies_command: Notify protocol spies start", 0);
	if ( env ) {
		int i = 0;
		for ( ; i < MAX_SPY_NUM; ++i ) {
			if ( waveProtocol->paramSpies[i] ) {
				jclass classID = (*env)->GetObjectClass( env, (jobject)waveProtocol->paramSpies[i]);
				if ( classID != NULL ) {
					const char* methodEventId = sendEvent ? "sentRadioExchangeCommand" : "receivedRadioExchangeCommand";
					jmethodID methodID = (*env)->GetMethodID( env, classID, methodEventId, "(Lcom/coronis/sdk/protocol/api/RadioExchangeCommand;)V");

					if ( methodID != NULL ) {
						LOG_DEBUG(g_Logger, "jni_notify_spies_command: Notify protocol spies call handler", 0);
						(*env)->CallVoidMethod( env, (jobject)waveProtocol->paramSpies[i], methodID, command );
						LOG_DEBUG(g_Logger, "jni_notify_spies_command: Notify protocol spies hanler call end", 0);
					}
					(*env)->DeleteLocalRef( env, classID);
				}
			}
		}
		(*g_jvm)->DetachCurrentThread( g_jvm );
	}
	LOG_DEBUG(g_Logger, "jni_notify_spies_command: Notify protocol spies end", 0);
}

/**
 * Function used to notify sent/receive spies about sent/received frames.
 * @param waveProtocol - a waveport protocol instance.
 * @param frame - a radio frame about witch spies will be notified.
 * @param sendEvent - a flag that manage witch of send or receive spies to be notified.
 * @return error code obtained from frame.
 */
void jni_notify_spies_hex( const ProtocolInfo* waveProtocol, const char* frame, _bool sendEvent ) {
	SerialFrame sf = { {0}, strlen(frame)/2 };
	char *stopstring;
	size_t i = 0;

	for ( i = 0; i < sf.len*2; i+=2 ) {
		char buff[3] = { frame[i], frame[i+1], 0 };
		sf.frame[i/2] = (byte)strtol( buff, &stopstring, 16 );
	}

	RadioExchangeCommand cmd;
	LOG_DEBUG(g_Logger,"jni_notify_spies_hex: Call frame parsing for "
			"len %d and first byte 0x%02X", sf.len, sf.frame[0]);
	parse_frame((ProtocolInfo*)waveProtocol, &sf, &cmd, NULL, NULL, NULL, NULL );
	jni_notify_spies_native_command( waveProtocol, &cmd, sendEvent );
}

/**
 * Function used to notify sent/receive spies about sent/received frames.
 * @param waveProtocol - a waveport protocol instance.
 * @param command - a radio exchange frame about witch spies will be notified.
 * @param sendEvent - a flag that manage witch of send or receive spies to be notified.
 * @return error code obtained from frame.
 */
void jni_notify_spies_native_command( const ProtocolInfo* waveProtocol, const RadioExchangeCommand* command, _bool sendEvent ) {
	JNIEnv* env = get_jni_env();
	if ( env ) {

		WavePortResponses response = -1;
		if (waveProtocol && waveProtocol->receivedFrame.len > 0) {
			response = (WavePortResponses)waveProtocol->receivedFrame.frame[0];
		}

		LOG_DEBUG(g_Logger,"create_radio_exchange_command_object: Called from jni_notify_spies_native_command !", 0);
		jobject reCommand = create_radio_exchange_command_object( env,
				g_classPayload, g_classRadioExchangeCommand,
				(RadioExchangeCommand*)command, NULL, NULL, response);
		if ( reCommand ) {
			jni_notify_spies_command( waveProtocol, reCommand, sendEvent );
			(*env)->DeleteLocalRef( env, reCommand );
		}
		(*g_jvm)->DetachCurrentThread( g_jvm );
	}
}

/**
 * Function used to notify sent/receive spies about sent/received frames.
 * @param waveProtocol - a waveport protocol instance.
 * @param eventCode - a flag that indicate type of event START_LISTENING,STOP_LISTENING,...etc.
 * @return error code obtained from frame.
 */
void jni_notify_event_native( const ProtocolInfo* waveProtocol, int eventCode ) {
	JNIEnv* env = get_jni_env();

	LOG_DEBUG(g_Logger, "jni_notify_event_native: Notify protocol start", 0);
	if ( env && waveProtocol->paramProtocolListener ) {
		jobject protocolEvent = NULL;
		if ( g_classProtocolEvent ) {
			const char* fieldIDName = NULL;
			switch( eventCode ) {
			case PE_START_LISTENING:            fieldIDName = "START_LISTENING";            break;
			case PE_STOP_LISTENING:             fieldIDName = "STOP_LISTENING";             break;
			case PE_SERIALDRIVER_DISCONNECTION: fieldIDName = "SERIALDRIVER_DISCONNECTION"; break;
			case PE_IGNORED_COMMAND_RECEIVED:
			default:					        fieldIDName = "IGNORED_COMMAND_RECEIVED";   break;
			}
			jfieldID id = (*env)->GetStaticFieldID( env, g_classProtocolEvent, fieldIDName, "Lcom/coronis/sdk/protocol/api/ProtocolEvent;");
			if ( id != NULL )
				protocolEvent = (*env)->GetStaticObjectField( env, g_classProtocolEvent, id);
		}
		if ( protocolEvent != NULL ) {
			jclass classID = (*env)->GetObjectClass( env, (jobject)waveProtocol->paramProtocolListener);
			if ( classID != NULL ) {
				jmethodID methodID = (*env)->GetMethodID( env, classID, "receiveEvent", "(Lcom/coronis/sdk/protocol/api/ProtocolEvent;)V");
				if ( methodID != NULL ) {
					LOG_DEBUG(g_Logger, "jni_notify_event_native: Notify protocol call handler", 0);
					(*env)->CallVoidMethod( env, (jobject)waveProtocol->paramProtocolListener, methodID, protocolEvent );
					LOG_DEBUG(g_Logger, "jni_notify_event_native: Notify protocol handler call end", 0);

				}
				(*env)->DeleteLocalRef( env, classID);
			}
			(*env)->DeleteLocalRef( env, protocolEvent);
		}
		(*g_jvm)->DetachCurrentThread( g_jvm );
	}
	LOG_DEBUG(g_Logger, "jni_notify_event_native: Notify protocol end", 0);

}

JNIEnv* get_jni_env() {
	JNIEnv* env = NULL;
	JavaVMAttachArgs args;
	args.version = JNI_VERSION_1_2;
	args.name = "user";
	args.group = NULL;

	if ( g_jvm )
		(*g_jvm)->AttachCurrentThread( g_jvm, (void **)&env, &args);
	return env;
}


jobject sendMessage(JNIEnv *env, jobject jniThis, jobject command, byte commandId, byte responseId ) {
	RadioExchangeCommand reCommand;
	CommandResponse rsp;
	memset( &reCommand, 0, sizeof(reCommand) );
	memset( &rsp, 0, sizeof(rsp) );

	LOG_DEBUG(g_Logger, "JNI sendMessage: Start sending message", 0);
	ProtocolInfo* waveProtocol = get_protocol_from_java( env, jniThis );
	// TODO TCD
//	_bool notWaitedFrameReceived = _false;
	if ( !waveProtocol ) {
		rsp.status = KO_PROTOCOL_NOT_OPEN;
		return create_command_response( env, g_classSerialFrame, &rsp );
	}

	if ( JNIFUNCTION(isOpen)( env,  jniThis) == JNI_FALSE) {
		rsp.status = KO_PROTOCOL_NOT_OPEN;
		return create_command_response( env, g_classSerialFrame, &rsp );
	}

	// 1. A command session is already in progress?
	if ( waveProtocol->status == PS_PROCESS_IN_PROGRESS ) {
		rsp.status = KO_PROTOCOL_ALREADY_IN_USE;
		throw_exception_id( env, rsp.status);
		return create_command_response( env, g_classSerialFrame, &rsp );
	}

	// put protocol status in busy mode
	waveProtocol->status = PS_PROCESS_IN_PROGRESS;
	LOG_DEBUG(g_Logger, "JNI sendMessage: Get parameters", 0);
	jobject jDynamicStandardParams = NULL;
	jobject jDynamicExpertParams   = NULL;
	rsp.status = KO_PROTOCOL_NO_VALID_PARAMETER;
	jclass classID = (*env)->GetObjectClass( env, command);
	if ( classID != NULL ) {

		LOG_DEBUG(g_Logger, "sendMessage: Getting dynamicStandardParams ... ", 0);
		rsp.status = get_dynamic_standard_params( env, classID, command, &reCommand.dynamicStandardParams );
		if ( rsp.status == OK_PROTOCOL_SUCCESS ) {
			LOG_DEBUG(g_Logger, "JNI sendMessage: Got dynamicStandardParams, now getting dynamicExpertParams ... ", 0);
			rsp.status = get_dynamic_expert_params( env, classID, command, &reCommand.dynamicExpertParams );
		}

		if ( rsp.status == OK_PROTOCOL_SUCCESS ) {
			LOG_DEBUG(g_Logger, "JNI sendMessage: Got dynamicExpertParams, getting payload ..", 0);
			rsp.status = KO_PAYLOAD_REQUIRED;
			jfieldID fieldID = (*env)->GetFieldID( env, classID, "payload", "Lcom/coronis/sdk/protocol/api/IPayload;");
			if ( fieldID != NULL ) {
				jobject serialFrame = (*env)->GetObjectField( env, command, fieldID );
				if ( serialFrame != NULL ) {
					if ( get_payload_from_object( env, serialFrame, (Payload*)&reCommand.command.payload ) ){
						rsp.status = OK_PROTOCOL_SUCCESS;
						LOG_DEBUG(g_Logger, "JNI sendMessage: Got payload from Java [+] ", 0);
						protocol_dump_payload((Payload*)&reCommand.command.payload);
					}
					else
						rsp.status = KO_PAYLOAD_NOT_VALID;

					(*env)->DeleteLocalRef( env, serialFrame);
				}
			}
			fieldID = (*env)->GetFieldID( env, classID, "dynamicStandardParams", "Lcom/coronis/sdk/protocol/api/DynamicStandardParams;");
			if ( fieldID != NULL )
				jDynamicStandardParams = (*env)->GetObjectField( env, command, fieldID );
			fieldID = (*env)->GetFieldID( env, classID, "dynamicExpertParams", "Lcom/coronis/sdk/protocol/api/DynamicExpertParams;");
			if ( fieldID != NULL )
				jDynamicExpertParams = (*env)->GetObjectField( env, command, fieldID );
		}
		(*env)->DeleteLocalRef( env, classID);
	}

	//--------------------------------------------------------------------------
	//
	LOG_DEBUG(g_Logger, "sendMessage: Try to send data", 0);
	if ( rsp.status == OK_PROTOCOL_SUCCESS ) {
		// put protocol status in idle mode
		waveProtocol->status = PS_IDLE;
		// read dynamic expert params in "C" way
		LOG_DEBUG(g_Logger, "JNI sendMessage: Try to send data in C", 0);
		protocol_send_application_message( waveProtocol, &reCommand, &rsp);
		// put protocol status in busy mode
		waveProtocol->status = PS_PROCESS_IN_PROGRESS;
		if ( rsp.status == KO_PROTOCOL_NOT_OPEN ) {

			LOG_DEBUG(g_Logger, "JNI sendMessage: Send of data in C, failed, senging using Java loaded driver", 0);
			// fail to send it in "C" way, so then try Java way

			jmethodID methodID = NULL;
			jobject serialDriver = waveProtocol->jniSerialDriver;
			jclass serialDriverClassID = (*env)->GetObjectClass( env, serialDriver);
			if ( serialDriverClassID != NULL ) {
				methodID = (*env)->GetMethodID( env, serialDriverClassID, "send", "(Lcom/coronis/sdk/serialdriver/api/ISerialFrame;)Lcom/coronis/sdk/serialdriver/api/SendingStatus;");
				(*env)->DeleteLocalRef( env, serialDriverClassID);
			}

			if ( methodID != NULL  ) {
				// 3. Payload filled in?
				if ( ((Payload*)&reCommand.command.payload)->len > 0 && ((Payload*)&reCommand.command.payload)->len <= MAX_REQ_SEND_FRAME_BYTES_LEN ) {

					// 3. Node address filled in?
					LOG_DEBUG(g_Logger, "JNI sendMessage: Check node address ....", 0);
					if ( strlen( ((RadioAddress*)&reCommand.dynamicStandardParams.nodeAddress)->addressValue ) == (MAX_RADIOADDRESS_LEN-1) ) {

						if (is_node_address_valid(((RadioAddress*)&reCommand.dynamicStandardParams.nodeAddress)->addressValue)) {

							LOG_DEBUG(g_Logger, "JNI sendMessage: Node address is filled in, use it", 0);
							if ((*env)->ExceptionCheck(env)) {
									LOG_DEBUG(g_Logger, "JNI sendApplicationMessage: Java exception occurred 1", 0);
							}

							jobject cmd = create_radio_settings_command( env, g_classPayload, jDynamicStandardParams, jDynamicExpertParams, NULL, NULL, NULL);

							if ((*env)->ExceptionCheck(env)) {
									LOG_DEBUG(g_Logger, "JNI sendApplicationMessage: Java exception occurred 2", 0);
							}

							if ( cmd != NULL ) {
								// put protocol status in idle mode
								waveProtocol->status = PS_IDLE;
								rsp.status = KO_PARAM_WRITING_FAILURE;
								jobject commandResponse = JNIFUNCTION(writeRadioParams)( env, jniThis, cmd);

								// put protocol status in busy mode
								waveProtocol->status = PS_PROCESS_IN_PROGRESS;
								if ( commandResponse != NULL ) {
									rsp.status = get_response_status( env, commandResponse );
									LOG_DEBUG( g_Logger, "JNI sendApplicationMessage -> writeRadioParams: %d", rsp.status);
									(*env)->DeleteLocalRef( env, commandResponse );
								}

								if (rsp.status ==  KO_PROTOCOL_NO_VALID_PARAMETER) {
									LOG_DEBUG (g_Logger, "JNI sendMessage: There was an error writing parameters, but it is OK when no parameter are specified on port open", 0);
									rsp.status = OK_PROTOCOL_SUCCESS;
								}

								(*env)->DeleteLocalRef( env, cmd );
							}

							LOG_DEBUG( g_Logger, "sendApplicationMessage -> Check status %d", rsp.status);
							if ( rsp.status == OK_PROTOCOL_SUCCESS )  {
								char payloadString[MAX_SERIALFRAME_LEN*2+1] = "";
								char data[MAX_SERIALFRAME_LEN * 2 ] = "";
								snprintf( data, sizeof(data) -1, "%02X%s", commandId, ((RadioAddress*)&reCommand.dynamicStandardParams.nodeAddress)->addressValue );
								snprintf( data + strlen(data), sizeof(data) - 1, "%s",
										get_hex_payload(payloadString, sizeof(payloadString), (Payload*)&reCommand.command.payload) );

								LOG_DEBUG( g_Logger, "JNI sendApplicationMessage -> start jni_send_frame_from_hex", 0);
								jni_send_frame_from_hex( waveProtocol, env, serialDriver, methodID, data, responseId, &rsp, NULL  );
								waveProtocol->receivedFrame.len = 0;
								if ((*env)->ExceptionCheck(env)) {
										LOG_DEBUG(g_Logger, "sendApplicationMessage: Java exception occurred 3", 0);
								}

								if ( rsp.status == OK_PROTOCOL_SUCCESS ) {
									WavePortResponses wprCode = -1;
									DynamicExpertParams* dynamicExpertParams = &waveProtocol->params.dynamicExpertParams;
									if ( reCommand.dynamicExpertParams.isSet & FLAG_WAKEUP_LENGTH )
										dynamicExpertParams = &reCommand.dynamicExpertParams;
									wait_for_frame( waveProtocol, dynamicExpertParams,_true, _false, &rsp, &wprCode);
									if ( rsp.status == OK_PROTOCOL_SUCCESS ) {

										char frameString[MAX_SERIALFRAME_LEN*2+1] = "";

										switch( wprCode ) {
										case END_MESSAGE_EXCHANGE:
											jni_notify_spies_hex( waveProtocol, data, _true );
											rsp.status = OK_PROTOCOL_SUCCESS;
											rsp.statusDescription[0] = 0;
											break;
										case RECEPTION_ERROR:
											rsp.status = parse_reception_error( &waveProtocol->receivedFrame );
											if ( rsp.status == KO_PROTOCOL_UNEXPECTED_FRAME  )
												sprintf( rsp.statusDescription, "unexpected received frame during waiting for waveport END_MESSAGE_EXCHANGE %s",
														get_hex_serialframe(frameString, sizeof(frameString), &waveProtocol->receivedFrame) );
											else
												sprintf( rsp.statusDescription, "waveport RECEPTION_ERROR received %s",
														get_hex_serialframe(frameString, sizeof(frameString), &waveProtocol->receivedFrame) );
											rsp.status = KO_RADIO_TRANSMISSION_FAILURE;
											break;
										default:
											sprintf( rsp.statusDescription, "unexpected received frame during waiting for waveport END_MESSAGE_EXCHANGE %s",
													get_hex_serialframe(frameString, sizeof(frameString), &waveProtocol->receivedFrame) );
											rsp.status = KO_RADIO_TRANSMISSION_FAILURE;
											// TODO launch spontaneous message transmission ?
//											notWaitedFrameReceived = _true;
											break;
										}
									}
								}else {
									// override sending error status code
									rsp.status = KO_PROTOCOL_PROCESS_FAILURE;
								}

							} else {
								sprintf( rsp.statusDescription, "signals param in error %s", get_response_status_name( rsp.status ) );
							}
						} else
							rsp.status = KO_NODE_ADDRESS_NOT_VALID;

					} else {
						LOG_DEBUG(g_Logger, "sendMessage: Node address is NOT filled in, set error code ", 0);
						rsp.status = KO_NODE_ADDRESS_REQUIRED;
					}

				} else // 3. Payload filled in? -> NO
					rsp.status = KO_PAYLOAD_REQUIRED;
			}
		} else {
			LOG_DEBUG(g_Logger, "JNI sendMessage: Data Were sent in C with code %d ", rsp.status );
		}
	} else {
		LOG_DEBUG(g_Logger, "JNI sendMessage: Failed to send data with code %d ", rsp.status );

	}

	if ( jDynamicStandardParams != NULL )
		(*env)->DeleteLocalRef( env, jDynamicStandardParams);
	if ( jDynamicExpertParams != NULL )
		(*env)->DeleteLocalRef( env, jDynamicExpertParams);

	// put protocol status in idle mode
	waveProtocol->status = PS_IDLE;
	// TODO TCD
//	if (notWaitedFrameReceived) {
//		LOG_DEBUG( g_Logger, "JNI sendMessage not waited frame has been received and must be process", 0);
//		receiveInternalFrame(env, g_classSerialFrame);
//	}
	return create_command_response( env, g_classSerialFrame, &rsp );
}

jobject sendRequest(JNIEnv *env, jobject jniThis, jobject command, byte commandId, byte responseId ) {
	jobject radioExchangeCommand = NULL;
	RadioExchangeCommand reCommand;
	RequestResponse rsp;
	WavePortResponses wprCode = -1;

	memset( &reCommand, 0, sizeof(reCommand) );
	memset( &rsp, 0, sizeof(rsp) );
	rsp.response.status = KO_PROTOCOL_NOT_OPEN;

	ProtocolInfo* waveProtocol = get_protocol_from_java( env, jniThis );
	if (! waveProtocol ) {
		return create_request_response( env, g_classSerialFrame, &rsp.response, radioExchangeCommand );
	}

	if ( JNIFUNCTION(isOpen)( env,  jniThis) == JNI_FALSE) {
		return create_request_response( env, g_classSerialFrame, &rsp.response, radioExchangeCommand );
	}

	// 1. A command session is already in progress?
	if ( waveProtocol->status == PS_PROCESS_IN_PROGRESS ) {
		// 1. A command session is already in progress? -> NO
		rsp.response.status = KO_PROTOCOL_ALREADY_IN_USE;
		throw_exception_id( env, rsp.response.status);
		return create_request_response( env, g_classSerialFrame, &rsp.response, radioExchangeCommand );
	}

	// put protocol status in busy mode
	waveProtocol->status = PS_PROCESS_IN_PROGRESS;

	jobject jDynamicStandardParams = NULL;
	jobject jDynamicExpertParams   = NULL;
	rsp.response.status = KO_PROTOCOL_NO_VALID_PARAMETER;
	jclass classID = (*env)->GetObjectClass( env, command);

	if ( classID != NULL ) {
		rsp.response.status = get_dynamic_standard_params( env, classID, command, &reCommand.dynamicStandardParams );
		LOG_DEBUG(g_Logger, "sendRequest: read of dynamic standard parameters finished with status %d", rsp.response.status);
		if ( rsp.response.status == OK_PROTOCOL_SUCCESS )
			rsp.response.status = get_dynamic_expert_params( env, classID, command, &reCommand.dynamicExpertParams );
		if ( rsp.response.status == OK_PROTOCOL_SUCCESS ) {
			rsp.response.status = KO_PAYLOAD_REQUIRED;
			jfieldID fieldID = (*env)->GetFieldID( env, classID, "payload", "Lcom/coronis/sdk/protocol/api/IPayload;");
			if ( fieldID != NULL ) {
				jobject serialFrame = (*env)->GetObjectField( env, command, fieldID );
				if ( serialFrame != NULL ) {
					if ( get_payload_from_object( env, serialFrame, (Payload*)&reCommand.command.payload ) )
						rsp.response.status = OK_PROTOCOL_SUCCESS;
					(*env)->DeleteLocalRef( env, serialFrame);
				}
			}
			fieldID = (*env)->GetFieldID( env, classID, "dynamicStandardParams", "Lcom/coronis/sdk/protocol/api/DynamicStandardParams;");
			if ( fieldID != NULL )
				jDynamicStandardParams = (*env)->GetObjectField( env, command, fieldID );
			fieldID = (*env)->GetFieldID( env, classID, "dynamicExpertParams", "Lcom/coronis/sdk/protocol/api/DynamicExpertParams;");
			if ( fieldID != NULL )
				jDynamicExpertParams = (*env)->GetObjectField( env, command, fieldID );
		}
		(*env)->DeleteLocalRef( env, classID);
	} else {
		LOG_ERROR(g_Logger, "sendRequest: Dynamic standard parameters are note specified", 0);
	}

	LOG_DEBUG(g_Logger, "sendRequest: Done with params, try to send request ", 0);
	if ( rsp.response.status == OK_PROTOCOL_SUCCESS ) {

		// put protocol status in idle mode
		waveProtocol->status = PS_IDLE;
		// send application request in "C" way
		LOG_DEBUG( g_Logger, "sendRequest JNI: Sending request ...", 0);

		protocol_send_application_request( waveProtocol, &reCommand, &rsp);
		// put protocol status in busy mode
		waveProtocol->status = PS_PROCESS_IN_PROGRESS;
		if ( rsp.response.status == KO_PROTOCOL_NOT_OPEN ) {
			// fail to read it in "C" way, so then try Java way
			LOG_DEBUG( g_Logger, "sendRequest JNI: C layer failed, try using Java", 0);

			jmethodID methodID = NULL;
			jobject serialDriver = waveProtocol->jniSerialDriver;
			jclass serialDriverClassID = (*env)->GetObjectClass( env, serialDriver);
			if ( serialDriverClassID != NULL ) {
				methodID = (*env)->GetMethodID( env, serialDriverClassID, "send", "(Lcom/coronis/sdk/serialdriver/api/ISerialFrame;)Lcom/coronis/sdk/serialdriver/api/SendingStatus;");
				(*env)->DeleteLocalRef( env, serialDriverClassID);
			}

			if ( methodID != NULL  ) {
				// 3. Payload filled in?
				if ( ((Payload*)&reCommand.command.payload)->len > 0 && ((Payload*)&reCommand.command.payload)->len <= MAX_REQ_SEND_FRAME_BYTES_LEN ) {

					// 3. Node address filled in?
					if ( strlen( ((RadioAddress*)&reCommand.dynamicStandardParams.nodeAddress)->addressValue ) == (MAX_RADIOADDRESS_LEN-1) ) {

						if (is_node_address_valid(((RadioAddress*)&reCommand.dynamicStandardParams.nodeAddress)->addressValue)) {

							jobject cmd = create_radio_settings_command( env, g_classPayload, jDynamicStandardParams, jDynamicExpertParams, NULL, NULL, NULL);
							if ( cmd != NULL ) {

								// put protocol status in idle mode
								waveProtocol->status = PS_IDLE;
								jobject commandResponse = JNIFUNCTION(writeRadioParams)( env, jniThis, cmd);
								// put protocol status in busy mode
								waveProtocol->status = PS_PROCESS_IN_PROGRESS;
								if ( commandResponse != NULL ) {
									rsp.response.status = get_response_status( env, commandResponse );
									LOG_DEBUG( g_Logger, "sendApplicationRequest -> writeRadioParams: %d", rsp.response.status);
									(*env)->DeleteLocalRef( env, commandResponse );
								}
								(*env)->DeleteLocalRef( env, cmd );
							}

							// After writeRadioParams
							if ( rsp.response.status == OK_PROTOCOL_SUCCESS )  {
								char payloadString[MAX_SERIALFRAME_LEN*2+1] = "";
								char data[MAX_SERIALFRAME_LEN] = "";
								snprintf( data, sizeof(data) -1, "%02X%s", commandId, ((RadioAddress*)&reCommand.dynamicStandardParams.nodeAddress)->addressValue );
								snprintf( data+strlen(data),sizeof(data) -1, "%s",
										get_hex_payload(payloadString, sizeof(payloadString), (Payload*)&reCommand.command.payload) );
								jni_send_payload_from_hex( waveProtocol, env, serialDriver,
										methodID, data, responseId, &rsp.response,
										(Payload*)&rsp.command.command.payload );
								waveProtocol->receivedFrame.len = 0;

								if ( rsp.response.status == OK_PROTOCOL_SUCCESS ) {

									DynamicExpertParams* dynamicExpertParams = &waveProtocol->params.dynamicExpertParams;
									if ( reCommand.dynamicExpertParams.isSet & FLAG_WAKEUP_LENGTH )
										dynamicExpertParams = &reCommand.dynamicExpertParams;

									LOG_DEBUG( g_Logger, "sendRequest JNI: Start frame waiting ...", 0);
									wait_for_frame( waveProtocol, dynamicExpertParams,_true, _false, &rsp.response, &wprCode);
									LOG_DEBUG( g_Logger, "sendRequest JNI: Frame waiting ended with response status ", rsp.response.status);

									if ( rsp.response.status == OK_PROTOCOL_SUCCESS ) {

										jni_notify_spies_hex( waveProtocol, data, _true );
										rsp.response.statusDescription[0] = 0;
										LOG_DEBUG( g_Logger, "sendRequest JNI: parsing frame ...", 0);
										rsp.response.status = parse_frame(waveProtocol, &waveProtocol->receivedFrame, &rsp.command, NULL, NULL, NULL, NULL );

										if (rsp.response.status == (ResponseStatus)OK_PROTOCOL_SUCCESS) {

											LOG_DEBUG( g_Logger, "sendRequest JNI: Good frame received", 0);

											//memcpy( &rsp.command.command.payload,
											//		&waveProtocol->receivedFrame, sizeof(rsp.command.command.payload) );

											//((Payload *)&rsp.command.command.payload)->len = waveProtocol->receivedFrame.len;

											memcpy( &rsp.response.serialFrame,
													&waveProtocol->receivedFrame, sizeof(rsp.response.serialFrame) );

											((SerialFrame *)&rsp.response.serialFrame)->len = waveProtocol->receivedFrame.len;
										} else {

											LOG_ERROR( g_Logger, "sendRequest JNI: Error frame received!", 0);

											((Payload *)&rsp.command.command.payload)->len = 0;
											((SerialFrame *)&rsp.response.serialFrame)->len = 0;
										}

										char frameString[MAX_SERIALFRAME_LEN*2+1] = "";
										if ( rsp.response.status == KO_PROTOCOL_UNEXPECTED_FRAME ) {
											rsp.response.status = KO_RADIO_TRANSMISSION_FAILURE;
											sprintf( rsp.response.statusDescription, "unexpected received frame during waiting for waveport END_MESSAGE_EXCHANGE %s",
													get_hex_serialframe(frameString, sizeof(frameString), &waveProtocol->receivedFrame) );
										} else if ( rsp.response.status != OK_PROTOCOL_SUCCESS )
											sprintf( rsp.response.statusDescription, "waveport RECEPTION_ERROR received %s",
													get_hex_serialframe(frameString, sizeof(frameString), &waveProtocol->receivedFrame) );
									}


								} else {
									// description is propagated from jni_send_frame_from_hex
									rsp.response.status = KO_PROTOCOL_PROCESS_FAILURE;
								}
							} else {
								sprintf( rsp.response.statusDescription, "signals param in error %s", get_response_status_name( rsp.response.status ) );
							}
						}else
							rsp.response.status = KO_NODE_ADDRESS_NOT_VALID;
					} else
						rsp.response.status = KO_NODE_ADDRESS_REQUIRED;

				} else // 3. Payload filled in? -> NO
					rsp.response.status = KO_PAYLOAD_REQUIRED;
			}
		}

		LOG_DEBUG(g_Logger,"create_radio_exchange_command_object: Called from sendRequest", 0);


		WavePortResponses response = -1;
		if (((SerialFrame *)&rsp.response.serialFrame)->len > 0) {
			response = (WavePortResponses)((SerialFrame *)&rsp.response.serialFrame)->frame[0];
		}

		radioExchangeCommand = create_radio_exchange_command_object( env,
				g_classPayload, g_classRadioExchangeCommand, &rsp.command,
				jDynamicStandardParams, jDynamicExpertParams, response);
		if ( rsp.response.status == OK_PROTOCOL_SUCCESS ) {
			if ( radioExchangeCommand )
				jni_notify_spies_command( waveProtocol, radioExchangeCommand, _false );
		}
	} else {
		LOG_ERROR(g_Logger,"create_radio_exchange_command_object: Failed to get params", 0);
	}

	if ( jDynamicStandardParams != NULL )
		(*env)->DeleteLocalRef( env, jDynamicStandardParams);
	if ( jDynamicExpertParams != NULL )
		(*env)->DeleteLocalRef( env, jDynamicExpertParams);

	// put protocol status in idle mode
	waveProtocol->status = PS_IDLE;
	LOG_DEBUG(g_Logger,"create_request_response: Called from sendRequest", 0);
	return create_request_response( env, g_classSerialFrame, &rsp.response, radioExchangeCommand );
}

/* caching the class references, or method and field IDs
 * definition of this function*/
void initialize_global_class_references( JavaVM *jvm, jint jniVersion ) {
	JNIEnv *env;
	if ( !(*jvm)->GetEnv(jvm, (void **)&env, jniVersion) ) {
		jclass classID = (*env)->FindClass( env, "com/coronis/sdk/protocol/api/ProtocolEvent");
		if ( classID )
			g_classProtocolEvent = (*env)->NewWeakGlobalRef(env, classID);
		classID = (*env)->FindClass( env, "com/coronis/sdk/protocol/api/RadioExchangeCommand");
		if ( classID )
			g_classRadioExchangeCommand = (*env)->NewWeakGlobalRef(env, classID);
		classID = (*env)->FindClass( env, "com/coronis/sdk/domain/SerialFrame");
		if ( classID )
			g_classSerialFrame = (*env)->NewWeakGlobalRef(env, classID);
		classID = (*env)->FindClass( env, "com/coronis/sdk/domain/Payload");
		if ( classID )
			g_classPayload = (*env)->NewWeakGlobalRef(env, classID);
	}
}

/* uncaching the class references, or method and field IDs
* definition of this function*/
void uninitialize_global_class_references( JavaVM *jvm, jint jniVersion ) {
	JNIEnv *env;
	if ( !(*jvm)->GetEnv(jvm, (void **)&env, jniVersion) ) {
		if ( g_classProtocolEvent )
			(*env)->DeleteWeakGlobalRef(env, g_classProtocolEvent);
		g_classProtocolEvent = NULL;
		if ( g_classRadioExchangeCommand )
			(*env)->DeleteWeakGlobalRef(env, g_classRadioExchangeCommand);
		g_classRadioExchangeCommand = NULL;
		if ( g_classSerialFrame )
			(*env)->DeleteWeakGlobalRef(env, g_classSerialFrame);
		g_classSerialFrame = NULL;
		if ( g_classPayload )
			(*env)->DeleteWeakGlobalRef(env, g_classPayload);
		g_classPayload = NULL;
	}
}
#endif //_JAVA_WRAPPER
