/*******************************************************************************
 * Copyright (c) 2012 Coronis
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Contributors:
 *     Coronis - initial API and implementation
 *******************************************************************************/

/** @file i_serial_driver_service.h
 * @brief This interface represents a Coronis SDK SerialDriver
 * It generalizes and standardizes the available services
 * to exchanged serial frame,
 * throw a serial port connection,
 * via a low level serial protocol
 */
#ifndef _I_SERIAL_DRIVER_SERVICE_H_
#define _I_SERIAL_DRIVER_SERVICE_H_

#include <i_serial_frame.h>
#include <i_serial_driver_listener_types.h>

#ifdef  _WINDOWS
#define CORONIS_API __declspec(dllexport)
#define CORONIS_CALL __cdecl
#else
#define CORONIS_API extern
#define CORONIS_CALL
#endif//_WINDOWS

/**
 * The API version of the serial driver
 *
 */
#define SERIAL_DRIVER_API_VERSION  0x02

/**
 * 	Macros used to extract API, MAX and MIN version from protocol version
 * 	encoded value
 */
#define GET_SERIAL_DRIVER_MIN_VERSION(_X_) ((_X_ >> 0) & 0x0000FF)
#define GET_SERIAL_DRIVER_MAX_VERSION(_X_) ((_X_ >> 8) & 0x0000FF)
#define GET_SERIAL_DRIVER_API_VERSION(_X_) ((_X_ >>16) & 0x0000FF)

/**
 * Get driver version
 * @return The driver version MAX.MID.MIN encoded in a 8 bytes integer (2 bytes
 *  per number)
 */
CORONIS_API int CORONIS_CALL driver_get_version();

/**
 * Create a new serial driver instance
 * @return a pointer to created serial driver.<code>NULL</code> if driver failed to be created.
 */
CORONIS_API void* CORONIS_CALL driver_create();

/**
 * Destroy the serial driver instance. It will automatically close the current connection
 * @param driver The SerialDriver instance
 * @return <code>0</code> if succeed, a positive number otherwise
 */
CORONIS_API int  CORONIS_CALL driver_destroy( const void* driver );

/**
 * Initialize the serial driver settings if not already configured (serialPortId, baudRate, etc.)
 * @param driver The SerialDriver instance
 * @return the init status (serial_driver_status)
 *         OK_SERIALDRIVER_SUCCESS if initialization succeed
 *         KO_SERIALDRIVER_ALREADY_OPEN if SerialDriver is already open
 */
CORONIS_API int  CORONIS_CALL driver_init( const void* driver );

/**
 * Initialize the serial driver settings if not already configured (serialPortId, baudRate, etc.)
 * @param driver The SerialDriver instance
 * @param file The serial driver property file URL
 * @return the init status (serial_driver_status)
 *         OK_SERIALDRIVER_SUCCESS if initialization succeed
 *         KO_SERIALDRIVER_ALREADY_OPEN if SerialDriver is already open
 */
CORONIS_API int CORONIS_CALL driver_init_from_file( const void* driver, const char* file );

/**
 * Open the SerialDriver connection with the serial port id,
 * and configure the serial port parameters
 * @param driver The SerialDriver instance to open
 * @return the open status (serial_driver_status)
 *         OK_PROTOCOL_SUCCESS if opening succeed
 *         KO_SERIALPORTID_NOT_EXIST if id does not exist from the OS point of view
 *         KO_SERIALPORT_NOT_AVAILABLE if port is not available (already in use, broken down, etc.)
 *         KO_PARAM_XXXX_NOT_VALID if parameters configuration fails
 */
CORONIS_API int CORONIS_CALL driver_open( const void* driver );

/**
 * Close the SerialDriver connection,
 * @param driver The SerialDriver instance to close
 * @return the close status (serial_driver_status)
 *         OK_SERIALDRIVER_SUCCESS if closing succeed
 *         KO_SERIALPORT_UNEXPECTED_ERROR if error occurs during closing process
 */
CORONIS_API int CORONIS_CALL driver_close( const void* driver );

/**
 * Test if the SerialDriver connection is open are not
 * @param driver The SerialDriver instance
 * @return <code>1</code> if the serial driver is connected, <code>0</code> otherwise
 */
CORONIS_API int CORONIS_CALL driver_is_open( const void* driver );

/**
 * Get the serial port ID, if previously setted
 * @param driver The SerialDriver instance
 * @param Setted literal value of serial port ID (dependent from OS)
 * @param length of the setted literal value, <code>0</code> if not setted
 */
CORONIS_API int CORONIS_CALL driver_get_port_id( const void* driver, char* id, int idLen );

/**
 * Set the serial port ID to which SerialDriver must be connected
 * @param driver The SerialDriver instance
 * @param id The serial port identifier to set
 * @return the set status (serial_driver_status)
 *         OK_SERIALDRIVER_SUCCESS if setting succeed
 *         KO_SERIALDRIVER_ALREADY_OPEN if driver is already open
 *         KO_SERIALPORTID_NOT_EXIST if id is not valid for the OS
 */
CORONIS_API int CORONIS_CALL driver_set_port_id( const void* driver, const char* id );

/**
 * Set the standard serial port parameters
 * Valid serial port parameters are listed in serial_port_constants.h
 * @param driver The SerialDriver instance
 * @param baudRate The baud rate
 * @param dataBits The number of data bits
 * @param stopBits The number of stop bits
 * @param parity The type of parity
 * @return the set status (serial_driver_status)
 *         OK_SERIALDRIVER_SUCCESS if setting succeed
 *         KO_SERIALDRIVER_ALREADY_OPEN if driver is already open
 *         KO_PARAM_xxx_NOT_VALID if param xxx is not valid
 */
CORONIS_API int CORONIS_CALL driver_set_port_params( const void* driver, int baudRate, int dataBit, int stopBit, int parity );

/**
 * Set the listening period parameter of the serial port listener
 * @param driver The SerialDriver instance
 * @param receiveTimeout Listening period to set (between 1 ms and 100 ms)
 * @return the set status (serial_driver_status)
 *         OK_SERIALDRIVER_SUCCESS if setting succeed
 *         KO_SERIALDRIVER_ALREADY_OPEN if driver is already open
 *         KO_PARAM_LISTENINGPERIOD_NOT_VALID if period is not valid for the OS
 */
CORONIS_API int CORONIS_CALL driver_set_listening_period_param( const void* driver, int receiveTimeout );

/**
 * Set the flowcontrol mode parameter of the serial port
 * Valid serial port parameters are listed in serial_port_constants.h
 * @param driver The SerialDriver instance
 * @param flowControlIn Type of flow control for receiving (SerialPort constants)
 * @param flowControlOut Type of flow control for sending (SerialPort constants)
 * @return the set status (serial_driver_status)
 *         OK_SERIALDRIVER_SUCCESS if setting succeed
 *         KO_SERIALDRIVER_ALREADY_OPEN if driver is already open
 *         KO_PARAM_xxx_NOT_VALID if param xxx is not valid
 */
CORONIS_API int CORONIS_CALL driver_set_flow_control_mode_params( const void* driver, int flowControlIn, int flowControlOut );

/**
 * Send a serial frame via the SerialDriver
 * @param driver The SerialDriver instance
 * @param frame The ISerialFrame to send
 * @return sending_status of the send process
 *         OK_SERIALDRIVER_SUCCESS if sending succeed
 *         KO_SERIALFRAME_REQUIRED if frame is <code>empty</code>
 *         KO_SERIALDRIVER_NOT_OPEN if SerialDriver is not opened
 *         KO_SENDING_FAILURE if sending failed
 *         KO_SERIALDRIVER_ALREADY_IN_USE if SerialDriver is already in process
 *         KO_SERIALPORT_DISCONNECTED if SerialDriver disconnection is observed during the process
 */
CORONIS_API int CORONIS_CALL driver_send( const void* driver, const pSerialExchangeData frame);

/**
 * Register a listener to be notify of spontaneous received frame and, a listener
 * to be notify of SerialDriver event
 * The receiveFrameListener listener have to implement the serial_driver_listener_t
 * interface, and eventListener have to implement the serial_driver_listener_notify_t interface.
 * SerialDriver instance must allow only one listener.
 * @param driver The SerialDriver instance
 * @param receiveFrameListener The frame listener to register
 * @param eventListener The event listener to register
 * @param param A user define data which is passed back to listener when a frame is received
 * @return the registration status (serial_driver_status)
 * 		   OK_SERIALDRIVER_SUCCESS if registration succeed
 *         KO_PARAM_NOT_VALID - if driver descriptor is invalid
 */
CORONIS_API int CORONIS_CALL driver_add_listener( const void* driver,
		serial_driver_listener_t receiveFrameListener,
		serial_driver_listener_notify_t notifyListener,
		void* param );

/**
 * Remove all listeners from being notify of spontaneous received frame or events.
 * @param driver The SerialDriver instance
 * @return the operation status (serial_driver_status)
 * 		   OK_SERIALDRIVER_SUCCESS if unregistration succeed
 *         KO_PARAM_NOT_VALID - if driver descriptor is invalid
 */
CORONIS_API int CORONIS_CALL driver_remove_listener( const void* driver );

/**
 * Register a spy listener to be notify of sent and received frame via the SerialDriver.
 * The listeners have to implement the serial_driver_spy_t interface.
 * SerialDriver instance can allow many listeners.
 * @param driver The SerialDriver instance
 * @param spySentFrameListener The listener to by notify of sent frame
 * @param spyRecvFrameListener The listener to by notify of received frame
 * @param param A user define data which is passed back to listener when a frame is exchanged
 * @param spyIndex The index of the Spy listener
 * @return the registration status (serial_driver_status)
 * 		   OK_PROTOCOL_SUCCESS if registration succeed
 *         KO_SERIALDRIVER_LISTENER_ALREADY_REGISTERED if the spyIndex is already registered
 *         KO_SERIALDRIVER_MAX_SPY_LISTENER_EXCEED if the maximum number of spy listener
 */
CORONIS_API int CORONIS_CALL driver_add_spy_listener( const void* driver, serial_driver_spy_t spySentFrameListener, serial_driver_spy_t spyRecvFrameListener, void* param, int* spyIndex );

/**
 * Unregister a spy listener from being notify of sent and received frame via the SerialDriver.
 * @param driver The SerialDriver instance
 * @param spyIndex The index of the Spy listener
 * @return the unregistration status (serial_driver_status)
 * 		   OK_SERIALDRIVER_SUCCESS if unregistration succeed
 *         KO_SERIALDRIVER_LISTENER_NOT_REGISTERED if the spyIndex is not registered
 */
CORONIS_API int CORONIS_CALL driver_remove_spy_listener( const void* driver, int spyIndex );


/**
 * Get the serial driver instance into a single structure that contains internal driver state and
 * pointers to all serial driver functions.
 *
 * @param driver The SerialDriver handler
 * @return the serial driver instance pointer or 0 if failed
 *
 * @note returned pointer points to a static internal structure, that is going to be the
 * basically the same during entire life time cycle of the serial driver.
 */
CORONIS_API void* CORONIS_CALL  driver_get_serial_driver_instance(const void* driver );

#endif//_I_SERIAL_DRIVER_SERVICE_H_
