/*******************************************************************************
 * Copyright (c) 2012 Coronis
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Contributors:
 *     Coronis - initial API and implementation
 *******************************************************************************/

/**
 * @file protocol_types_utils.h
 * @brief Declaration of protocol types utility methods to manipulate
 * Coronis implementation of Protocol types
 */
#ifndef PROTOCOL_TYPES_UTILS_H_
#define PROTOCOL_TYPES_UTILS_H_

#include "protocol_types.h"
#include <i_protocol_types.h>
#include <protocol_command_types.h>
#include <protocol_constants.h>

/**
 * Function determines if passed node address has a valid length and format.
 *
 * @param node_address - a hex string that represents a node address
 * @return _true for a valid node address
 */
_bool is_node_address_valid(const char *node_address);

/**
 * 	Initialize LocalNodeServiceCommand structure and set the service ID
 * @param serviceId - the service Id to set after structure initialization
 * @return - the initialized structure.
 */
LocalNodeServiceCommand create_local_node_service_command(NodeServiceId serviceId, Payload payload);

/**
 *  Initialize LocalNodeServiceCommand structure and set the service ID
 * @param serviceId - the service Id to set after structure initialization
 * @param payload_string - hex string value of payload to copy in initialized structure
 * @return - the initialized structure.
 */
LocalNodeServiceCommand create_local_node_service_command_from_string(NodeServiceId serviceId, const char* payload_string);

/**
 *	Initialize a RadioExchangeCommand structure. Setup payload and node address.
 * @param payload - payload to copy in initialized structure
 * @param destAddr - destination address to be set in this structure
 * @return an initialized structure
 */
RadioExchangeCommand create_radio_exchange_command(Payload payload, RadioAddress destAddr);

/**
 *	Initialize a RadioExchangeCommand structure from Hex String value of payload and node address.
 * @param payload_string - hex string value of payload to copy in initialized structure
 * @param destinationAddress - destination address to be set in this structure
 * @return an initialized structure
 */
RadioExchangeCommand create_radio_exchange_command_from_string(const char* payload_string, RadioAddress destinationAddress);

/**
 *	Initialize a RadioExchangeCommand structure. Setup payload, node address and relay path
 * @param payload - payload to copy in initialized structure
 * @param destAddr - destination address to be set in this structure
 * @param relayPath - the relay path to destination node
 * @return an initialized structure
 */
RadioExchangeCommand create_radio_exchange_command_with_relay(Payload payload, RadioAddress destAddr, RadioAddress* relayPath);

/**
 *	Initialize a RadioExchangeCommand structure from Hex String value of payload, node address and relay path
 * @param payload_string - hex string value of payload to copy in initialized structure
 * @param destAddr - destination address to be set in this structure
 * @param relayPath - the relay path to destination node
 * @return an initialized structure
 */
RadioExchangeCommand create_radio_exchange_command_with_relay_from_string(const char* payload_string, RadioAddress destAddr, RadioAddress* relayPath);

/**
 *	Copy payload data from a concrete Payload implementation into a generic Payload data type.
 * @param dest - generic payloada data type where to copy payload
 * @param src - the concrete source payload implementation
 */
void set_payload_value(pIPayload dest, const Payload *src);

/**
 * Function used to obtain hex string of payload
 * @param payload - pointer to payload.
 * @return hex string of  given payload.
 */
const char* get_hex_payload(char* payloadString, const int payloadStringSize, const Payload* pl );

/**
 * Function used to obtain string name of ResponceStatus enumeration value.
 * @param id - ResponceStatus enumeration value
 * @return string name of id value or empty string if given value not in ResponceStatus enumeration range.
 */
const char* get_response_status_name( ResponseStatus rs );

/**
 * Function used to obtain string name of NodeServiceId enumeration value.
 * @param nsid - NodeServiceId enumeration value
 * @return string name of id value or empty string if given value not in NodeServiceId enumeration range.
 */
const char* get_local_node_service_name( NodeServiceId nsid );

#endif /* PROTOCOL_TYPES_UTILS_H_ */
