/*******************************************************************************
 * Copyright (c) 2012 Coronis
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Contributors:
 *     Coronis - initial API and implementation
 *******************************************************************************/

/** @file main.c
 * @brief Main Class of Waveport protocol example
 * This example will demonstrate sequentially how to
 * - Read the radio address of connected radio modem
 * - Send an application request to a distant device
 * - Send an application message to a distant device (with 1 relay)
 */

// serial driver api include
#include <serial_driver_event.h>
#include <serial_driver_status.h>
// protocol api include
#include <protocol_command_types.h>
#include <protocol_response_types.h>
// domain include
#include "loader_serial_driver.h"
#include "loader_protocol.h"
#include "serial_driver_types.h"
#include "protocol_types.h"
#include "protocol_types_utils.h"

#include "utils.h"

SerialDriverLoader driverHandler;
ProtocolLoader protocolHandler;
void*              rs232DriverInstance;
void*              waveportInstance;

#ifdef _WINDOWS
char portId[256] = "COM4";
#else
char portId[256] = "/dev/ttyUSB0";
#endif //_WINDOWS

char driverId[256] = "csrs232driver";
char protocolId[256] = "cswaveportprotocol";
RadioAddress destinationAddress = (RadioAddress){"10194C300559"};
RadioAddress relayPath[3] = {(RadioAddress){"10194C300559"}, (RadioAddress){""}, (RadioAddress){""}};
Payload payloadGetType = (Payload){ {0x20}, 1 };
Payload payloadGetTemp = (Payload){ {0x01}, 1 }; 		// no precision for Dallas
Payload payloadGetFirm = (Payload){ {0x28}, 1 };
Payload payloadGetDate = (Payload){ {0x12}, 1 };
Payload payloadInitSensor = (Payload){ {0x04}, 1 };		//Wavetherm
Payload payloadStackService_READ_RADIOADDRESS = (Payload){ {0x50, 0x05}, 2 };
Payload payloadStackService_READ_PHYCONFIG = (Payload){ {0x66}, 1 };


void process_local_node_service();
void process_send_request();
void process_send_message();

const char* get_local_node_service_name( NodeServiceId nsid );
const char* GetCurrentTimeString();

int main(void) {
	printf( "%s cswaveportprotocol example BEGIN. \n", GetCurrentTimeString() );
    if ( OK == load_serial_driver( driverId, &driverHandler) &&
    		OK == load_protocol( protocolId, &protocolHandler))
    {
    	// 1. SerialDriver instantiation
        rs232DriverInstance = driverHandler.create();

        // 2. Protocol instantiation
        waveportInstance = protocolHandler.create();
        if ( rs232DriverInstance && waveportInstance)
        {
            printf( "%s csrs232driver & cswaveportprotocol creation success. \n",
                GetCurrentTimeString() );

            // 3. Set port Id to use via SerialDriver level
            driverHandler.set_port_id( rs232DriverInstance, portId);
            int status = OK_SERIALDRIVER_SUCCESS;

            // 4. Protocol initialization (use the define serialDriver)
            protocolHandler.set_serial_driver_instance(waveportInstance, driverHandler.get_serial_driver_instance(rs232DriverInstance));
            //protocolHandler.set_serial_driver_id(waveportInstance, driverId);
            //char driverId[256] = "";
            //printf( "getSerialDriverId before init: %s \n", protocolHandler.get_serial_driver_id(waveportInstance, driverId, 256));
            status = protocolHandler.init( waveportInstance );
            //printf( "getSerialDriverId after init: %s \n", protocolHandler.get_serial_driver_id(waveportInstance, driverId, 256));
            printf( "%s cswaveportprotocol init status: %i \n",
                GetCurrentTimeString(), status );

            // 5. Open the protocol connection
            status = protocolHandler.open( waveportInstance );
            printf( "%s cswaveportprotocol open status: %i \n",
                GetCurrentTimeString(), status );

            // 6. Process 3 different protocol services
//            process_local_node_service();
            process_send_request();
//            process_send_message();

            // 7. Close the protocol connection
            protocolHandler.close( waveportInstance );

            // 8. Protocol and Serial driver instances suppression
			protocolHandler.destroy( waveportInstance );
			driverHandler.destroy( rs232DriverInstance );
        }

        free_protocol( &protocolHandler );
        free_serial_driver( &driverHandler);
    } else {
    	printf( "%s cswaveportprotocol loading failure. \n", GetCurrentTimeString() );
    }
    printf( "%s cswaveportprotocol example END. \n", GetCurrentTimeString() );
    return 0;
}

void process_local_node_service() {
	printf( "%s process local node service: start \n",
						GetCurrentTimeString());
	// Prepare the request to send (throw 1 relay)
	LocalNodeServiceCommand command = create_local_node_service_command(STACK_SERVICE, payloadStackService_READ_RADIOADDRESS);
	LocalNodeServiceResponse rsp;
	memset( &rsp, 0, sizeof(rsp) );
	// Execute the application message
	protocolHandler.request_local_node_service( waveportInstance, &command, &rsp );
	char hexpayload[MAX_PAYLOAD_LEN*2+1];
	printf( "%s process local node service: %s;%s;%s \n",
					GetCurrentTimeString(),
					get_response_status_name(rsp.response.status),
					get_local_node_service_name(*(&rsp.command.serviceId)),
					get_hex_payload(hexpayload, sizeof(hexpayload), (Payload*)&rsp.command.command.payload));
}

void process_send_request() {
	printf( "%s process send request: start \n",
						GetCurrentTimeString());
	// Prepare the request to send (throw 1 relay)
	RadioExchangeCommand command = create_radio_exchange_command(payloadGetTemp, destinationAddress);
	RequestResponse rsp;
	memset( &rsp, 0, sizeof(rsp) );
	// Execute the application message
	protocolHandler.send_application_request( waveportInstance, &command, &rsp );
	printf( "%s process send request status: %s \n",
					GetCurrentTimeString(), get_response_status_name(rsp.response.status) );

	char hexpayload[MAX_PAYLOAD_LEN*2+1];
	printf( "%s process send request response: %s;%s \n",
						GetCurrentTimeString(),
						((RadioAddress*)&rsp.command.dynamicStandardParams.nodeAddress)->addressValue,
						get_hex_payload(hexpayload, sizeof(hexpayload), (Payload*)&rsp.command.command.payload));
}

void process_send_message() {
	printf( "%s process send message: start \n",
						GetCurrentTimeString());
	// Prepare the message to send (throw 1 relay)
	CommandResponse rsp;
	memset( &rsp, 0, sizeof(rsp) );
	RadioExchangeCommand command = create_radio_exchange_command_with_relay(payloadGetTemp, destinationAddress, relayPath);
	// Execute the application message
	protocolHandler.send_application_message( waveportInstance, &command, &rsp );
	printf( "%s process send message status: %s \n",
					GetCurrentTimeString(), get_response_status_name(rsp.status) );
}

const char* GetCurrentTimeString()
{
    static char currTime[100] = "";
    currTime[0] = 0;

#ifdef _WINDOWS
	SYSTEMTIME st = {0};
	GetLocalTime(&st);
	sprintf( currTime, "[%02d:%02d:%02d:%03d]", st.wHour, st.wMinute, st.wSecond, st.wMilliseconds );
#else
	struct timeval currentTime;
	gettimeofday( &currentTime, 0 );
	struct tm *tm = localtime(&currentTime.tv_sec);
	sprintf( currTime, "[%02d:%02d:%02d:%03d] ", tm->tm_hour, tm->tm_min, tm->tm_sec, (int)currentTime.tv_usec / 1000 );
#endif

    return currTime;
}
