/*******************************************************************************
 * Copyright (c) 2012 Coronis
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Contributors:
 *     Coronis - initial API and implementation
 *******************************************************************************/

/** @file i_protocol_service.h
 * @brief This interface represents a Coronis SDK Protocol
 * It generalizes and standardizes the available services\n
 * to exchanged radio command,\n
 * with a high level serial protocol,\n
 * throw a serial driver link
 */
#ifndef I_PROTOCOL_SERVICE_H_
#define I_PROTOCOL_SERVICE_H_

#include <i_protocol_listener_types.h>
#include <i_protocol_types.h>

#ifdef  _WINDOWS
#define CORONIS_API __declspec(dllexport)
#define CORONIS_CALL __cdecl

#else
#define CORONIS_API extern
#define CORONIS_CALL
#endif//_WINDOWS


/**
 * The API version of the protocol driver
 *
 */
#define PROTOCOL_API_VERSION 0x02

/**
 * 	Macros used to extract API, MAX and MIN version from protocol version
 * 	encoded value
 */
#define GET_PROTOCOL_MIN_VERSION(_X_) ((_X_ >> 0) & 0x0000FF)
#define GET_PROTOCOL_MAX_VERSION(_X_) ((_X_ >> 8) & 0x0000FF)
#define GET_PROTOCOL_API_VERSION(_X_) ((_X_ >>16) & 0x0000FF)

/**
 * Get protocol version
 * @return The protocol version MAX.MID.MIN encoded in a 8 bytes integer (2 bytes
 *  per number)
 */
CORONIS_API  int CORONIS_CALL protocol_get_version();

/**
 * Create a new protocol instance
 * @return a pointer to created protocol.<code>NULL</code> if protocol failed to be created.
 */
CORONIS_API void* CORONIS_CALL protocol_create();

/**
 * Destroy the protocol instance. It will automatically close the current connection
 * @param protocol The Protocol instance
 * @return <code>0</code> if succeed, a positive number otherwise
 */
CORONIS_API int CORONIS_CALL protocol_destroy( const void* protocol );

/**
 * Initialize the protocol settings if not already configured (serialDriverId, etc.)
 * @param protocol The Protocol instance to open
 * @return the init status (response_status)
 *         OK_PROTOCOL_SUCCESS if initialization succeed
 *         KO_PROTOCOL_ALREADY_OPEN if Protocol is already open
 * 	       KO_SERIALDRIVER_ID_NOT_EXIST if Serial driver id is not resolved on the OS
 *         KO_SERIALDRIVER_NOT_AVAILABLE if connection of Serial driver link fails (already open, etc..)
 */
CORONIS_API int CORONIS_CALL protocol_init( const void* protocol );

/**
 * Initialize the protocol settings if not already configured (serialDriverId, etc.)
 * @param protocol The Protocol instance to open
 * @param file The serial driver property file URL
 * @return the init status (response_status)
 *         OK_PROTOCOL_SUCCESS if initialization succeed
 *         KO_PROTOCOL_ALREADY_OPEN if Protocol is already open
 * 	       KO_SERIALDRIVER_ID_NOT_EXIST if Serial driver id is not resolved on the OS
 *         KO_SERIALDRIVER_NOT_AVAILABLE if connection of Serial driver link fails (already open, etc..)
 */
CORONIS_API int CORONIS_CALL protocol_init_from_file( const void* protocol, const char* file );

/**
 * Open the Protocol connection via a serial driver link (identified by an ID),
 * and configure the protocol parameters
 * @param protocol The Protocol instance to open
 * @return the open status (serial_driver_status)
 *         OK_PROTOCOL_SUCCESS if opening succeed
 *         KO_SERIALDRIVER_ID_NOT_EXIST if Serial driver id is not resolved on the OS
 *         KO_SERIALDRIVER_NOT_AVAILABLE if connection of Serial driver link fails (already open, etc..)
 *         KO_PROTOCOL_SETTING_ERROR if parameters configuration fails
 */
CORONIS_API int CORONIS_CALL protocol_open( const void* protocol );

/**
 * Close the Protocol connection
 * @param protocol The Protocol instance to close
 * @return the close status (response_status)
 *         OK_PROTOCOL_SUCCESS if closing succeed
 *         KO_UNEXPECTED_ERROR if error occurs during closing process
 */
CORONIS_API int CORONIS_CALL protocol_close( const void* protocol );

/**
 * Test if the Protocol connection is open are not
 * @param protocol The Protocol instance
 * @return <code>1</code> if the protocol is connected, <code>0</code> otherwise
 */
CORONIS_API int CORONIS_CALL protocol_is_open( const void* protocol );

/**
 * Get the serial driver link ID, if previously setted
 * @param protocol The Protocol instance
 * @param Setted literal value of serial driver ID
 * @param length of the setted literal value, <code>0</code> if not setted
 * @return <code>0</code> if success or positive number if error occurs during process
 */
CORONIS_API int CORONIS_CALL protocol_get_serial_driver_id( const void* protocol, char* serialDriverId, int idLen );

/**
 * Set the serial driver link ID which Protocol must used as Serial driver link
 * @param protocol The Protocol instance to open
 * @param serialDriverId serial driver ID (must be an available ISerialDriverService implementation ID)
 * @return the set status (response_status)
 *         OK_PROTOCOL_SUCCESS <code>0</code> if setting succeed
 *         KO_PROTOCOL_ALREADY_OPEN if Protocol is already open
 *         KO_SERIALDRIVER_ID_NOT_EXIST if id not exist in the OS
 */
CORONIS_API int CORONIS_CALL protocol_set_serial_driver_id( const void* protocol, const char* serialDriverId);

/**
 * Set the serial driver link which Protocol must used as Serial driver link
 * @param protocol The Protocol instance
 * @param serialDriver SerialDriver instance (must implementation i_serial_driver_service)
 * @return the set status (response_status)
 *         OK_PROTOCOL_SUCCESS <code>0</code> if setting succeed
 * 		   KO_PROTOCOL_NO_VALID_PARAMETER if ivalid parameters were passed
 */
CORONIS_API int CORONIS_CALL protocol_set_serial_driver_instance( const void* protocol, const void* serial_driver_instance);

/**
 * Register a listener to be notify of spontaneous received command
 * and Protocol event.
 * The listeners have to implement the protocol_listener_t
 * and protocol_listener_event_t interfaces.
 * Protocol instance must allow only one listener.
 * @param protocol The Protocol instance
 * @param receiveRadioExchangeCommand The listener of spontaneous received command to register
 * @param receiveEvent The listener of Protocol event to register
 * @param param A user define data which is passed back to listener when a frame is received
 * @return the registration status (response_status)
 *         OK_PROTOCOL_SUCCESS <code>0</code> if registration succeed
 *         KO_PROTOCOL_LISTENER_ALREADY_REGISTERED if a listener is already registered
 */
CORONIS_API int CORONIS_CALL protocol_add_message_listener( const void* protocol,
		protocol_listener_t receiveRadioExchangeCommand,
		protocol_listener_event_t receiveEvent,
		void* param);

/**
 * Unregister the listener from being notify of spontaneous received command
 * and Protocol event.
 * @param protocol The Protocol instance
 * @return the unregistration status (response_status)
 *         OK_PROTOCOL_SUCCESS <code>0</code> if unregistration succeed
 *         KO_PROTOCOL_LISTENER_NOT_REGISTERED if no listener is registered
 */
CORONIS_API int CORONIS_CALL protocol_remove_message_listener( const void* protocol );

/**
 * Register a spy listener to be notify of sent and received radio command
 * via the Protocol instance.
 * The listeners have to implement the protocol_spy_listener_t interface.
 * Protocol instance can allow many listeners.
 * @param protocol The Protocol instance
 * @param spyRecvRadioCommandListener The listener to by notify of received radio command
 * @param spySentRadioCommandListener The listener to by notify of sent radio command
 * @param param A user define data which is passed back to listener when a radio command is exchanged
 * @param spyIndex The index of the Spy listener
 * @return the registration status (response_status)
 * 		   OK_PROTOCOL_SUCCESS <code>0</code> if registration succeed
 *         KO_PROTOCOL_LISTENER_ALREADY_REGISTERED if the spyIndex is already registered
 *         KO_PROTOCOL_MAX_SPY_LISTENER_EXCEED if the maximum number of spy listener
 */
CORONIS_API int CORONIS_CALL protocol_add_spy_message_listener( const void* protocol,
		protocol_spy_listener_t spyRecvRadioCommandListener,
		protocol_spy_listener_t spySentRadioCommandListener,
		void* param,
		int* spyIndex );

/**
 * Unregister a spy listener from being notify of sent and received radio command
 * via the Protocol instance.
 * @param protocol The Protocol instance
 * @param spyIndex The index of the Spy listener
 * @return the unregistration status (response_status)
 * 		   OK_PROTOCOL_SUCCESS <code>0</code> if registration succeed
 *         KO_PROTOCOL_LISTENER_NOT_REGISTERED if the spyIndex is not registered
 */
CORONIS_API int  CORONIS_CALL protocol_remove_spy_message_listener( const void* protocol, int spyIndex );

/**
 * Send a command in transparent mode.
 * With this method, the protocol will directly send the given serial frame
 * to the serial driver without any process on Protocol level.
 * @param protocol The Protocol instance
 * @param frame The serial frame which contains the information of the message to send
 * @param response The CommandResponse of the send process which include the ResponseStatus
 * @return 0 if success or positive number if unexpected error occurs during sending process
 */
CORONIS_API int CORONIS_CALL protocol_send_transparent_message( const void* protocol, const pSerialExchangeData frame, CommandResponse* response );

/**
 * Send an application message to a radio device
 * without waiting for a device responses.
 * @param protocol The Protocol instance
 * @param command The RadioExchangeCommand which contains the information of the message to send
 * @param response The CommandResponse of the send process which include the ResponseStatus
 * @return 0 if success or positive number if unexpected error occurs during sending process
 */
CORONIS_API int CORONIS_CALL protocol_send_application_message( const void* protocol, const RadioExchangeCommand* command, CommandResponse* response );

/**
 * Send an application message in a radio broadcast mode
 * without waiting for device responses.
 * @param protocol The Protocol instance
 * @param command The RadioExchangeCommand which contains the information of the message to send
 * @param response The CommandResponse of the send process which include the ResponseStatus
 * @return 0 if success or positive number if unexpected error occurs during sending process
 */
CORONIS_API int CORONIS_CALL protocol_send_application_broadcast_message( const void* protocol, const RadioExchangeCommand* command, CommandResponse* response );

/**
 * Send an application request to a radio device
 * and waiting for the device response.
 * @param protocol The Protocol instance
 * @param command The RadioExchangeCommand which contains the information of the request to send
 * @param response The RequestResponse of the send process which include the ResponseStatus
 * @return 0 if success or positive number if unexpected error occurs during sending process
 */
CORONIS_API int CORONIS_CALL protocol_send_application_request( const void* protocol, const RadioExchangeCommand* command, RequestResponse* response );

/**
 * Send an application request in a radio broadcast mode
 * and waiting for devices responses.
 * @param protocol The Protocol instance
 * @param command The RadioExchangeCommand which contains the information of the request to send
 * @param response The MulticastRequestResponse of the send process which include the ResponseStatus
 * @return 0 if success or positive number if unexpected error occurs during sending process
 */
CORONIS_API int CORONIS_CALL protocol_send_application_broadcast_request( const void* protocol, const RadioExchangeCommand* command, MulticastRequestResponse* response );

/**
 * Send an application response to a radio device.
 * @param protocol The Protocol instance
 * @param command The pointer to command that contains Payload exchange data structure of the application response to send
 * @param response The CommandResponse of the send process which include the ResponseStatus
 * @return 0 if success or positive number if unexpected error occurs during sending process
 */
CORONIS_API int CORONIS_CALL  protocol_send_application_response( const void* protocol, const Command *command, CommandResponse* response );

/**
 * Send a service message to a radio device
 * without waiting for a device responses.
 * @param protocol The Protocol instance
 * @param command The RadioExchangeCommand which contains the information of the message to send
 * @param response The CommandResponse of the send process which include the ResponseStatus
 * @return 0 if success or positive number if unexpected error occurs during sending process
 */
CORONIS_API int CORONIS_CALL protocol_send_service_message( const void*	protocol, const NodeServiceId serviceId, const RadioExchangeCommand* command, CommandResponse* response );

/**
 * Send a service request to a radio device for a specific Service Id
 * and waiting for the device response.
 * @param protocol The Protocol instance
 * @param command The RadioExchangeCommand which contains the information of the request to send
 * @param response The RequestResponse of the send process which include the ResponseStatus
 * @return 0 if success or positive number if unexpected error occurs during sending process
 */
CORONIS_API int CORONIS_CALL  protocol_send_service_request( const void* protocol, const NodeServiceId serviceId, const RadioExchangeCommand* command, RequestResponse* response );

/**
 * Write the radio parameters of the connected Radio modem
 * @param protocol The Protocol instance
 * @param command The RadioSettingsCommand to configure
 *        RadioSettingsCommand could include the DynamicExpertParams, StaticStandardParams,
 *        StaticExpertParams or many of them
 * @param response The CommandResponse of the write process which include the ResponseStatus
 * @return 0 if success or positive number if unexpected error occurs during sending process
 */
CORONIS_API int CORONIS_CALL protocol_write_radio_params( const void* protocol, const RadioSettingsCommand* command, CommandResponse* response );

/**
 * Read the dynamic expert radio parameters of the connected Radio modem
 * @param protocol The Protocol instance
 * @param response The RadioSettingsResponse which include the readied DynamicExpertParams
 * @return 0 if success or positive number if unexpected error occurs during sending process
 */
CORONIS_API int CORONIS_CALL  protocol_read_dynamic_expert_params( const void* protocol, RadioSettingsResponse* response);

/**
 * Read the static standard radio parameters of the connected Radio modem
 * @param protocol The Protocol instance
 * @param response The RadioSettingsResponse which include the readied StaticStandardParams
 * @return 0 if success or positive number if unexpected error occurs during sending process
 */
CORONIS_API int CORONIS_CALL  protocol_read_static_standard_params( const void* protocol, RadioSettingsResponse* response);

/**
 * Read the static expert radio parameters of the connected Radio modem
 * @param protocol The Protocol instance
 * @param response The RadioSettingsResponse which include the readied StaticExpertParams
 * @return 0 if success or positive number if unexpected error occurs during sending process
 */
CORONIS_API int CORONIS_CALL protocol_read_static_expert_params( const void* protocol, RadioSettingsResponse* response);

/**
 * Send a local service request to the connected Radio modem
 * @param protocol The Protocol instance
 * @param command The LocalNodeServiceCommand of the local service to request
 * @param response The LocalNodeServiceResponse which include the request response
 * @return 0 if success or positive number if unexpected error occurs during sending process
 */
CORONIS_API int CORONIS_CALL protocol_request_local_node_service( const void* protocol, const LocalNodeServiceCommand* command, LocalNodeServiceResponse* response );

#endif /* I_PROTOCOL_SERVICE_H_ */
