/*******************************************************************************
 * Copyright (c) 2012 Coronis
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Contributors:
 *     Coronis - initial API and implementation
 *******************************************************************************/

/**
 * @file protocol_command_types.h
 * @brief Definition of command structures used to process commands
 * with Protocol service
 */
#ifndef PROTOCOL_COMMAND_TYPES_H_
#define PROTOCOL_COMMAND_TYPES_H_

#include <protocol_params_types.h>
#include <i_serial_frame.h>
#include <i_protocol_types.h>
/**
 * Parent structure used to process a command
 * with a i_protocol_service implementation
 */
typedef struct tagICommand {
	/**
	 * The payload exchanged between
	 *        the Protocol and the SerialDriver levels
	 */
	pIPayload payload;
} Command;

/**
 * Structure that contains the command information
 * relative to a data exchanged by radio mode
 * This structure is used by the Protocol service which
 * offers the capabilities to exchange radio data between
 * the connected Radio modem and a distant radio device.
 */
typedef struct tagIRadioExchangeCommand {
	/**
	 * Inherited parent Command
	 */
	Command command;

	/**
	 * Payload data of the RadioExchangeCommand
	 */
	// removed because it is contained in parrent command structure
	//pIPayload payload;

	/**
	 * Configuration of the dynamic standard radio settings
	 * used for the exchange radio
	 * This parameters are not persisted by the Radio modem
	 */
	DynamicStandardParams dynamicStandardParams;

	/**
	 * Configuration of the dynamic expert radio parameters
	 * used for the exchange radio
	 * This parameters are persisted by the Radio modem
	 */
	DynamicExpertParams dynamicExpertParams;
} RadioExchangeCommand;

/**
 * Structure that contains the command information
 * relative to a radio settings command
 * This structure is used by the Protocol service which
 * offers the capabilities to read/write radio parameters
 * on the connected Radio modem.
 */
typedef struct tagIRadioSettingsCommand {
	/**
	 * Inherited parent Command
	 */
	Command command;

	/**
	 * Configuration of the dynamic standard radio settings
	 * This parameters are not persisted by the Radio modem
	 */
	DynamicStandardParams dynamicStandardParams;

	/**
	 * Configuration of the dynamic expert radio parameters
	 * This parameters are persisted by the Radio modem
	 */
	DynamicExpertParams dynamicExpertParams;

	/**
	 * Configuration of the static standard radio parameters
	 * This parameters are persisted by the Radio modem
	 */
	StaticStandardParams staticStandardParams;

	/**
	 * Configuration of the static expert radio parameters
	 * This parameters are persisted by the Radio modem
	 */
	StaticExpertParams staticExpertParams;

} RadioSettingsCommand;

/**
 * Structure that contains the command information
 * relative to a service data of the connected Radio modem
 * This structure is used by the Protocol service which
 * offers the capabilities to request local service
 * with the connected Radio modem.
 */
typedef struct tagILocalNodeServiceCommand {

	/**
	 * Inherited parent Command
	 */
	 Command command;

	 /**
	 * Service identifier of the LocalNodeServiceCommand
	 */
	NodeServiceId serviceId;

	// contained in parent command
	//pIPayload	 serviceValue;

} LocalNodeServiceCommand;

#endif /* PROTOCOL_COMMAND_TYPES_H_ */
