/*******************************************************************************
 * Copyright (c) 2012 Coronis
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Contributors:
 *     Coronis - initial API and implementation
 *******************************************************************************/

/*
 * driver.c
 *
 * This file contain definitions of all functions from Serial Driver Wavenis API.
 */

#include <i_serial_frame.h>
#include "listeners.h"
#include "driver_common.h"
#include <loader_logger.h>
#include "serial_driver_instance.h"
#include "utils.h"

DriverInfo drivers[MAX_DRIVER_NUM];

/****************************************************************************************************
 *                        MAIN SECTION - ISerialDriverService FUNCTIONS DEFINITION
 ****************************************************************************************************/
/****************************************************************************************************/
int driver_get_version( ) {
	return ( ((SERIAL_DRIVER_API_VERSION << 16) & 0xFF0000) |
			 ((SERIAL_DRIVER_MAX_VERSION << 8) & 0x00FF00) |
			 ((SERIAL_DRIVER_MIN_VERSION << 0) & 0x0000FF));
}

/****************************************************************************************************/
void* driver_create( ) {
	size_t freeIndex = 0;
	size_t i = 0;
	for ( ; i < ARRAYSIZE(drivers); ++i )
		if ( drivers[i].integrity != SERIAL_DRIVER_INTEGRITY ) {
			freeIndex = i+1;
			memset( &(drivers[i]), 0, sizeof(drivers[i]) );
			drivers[i].integrity = SERIAL_DRIVER_INTEGRITY;
			pthread_mutex_init( &drivers[i].mutexProcessing, NULL );
			break;
		} else
			LOG_ERROR( g_Logger, "driver_create: %d %X", i, drivers[i].integrity);
	return (void*)freeIndex;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/****************************************************************************************************/
int driver_init_from_string( DriverInfo* driver, const char* initString ) {
	get_all_properties( initString, &driver->params );
	return OK_SERIALDRIVER_SUCCESS;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/****************************************************************************************************/
int driver_init_from_file( const void* driver, const char* file ) {
	char data[10240] = "";
	char path[MAX_PATH] = "";
	int iRet = KO_PARAM_NOT_VALID;

	DriverInfo* serialDriver = get_driver( driver );

	if ( serialDriver != NULL ) {
		iRet = KO_SERIALDRIVER_ALREADY_OPEN;
		if ( !driver_is_open( driver ) ) {

			if ( access( file, 0 ) == -1 ) {
				if ( get_module_path( g_ModuleHandle, path, ARRAYSIZE(path) ) )
					strcat( path, file );
			} else
				strcpy( path, file );
			int fd = open( path, O_RDONLY, 0);
			if ( fd != 0 ) {
				read( fd, data, sizeof(data) );
				close( fd );
			}

			iRet = driver_init_from_string( serialDriver, data );
		}
	}

	if ( g_Logger.loggerHandle == NULL ){
		if ( get_module_path( g_ModuleHandle, path, ARRAYSIZE(path) ) ) {
			strcat( path, DOMAIN_NAME );
			if ( OK_SERIALDRIVER_SUCCESS == load_logger( path, &g_Logger) ) {
				g_Logger.init_log_from_property_file (g_loger_rs232_properties_file_name);
				LOG_DEBUG( g_Logger, "RS232DRIVER LOGGING STARTED", 0);
			}
		}
	}

	return iRet;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/****************************************************************************************************/
int driver_init( const void* driver ) {

	return driver_init_from_file( driver, SERIAL_DRIVER_PROPERTIES_FILE );
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/****************************************************************************************************/
int driver_open( const void* driver ) {
	int iRet = KO_PARAM_NOT_VALID;


	//printf("driver_open:: opening driver ... %d\n", (int)driver);
	DriverInfo* serialDriver = get_driver( driver );
	LOG_DEBUG( g_Logger, "driver_open: Called get_driver, check driver descriptor ... ", 0);
	if (!serialDriver) {
		//printf("driver_open:: Could not get driver for id  %d\n", (int)driver);

		LOG_ERROR( g_Logger, "driver_open: Driver is null", 0);
		return iRet;
	}

	iRet = KO_SERIALPORT_NOT_AVAILABLE;
	//printf("driver_open:: opening id %s\n", serialDriver->params.id);

	if (strlen(serialDriver->params.id) > 3 ) {
		LOG_DEBUG( g_Logger, "driver_open: serialDriver descriptor is ok, call port_open", 0);
		return port_open( serialDriver );
	}else {
		//printf("driver_open:: Port id not set!\n");
		LOG_ERROR( g_Logger, "driver_open: Port ID is not set", 0);
	}

	LOG_ERROR( g_Logger, "driver_open: Exit with error code", 0);
	return iRet;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/****************************************************************************************************/
int driver_close( const void* driver ) {
	int iRet = KO_PARAM_NOT_VALID;
	DriverInfo* serialDriver = get_driver( driver );

	if ( serialDriver != NULL ) {
		iRet = OK_SERIALDRIVER_SUCCESS;
		if ( driver_is_open(driver) ) {
			if ( serialDriver->notifyListener )
				serialDriver->notifyListener( serialDriver->paramNotify, STOP_LISTENING);
			serialDriver->startNotifyListenerEvent = _false;
			iRet = port_close( serialDriver );
		}
	} else
		LOG_ERROR( g_Logger, "driver_close: UNEXPECTED ERROR!!!", 0);
	return iRet;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/****************************************************************************************************/
int driver_destroy( const void* driver ) {
	int iRet = KO_PARAM_NOT_VALID;
	DriverInfo* serialDriver = get_driver( driver );

	if ( serialDriver != NULL ) {
		iRet = driver_close( driver );
		pthread_mutex_destroy( &serialDriver->mutexProcessing );
		memset( serialDriver, 0, sizeof(DriverInfo) );
		LOG_ERROR( g_Logger, "driver_destroy: %X", serialDriver);
	}
	return iRet;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/****************************************************************************************************/
int driver_is_open( const void* driver ) {
	DriverInfo* serialDriver = get_driver( driver );

	if ( serialDriver != NULL )
		return ( serialDriver->port != NULL );

	return 0;/*false*/
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/****************************************************************************************************/
int driver_get_port_id( const void* driver, char* id, int idLen ) {
	int iRet = KO_PARAM_NOT_VALID;
	DriverInfo* serialDriver = get_driver( driver );

	if ( serialDriver != NULL && id && idLen >= DRIVER_ID_LEN ) {
		strcpy( id, serialDriver->params.id );
		iRet = OK_SERIALDRIVER_SUCCESS;
	}

	return iRet;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/****************************************************************************************************/
int is_numeric(const char *str){

	if (str == NULL || *str == '\0' /*|| isspace(*str)*/) return 0;
	char * str_ptr;
	strtod (str, &str_ptr);
	return *str_ptr == '\0';
}

int driver_set_port_id( const void* driver, const char* id ) {
	// printf("\ndriver_set_port_id:: entered!\n");
#ifdef  _WINDOWS
#define PORT_NAME_LEN 3
#define PORT_NAME     "COM"
#else
#define PORT_NAME_LEN 8
#define PORT_NAME     "/dev/tty"
#endif//_WINDOWS
	int iRet = KO_PARAM_NOT_VALID;
	DriverInfo* serialDriver = get_driver( driver );

	// printf("driver_set_port_id:: set port Id %s\n", id);

	if ( serialDriver != NULL && id && strlen(id) > PORT_NAME_LEN ) {// COM,ttyS
		iRet = KO_SERIALDRIVER_ALREADY_OPEN;

		if ( !driver_is_open( driver ) ) {
			iRet = KO_PARAM_SERIALPORTID_NOT_EXIST;

			if ( 0 == strncmp( id, PORT_NAME, PORT_NAME_LEN ) ) {
				// check the format of setrial port

				#ifdef  _WINDOWS
				if (!is_numeric(id + PORT_NAME_LEN)) {
					return iRet;
				}
				#else

				#endif

				strncpy( serialDriver->params.id, id, DRIVER_ID_LEN-1 );
				serialDriver->params.isSet |= FLAG_SERIAL_PORT_ID;
				iRet = OK_SERIALDRIVER_SUCCESS;
			}
		}

		// printf("driver_set_port_id:: serialDriver->params.id = %s\n",serialDriver->params.id);
	}

	return iRet;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/****************************************************************************************************/
int driver_set_port_params( const void* driver, int baudRate, int dataBit, int stopBit, int parity ) {
	int iRet = KO_PARAM_NOT_VALID;
	DriverInfo* serialDriver = get_driver( driver );

	if ( serialDriver != NULL ) {
		iRet = KO_SERIALDRIVER_ALREADY_IN_USE;
		if ( serialDriver->status == STATUS_IDLE ) {
			if ( ( ( iRet = driver_verify_baudrate( baudRate ) ) == OK_SERIALDRIVER_SUCCESS ) &&
				 ( ( iRet = driver_verify_databit( dataBit )   ) == OK_SERIALDRIVER_SUCCESS ) &&
			     ( ( iRet = driver_verify_stopbit( stopBit )   ) == OK_SERIALDRIVER_SUCCESS ) &&
			     ( ( iRet = driver_verify_parity( parity )     ) == OK_SERIALDRIVER_SUCCESS ) ) {

				serialDriver->params.baudRate = baudRate;
				serialDriver->params.isSet |= FLAG_BAUD_RATE;
				serialDriver->params.dataBit = dataBit;
				serialDriver->params.isSet |= FLAG_DATA_BIT;
				serialDriver->params.parity = parity;
				serialDriver->params.isSet |= FLAG_PARITY;
				switch (stopBit) {
				case 2:	serialDriver->params.stopBit = TWOSTOPBITS; break;
				case 1:
				default: serialDriver->params.stopBit = ONESTOPBIT; break;
				}
				serialDriver->params.isSet |= FLAG_STOP_BIT;
				if (driver_is_open(driver))
					iRet = serial_port_set_params(serialDriver->port,
							serialDriver->params.baudRate, serialDriver->params.dataBit,
							serialDriver->params.stopBit, serialDriver->params.parity,
							serialDriver->params.radioAckTimeout,
							serialDriver->params.flowControlMode);
			}
		}
	}

	return iRet;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/****************************************************************************************************/
int driver_set_listening_period_param( const void* driver, int listeningPeriod ) {
	int iRet = KO_PARAM_NOT_VALID;
	DriverInfo* serialDriver = get_driver( driver );

	if ( serialDriver != NULL ) {
		iRet = KO_SERIALDRIVER_ALREADY_OPEN;
		if ( !driver_is_open( driver ) ) {
			if ( listeningPeriod < 0 )
				iRet = KO_PARAM_LISTENINGPERIOD_NOT_VALID;
			else {
				serialDriver->params.listeningPeriod = listeningPeriod;
				serialDriver->params.isSet |= FLAG_LISTENING_PERIOD;
				iRet = OK_SERIALDRIVER_SUCCESS;
			}
		}
	}

	return iRet;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/****************************************************************************************************/
int driver_set_flow_control_mode_params( const void* driver, int flowControlIn, int flowControlOut ) {
	int iRet = KO_PARAM_NOT_VALID;
	DriverInfo* serialDriver = get_driver( driver );

	if ( serialDriver != NULL ) {
		iRet = KO_SERIALDRIVER_ALREADY_OPEN;
		if ( !driver_is_open( driver ) ) {
			if ( ( ( iRet = driver_verify_flowcontrolmode_in( flowControlIn )   ) == OK_SERIALDRIVER_SUCCESS ) &&
			     ( ( iRet = driver_verify_flowcontrolmode_out( flowControlOut ) ) == OK_SERIALDRIVER_SUCCESS ) ) {
				serialDriver->params.flowControlMode = flowControlIn|flowControlOut;
				serialDriver->params.isSet |= FLAG_FLOWCONTROL_MODE;
			}
		}
	}
	return iRet;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/****************************************************************************************************/
int driver_send( const void* driver, const pSerialExchangeData exchangeDataFrame ) {

	int iRet = KO_PARAM_NOT_VALID;
	DriverInfo* serialDriver = get_driver( driver );
	if ( serialDriver != NULL ) {
		iRet = KO_SERIALFRAME_REQUIRED;
		if ( exchangeDataFrame && ((SerialFrame*)&exchangeDataFrame->serialFrame)->len > 0 ) {
			iRet = KO_SERIALDRIVER_NOT_OPEN;
			if ( driver_is_open(driver) )
				iRet = port_send( serialDriver,
						((SerialFrame*)&exchangeDataFrame->serialFrame)->frame,
						((SerialFrame*)&exchangeDataFrame->serialFrame)->len );
		}
	}

	return iRet;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/****************************************************************************************************/
int driver_add_listener( const void* driver,
		serial_driver_listener_t receiveFrame,
		serial_driver_listener_notify_t notifyListener,
		void* param ) {

	DriverInfo* serialDriver = get_driver( driver );
	if ( serialDriver == NULL )  return KO_PARAM_NOT_VALID;

	driver_remove_listener(driver);

	//register frame listener
	if (receiveFrame != NULL) {
		serialDriver->realListeners.receiveFrame = receiveFrame;
		serialDriver->realListeners.paramReceiveFrame = param;
		serialDriver->receiveFrame = serial_driver_frame_listener;
		serialDriver->paramReceiveFrame = serialDriver;
	}

	// register notifyListener
	if (notifyListener != NULL) {
		serialDriver->realListeners.notifyListener = notifyListener;
		serialDriver->realListeners.paramNotify = param;
		serialDriver->notifyListener = serial_driver_notify_listener;
		serialDriver->paramNotify = serialDriver;
		if ( serialDriver->notifyListener ) {
			serialDriver->startNotifyListenerEvent = _true;
			serialDriver->notifyListener( serialDriver->paramNotify, START_LISTENING);
		}
	}

	return OK_SERIALDRIVER_SUCCESS;;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/****************************************************************************************************/
int driver_remove_listener( const void* driver ) {
	DriverInfo* serialDriver = get_driver( driver );
	if ( serialDriver == NULL ) return KO_PARAM_NOT_VALID;;

	// unregister frame listener
	serialDriver->receiveFrame = NULL;
	serialDriver->paramReceiveFrame = NULL;
	serialDriver->realListeners.receiveFrame = NULL;
	serialDriver->realListeners.paramReceiveFrame = NULL;

	// unregister event listener
	if ( serialDriver->notifyListener )
		serialDriver->notifyListener( serialDriver->paramNotify, STOP_LISTENING);

	serialDriver->realListeners.paramNotify = NULL;
	serialDriver->realListeners.notifyListener = NULL;
	serialDriver->paramNotify = NULL;
	serialDriver->notifyListener = NULL;
	serialDriver->startNotifyListenerEvent = _false;

	return OK_SERIALDRIVER_SUCCESS;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/****************************************************************************************************/
int driver_add_spy_listener( const void* driver, serial_driver_spy_t spySend,
		serial_driver_spy_t spyRecv, void* param, int* spyIndex ) {
	int iRet = KO_PARAM_NOT_VALID;
	DriverInfo* serialDriver = get_driver( driver );

	if ( serialDriver != NULL ) {
		int i = 0;
		// verify if given spy listener has already registered
		for ( ; i < MAX_SPY_NUM; ++i )
			if ( serialDriver->realListeners.spiesSend[i] == spySend ||
					serialDriver->realListeners.spiesRecv[i] == spyRecv )
				return KO_SERIALDRIVER_LISTENER_ALREADY_REGISTERED;

		iRet = KO_SERIALDRIVER_MAX_SPY_LISTENER_EXCEED;
		for ( i = 0; i < MAX_SPY_NUM; ++i ) {
			if ( serialDriver->spiesSend[i] == NULL &&
					serialDriver->spiesRecv[i] == NULL ) {
				serialDriver->realListeners.spiesSend[i] = spySend;
				serialDriver->realListeners.spiesRecv[i] = spyRecv;
				serialDriver->realListeners.paramSpies[i] = param;
				serialDriver->spiesSend[i] = serial_driver_spy_send_listener;
				serialDriver->spiesRecv[i] = serial_driver_spy_recv_listener;
				serialDriver->paramSpies[i] = serialDriver;
				if ( spyIndex )
					*spyIndex = i;
				iRet = OK_SERIALDRIVER_SUCCESS;
				break;
			}
		}
	}

	return iRet;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/****************************************************************************************************/
int driver_remove_spy_listener( const void* driver, int spyIndex ) {
	int iRet = KO_PARAM_NOT_VALID;
	DriverInfo* serialDriver = get_driver( driver );

	if ( serialDriver != NULL  && spyIndex >= 0 && spyIndex < MAX_SPY_NUM ) {
		iRet = KO_SERIALDRIVER_LISTENER_NOT_REGISTERED;
		if ( serialDriver->spiesSend[spyIndex] || serialDriver->spiesRecv[spyIndex] )
			iRet = OK_SERIALDRIVER_SUCCESS;
		serialDriver->spiesSend[spyIndex] = NULL;
		serialDriver->spiesRecv[spyIndex] = NULL;
		serialDriver->paramSpies[spyIndex] = NULL;
	}

	return iRet;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


/****************************************************************************************************
 *                        HELPER SECTION - OTHER FUNCTIONS DEFINITION
 ****************************************************************************************************/
_bool is_port_already_initialized( const char* id ) {
	if ( id && strlen(id) > 0 ) {
		int i = 0;
		for ( ; i < ARRAYSIZE(drivers); ++i )
			if ( drivers[i].params.id[0] != '\0' && 0 == strcmp( id, drivers[i].params.id ) )
				return _true;// we already have the same port name in array

	}
	return _false;
}

//*****************************************************************************
// prototype methods definition to solve problem with serial driver instance
//	1. create a structure that will hold two types of information
//     a) - the DLL handler(?) and pointers to each DLL exported functions (this is possible). Probably the handler to DLL
//		    function is not required (and is not possible to obtain) because the DLL is controlled by the owner, and
//			and this information is obtained from DLL that is used (i.e. the dll does not known it's handler)
//     b) - the internal state information. It contains all the CFG, current status and data frames of the serial driver.
//
//     There are already two structures which satisfy our needs (a) SerialDriverLoader and (b) DriverInfo.
//      The task is to join these two structures and to request this information from serial driver
//
//   2.  Update all internal structures to use this renewed information.
//

static SerialDriverInstance driver_instances[MAX_DRIVER_NUM];

void* driver_get_serial_driver_instance(const void* driver ) {

	// O.K. for static, because the functions addresses will not be modified
	// during DLL lifetime

	// printf("driver_get_serial_driver_instance:: get serial driver ID %d", (int) driver);
	if (get_driver(driver) != 0) {

		DriverInfo *info = get_driver(driver);

		SerialDriverInstance *sdi = &driver_instances[(int)driver];
		sdi->driverInfo = (void* )driver; //(void*)get_driver(driver);

		sdi->loader.add_listener 				= &driver_add_listener;
		sdi->loader.add_spy_listener 			= &driver_add_spy_listener;
		sdi->loader.close 						= &driver_close;
		sdi->loader.create 						= &driver_create;
		sdi->loader.destroy 						= &driver_destroy;
		sdi->loader.get_port_id 					= &driver_get_port_id;
		sdi->loader.get_version 					= &driver_get_version;
		sdi->loader.init 						= &driver_init;
		sdi->loader.init_from_file				= &driver_init_from_file;
		sdi->loader.is_open						= &driver_is_open;
		sdi->loader.open							= &driver_open;
		sdi->loader.remove_listener  			= &driver_remove_listener;
		sdi->loader.remove_spy_listener  		= &driver_remove_spy_listener;
		sdi->loader.send  						= &driver_send;
		sdi->loader.set_flow_control_mode_params = &driver_set_flow_control_mode_params;
		sdi->loader.set_listening_period_param 	= &driver_set_listening_period_param;
		sdi->loader.set_port_id 					= &driver_set_port_id;
		sdi->loader.set_port_params 				= &driver_set_port_params;


		sdi->loader.get_serial_driver_instance   = &driver_get_serial_driver_instance;



		sdi->loader.serialDriverHandle 			= 0;
		sdi->loader.serialDriverId[0]			= '\0';

		//printf("\ndriver_get_serial_driver_instance:: Serial driver INFO %X for ID %d\n",
		//		sdi->driverInfo,
		//		(int) driver);

		//printf("driver_get_serial_driver_instance:: Driver info: port %X, BD %d, DataB %d, ST %d, Parity %d\n",
		//		info->port,  info->params.baudRate, info->params.dataBit, info->params.stopBit,
		//		info->params.parity);


		return (void* )sdi;
	}

	return 0;

}
