/*******************************************************************************
 * Copyright (c) 2012 Coronis
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Contributors:
 *     Coronis - initial API and implementation
 *******************************************************************************/
#include <stdlib.h>

#include "listeners.h"
#include "loader_logger.h"
#include "loader_serial_driver.h"

// coronis domain includes
#include "serial_driver_types_utils.h"

#ifndef _WINDOWS
#include <bits/local_lim.h>
	#define MAX_LISTENER_THREADS (_POSIX_THREAD_THREADS_MAX-2)
#else
	#define MAX_LISTENER_THREADS (PTHREAD_THREADS_MAX-2)
#endif

typedef struct tagListenerThreadParam {
	RealListeners realListeners;
	pthread_t threadHandle;
	SerialFrame frame;
	/*
	 * thread index
	 * */
	int threadId;
} ListenerThreadParam;

static ListenerThreadParam g_ltp[MAX_LISTENER_THREADS];

void* listener_thread( void *param );
int serial_driver_listeners( void* param, const SerialFrame* frame );

int serial_driver_frame_listener( void* param, const pSerialExchangeData frame ) {
	if (frame == 0 ) return -1;
	((DriverInfo*)param)->realListeners.eventId = FRAME_EVENT;
	return serial_driver_listeners( param, (SerialFrame*)frame->serialFrame );
}

int serial_driver_notify_listener( void* param, int notifyCode ) {
	((DriverInfo*)param)->realListeners.eventId = NOTIFY_EVENT;
	((DriverInfo*)param)->realListeners.id = notifyCode;
	return serial_driver_listeners( param, NULL );
}

int serial_driver_spy_send_listener( void* param, int spyIndex, const pSerialExchangeData frame ) {
	if (frame == 0 ) return -1;
	((DriverInfo*)param)->realListeners.eventId = SPY_SEND_EVENT;
	((DriverInfo*)param)->realListeners.id = spyIndex;
	return serial_driver_listeners( param, (SerialFrame*)frame->serialFrame );
}

int serial_driver_spy_recv_listener( void* param, int spyIndex, const pSerialExchangeData frame ) {
	if (frame == 0 ) return -1;
	((DriverInfo*)param)->realListeners.eventId = SPY_RECV_EVENT;
	((DriverInfo*)param)->realListeners.id = spyIndex;
	return serial_driver_listeners( param, (SerialFrame*)frame->serialFrame );
}

void cleanup() {
	int i = 0;
	int cnt = 0;
	for( ; i < MAX_LISTENER_THREADS; ++i ) {

#ifndef _WINDOWS
		if ( g_ltp[i].threadId == 0 && g_ltp[i].threadHandle) {
#else
			if ( g_ltp[i].threadId == 0 && g_ltp[i].threadHandle.p ) {
#endif
			pthread_join( g_ltp[i].threadHandle, NULL );
			memset( &g_ltp[i], 0, sizeof(g_ltp[i]) );
			cnt += 1;
		}
	}

	LOG_DEBUG( g_Logger, "CLEANUP: Cleared %d threads (back to pool)", cnt);
}

int serial_driver_listeners( void* param, const SerialFrame* frame ) {

	DriverInfo* di = (DriverInfo*)param;
	cleanup();
	int i = 0;

	for( ; i < MAX_LISTENER_THREADS; ++i ) {
#ifndef _WINDOWS
		if ( g_ltp[i].threadId == 0 && g_ltp[i].threadHandle == 0 ) {
#else
			if ( g_ltp[i].threadId == 0 && g_ltp[i].threadHandle.p == 0 ) {
#endif
			g_ltp[i].threadId = i+1;
			g_ltp[i].realListeners = di->realListeners;
			g_ltp[i].frame.len = 0;

			if ( frame )
				memcpy( &g_ltp[i].frame, frame, sizeof(SerialFrame) );

#ifdef _JAVA_WRAPPER
			if ( di->java )
				create_jni_global_ref_to_param( &g_ltp[i].realListeners );
#endif

			if (pthread_create( &g_ltp[i].threadHandle, NULL, listener_thread, &g_ltp[i] ) != 0) {
				LOG_ERROR( g_Logger, "serial_driver_listeners: Could not create a thread to call a listener", "");
			}
			break;
		}
	}

	int j = 0;
	int cnt = 0;
	for(; j < MAX_LISTENER_THREADS; ++j ) {
#ifndef _WINDOWS
		if ( g_ltp[j].threadId != 0 || g_ltp[j].threadHandle != 0 )
#else
			if ( g_ltp[j].threadId != 0 || g_ltp[j].threadHandle.p != 0 )
#endif
			cnt += 1;
	}

	LOG_DEBUG( g_Logger, "Pool is %d% % full", (int)((cnt*100)/MAX_LISTENER_THREADS));
	if (i >= MAX_LISTENER_THREADS) {
		LOG_ERROR(g_Logger, "No more threads in pool. Java listener has not been called!", "");
	}

	return 0;
}

void* listener_thread( void *param ) {
	ListenerThreadParam* ltp = (ListenerThreadParam*)param;

	if (ltp == 0) {
		LOG_ERROR(g_Logger, "listener_thread: passed param is NULL!", 0);
		return 0;
	}
	RealListeners* rl = &ltp->realListeners;

	if (rl == 0) {
		LOG_ERROR(g_Logger, "listener_thread: RealListener is NULL!", 0);
		return 0;
	}

	LOG_DEBUG( g_Logger, "listener_thread %d-0x%x: Preparing to notify listener ",
			ltp->threadId, ltp->threadHandle);

	SerialExchangeData sed = create_serial_exchange_data(ltp->frame);
	LOG_DEBUG( g_Logger, "listener_thread %d-0x%x: Frame copied! ", ltp->threadId, ltp->threadHandle);

	LOG_DEBUG( g_Logger, ".", 0);

	switch ( rl->eventId ) {
	case NOTIFY_EVENT:		// 0
		LOG_DEBUG( g_Logger, "listener_thread %d-0x%x: Notify event ... ", ltp->threadId, ltp->threadHandle);
		rl->notifyListener( rl->paramNotify, rl->id );
		LOG_DEBUG( g_Logger, "listener_thread %d-0x%x: Event notified!", ltp->threadId, ltp->threadHandle);
		break;
	case FRAME_EVENT:		// 1
		LOG_DEBUG( g_Logger, "listener_thread %d-0x%x: Notify frame event ... ", ltp->threadId, ltp->threadHandle);
		rl->receiveFrame( rl->paramReceiveFrame, &sed );
		LOG_DEBUG( g_Logger, "listener_thread %d-0x%x: Frame event notified... ", ltp->threadId, ltp->threadHandle);
		break;
	case SPY_SEND_EVENT:	// 2
		LOG_DEBUG( g_Logger, "listener_thread %d-0x%x: Notify spy send event ... ", ltp->threadId, ltp->threadHandle);
		rl->spiesSend[rl->id]( rl->paramSpies[rl->id], rl->id, &sed );
		LOG_DEBUG( g_Logger, "listener_thread %d-0x%x: Spy send event notified!", ltp->threadId, ltp->threadHandle);
		break;
	case SPY_RECV_EVENT:	// 3
		LOG_DEBUG( g_Logger, "listener_thread %d-0x%x: Notify recv event ... ", ltp->threadId, ltp->threadHandle);
		rl->spiesRecv[rl->id]( rl->paramSpies[rl->id], rl->id, &sed );
		LOG_DEBUG( g_Logger, "listener_thread %d-0x%x: Recv event notified", ltp->threadId, ltp->threadHandle);
		break;
	default: break;
	}

	LOG_DEBUG( g_Logger, "listener_thread EXIT thread:%d event:%d other:%d", ltp->threadId, rl->eventId, rl->id  );
	ltp->threadId = 0;
	return 0;
}
