/*******************************************************************************
 * Copyright (c) 2012 Coronis
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Contributors:
 *     Coronis - initial API and implementation
 *******************************************************************************/

/**
 * The serial.c basically is an abstraction over concrete platform, windows or linux.
 */
#include "listeners.h"

// coronis domain includes
#include "loader_logger.h"
#include "loader_serial_driver.h"
#include "utils.h"
#include "serial_driver_types_utils.h"

#ifndef CORONIS_SDK_SERIAL_DRIVER_CONSTANTS_V_1_0
	#error "Please make sure that right serial_driver_types.h is included in project. Currently is supported v1.0"
#endif // CORONIS_SDK_SERIAL_DRIVER_CONSTANTS_V_1_0

// Acknowledged frame:             | SYNC & STX | LEN | FRAME |    CRC     | ETX |
static const byte ACKNOWLEDGE[]  = { SYNC,  STX , 0x04,  0x06 , 0x56, 0x02 , ETX };

// Not Acknowledged frame          | SYNC & STX | LEN | FRAME |    CRC     | ETX |
static const byte NACKNOWLEDGE[] = { SYNC,  STX , 0x04,  0x15 , 0x4C, 0x20 , ETX };

// Error frame                     | SYNC & STX | LEN | FRAME |    CRC     | ETX |
static const byte ERRORFRAME[]   = { SYNC,  STX , 0x04,  0x00 , 0x60, 0x67 , ETX };

static _bool toWork = _false;

static int listeningPeriod = DEFAULT_LISTENING_PERIOD_TIMEOUT;

static pthread_t mainThreadHandle = {0};

void* main_thread(void *argument);


/**
 *	Open the serial driver communication port with following actions:
 *     - call low level open port (the low level is a call to windows or linux api)
 *	   - set port parameters
 *	   - call registered events
 *	   - start serial driver state machine thread in case if this is first type
 *	   a serial port is opened.
 *
 *	  @param - serial driver descriptor used for multiple serial driver instances
 *	  @return - status or error code
 */
int port_open( DriverInfo* driver ) {

	int iRet = OK_SERIALDRIVER_SUCCESS;
	pthread_mutex_lock( &driver->mutexProcessing );

	//printf("port_open: Open port ... \n");
	if ( driver->port == NULL ) {
		LOG_DEBUG( g_Logger, "port_open: port is null call serial_port_open", 0);
		iRet = serial_port_open( driver->params.id, &driver->port);
	}

	if ( OK_SERIALDRIVER_SUCCESS == iRet ) {

		//printf("port_open: Port opened\n");

		if ( driver->notifyListener && !driver->startNotifyListenerEvent ) {
			driver->startNotifyListenerEvent = _true;
			driver->notifyListener( driver->paramNotify, START_LISTENING);
		}

		LOG_DEBUG( g_Logger, "port_open: OK, set port parameters", 0);
		LOG_DEBUG( g_Logger, "port_open: port %X\n", driver->port);
		LOG_DEBUG( g_Logger, "port_open: baudrate %d\n", driver->params.baudRate);
		LOG_DEBUG( g_Logger, "port_open: databit %d\n", driver->params.dataBit);
		LOG_DEBUG( g_Logger, "port_open: stop bit %d\n", driver->params.stopBit);
		LOG_DEBUG( g_Logger, "port_open: parity %d\n", driver->params.parity);
		LOG_DEBUG( g_Logger, "port_open: ACK timeout %d\n", driver->params.radioAckTimeout);
		LOG_DEBUG( g_Logger, "port_open: Flow control mode %d\n", driver->params.flowControlMode);

		iRet = serial_port_set_params( driver->port, driver->params.baudRate, driver->params.dataBit,
				driver->params.stopBit, driver->params.parity, driver->params.radioAckTimeout, driver->params.flowControlMode );

		if ( OK_SERIALDRIVER_SUCCESS == iRet ) {
			listeningPeriod = DEFAULT_LISTENING_PERIOD_TIMEOUT + driver->params.listeningPeriod;
			if ( !toWork ) {
				toWork = _true;
				iRet = pthread_create( &mainThreadHandle, NULL, main_thread, (void *)&toWork );
				if ( iRet != 0 )
					port_close( driver );
			}

		} else {

			LOG_DEBUG( g_Logger, "port_open: Failed  to set port params: %d, close port", iRet);
			port_close( driver );
		}

	} else
	{
		// printf("port_open: Failed to open port!\n");

		LOG_ERROR( g_Logger, "port_open: Failed to open port", 0);
	}

	LOG_DEBUG( g_Logger, "port_open: %s %d", driver->params.id, get_sending_status_name(iRet) );
	pthread_mutex_unlock( &driver->mutexProcessing );
	return iRet;
}

/**
 *	Closes the state machine session and low level com port. Check if this is the last port closed,
 *	terminate the main thread that processes serial driver state machine.
 *	@param - the driver descriptor
 *	@return the result code
 */
int port_close( DriverInfo* driver ) {
	LOG_DEBUG( g_Logger, "port_close %s", driver->params.id );

	int iRet = OK_SERIALDRIVER_SUCCESS;

	if ( driver->status != STATUS_IDLE ) {
		if ( driver->params.hostAckStatus )
			serial_port_send(driver->port, ACKNOWLEDGE, ARRAYSIZE(ACKNOWLEDGE) );
		driver->status = STATUS_IDLE;
		LOG_DEBUG( g_Logger, "port_close cleanup - status: %d", driver->status );
	}


	if ( driver->port != NULL ) {
		LOG_DEBUG( g_Logger, "port_close lock and close mutex %X", driver->mutexProcessing );
		pthread_mutex_lock( &driver->mutexProcessing );
		iRet = serial_port_close( driver->port );
		pthread_mutex_unlock( &driver->mutexProcessing );
		driver->port = NULL;
		LOG_DEBUG( g_Logger, "port_close closed!", 0 );
	}

	LOG_DEBUG( g_Logger, "port_close clean up", 0 );
	memset( &driver->sendFrame, 0, sizeof(driver->sendFrame) );

	_bool hasAnyPortOpen = _false;
	int i = 0;
	for ( i = 0; i < MAX_DRIVER_NUM; ++i )
		if ( drivers[i].port != NULL ) {
			hasAnyPortOpen = _true;
			break;
		}

	if ( !hasAnyPortOpen && toWork ) { // no any port opened we can stop thread
		LOG_DEBUG( g_Logger, "port_close no more ports opened, close thread ...", 0 );
		toWork = _false;
		pthread_join( mainThreadHandle, NULL);
		memset( &mainThreadHandle, 0, sizeof(mainThreadHandle) );
		LOG_DEBUG( g_Logger, "port_close Thread closed", 0 );
	}

	LOG_DEBUG( g_Logger, "port_close %s", get_sending_status_name( iRet ) );
	return iRet;
}

void printf_serial_frame (char *info, char *port, SerialFrame *frame) {

	//if (frame->len > 0 )
	{

		int i = 0;
		char buff [300 + 1];
		for( ; (i < frame->len) && (i < 300);  i++) {
			sprintf(&buff[i *2], "%02X", frame->frame[i] & 0xFF);
		}

		buff[i *2] = 0;
		//printf("** %s::  %s -> frame %s\n", info, port, buff);
	}
}

/**
 * 	The function does not actually send data, but schedules data for sending.
 * 	 - It checks the state machine to see if it is in IDLE state.
 * 	 - composes the final serial driver frame by adding length, sync characters, calculate and add CRC if
 * 	   configuration file allows this.
 * 	 After data have been scheduled for sending, a waiting wail is execute, the function exits.
 *
 */
int port_send( DriverInfo* driver, const byte* frame, int frameLen ) {

	static long index = 0;
	int iRet = KO_SERIALDRIVER_ALREADY_IN_USE;

	LOG_DEBUG( g_Logger, "port_send: Enter function", 0 );

	if ( driver->status == STATUS_IDLE ) {
			driver->status = STATUS_BUSY;

		iRet = KO_SERIALPORT_UNEXPECTED_ERROR;
		LOG_DEBUG( g_Logger, "port_send [%d] port_send START sending Status:%d and status:%d",
				++index, driver->sendingStatus, driver->status );

		driver->retryCount = driver->params.retryCountDef;

		memset( &driver->sendFrame, 0, sizeof(driver->sendFrame) );

		driver->sendFrame.len = frameLen;
		driver->sendFrame.len += driver->params.lengthActivation ? 1 : 0; // + len byte
		driver->sendFrame.len += driver->params.syncCharacterActivation ? 1 : 0; // + SYNC byte
		driver->sendFrame.len += driver->params.startEndCharactersActivation ? 2 : 0; // + STX & ETX bytes
		driver->sendFrame.len += driver->params.crc16Activation ? 2 : 0; // + CRC 2 bytes

		int i = -1;
		if ( driver->params.syncCharacterActivation )
			driver->sendFrame.frame[++i] = SYNC;
		if ( driver->params.startEndCharactersActivation )
			driver->sendFrame.frame[++i] = STX;

		if (driver->params.lengthActivation)
			driver->sendFrame.frame[++i] = frameLen+1+(driver->params.crc16Activation ? 2 : 0);// write length

		memcpy(driver->sendFrame.frame + (++i), frame, frameLen);

		if ( driver->params.crc16Activation ) {
			unsigned int crc = calc_crc(driver->sendFrame.frame + (i-1), frameLen + 1);
			driver->sendFrame.frame[i + frameLen ] = crc & 0x00FF;
			driver->sendFrame.frame[i + frameLen + 1] = (crc & 0xFF00) >> 8;
		}

		if ( driver->params.startEndCharactersActivation )
			driver->sendFrame.frame[driver->sendFrame.len - 1] = ETX;

		char frameString[MAX_SERIALFRAME_LEN*2+1] = "";
		LOG_DEBUG( g_Logger, "port_send RS232Driver send: %s",
				get_hex_serialframe(frameString, sizeof(frameString), &driver->sendFrame ) );

		// frame has been prepared so we could notify thread about it!
		driver->status = STATUS_SEND_COMMAND;
		driver->sendingStatus = SENDING_STATUS_UNKNOWN;


		// printf_serial_frame("SEND", driver->params.id, &driver->sendFrame);
		while ( driver->sendingStatus == SENDING_STATUS_UNKNOWN ) {
			SLEEP( DEFAULT_LISTENING_PERIOD_TIMEOUT );
		}

		iRet = driver->sendingStatus;

		memset( &driver->sendFrame, 0, sizeof(driver->sendFrame) );
		LOG_DEBUG( g_Logger, "port_send [%d] port_send END sending Status:%s and status:%d", index, get_sending_status_name( driver->sendingStatus ), driver->status );
	}

	return iRet;
}



/**
 *	 Read data from serial port and compose final serial frame as requried by configuration options
 *	  (check crc, len, command).
 */
int read_data_from_serial_port( DriverInfo* driver ) {

	unsigned long amountInInQueue = 0;
	int iRet = OK_SERIALDRIVER_SUCCESS;
	unsigned int len = 0;
	struct timeval startTime;
	struct timeval endTime;
	int    bytes_to_read = 0;

	driver->frame.len = 0;

	//--------------------------------------------------------------------------
	// check if synchronization bytes are required
	if ( driver->params.syncCharacterActivation )
		bytes_to_read += 1;

	// both characters STX and ETX, but ETX is examined at the end of frame
	if ( driver->params.startEndCharactersActivation )
		bytes_to_read += 1;

	//--------------------------------------------------------------------------
	// data is available, firstly read | SYNC & STX | bytes!!! (if required)
	if(bytes_to_read)
		iRet = serial_port_recv(driver->port, driver->frame.frame, bytes_to_read, &driver->frame.len);

	if ( OK_SERIALDRIVER_SUCCESS != iRet || driver->frame.len != bytes_to_read ) {
		memset( &driver->frame.frame, 0, sizeof(driver->frame.frame) );
		return iRet;
	}

	int i = -1;
	_bool wavePortMsg = _true;
	if ( driver->params.syncCharacterActivation )
		wavePortMsg = (SYNC == (int)driver->frame.frame[++i]);

	if ( driver->params.startEndCharactersActivation && wavePortMsg )
		wavePortMsg = (STX == driver->frame.frame[++i] );

	//--------------------------------------------------------------------------
	if ( !wavePortMsg ) {
		// wrong format!!!
		iRet = KO_SERIALFRAME_REQUIRED;
		memset( &driver->frame.frame, 0, sizeof(driver->frame.frame) );
		return iRet;
	}

	//--------------------------------------------------------------------------
	// it is ok, this is waveport message
	// some startup characters are read, go next
	if ( i == bytes_to_read ) {
		gettimeofday( &startTime, 0 );
		do { // wait for length byte
			iRet = serial_port_check(driver->port, &amountInInQueue, 0);
			if ( iRet == OK_SERIALDRIVER_SUCCESS && amountInInQueue == 0 ) {
				gettimeofday( &endTime, 0 );
				if ( millisecondtimediff( startTime, endTime ) > driver->params.radioAckTimeout )
					iRet = KO_SENDING_FAILURE; // timeout expired
				SLEEP( 10 );
			}
		} while( iRet == OK_SERIALDRIVER_SUCCESS && amountInInQueue == 0 );
	}

	//--------------------------------------------------------------------------
	if ( iRet != OK_SERIALDRIVER_SUCCESS ) {
		memset( &driver->frame.frame, 0, sizeof(driver->frame.frame) );
		return iRet;
	}

	//--------------------------------------------------------------------------
	// try to read length (if first, we will stop here until a byte is received)
	// Read data into next available memory location
	if (driver->params.lengthActivation) {

		if ( i == (bytes_to_read - 1) ) // read length byte of response frame
			iRet = serial_port_recv( driver->port, driver->frame.frame + (++i), 1, &len);
		else
			if ( i == (bytes_to_read - 2) ) i = 0;

		//--------------------------------------------------------------------------
		if ( iRet != OK_SERIALDRIVER_SUCCESS/* || i <= 0 */) {
			memset( &driver->frame.frame, 0, sizeof(driver->frame.frame) );
			return iRet;
		}

		//--------------------------------------------------------------------------
		// receive message body
		int messageBodyLen = driver->frame.frame[i]-1+( driver->params.startEndCharactersActivation ? 1 : 0);
		driver->frame.len += len;
		gettimeofday( &startTime, 0 );

		do { // wait for message body
			iRet = serial_port_check(driver->port, &amountInInQueue, 0);
			if ( iRet == OK_SERIALDRIVER_SUCCESS && amountInInQueue == 0 ) {
				gettimeofday( &endTime, 0 );
				if ( millisecondtimediff( startTime, endTime ) > driver->params.radioAckTimeout )
					iRet = KO_SENDING_FAILURE; // timeout expired
				SLEEP( 10 );
			} else
				gettimeofday( &startTime, 0 );
		} while( iRet == OK_SERIALDRIVER_SUCCESS && amountInInQueue < messageBodyLen );

		//--------------------------------------------------------------------------
		// read response frame with crc + ETX bytes
		if ( iRet == OK_SERIALDRIVER_SUCCESS ) {
			iRet = serial_port_recv(driver->port, driver->frame.frame+(++i), messageBodyLen, &len);
			if ( OK_SERIALDRIVER_SUCCESS == iRet && len == messageBodyLen ) {
				driver->frame.len += messageBodyLen;
				// printf_serial_frame("RECV", driver->params.id, &driver->frame);
				return iRet;
			}
		}

		//----------------------------------------------------------------------
		// fail to read data from port, cleanup garbage
		memset( &driver->frame.frame, 0, sizeof(driver->frame.frame) );
		return iRet;
	}

	//--------------------------------------------------------------------------
	// if length is not specified, consider to read all bytes (as much as
	// there are in RS232 buffer.
	i = len = bytes_to_read = 0;
	driver->frame.len = 0;

	if (driver->params.crc16Activation)
		bytes_to_read 		+= 2;

	if (driver->params.startEndCharactersActivation) {
		driver->frame.len 	+= 1;
		bytes_to_read 		+= 1;
		i += 1;
	}

	if (driver->params.syncCharacterActivation) {
		driver->frame.len 	+= 1;
		i += 1;
	}

	gettimeofday( &startTime, 0 );

	int recv_buffer_timeout = (int)((double)(((double)(MAX_SERIALFRAME_LEN * 8))/driver->params.baudRate ) * 1000);
	// wait for message body
	do {
		gettimeofday( &endTime, 0 );
		if ( millisecondtimediff( startTime, endTime ) > recv_buffer_timeout )
			break;
		SLEEP( 10 );
	} while(1);

	iRet = serial_port_recv( driver->port, driver->frame.frame + i,
			sizeof(driver->frame.frame) - i, &len);

	if (iRet != OK_SERIALDRIVER_SUCCESS) {
		memset( &driver->frame.frame, 0, sizeof(driver->frame.frame) );
		return iRet;
	}

	if ((len < bytes_to_read) || (len > (sizeof (driver->frame.frame) - driver->frame.len))) {
		memset( &driver->frame.frame, 0, sizeof(driver->frame.frame) );
		return KO_SERIALFRAME_REQUIRED;
	}

	driver->frame.len += len;

	//printf_serial_frame("RECV", driver->params.id, &driver->frame);
	return iRet;
}

/**
 * 	this is the serial driver state machine that runs in a separate thread
 * 	and is called for every serial driver instance by passing serial driver
 * 	descriptor.
 *
 *  The state machine processor check the status of passed serial driver
 *  instance and update it's state (like sending, sent, timeout, idle, etc)
 *
 *  The function is thread-safe because of mutexes usage.
 *
 *  @return - a boolean that indicates that something has changed in state
 *  machine, and there are some future processing. If the function returns
 *  false for all driver instances, the state machine thread will go into a
 *  short sleep to reduce the CPU usage.
 */
_bool process_port( DriverInfo* driver ) {
	pthread_mutex_lock( &driver->mutexProcessing );
	_bool processedOk = _false;
	int errorCode = OK_SERIALDRIVER_SUCCESS;
	//SerialFrame frame;
	SerialExchangeData frame;
	unsigned long amountInInQueue = 0;
	struct timeval endTime;
	int min_bytes_required;
	char frameString[MAX_SERIALFRAME_LEN*2+1] = "";

	LOG_DEBUG( g_Logger, "process_port: %s, -----------------------------------------", driver->params.id);

	memset( &driver->frame, 0, sizeof(driver->frame) );
	if ( driver->port != NULL ) {
		errorCode = serial_port_check( driver->port, &amountInInQueue, 0);
	} else {
		LOG_ERROR( g_Logger, "Port is NULL for %s", driver->params.id);
	}

	LOG_DEBUG( g_Logger, "process_port: %s, amountInInQueue = %d", driver->params.id, amountInInQueue);


	switch ( driver->status ) {

	case STATUS_SEND_COMMAND:
		LOG_DEBUG( g_Logger, "process_port: STATUS_SEND_COMMAND", "");
		errorCode = serial_port_send(driver->port, driver->sendFrame.frame, driver->sendFrame.len );
		if ( OK_SERIALDRIVER_SUCCESS == errorCode ) {

			gettimeofday( &driver->startTime, 0 );
			// if radioBoardAckStatus false we not expect acknowledge frame!

			if ( !driver->params.radioBoardAckStatus) {
				driver->status = STATUS_IDLE;
				driver->sendingStatus = OK_SERIALDRIVER_SUCCESS;
				LOG_DEBUG( g_Logger, "process_port: %s - ACK is not required!", driver->params.id);
			} else {
				driver->status = STATUS_WAIT_ACKNOWLEDGE;
			}
		}

		LOG_DEBUG( g_Logger, "process_port: %s - STATUS_SEND_COMMAND: %d", driver->params.id, errorCode );
		break;

	case STATUS_SEND_ACKNOWLEDGE:
		LOG_DEBUG( g_Logger, "process_port: STATUS_SEND_ACKNOWLEDGE", "");
		errorCode = OK_SERIALDRIVER_SUCCESS;

		if ( driver->params.hostAckStatus )
			errorCode = serial_port_send(driver->port, ACKNOWLEDGE, ARRAYSIZE(ACKNOWLEDGE) );

		if ( OK_SERIALDRIVER_SUCCESS == errorCode )
			driver->status = STATUS_IDLE;
		LOG_DEBUG( g_Logger, "process_port: %s - STATUS_SEND_ACKNOWLEDGE: %d", driver->params.id, errorCode );
		break;

	case STATUS_WAIT_RESPONSE:
		LOG_DEBUG( g_Logger, "process_port: STATUS_WAIT_RESPONSE", "");

		errorCode = OK_SERIALDRIVER_SUCCESS;
		if ( amountInInQueue >= 2 ) {
			errorCode = read_data_from_serial_port(driver);
			LOG_DEBUG( g_Logger, "process_port: %s - STATUS_WAIT_RESPONSE: %d", driver->params.id, errorCode );
			if ( OK_SERIALDRIVER_SUCCESS == errorCode ) {
				if ( 0 == memcmp( &driver->frame.frame, ACKNOWLEDGE, sizeof(ACKNOWLEDGE) ) ) {
					processedOk = _true;
					driver->status = STATUS_WAIT_RESPONSE;
					gettimeofday( &driver->startTime, 0 );
				} else if ( 0 == memcmp( &driver->frame.frame, NACKNOWLEDGE, sizeof(NACKNOWLEDGE) ) ) {
					errorCode = KO_SENDING_FAILURE;
				} else if ( 0 == memcmp( &driver->frame.frame, ERRORFRAME, sizeof(ERRORFRAME) ) ) {
					errorCode = KO_SENDING_FAILURE;
				} else {
					if ( driver->receiveFrame ) {
						copy_frame_body( driver, &driver->frame, (SerialFrame*)&frame.serialFrame );
						driver->receiveFrame( driver->paramReceiveFrame, &frame);
					}

					LOG_DEBUG( g_Logger, "RS232Driver recv: %s",
							get_hex_serialframe(frameString, sizeof(frameString), &driver->frame ) );
					processedOk = _true;
					driver->status = STATUS_SEND_ACKNOWLEDGE;
					driver->sendingStatus = OK_SERIALDRIVER_SUCCESS;
				}
			}
		}
		if ( !processedOk ) {

			gettimeofday( &endTime, 0 );
			long msDiff = millisecondtimediff( driver->startTime, endTime );

			LOG_DEBUG( g_Logger, "process_port: %s - STATUS_WAIT_RESPONSE: check timeout (%d)", driver->params.id, msDiff);

			// If the Acknowledge frame is not received by the initiator,
			// it can decide to re-send the frame several times (retry mechanism)

			if ( msDiff > driver->params.radioAckTimeout ) {
				LOG_ERROR( g_Logger, "process_port: %s - STATUS_WAIT_RESPONSE: Timeout expired!", driver->params.id);
				errorCode = KO_SENDING_FAILURE;
			}
		}
		break;

	case STATUS_WAIT_ACKNOWLEDGE:
		LOG_DEBUG( g_Logger, "process_port: STATUS_WAIT_ACKNOWLEDGE", "");

		LOG_DEBUG( g_Logger, "process_port: STATUS_WAIT_ACKNOWLEDGE - amountInInQueue : %d", amountInInQueue);
		if ( amountInInQueue >= 2 ) {
			errorCode = read_data_from_serial_port(driver);
			LOG_DEBUG( g_Logger, "process_port: %s - STATUS_WAIT_ACKNOWLEDGE: %d", driver->params.id, errorCode );
			if ( OK_SERIALDRIVER_SUCCESS == errorCode ) {

				LOG_DEBUG( g_Logger, "process_port: STATUS_WAIT_ACKNOWLEDGE - OK_SERIALDRIVER_SUCCESS == errorCode", "");

				if ( 0 == memcmp( &driver->frame.frame, ACKNOWLEDGE, sizeof(ACKNOWLEDGE) ) ) {
					LOG_DEBUG( g_Logger, "process_port: STATUS_WAIT_ACKNOWLEDGE - ACK Frame received!", "");
					processedOk = _true;
					driver->status = STATUS_IDLE; //STATUS_WAIT_RESPONSE;
					if (driver->sendingStatus == SENDING_STATUS_UNKNOWN)
						driver->sendingStatus = OK_SERIALDRIVER_SUCCESS;

				} else if ( 0 == memcmp( &driver->frame.frame, NACKNOWLEDGE, sizeof(NACKNOWLEDGE) ) ) {
					LOG_DEBUG( g_Logger, "process_port: STATUS_WAIT_ACKNOWLEDGE - NACK Frame received!", "");
					errorCode = KO_SENDING_FAILURE;
				} else if ( 0 == memcmp( &driver->frame.frame, ERRORFRAME, sizeof(ERRORFRAME) ) ) {
					LOG_DEBUG( g_Logger, "process_port: STATUS_WAIT_ACKNOWLEDGE - Error Frame received!", "");
					errorCode = KO_SENDING_FAILURE;
				} else { // unexpected frame received!!!

					LOG_DEBUG( g_Logger, "process_port: STATUS_WAIT_ACKNOWLEDGE - Unexpected Frame received!", "");
					if ( OK_SERIALDRIVER_SUCCESS == copy_frame_body( driver, &driver->frame, (SerialFrame*)&frame.serialFrame  ) ) {
						errorCode = KO_SENDING_FAILURE;
						if ( driver->receiveFrame )
							driver->receiveFrame( driver->paramReceiveFrame, &frame );
					}
					else
						errorCode = KO_SERIALFRAME_REQUIRED;
					if ( driver->params.hostAckStatus )
						serial_port_send( driver->port, ACKNOWLEDGE, ARRAYSIZE(ACKNOWLEDGE) );
				}
				memset( &driver->frame, 0, sizeof(driver->frame) );
			}
			driver->retryCount = driver->params.retryCountDef;
		}

		if ( !processedOk ) {
			// If the Acknowledge frame is not received by the initiator,
			// it can decide to re-send the frame several times (retry mechanism)
			gettimeofday( &endTime, 0 );
			if ( millisecondtimediff( driver->startTime, endTime ) > driver->params.radioAckTimeout ) {
				if ( --driver->retryCount >= 0 )
					driver->status = STATUS_SEND_COMMAND;
				else
					errorCode = KO_SENDING_FAILURE; // device did not acknowledge retryCountDef times!!!
			}

			LOG_DEBUG( g_Logger, "STATUS_WAIT_ACKNOWLEDGE - retryCount : %d", driver->retryCount);

		}
		break;

	case STATUS_IDLE:
		LOG_DEBUG( g_Logger, "process_port: STATUS_IDLE", "");

		min_bytes_required = 1;
		if (driver->params.syncCharacterActivation) {
			min_bytes_required = 2;
		}

		if ( amountInInQueue >= min_bytes_required && driver->params.listeningStatus ) {
			errorCode = read_data_from_serial_port(driver);

			LOG_DEBUG( g_Logger, "process_port: %s - STATUS_IDLE: %d", driver->params.id, errorCode );
			if ( OK_SERIALDRIVER_SUCCESS == errorCode ) {
//prd
				if ( driver->receiveFrame ) {
					copy_frame_body( driver, &driver->frame, (SerialFrame*)&frame.serialFrame);
					driver->receiveFrame( driver->paramReceiveFrame, &frame );
				}

				LOG_DEBUG( g_Logger, "RS232Driver IDLE recv: %s",
						get_hex_serialframe(frameString, sizeof(frameString), &driver->frame ) );
				processedOk = _true;

				// if hostAckStatus is true, send acknowledge after read spontaneous message
				driver->status = driver->params.hostAckStatus ? STATUS_SEND_ACKNOWLEDGE : STATUS_IDLE;
			}
		}
		break;

	default:
		LOG_DEBUG( g_Logger, "process_port: default", "");

		break;
	}

	LOG_DEBUG( g_Logger, "porcess_port: End states processing with processedOk=%d, errorCode=%d", processedOk, errorCode);

	if ( processedOk ) {
		LOG_DEBUG( g_Logger, "porcess_port: -> if ( processedOk )", "");
		SerialFrame* frameToNotify = ( driver->sendFrame.len > 0 ) ? &driver->sendFrame : &driver->frame;
		serial_driver_spy_t* spies = ( driver->sendFrame.len > 0 ) ? driver->spiesSend : driver->spiesRecv;
		int i = 0;
		if ( frameToNotify->len > 0 ) {
			LOG_DEBUG( g_Logger, "porcess_port, processedOk: frame len =%d", frameToNotify->len);
			for( i = 0; i < MAX_SPY_NUM; ++i ) {

				if (  spies[i] ) {
// prd
					LOG_DEBUG( g_Logger, "porcess_port, processedOk: Found spy %d, notify it", i);
					copy_frame_body( driver, frameToNotify, (SerialFrame*)&frame.serialFrame  );
					spies[i]( driver->paramSpies[i], i, &frame );
				}
			}
		}

		LOG_DEBUG( g_Logger, "porcess_port: -> processedOk, finished, clear frame", "");
		memset( frameToNotify, 0, sizeof(*frameToNotify) );
	} else if ( OK_SERIALDRIVER_SUCCESS != errorCode ) {

		LOG_DEBUG( g_Logger, "porcess_port: -> else if ( OK_SERIALDRIVER_SUCCESS != errorCode )", "");
		driver->status = STATUS_IDLE;
		if ( driver->notifyListener ) {
			switch ( errorCode ) {
			case KO_SENDING_FAILURE: break;
			case KO_SERIALFRAME_REQUIRED:
				driver->notifyListener( driver->paramNotify, IGNORED_FRAME_RECEIVED );
				errorCode = KO_SENDING_FAILURE;
				break;
			case KO_SERIALPORT_DISCONNECTED:
				driver->notifyListener( driver->paramNotify, SERIALPORT_DISCONNECTION );
				driver->notifyListener( driver->paramNotify, STOP_LISTENING );
				pthread_mutex_unlock( &driver->mutexProcessing );
				port_close( driver );
				pthread_mutex_lock( &driver->mutexProcessing );
				break;
			default:
				errorCode = KO_SERIALPORT_UNEXPECTED_ERROR;
				break;
			}
		}
		driver->sendingStatus = errorCode;
	}
	pthread_mutex_unlock( &driver->mutexProcessing );
	return processedOk;
}

/*
 * 	serial driver main thread that iterates trough all opened serial driver instances
 * 	and calls state machine processor  each of them.
 */
void* main_thread( void *argument ) {
	LOG_DEBUG( g_Logger, "main_thread start", 0 );

	_bool* work = (_bool*) argument;
	size_t i = 0;

	while ( *work ) {
		_bool isAnyWorkPerformed = _false;
		for ( i = 0; i < MAX_DRIVER_NUM; ++i )
			if ( driver_is_open( (void*)(i+1) ) )
				isAnyWorkPerformed |= process_port( &drivers[i] );
		if ( !isAnyWorkPerformed && *work ) // if no any work performed then reduce CPU load
			SLEEP( listeningPeriod );
	}

	LOG_DEBUG( g_Logger, "main_thread finish", 0 );
	pthread_exit(NULL);
	return NULL;
}
