/*******************************************************************************
 * Copyright (c) 2012 Coronis
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Contributors:
 *     Coronis - initial API and implementation
 *******************************************************************************/

#ifndef _WINDOWS

#include <stdlib.h>
#include <unistd.h>
#include <termios.h>
#include <sys/ioctl.h>

#include "u_serial.h"


int serial_port_open( const char* id, void** port ) {
	int iRet = OK_SERIALDRIVER_SUCCESS;
	struct termios cfg;

	/* Serial Port Handle */
	int fd = open( id, O_RDWR | O_NOCTTY | O_NDELAY);
	if (fd == -1)  {
	   /* Could not open the port. */
		switch( errno ) {
		case EACCES:
			iRet = KO_SERIALDRIVER_ALREADY_OPEN; // WA_ERROR_SERIALDRIVER_ALREADY_OPEN;
			break;
		case ENOENT:
			iRet = KO_PARAM_SERIALPORTID_NOT_EXIST; // WA_ERROR_PORT_ID_NOT_EXIST;
			break;
		default:
			iRet = KO_SERIALPORT_NOT_AVAILABLE; // WA_ERROR_PORT_NOT_AVAILABLE;
			break;
		}
	}  else {
		*port = (void*)fd;
		// default configuration
		tcgetattr( fd, &cfg);
		/* ignore bytes with parity errors , ignore break */
		cfg.c_iflag = IGNPAR | IGNBRK;
		/* Raw output. */
		cfg.c_oflag =0;
		/* set input mode (non-canonical, no echo,...) */
		cfg.c_lflag = 0;
		tcsetattr( fd, TCSANOW, &cfg);
	}

	return iRet;
}

int serial_port_close( const void* port ) {
	int iRet = OK_SERIALDRIVER_SUCCESS;

	close( (int)port );

	return iRet;
}

int serial_port_send( const void* port, const byte* frame, int frameLen ) {
	int iRet = OK_SERIALDRIVER_SUCCESS;

	int n = write( (int)port, frame, frameLen);
	if ( n < 0 )
		iRet = KO_SERIALPORT_UNEXPECTED_ERROR; //WA_ERROR_SERIALPORT_UNEXPECTED_ERROR;

	return iRet;
}

int serial_port_recv( const void* port, byte* frame, int frameLen, int* recvLen ) {
	int iRet = OK_SERIALDRIVER_SUCCESS;

	int n = read( (int)port, frame, frameLen );
	if ( n >= 0 )
		*recvLen = n;
	else
		iRet = KO_SERIALPORT_UNEXPECTED_ERROR; //WA_ERROR_SERIALPORT_UNEXPECTED_ERROR;

	return iRet;
}

int serial_port_check( const void* port, unsigned long* amountInInQueue, unsigned long* amountInOutQueue) {
	int iRet = OK_SERIALDRIVER_SUCCESS;

	int bytes = 0;
	ioctl( (int)port, FIONREAD, &bytes);
	*amountInInQueue = bytes;

	return iRet;
}

int serial_port_set_params( const void* port, int baudRate, int dataBit, int stopBit, int parity, int timeout, int flowControl ) {
	int iRet = OK_SERIALDRIVER_SUCCESS;

	iRet = driver_verify_baudrate( baudRate );
	if ( iRet != OK_SERIALDRIVER_SUCCESS )
		return iRet;
	iRet = driver_verify_databit( dataBit );
	if ( iRet != OK_SERIALDRIVER_SUCCESS )
		return iRet;
	iRet = driver_verify_stopbit( stopBit );
	if ( iRet != OK_SERIALDRIVER_SUCCESS )
		return iRet;
	iRet = driver_verify_parity( parity );
	if ( iRet != OK_SERIALDRIVER_SUCCESS )
		return iRet;

	fcntl( (int)port, F_SETFL, FNDELAY);
	struct termios options;
	tcgetattr( (int)port, &options);
	int unixBaudrate = B9600;
	switch ( baudRate ) {
	case CBR_4800: unixBaudrate = B4800; break;
	case CBR_19200: unixBaudrate = B19200; break;
	case CBR_38400: unixBaudrate = B38400; break;
	case CBR_57600: unixBaudrate = B57600; break;
	case CBR_115200: unixBaudrate = B115200; break;
	case CBR_9600:
	default:
		unixBaudrate = B9600;
		break;
	}
	cfsetispeed( &options, unixBaudrate);
	cfsetospeed( &options, unixBaudrate);

	options.c_cflag |= (CLOCAL | CREAD);

	switch( parity ) {
	case ODDPARITY:
		 options.c_cflag |=(PARENB|PARODD);
		 break;
	case EVENPARITY:
		 options.c_cflag &= (~PARODD);
		 options.c_cflag |= PARENB;
		 break;
	case NOPARITY:
	default:
		options.c_cflag &= ~PARENB;
		break;
	}

	switch (stopBit) {
	case TWOSTOPBITS:
		options.c_cflag |= CSTOPB;
		break;
	case ONESTOPBIT:
	default:
		options.c_cflag &= ~CSTOPB;
		break;
	}

	options.c_cflag &= ~CSIZE;
	int unixDataBit = CS8;
	switch( dataBit ) {
	case DATABITS_FIVE: unixDataBit = CS5; break;
	case DATABITS_SIX: unixDataBit = CS6; break;
	case DATABITS_SEVEN: unixDataBit = CS7; break;
	case DATABITS_EIGHT:
	default:
		unixDataBit = CS8;
		break;
	}
	options.c_cflag |= unixDataBit;

	if( flowControl == FLOWCONTROL_NONE ) {
		options.c_cflag&=(~CRTSCTS);
		options.c_iflag&=(~(IXON|IXOFF|IXANY));
	} else {
		if ( ( flowControl & FLOWCONTROL_RTSCTS_IN ) ||
			 ( flowControl & FLOWCONTROL_RTSCTS_OUT ) ) {
			options.c_cflag |= CRTSCTS;
			options.c_iflag &= (~(IXON|IXOFF|IXANY));
		} else
		if ( ( flowControl & FLOWCONTROL_XONXOFF_IN ) ||
		     ( flowControl & FLOWCONTROL_XONXOFF_OUT ) ) {
			options.c_cflag &= (~CRTSCTS);
			options.c_iflag |= (IXON|IXOFF|IXANY);
		}
	}

	tcsetattr( (int)port, TCSANOW, &options);

	return iRet;
}

#endif /* _WINDOWS */
