/*******************************************************************************
 * Copyright (c) 2012 Coronis
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Contributors:
 *     Coronis - initial API and implementation
 *******************************************************************************/

#ifdef _WINDOWS

#include "w_serial.h"
#include "loader_logger.h"
#include "utils.h"

#define DELTA_EPOCH_IN_MICROSECS  11644473600000000UL

// Macro definition for defining IOCTL and FSCTL function control codes.
#ifndef CTL_CODE
#define CTL_CODE(DeviceType, Function, Method, Access) (((DeviceType) << 16) | ((Access) << 14) | ((Function) << 2) | (Method))
#endif//CTL_CODE

// Define the serial device type value
#ifndef FILE_DEVICE_SERIAL_PORT
#define FILE_DEVICE_SERIAL_PORT         0x0000001b
#endif//FILE_DEVICE_SERIAL_PORT

// Define the method codes for how buffers are passed for I/O and FS controls
#ifndef METHOD_BUFFERED
#define METHOD_BUFFERED                 0
#endif//METHOD_BUFFERED

// Define the access check value for any access
#ifndef FILE_ANY_ACCESS
#define FILE_ANY_ACCESS                 0
#endif//FILE_ANY_ACCESS

#ifndef IOCTL_SERIAL_GET_COMMSTATUS
#define IOCTL_SERIAL_GET_COMMSTATUS     CTL_CODE(FILE_DEVICE_SERIAL_PORT,27,METHOD_BUFFERED,FILE_ANY_ACCESS)
#endif//IOCTL_SERIAL_GET_COMMSTATUS

//
// This structure is used to get the current error and
// general status of the driver.
//
typedef struct _SERIAL_STATUS {
	ULONG Errors;
	ULONG HoldReasons;
	ULONG AmountInInQueue;
	ULONG AmountInOutQueue;
	BOOLEAN EofReceived;
	BOOLEAN WaitForImmediate;
} SERIAL_STATUS;

int serial_port_open( const char* id, void** port ) {
	int iRet = OK_SERIALDRIVER_SUCCESS;
	DWORD lastError = ERROR_SUCCESS;
	char fullId[100] = "";
	sprintf( fullId, "\\\\.\\%s", id );

	LOG_DEBUG(g_Logger, "serial_port_open: enter, open device", 0);

	// Open the device
	HANDLE hComm = CreateFileA( fullId, GENERIC_READ|GENERIC_WRITE, 0, 0, OPEN_EXISTING, 0, 0);

	LOG_DEBUG(g_Logger, "serial_port_open: enter, Handle created, check it's value %X", hComm);
	if ( hComm != INVALID_HANDLE_VALUE )
		*port = hComm;
	else { // ( hComm != INVALID_HANDLE_VALUE)
		lastError = GetLastError();
		LOG_ERROR(g_Logger, "serial_port_open: enter, Failed to create handle with error code %d", lastError);
	}

	switch( lastError ) {
	case ERROR_SUCCESS: iRet = OK_SERIALDRIVER_SUCCESS; break;
	case ERROR_SEM_IS_SET:
	case ERROR_FILE_NOT_FOUND: iRet = KO_PARAM_SERIALPORTID_NOT_EXIST; break;
	case ERROR_TOO_MANY_SEMAPHORES:
	case ERROR_ACCESS_DENIED: iRet = KO_SERIALDRIVER_ALREADY_OPEN; break;
	default: iRet = KO_SERIALPORT_NOT_AVAILABLE; break;
	}

	LOG_DEBUG(g_Logger, "serial_port_open: Exiting with code %d", iRet);
	return iRet;
}

int serial_port_close( const void* port ) {
	int iRet = OK_SERIALDRIVER_SUCCESS;

	LOG_DEBUG( g_Logger, "serial_port_close WIN Close serial port ... ", 0 );

	EscapeCommFunction( (HANDLE)port, CLRRTS );
	EscapeCommFunction( (HANDLE)port, CLRDTR );

	LOG_DEBUG( g_Logger, "serial_port_close WIN purge port ", 0 );
	PurgeComm( (HANDLE)port, PURGE_TXABORT | PURGE_TXCLEAR | PURGE_RXABORT | PURGE_RXCLEAR );
	LOG_DEBUG( g_Logger, "serial_port_close close handle ", 0 );

	if ( !CloseHandle( (HANDLE)port ) )

		LOG_DEBUG( g_Logger, "serial_port_close Get error code ", 0 );

		switch( GetLastError() ) {
		case ERROR_SUCCESS: iRet = OK_SERIALDRIVER_SUCCESS; break;
		default: iRet = KO_SERIALPORT_UNEXPECTED_ERROR; break;
		}

	LOG_DEBUG( g_Logger, "serial_port_close Done", 0 );
	return iRet;
}

int serial_port_send( const void* port, const byte* frame, int frameLen ) {
	int iRet = OK_SERIALDRIVER_SUCCESS;
	DWORD bytesWritten = 0;

	PurgeComm( (HANDLE)port, PURGE_RXCLEAR | PURGE_TXCLEAR | PURGE_RXABORT | PURGE_TXABORT );
	if ( !WriteFile( (HANDLE)port, frame, frameLen, &bytesWritten, NULL ) )
	{
		switch( GetLastError() ) {
		case ERROR_SUCCESS:       iRet = OK_SERIALDRIVER_SUCCESS; break;
		case ERROR_ACCESS_DENIED: iRet = KO_SERIALPORT_DISCONNECTED; break;
		default:  			      iRet = KO_SERIALPORT_UNEXPECTED_ERROR; break;
		}
	}
	return iRet;
}

int serial_port_recv( const void* port, byte* frame, int frameLen, unsigned int* recvLen ) {
	int iRet = OK_SERIALDRIVER_SUCCESS;
	DWORD bytesRead = 0;

	if ( !ReadFile( (HANDLE)port, frame, frameLen, &bytesRead, NULL ) )
		switch( GetLastError() ) {
		case ERROR_SUCCESS:       iRet = OK_SERIALDRIVER_SUCCESS; break;
		case ERROR_ACCESS_DENIED: iRet = KO_SERIALPORT_DISCONNECTED; break;
		default:  			      iRet = KO_SERIALPORT_UNEXPECTED_ERROR; break;
		}

	if ( recvLen )
		*recvLen = bytesRead;

	return iRet;
}

int serial_port_check( const void* port, unsigned long* amountInInQueue, unsigned long* amountInOutQueue) {

	int iRet = OK_SERIALDRIVER_SUCCESS;
	DWORD dwBytesReturned = 0;
	SERIAL_STATUS serialStatus = {0};

	if ( !DeviceIoControl( (HANDLE)port, IOCTL_SERIAL_GET_COMMSTATUS,
			NULL, 0, &serialStatus, sizeof(serialStatus), &dwBytesReturned, NULL) ) {

		switch( GetLastError() ) {
		case ERROR_SUCCESS:       iRet = OK_SERIALDRIVER_SUCCESS; break;
		case ERROR_ACCESS_DENIED: iRet = KO_SERIALPORT_DISCONNECTED; break;
		default:                  iRet = KO_SERIALPORT_UNEXPECTED_ERROR; break;
		}
	}

	if ( amountInInQueue )
		*amountInInQueue = serialStatus.AmountInInQueue;
	if ( amountInOutQueue )
		*amountInOutQueue = serialStatus.AmountInOutQueue;

	return iRet;
}

int serial_port_set_params( const void* port, int baudRate, int dataBit, int stopBit, int parity, int timeout, int flowControl ) {
	int iRet = OK_SERIALDRIVER_SUCCESS;

	iRet = driver_verify_baudrate( baudRate );
	if ( iRet != OK_SERIALDRIVER_SUCCESS )
		return iRet;
	iRet = driver_verify_databit( dataBit );
	if ( iRet != OK_SERIALDRIVER_SUCCESS )
		return iRet;
	iRet = driver_verify_stopbit( stopBit );
	if ( iRet != OK_SERIALDRIVER_SUCCESS )
		return iRet;
	iRet = driver_verify_parity( parity );
	if ( iRet != OK_SERIALDRIVER_SUCCESS )
		return iRet;


	DCB dcbSerial = {0};
	dcbSerial.DCBlength = sizeof(dcbSerial);
	dcbSerial.BaudRate = baudRate;

	dcbSerial.ByteSize = dataBit;
	dcbSerial.StopBits = stopBit;
	dcbSerial.Parity = parity;

	dcbSerial.EofChar = 0x00;
	dcbSerial.ErrorChar = 0x00;
	dcbSerial.EvtChar = 0x00;
	dcbSerial.XonChar = 0x11;
	dcbSerial.XoffChar = 0x13;

	dcbSerial.XonLim = 64;
	dcbSerial.XoffLim = 64;

	COMMTIMEOUTS timeouts = {0};
	timeouts.ReadIntervalTimeout = -1;
	timeouts.ReadTotalTimeoutMultiplier = 0;
	timeouts.ReadTotalTimeoutConstant = timeout;
	timeouts.WriteTotalTimeoutMultiplier = 0;
	timeouts.WriteTotalTimeoutConstant = timeout;

	if ( SetCommState( (HANDLE)port, &dcbSerial ) ) {
		SetupComm( (HANDLE)port, MAX_PATH, MAX_PATH );
		SetCommTimeouts( (HANDLE)port, &timeouts);
		PurgeComm( (HANDLE)port, PURGE_RXCLEAR | PURGE_TXCLEAR | PURGE_RXABORT | PURGE_TXABORT );
	}

	return iRet;
}
/*
int gettimeofday(struct timeval *tv, struct timezone *tz)
{
	// Define a structure to receive the current Windows filetime
	FILETIME ft;

	// Initialize the present time to 0 and the timezone to UTC
	unsigned __int64 tmpres = 0;
	static int tzflag = 0;

	if (NULL != tv)
	{
		GetSystemTimeAsFileTime(&ft);

		// The GetSystemTimeAsFileTime returns the number of 100 nanosecond
		// intervals since Jan 1, 1601 in a structure. Copy the high bits to
		// the 64 bit tmpres, shift it left by 32 then or in the low 32 bits.
		tmpres |= ft.dwHighDateTime;
		tmpres <<= 32;
		tmpres |= ft.dwLowDateTime;

		// Convert to microseconds by dividing by 10
		tmpres /= 10;

		// The Unix epoch starts on Jan 1 1970.  Need to subtract the difference
		// in seconds from Jan 1 1601.
		tmpres -= DELTA_EPOCH_IN_MICROSECS;

		// Finally change microseconds to seconds and place in the seconds value.
		// The modulus picks up the microseconds.
		tv->tv_sec = (long)(tmpres / 1000000UL);
		tv->tv_usec = (long)(tmpres % 1000000UL);
	}

	if (NULL != tz)
	{
		if (!tzflag)
		{
			_tzset();
			tzflag++;
		}

		// Adjust for the timezone west of Greenwich
		tz->tz_minuteswest = _timezone / 60;
		tz->tz_dsttime = _daylight;
	}

	return 0;
}
*/
#endif /* _WINDOWS */
