/*******************************************************************************
 * Copyright (c) 2012 Coronis
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Contributors:
 *     Coronis - initial API and implementation
 *******************************************************************************/

/**
 * @file protocol_params_types.h
 * @brief Definition of radio parameters structures used by the Protocol service
 * to indicate the radio setting used by the connected radio modem
 */
#ifndef PROTOCOL_PARAMS_TYPES_H_
#define PROTOCOL_PARAMS_TYPES_H_

#include <protocol_constants.h>
#include <i_protocol_types.h>

/**
 * Structure that contains the dynamic standard radio parameters
 * relative to a radio settings command
 * This structure is used by the Protocol service to indicate
 * the dynamic standard radio parameters used by the connected radio modem
 *
 * Note: These parameters could be write dynamically
 * at the same time a radio frame is sent/received.
 * This parameters are not persisted by the Radio modem
 */
typedef struct tagIDynamicStandardParams {
	/**
	 * Radio address of the device to which the DynamicStandardParams is associated
	 * (sent or received RadioExchangedCommand).
	 * Can not be <code>empty</code> when accessed from a RadioExchangedCommand
	 * extract from a UNICAST radio command
	 */
	pIRadioAddress nodeAddress;

	/**
	 * Radio addresses of the router path (radio relay) used to reach the device
	 * to which the DynamicStandardParams is associated
	 * (sent or received RadioExchangedCommand).
	 * Can be <code>empty</code> if no router path is used
	 * The router path is ordering from the Radio Modem point of view
	 * it means that the first radio address must correspond to the device
	 * which is acccessible by the connection Radio modem.
	 */
	 pIRadioAddress routerPath[MAX_ROUTER_PATH];

	/**
	 * Group Id (decimal format) of the devices to which
	 * the DynamicStandardParams is associated (sent or received RadioExchangedCommand).
	 * Can not be <code>null</code> when accessed from a RadioExchangedCommand
	 * extract from a MULTICAST radio command
	 * <code>-1</code> means that this parameter is not initialized
	 */
	int groupId;

} DynamicStandardParams;

/**
 * Structure that contains the dynamic expert radio parameters
 * relative to a radio settings command
 * This structure is used by the Protocol service to indicate
 * the dynamic expert radio parameters used by the connected radio modem
 *
 * Note: These parameters could be write dynamically
 * at the same time a radio frame is sent/received.
 * This parameters are persisted by the Radio modem
 *
 * Important: The support of pseudo static parameters is dependent from the connected radio modem.
 * Refer the radio modem documentation to have more information on supported parameters and values
 */
typedef struct tagIDynamicExpertParams {
	/*
	 * Set flag values
	 * FLAG_WAKEUP_LENGTH
	 * FLAG_TRANSMISSION_POWER_LEVEL
	 * FLAG_TRANSMISSION_RADIO_CHANNEL
	 * FLAG_RESPONSE_TIMEOUT
	 * FLAG_BROADCAST_SESSION_MAX_RESPONSES_NB
	 * FLAG_BROADCAST_SESSION_TIMEOUT
	 */
	uint64 isSet;

	/**
	 * Length of the wake up preamble sent (ms).
	 * Used for all radio exchanges
	 */
	unsigned int wakeupLength;

	/**
	 * Reception mode of the radio communication.
	 * ReadOnly parameters (received command)
	 */
	ReceptionMode receptionMode;

	/**
	 * Quality of service (RSSI) of the receive frame (dB)
	 * ReadOnly parameters (received command)
	 */
	unsigned int rssi;

	/**
	 * Radio channel used for the radio transmission
	 * Used for all radio exchanges
	 */
	RadioChannel transmissionRadioChannel;

	/**
	 * Response reception timeout when waiting for an radio response (ms)
	 * Used only for UNICAST radio request
	 */
	unsigned int responseTimeout;

	/**
	 * Session timeout for waiting responses after a broadcast or multicast request
	 * Used only for MULTICAST or BROADCAST radio request
	 */
	unsigned int broadcastSessionTimeout;

	/**
	 * Maximum number of responses accepted for broadcast or multicast request
	 * Used only for MULTICAST or BROADCAST radio request
	 */
	unsigned int broadcastSessionMaxResponsesNb;

	/**
	 * Power level for the radio transmission (dB)
	 * Used for all radio exchanges
	 */
	float transmissionPowerLevel;

} DynamicExpertParams;

/**
 * Structure that contains the static standard radio parameters
 * relative to a radio settings command
 * This structure is used by the Protocol service to indicate
 * the static standard radio parameters used by the connected radio modem
 *
 * Note: These parameters could be write only separately
 * from the send of radio frame, with IProtocolService.writeRadioParams method.
 * This parameters are persisted by the Radio modem
 *
 * Important: The support of static parameters is dependent from the connected radio modem.
 * Refer the radio modem documentation to have more information on supported parameters and values.
 * If you don't know the precise signification of one of them, don't change it!
 */
typedef struct tagIStaticStandardParams {
	/*
	 * Set flag values
	 * FLAG_RADIO_ACK_ACTIVATION
	 * FLAG_FIRST_SCAN_ENABLE_ACTIVATION
	 * FLAG_ACTIVITY_PERIOD_ON_FIRST_SCAN
	 * FLAG_SWITCHING_MODE_ON_RADIO_ADDRESS_ACTIVATION
	 * FLAG_RECEPTION_GROUPID
	 */
	uint64 isSet;

	/**
	 * Radio acknowledgment activation
	 * Used for radio transmission
	 */
	_bool radioAckActivation;

	/**
	 * First radio scan activation (MAC layer)
	 * Used for radio reception
	 */
	_bool firstScanEnableActivation;

	/**
	 * First radio scan activity period (ms)
	 * Used for radio reception
	 */
	unsigned int firstScanActivityPeriod;

	/**
	 * Double radio scan activation in dependence with radio address
	 * Used for radio reception
	 */
	_bool switchingModeOnRadioAddressActivation;

	/**
	 * Group ID of connected Radio modem for MULTICAST radio request
	 * receive by the Radio modem (decimal format)
	 * Used for radio reception
	 */
	unsigned int receptionGroupId;

} StaticStandardParams;

/**
 * Structure that contains the static expert radio parameters
 * relative to a radio settings command
 * This structure is used by the Protocol service to indicate
 * the static expert radio parameters used by the connected radio modem
 *
 * Note: These parameters could be write only separately
 * from the send of radio frame, with IProtocolService.writeRadioParams method;
 * This parameters are persisted by the Radio modem
 *
 * Important: The support of static parameters is dependent from the connected radio modem.
 * Refer the radio modem documentation to have more information on supported parameters and values.
 * If you don't know the precise signification of one of them, don't change it!
 */
typedef struct tagIStaticExpertParams {
	/*
	 * Set flag values
	 * FLAG_SECOND_SCAN_ENABLE_ACTIVATION
	 * FLAG_COMPLEX_SCAN_ENABLE_ACTIVATION
	 * FLAG_RADIO_CRC_ACTIVATION
	 * FLAG_CWUP_TRANSMISSION_ACTIVATION
	 * FLAG_CWUP_AUTOSELECTION_ACTIVATION
	 * FLAG_CWUP_REQUEST_ONLY_ACTIVATION
	 * FLAG_CWUP_ONLY_ACTIVATION
	 * FLAG_PHYSICAL_CONFIG_ON_FIRST_SCAN
	 * FLAG_RADIO_CHANNEL_ON_FIRST_SCAN
	 * FLAG_SENSITIVITY_REDUCTION_FACTOR_ON_FIRST_SCAN
	 * FLAG_ACTIVITY_PERIOD_ON_SECOND_SCAN
	 * FLAG_PHYSICAL_CONFIG_ON_SECOND_SCAN
	 * FLAG_RADIO_CHANNEL_ON_SECOND_SCAN
	 * FLAG_SENSITIVITY_REDUCTION_FACTOR_ON_SECOND_SCAN
	 * FLAG_NB_SCAN_ON
	 * FLAG_NB_SCAN_OFF
	 * FLAG_FREQUENCY_HOPPING_CHANNEL_TABLE
	 */
	uint64 isSet;

	/**
	 * Second radio scan activation (MAC layer)
	 * Used for radio reception
	 */
	_bool secondScanEnableActivation;

	/**
	 * Complex scan activation
	 * Used for radio reception
	 */
	_bool complexScanEnableActivation;

	/**
	 * Radio CRC activation
	 * Used for all radio exchanges
	 */
	_bool radioCRCActivation;

	/**
	 * Complex Wakeup transmission activation
	 * Used for radio transmission
	 */
	_bool cwupTransmissionActivation;

	/**
	 * Complex Wakeup auto selection activation
	 * Used for radio transmission
	 */
	_bool cwupAutoSelectionActivation;

	/**
	 * Complex Wakeup only for request activation
	 * Used for radio transmission
	 */
	_bool cwupRequestOnlyActivation;

	/**
	 * Only Complex Wakeup activation
	 * Used for radio transmission
	 */
	_bool cwupOnlyActivation;

	/**
	 * Physical layer selection for first radio scan
	 * Used for radio reception
	 */
	PhysicalLayer firstScanPhysicalConfig;

	/**
	 * Radio channel used for the first radio scan
	 * Used for radio reception
	 */
	RadioChannel firstScanRadioChannel;

	/**
	 * Radio sensitivity reduction factor used for the first radio scan (dB)
	 * Used for radio reception
	 */
	unsigned int firstScanSensitivityReductionFactor;

	/**
	 * Second radio scan activity period (ms)
	 * Used for radio reception
	 */
	unsigned int secondScanActivityPeriod;

	/**
	 * Physical layer selection for second radio scan
	 * Used for radio reception
	 */
	PhysicalLayer secondScanPhysicalConfig;

	/**
	 * Radio channel used for the second radio scan
	 * Used for radio reception
	 */
	RadioChannel secondScanRadioChannel;

	/**
	 * Radio sensitivity reduction factor used for the second radio scan (dB)
	 * Used for radio reception
	 */
	unsigned int secondScanSensitivityReductionFactor;

	/**
	 * Number of scan ON (decimal value)
	 * Used for radio reception
	 */
	unsigned int nbScanOn;

	/**
	 * Number of scan OFF (decimal value)
	 * Used for radio reception
	 */
	unsigned int nbScanOff;

	/**
	 * Radio Frequency Channel Hopping table (hexadecimal value)
	 * Used for all radio exchanges
	 */
	char frequencyHoppingChannelTable[MAX_FREQUENCY_HOPPING_CHANNEL_TABLE_LENGTH];
} StaticExpertParams;


#endif /* PROTOCOL_PARAMS_TYPES_H_ */
