/*******************************************************************************
 * Copyright (c) 2012 Coronis
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Contributors:
 *     Coronis - initial API and implementation
 *******************************************************************************/
#include <generic_types.h>

#include "logger.h"
#include "utils.h"
#include "../include/property_reader.h"
#include <dirent.h>

#define LOGGER_FILE_NAME "csLogger.properties"
#define LOGGER_NAME      ".loggerName"
#define LOGGER_LEVEL     ".loggerLevel"
#define LOGGER_MAXSIZE   ".loggerMaxSize"


unsigned int get_next_file_index (const char *log_file_name);

pthread_mutex_t mutexLogUsage;
LogInfo linfo;

static char logBuffer[LOG_MAX_BUFFER_SIZE] = "";

void init_log_from_property_file( const char* propertyFile ) {
	if ( access( propertyFile, 0 ) == -1 ) {
		if ( get_module_path( g_ModuleHandle, logBuffer, sizeof(logBuffer) ) )
			strcat( logBuffer, propertyFile );
	} else
		strcpy( logBuffer, propertyFile );

	int fd = open( logBuffer, O_RDONLY, 0);
	memset( logBuffer, 0, sizeof(logBuffer) );
	if ( fd != 0 ) {
		read( fd, logBuffer, sizeof(logBuffer) );
		close( fd );
	}

	if ( linfo.fd != 0 ) close( linfo.fd );
	memset( &linfo, 0, sizeof(linfo) );

	char* buff = (char*)get_property( logBuffer, LOGGER_NAME );
	if ( buff ) {
		strcpy( linfo.logfile, buff );
		linfo.level = LL_FATAL;
		buff = (char*)get_property( logBuffer, LOGGER_LEVEL );
		if ( buff ) {
			if ( 0 == strcmp( "debug", buff ) ||
				 0 == strcmp( "all", buff ) )
				linfo.level = LL_DEBUG;
			else if ( 0 == strcmp( "info", buff ) )
				linfo.level = LL_INFO;
			else if ( 0 == strcmp( "warn", buff ) )
				linfo.level = LL_WARN;
			else if ( 0 == strcmp( "error", buff ) )
				linfo.level = LL_ERROR;
		}

		buff = (char*)get_property( logBuffer, LOGGER_MAXSIZE );
		if ( buff )
			linfo.maxFileSize = atol(buff)*1048576;

		linfo.fd = open( linfo.logfile, O_RDWR | O_APPEND, 0);
		if ( linfo.fd <= 0 ) {
			linfo.fd = creat( linfo.logfile, S_IRWXU );
			linfo.currentSize = 0;
			linfo.offset = 0;
		}
		else {
			struct stat st;
			stat( linfo.logfile, &st );
			linfo.offset = st.st_size;
			linfo.currentSize = linfo.offset;
		}

		debug( "DOMAIN LOGGER LOGGING STARTED" );
	}
}

void log_msg( const char* level) {

	if (linfo.fd == 0) return;

	// 0. write message
	char timestamp[124] = "";
	struct timeval currentTime;
	gettimeofday( &currentTime, 0 );
	struct tm *tm = localtime(&currentTime.tv_sec);
	sprintf( timestamp, "[%02d:%02d:%02d:%03d] ", tm->tm_hour, tm->tm_min, tm->tm_sec, (int)currentTime.tv_usec / 1000 );
	linfo.currentSize += write( linfo.fd, timestamp, strlen(timestamp) );
	linfo.currentSize += write( linfo.fd, level, strlen(level) );
	linfo.currentSize += write( linfo.fd, logBuffer, strlen(logBuffer) );
	char ch = '\n';
	linfo.currentSize += write( linfo.fd, &ch, 1 );

	// check if split is needed
	if ( linfo.maxFileSize <= 0 || linfo.currentSize < linfo.maxFileSize )
		return;

	int fileIndex = get_next_file_index (linfo.logfile) ;

	// 1. close and  rename currently opened file.
	close(linfo.fd);
	linfo.fd = 0;

	char new_file_name[MAX_PATH] = "";
	char file_ext[50] = "";
	sprintf(file_ext, ".%d", fileIndex);
	strcpy(new_file_name, linfo.logfile);
	strncat(new_file_name, file_ext, MAX_PATH);
	rename(linfo.logfile, new_file_name);

	// 3. create new file with the same name, and start writing to this file
	linfo.fd = creat( linfo.logfile, S_IRWXU );
	linfo.currentSize = 0;
	linfo.offset = 0;
}

void init_log() {

	init_log_from_property_file( LOGGER_FILE_NAME );
}

void debug( const char* format, ... ) {
	if ( linfo.fd > 0 && linfo.level <= LL_DEBUG ) {

		// print the trace message to buffer
#ifdef _WINDOWS
		if ( mutexLogUsage )
#endif//_WINDOWS
			pthread_mutex_lock( &mutexLogUsage );

		va_list args = { 0 };
		va_start(args, format);
		vsnprintf( logBuffer, LOG_MAX_BUFFER_SIZE, format, args);
		va_end(args);

		log_msg( "DEBUG ");
#ifdef _WINDOWS
		if ( mutexLogUsage )
#endif//_WINDOWS
			pthread_mutex_unlock( &mutexLogUsage );
	}
}

void info( const char* format, ... ) {
	if ( linfo.fd > 0 && linfo.level <= LL_INFO ) {
		// print the trace message to buffer
#ifdef _WINDOWS
		if ( mutexLogUsage )
#endif//_WINDOWS
			pthread_mutex_lock( &mutexLogUsage );

		va_list args = { 0 };
		va_start(args, format);
		vsnprintf( logBuffer, LOG_MAX_BUFFER_SIZE, format, args);
		va_end(args);

		log_msg( "INFO  ");
#ifdef _WINDOWS
		if ( mutexLogUsage )
#endif//_WINDOWS
			pthread_mutex_unlock( &mutexLogUsage );
	}
}

void warn( const char* format, ... ) {
	if ( linfo.fd > 0 && linfo.level <= LL_WARN ) {
		// print the trace message to buffer
#ifdef _WINDOWS
		if ( mutexLogUsage )
#endif//_WINDOWS
			pthread_mutex_lock( &mutexLogUsage );

		va_list args = { 0 };
		va_start(args, format);
		vsnprintf( logBuffer, LOG_MAX_BUFFER_SIZE, format, args);
		va_end(args);

		log_msg( "WARN  ");
#ifdef _WINDOWS
		if ( mutexLogUsage )
#endif//_WINDOWS
			pthread_mutex_unlock( &mutexLogUsage );
	}
}

void error( const char* format, ... ) {
	if ( linfo.fd > 0 && linfo.level <= LL_ERROR ) {
		// print the trace message to buffer
#ifdef _WINDOWS
		if ( mutexLogUsage )
#endif//_WINDOWS
			pthread_mutex_lock( &mutexLogUsage );

		va_list args = { 0 };
		va_start(args, format);
		vsnprintf( logBuffer, LOG_MAX_BUFFER_SIZE, format, args);
		va_end(args);

		log_msg( "ERROR ");
#ifdef _WINDOWS
		if ( mutexLogUsage )
#endif//_WINDOWS
			pthread_mutex_unlock( &mutexLogUsage );
	}
}

void fatal( const char* format, ... ) {
	if ( linfo.fd > 0 && linfo.level <= LL_FATAL ) {
		// print the trace message to buffer
#ifdef _WINDOWS
		if ( mutexLogUsage )
#endif//_WINDOWS
			pthread_mutex_lock( &mutexLogUsage );

		va_list args = { 0 };
		va_start(args, format);
		vsnprintf( logBuffer, LOG_MAX_BUFFER_SIZE, format, args);
		va_end(args);

		log_msg( "FATAL ");
#ifdef _WINDOWS
		if ( mutexLogUsage )
#endif//_WINDOWS
			pthread_mutex_unlock( &mutexLogUsage );
	}
}

unsigned int get_next_file_index (const char *log_file_name) {

	char *ptr = strrchr(log_file_name, '/');
	char path[256] = "./";

	if (ptr != NULL && *ptr != 0 && (ptr > log_file_name))
		strncpy(path, log_file_name, ptr - log_file_name);

	DIR *mydir = opendir(path);
	struct dirent *entry = NULL;
	unsigned int fileIndex = 0;

	if (mydir == NULL) return 0;

	while((entry = readdir(mydir))) {

		ptr = strrchr(entry->d_name, '.');
		if (ptr != NULL && ptr > entry->d_name) {
			ptr += 1;

			unsigned int number = atoi(ptr);
			if (number > fileIndex) {
				fileIndex  = number;
			}
		}
	}

	closedir(mydir);
	return fileIndex  + 1;
}
