/*******************************************************************************
 * Copyright (c) 2012 Coronis
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Contributors:
 *     Coronis - initial API and implementation
 *******************************************************************************/


#include <serial_driver_event.h>
#include <loader_logger.h>

#include "protocol.h"
#include "protocol_types.h"
#include "protocol_types_utils.h"
#include "serial_driver_types.h"
#include "serial_driver_types_utils.h"
#include "logger_utils.h"
#include "serial_driver_instance.h"
#include "utils.h"

#include <loader_protocol.h>

#define PROTOCOL_INTEGRITY 0xFEDCCDEF

#ifndef CORONIS_SDK_PROTOCOL_CONSTANTS_V_1_0
	#error "Please make sure that right protocol_types.h is included in project. Currently is supported v1.0"
#endif // CORONIS_SDK_PROTOCOL_CONSTANTS_V_1_0


ProtocolInfo g_WaveProtocol[MAX_PROTOCOL_NUM];

/****************************************************************************************************/
int protocol_get_version() {

	return ( ((PROTOCOL_API_VERSION << 16) & 0xFF0000) |
			 ((PROTOCOL_MAX_VERSION << 8) & 0x00FF00) |
			 ((PROTOCOL_MIN_VERSION << 0) & 0x0000FF));
}


/****************************************************************************************************/
void* protocol_create() {
	ProtocolInfo* waveProtocol = NULL;
	size_t i = 0;

	for ( ; i < ARRAYSIZE(g_WaveProtocol); ++i ) {
		if ( g_WaveProtocol[i].integrity != PROTOCOL_INTEGRITY ) {
			waveProtocol = &g_WaveProtocol[i];
			memset( waveProtocol, 0, sizeof(*waveProtocol) );
			waveProtocol->integrity = PROTOCOL_INTEGRITY;
			break;
		}
	}

	return (void*)waveProtocol;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/****************************************************************************************************/
int protocol_destroy( const void* protocol ) {
	ProtocolInfo* waveProtocol = get_protocol( protocol );
	if ( waveProtocol ) {
		if ( waveProtocol->serialDriver )
			waveProtocol->serialDriverHandler.destroy( waveProtocol->serialDriver );
		free_serial_driver( &waveProtocol->serialDriverHandler );
		memset( waveProtocol, 0, sizeof(*waveProtocol) );
		return OK_PROTOCOL_SUCCESS;
	}
	return KO_PROTOCOL_NO_VALID_PARAMETER;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


/****************************************************************************************************/
int protocol_is_open( const void* protocol ) {

	ProtocolInfo* waveProtocol = get_protocol( protocol );
	if ( waveProtocol ) {
		if ( waveProtocol->serialDriver ) {
			LOG_DEBUG( g_Logger, "protocol_is_open: Check if RS232 driver is opened", 0);
			return waveProtocol->serialDriverHandler.is_open( waveProtocol->serialDriver );
		}
	}

	return 0;
}


/****************************************************************************************************/
int protocol_init_from_file( const void* protocol, const char* file ) {

	ProtocolInfo* waveProtocol = get_protocol( protocol );

	int iRet = KO_PROTOCOL_NO_VALID_PARAMETER;
	char data[10340] = "";

	//char buff[20240] = "";
	//int fdlog = creat( "d:\\logfile.protocol_init_from_file.txt", _O_RDWR);
	//snprintf(buff, sizeof(buff), "Init entered ...\n");
	//write (fdlog, buff, strlen(buff));
	// close(fd);

	if ( !waveProtocol || !file || strlen(file) <= 3 ) {
		return iRet;
	}

	// 1. Protocol is already open ?
	if ( protocol_is_open( protocol ) ) {
		return iRet = KO_PROTOCOL_ALREADY_OPEN;
	}

	// read from specified properties file
	int fReadCount  = read_properties_file(g_ModuleHandle, data, sizeof(data), file);

	// Check if property file was valid, if not, read from default
	// properties file name
	if (fReadCount <= 0) {
		read_properties_file(g_ModuleHandle, data, sizeof(data), WAVEPORT_PROTOCOL_PROPERTIES_FILE);
	}

	// the result of  last read_properties_file is not verified, because the
	// function get_all_properties that follows will read default values
	// in case if no property is found.
	iRet = OK_PROTOCOL_SUCCESS;

	ProtocolParams* params = &waveProtocol->params;
	get_all_properties( data, params );

	// initialize logger
	if ( g_Logger.loggerHandle == NULL ) {
		if ( OK_PROTOCOL_SUCCESS == load_logger( DOMAIN_NAME, &g_Logger) ) {

			if (strlen(g_loger_protocol_prop_file_name) == 0)
				strcpy(g_loger_protocol_prop_file_name, "cslogger.properties");

			g_Logger.init_log_from_property_file(g_loger_protocol_prop_file_name);
			LOG_DEBUG( g_Logger, "WAVEPORT LOGGING STARTED", 0);
		}
	}

	LOG_DEBUG( g_Logger, "protocol_init_from_file :::: %d %s", waveProtocol->serialDriverHandler.serialDriverHandle,
			waveProtocol->params.serialDriverId);

	if ( waveProtocol->serialDriverHandler.serialDriverHandle == 0 &&
			waveProtocol->params.isSet & FLAG_SERIAL_DRIVER_ID &&
			waveProtocol->serialDriver == 0 ) {

		char serialDriverId[MAX_PATH] = "";
		strcpy( serialDriverId, waveProtocol->params.serialDriverId);
		waveProtocol->params.serialDriverId[0] = 0;

		// set serial driver id
		iRet = protocol_set_serial_driver_id( protocol, serialDriverId);
	}

	// initialize serial driver default instance
	if ( iRet == OK_PROTOCOL_SUCCESS && waveProtocol->serialDriverHandler.init)
	{
		iRet = waveProtocol->serialDriverHandler.init( waveProtocol->serialDriver );
	}

	return iRet;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/****************************************************************************************************/
int protocol_init( const void* protocol ) {

	return protocol_init_from_file( protocol, WAVEPORT_PROTOCOL_PROPERTIES_FILE );
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/****************************************************************************************************/
int  protocol_write_radio_params( const void* protocol, const RadioSettingsCommand* command, CommandResponse* response ) {
	ProtocolInfo* waveProtocol = get_protocol( protocol );

	if (! waveProtocol || !response || !command ) {
		return KO_PROTOCOL_NO_VALID_PARAMETER;
	}

	LOG_DEBUG( g_Logger, "protocol_write_radio_params: Entered", 0);

	memset( response, 0,  sizeof(*response) );

	// 0. Is protocol opened ?
	if ( !protocol_is_open( waveProtocol ) )
		return (response->status = KO_PROTOCOL_NOT_OPEN);

	// 1. A command session is already in progress?
	if ( waveProtocol->status == PS_PROCESS_IN_PROGRESS ) {
		response->status = KO_PROTOCOL_ALREADY_IN_USE;
		LOG_DEBUG( g_Logger, "protocol_write_radio_params status %d %s", response->status, get_response_status_name( response->status ) );
		return response->status;
	}

	waveProtocol->status = PS_PROCESS_IN_PROGRESS;

	//2.1 write dynamic standard parameters
	write_dynamic_standard_params( waveProtocol, (DynamicStandardParams*)&command->dynamicStandardParams, response );
	LOG_DEBUG( g_Logger, "protocol_write_radio_params: after write_dynamic_standard_params status: %d %s", response->status, get_response_status_name( response->status ));

	//2.2 write dynamic expert parameters
	if ( response->status == OK_PROTOCOL_SUCCESS ) {
		write_dynamic_expert_params( waveProtocol, (DynamicExpertParams*)&command->dynamicExpertParams, response );
		LOG_DEBUG( g_Logger, "protocol_write_radio_params: after write_dynamic_expert_params status: %d %s", response->status, get_response_status_name( response->status ));
	}
	//2.3 write static standard parameters
	LOG_DEBUG( g_Logger, "protocol_write_radio_params: Try to write static standard params", 0);
	if ( response->status == OK_PROTOCOL_SUCCESS  ){
		LOG_DEBUG( g_Logger, "protocol_write_radio_params: Proceeding with writting of write static standard params", 0);
		write_static_standard_params( waveProtocol, (StaticStandardParams*)&command->staticStandardParams, response );
		LOG_DEBUG( g_Logger, "write_static_standard_params: after write_dynamic_expert_params status: %d %s", response->status, get_response_status_name( response->status ));
	}else {
		LOG_DEBUG( g_Logger, "protocol_write_radio_params: Could not write static standard params with code %d", response->status);
	}

	//2.4 write static expert parameters
	if ( response->status == OK_PROTOCOL_SUCCESS ){
		write_static_expert_params( waveProtocol, (StaticExpertParams*)&command->staticExpertParams, response );
		LOG_DEBUG( g_Logger, "write_static_expert_params: after write_dynamic_expert_params status: %d %s", response->status, get_response_status_name( response->status ));
	}

	//2.4 write additional command
	LOG_DEBUG( g_Logger, "protocol_write_radio_params: write additional command", 0);
	if ( response->status == OK_PROTOCOL_SUCCESS /*|| response->status == KO_PROTOCOL_NO_VALID_PARAMETER */)
		if ( ((Payload*)&command->command.payload)->len > 0 ) {
			waveProtocol->receivedFrame.len = 0;
			LOG_DEBUG( g_Logger, "protocol_write_radio_params: write additional command - sending frame (Idx=1):  0x%02x",
					((Payload*)&command->command.payload)->frame[1]);

			if ( OK_PROTOCOL_SUCCESS != send_frame_from_payload( waveProtocol,
					(Payload*)&command->command.payload, RES_WRITE_RADIO_PARAM, response, 0) ) {

				LOG_DEBUG( g_Logger, "protocol_write_radio_params: write additional command - sending frame - failed with status status: %d %s", response->status, get_response_status_name( response->status ));
				response->status = KO_PROTOCOL_SETTING_ERROR;
			}
		}


	// do not filter unsupported parameter error messages
	if (response->status != OK_PROTOCOL_SUCCESS
		&& response->status != KO_PARAMETER_WAKEUP_LENGTH_NOT_VALID
		&& response->status != KO_PROTOCOL_NO_VALID_PARAMETER
		&& response->status != KO_PARAMETER_WAKEUP_LENGTH_VALUE_NOT_SUPPORTED
		&& response->status != KO_PARAMETER_RESPONSE_TIMEOUT_VALUE_NOT_SUPPORTED
		&& response->status != KO_PARAMETER_BROADCAST_SESSION_TIMEOUT_VALUE_NOT_SUPPORTED
		&& response->status != KO_PARAMETER_BROADCAST_SESSION_MAX_RESPONSES_NB_VALUE_NOT_SUPPORTED
		&& response->status != KO_PARAMETER_TRANSMISSION_RADIO_CHANNEL_VALUE_NOT_SUPPORTED
		&& response->status != KO_PARAMETER_TRANSMISSION_POWER_LEVEL_VALUE_NOT_SUPPORTED ) {

		response->status = KO_PROTOCOL_PROCESS_FAILURE;
	}

	waveProtocol->status = PS_IDLE;

	LOG_DEBUG( g_Logger, "protocol_write_radio_params status %d %s", response->status, get_response_status_name( response->status ) );
	return response->status;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


/****************************************************************************************************/
int protocol_open( const void* protocol ) {
	int iRet = KO_PROTOCOL_NO_VALID_PARAMETER;

	ProtocolInfo* waveProtocol = get_protocol( protocol );

	if ( waveProtocol ) {
		iRet = OK_PROTOCOL_SUCCESS;
		if ( !protocol_is_open( protocol ) ) {
			LOG_DEBUG (g_Logger, "protocol_open: Protocol not opened", 0);
			if ( (waveProtocol->serialDriverHandler.serialDriverHandle &&
					waveProtocol->serialDriverHandler.serialDriverId[0]) ||
					(!waveProtocol->serialDriverHandler.serialDriverHandle &&
					!waveProtocol->serialDriverHandler.serialDriverId[0])) {


				iRet = waveProtocol->serialDriverHandler.open( waveProtocol->serialDriver );

				if ( iRet == OK_PROTOCOL_SUCCESS ) {
					waveProtocol->serialDriverHandler.add_listener(waveProtocol->serialDriver, receive_frame, receive_event, (void*)protocol );
				}

			} else{ // waveProtocol->serialDriver.serialDriverHandle is 0! serial driver library is not initialized
				iRet = KO_PARAMETER_SERIAL_DRIVER_ID_NOT_VALID;
			}
		} else {
			LOG_DEBUG (g_Logger, "protocol_open: Protocol is already opened!", 0);
		}

		if ( iRet == OK_PROTOCOL_SUCCESS ) {
			CommandResponse response;
			RadioSettingsCommand cmd;
			memset( &cmd, 0, sizeof(cmd) );
			memset( &response, 0, sizeof(response) );
			cmd.dynamicStandardParams.groupId = -1;
			strcpy( ((RadioAddress*)&cmd.dynamicStandardParams.nodeAddress)->addressValue, "000000000000");

			if ( waveProtocol->params.writeDefaultRadioSettings ) {
				// write all default radio configuration params
				memcpy( &cmd.dynamicExpertParams, &waveProtocol->params.dynamicExpertParams, sizeof(cmd.dynamicExpertParams));
				memcpy( &cmd.staticStandardParams, &waveProtocol->params.staticStandardParams, sizeof(cmd.staticStandardParams));
				memcpy( &cmd.staticExpertParams, &waveProtocol->params.staticExpertParams, sizeof(cmd.staticExpertParams));
			}
			((Payload*)cmd.command.payload)->frame[0] = REQ_WRITE_RADIO_PARAM;
			((Payload*)cmd.command.payload)->frame[1] = EXCHANGE_STATUS;
			((Payload*)cmd.command.payload)->frame[2] = 0x03; // both status and error frames activated
			((Payload*)cmd.command.payload)->len = 3;

			iRet = protocol_write_radio_params( protocol, &cmd, &response );

			if (iRet  ==  KO_PROTOCOL_NO_VALID_PARAMETER) {
				LOG_DEBUG (g_Logger, "protocol_open: There was an error writing parameters, but it is OK when no parameter are specified on port open", 0);
				iRet  = OK_PROTOCOL_SUCCESS;
			}
		}
	}
	return iRet;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


/****************************************************************************************************/
int protocol_close( const void* protocol ) {
	ProtocolInfo* waveProtocol = get_protocol( protocol );
	if ( ! waveProtocol ) {
		return KO_PROTOCOL_NO_VALID_PARAMETER;
	}

	if ( waveProtocol->serialDriver ){
		if (waveProtocol->serialDriverHandler.close( waveProtocol->serialDriver ) !=
				OK_SERIALDRIVER_SUCCESS) {
			return KO_UNEXPECTED_ERROR;
		}
	}
	return OK_PROTOCOL_SUCCESS;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/****************************************************************************************************/
int protocol_get_serial_driver_id( const void* protocol, char* serialDriverId, int idLen ) {
	ProtocolInfo* waveProtocol = get_protocol( protocol );
	if ( waveProtocol && serialDriverId && idLen > 0 ) {
		strncpy( serialDriverId,
				waveProtocol->params.serialDriverId,
				min( idLen, strlen(waveProtocol->params.serialDriverId) ) );
		return OK_PROTOCOL_SUCCESS;
	}
	return KO_PROTOCOL_NO_VALID_PARAMETER;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/****************************************************************************************************/
int protocol_set_serial_driver_id( const void* protocol, const char* serialDriverId) {

	int iRet = KO_PROTOCOL_NO_VALID_PARAMETER;
	ProtocolInfo* waveProtocol = get_protocol( protocol );
	LOG_DEBUG(g_Logger, "set_serial_driver_id entered", 0);

	if ( waveProtocol && ( serialDriverId != NULL )  ) {
		// 1. Is protocol opened ?
		if ( protocol_is_open( waveProtocol ) )
		{
			return iRet = KO_PROTOCOL_ALREADY_OPEN;
		}

		// 2. SerialDriver Instance is already set --> unset it before
		if (waveProtocol->serialDriverHandler.serialDriverHandle )
		{
			if ( waveProtocol->serialDriver )
			{
				LOG_DEBUG(g_Logger, "set_serial_driver_id: free previous one: %s", waveProtocol->params.serialDriverId);
				if (waveProtocol->serialDriverHandler.is_open( waveProtocol->serialDriver )) {
					waveProtocol->serialDriverHandler.close( waveProtocol->serialDriver );
				}

				waveProtocol->serialDriverHandler.destroy( waveProtocol->serialDriver );
				waveProtocol->serialDriver = 0;
			}
			free_serial_driver( &waveProtocol->serialDriverHandler );
		}

		// load new serialDriverId
		LOG_DEBUG(g_Logger, "set_serial_driver_id: load serial driver: %s", serialDriverId);
		iRet = load_serial_driver( serialDriverId, &waveProtocol->serialDriverHandler );

		if ( iRet == OK_PROTOCOL_SUCCESS ) {
			LOG_DEBUG(g_Logger, "set_serial_driver_id: Serial driver loaded ", 0);

			strcpy( waveProtocol->params.serialDriverId, serialDriverId );
			waveProtocol->params.isSet |= FLAG_SERIAL_DRIVER_ID;

			// create an instance of serialDriver
			LOG_DEBUG(g_Logger, "set_serial_driver_id: Call creation of serialDriver", 0);
			if ( (waveProtocol->serialDriver = waveProtocol->serialDriverHandler.create( )) != 0 )
			{
				iRet = OK_PROTOCOL_SUCCESS;
			} else {
				LOG_ERROR(g_Logger, "Could not create serial driver instance %s ", serialDriverId);
				iRet = KO_SERIALDRIVER_ID_NOT_EXIST;
			}
		}
		else {
			LOG_ERROR(g_Logger, "Could not load serial driver %s ", serialDriverId);
			iRet = KO_SERIALDRIVER_ID_NOT_EXIST;
		}
	}
	return iRet;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/****************************************************************************************************/
int protocol_add_message_listener( const void* protocol,
		protocol_listener_t receiveRadioExchangeCommand,
		protocol_listener_event_t receiveEvent,
		void* param) {
	LOG_DEBUG(g_Logger, ">> 2 protocol_add_message_listener : adding message listener", 1);

	ProtocolInfo* waveProtocol = get_protocol( protocol );
	if ( waveProtocol ) {
		if ( !waveProtocol->paramProtocolListener )
			waveProtocol->paramProtocolListener = param;
		if ( !waveProtocol->receiveCommand )
			waveProtocol->receiveCommand = receiveRadioExchangeCommand;
		if ( !waveProtocol->receiveEvent )
			waveProtocol->receiveEvent = receiveEvent;
		return OK_PROTOCOL_SUCCESS;
	}
	return KO_PROTOCOL_NO_VALID_PARAMETER;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


/****************************************************************************************************/
int protocol_remove_message_listener( const void* protocol ) {
	ProtocolInfo* waveProtocol = get_protocol( protocol );
	if ( waveProtocol ) {
		waveProtocol->receiveCommand = NULL;
		waveProtocol->receiveEvent = NULL;
		waveProtocol->paramProtocolListener = NULL;
		return OK_PROTOCOL_SUCCESS;
	}
	return KO_PROTOCOL_NO_VALID_PARAMETER;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


/****************************************************************************************************/
int protocol_add_spy_message_listener( const void* protocol,
		protocol_spy_listener_t receivedRadioExchangeCommand,
		protocol_spy_listener_t sendRadioExchangeCommand,
		void* param,
		int* spyIndex ) {
	ProtocolInfo* waveProtocol = get_protocol( protocol );
	if ( waveProtocol ) {
		_bool registered = _false;
		int i = 0;
		for ( ; i < MAX_SPY_NUM; ++i ) {
			if ( waveProtocol->spiesSend[i] == NULL &&
				 waveProtocol->spiesRecv[i] == NULL &&
				 waveProtocol->paramSpies[i] == NULL ) {
				waveProtocol->paramSpies[i] = param;
				waveProtocol->spiesSend[i] = sendRadioExchangeCommand;
				waveProtocol->spiesRecv[i] = receivedRadioExchangeCommand;
				if ( spyIndex )
					*spyIndex = i;
				registered = _true;
				break;
			}
		}
		return registered ? OK_PROTOCOL_SUCCESS : KO_PROTOCOL_MAX_SPY_LISTENERS_EXCEEDED;
	}
	return KO_PROTOCOL_NO_VALID_PARAMETER;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


/****************************************************************************************************/
int protocol_remove_spy_message_listener( const void* protocol, int spyIndex ) {
	ProtocolInfo* waveProtocol = get_protocol( protocol );
	if ( waveProtocol ) {
		if ( spyIndex >= 0 && spyIndex < MAX_SPY_NUM ) {
			waveProtocol->spiesSend[spyIndex] = NULL;
			waveProtocol->spiesRecv[spyIndex] = NULL;
			waveProtocol->paramSpies[spyIndex] = NULL;
		}
		return OK_PROTOCOL_SUCCESS;
	}
	return KO_PROTOCOL_NO_VALID_PARAMETER;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


/****************************************************************************************************/
int protocol_send_transparent_message( const void* protocol, const pSerialExchangeData frame,
		CommandResponse* response ) {
	ProtocolInfo* waveProtocol = get_protocol( protocol );
	if ( waveProtocol && response && frame ) {
		memset( response, 0,  sizeof(*response) );
		// 0. Is protocol opened ?
		if ( !protocol_is_open( waveProtocol ) )
			return (response->status = KO_PROTOCOL_NOT_OPEN);

		// 1. A command session is already in progress?
		if ( waveProtocol->status != PS_PROCESS_IN_PROGRESS ) {
			waveProtocol->status = PS_PROCESS_IN_PROGRESS;

			//Point to Point mode  Max = 152 bytes of data
			//� Relay mode Max = 152 � (2 + 6 x Number of repeaters)
			//=> 1 repeater: 144 bytes of data
			//=> 2 repeaters: 138 bytes of data
			//=> 3 repeaters: 132 bytes of data

			// 2. Payload filled in?

			LOG_DEBUG( g_Logger, "protocol_send_transparent_message Sending frame with len %d and first byte 0x%02x ",
					((SerialFrame*)frame->serialFrame)->len, ((SerialFrame*)frame->serialFrame)->frame[0]);

			if ( ((SerialFrame*)frame->serialFrame)->len > 0 &&
					((SerialFrame*)frame->serialFrame)->len <= MAX_REQ_SEND_FRAME_BYTES_LEN ) {
				SerialFrame responseFrame;
				send_frame( waveProtocol, (SerialFrame*)frame->serialFrame, -1, response, &responseFrame);
				if ( responseFrame.len ) {
					set_serial_frame_value(frame->serialFrame, &responseFrame);
					//memcpy( &command->serialFrame, &responseFrame, sizeof(responseFrame));
					response->status = OK_PROTOCOL_SUCCESS;
				}
			} else {// 3. Payload filled in? -> NO
				response->status = KO_PAYLOAD_REQUIRED;
				LOG_DEBUG( g_Logger, "protocol_send_transparent_message Payload not filled in %d ", response->status );
			}

			// put protocol status in idle mode
			waveProtocol->status = PS_IDLE;
			LOG_DEBUG( g_Logger, "protocol_send_transparent_message status %d ", response->status );
			return response->status;
		} else // 1. A command session is already in progress? -> NO
			response->status = KO_PROTOCOL_ALREADY_IN_USE;
	}
	return KO_PROTOCOL_NO_VALID_PARAMETER;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


/****************************************************************************************************/
int protocol_send_application_message( const void* protocol, const RadioExchangeCommand* command, CommandResponse* response ) {
	return protocol_send_message( protocol, REQ_SEND_MESSAGE, RES_SEND_FRAME, command, response);
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/****************************************************************************************************/
int protocol_send_service_message( const void* protocol, const NodeServiceId serviceId, const RadioExchangeCommand* command, CommandResponse* response ) {
	// TODO test on service ID
	return protocol_send_message( protocol, REQ_SEND_SERVICE_MESSAGE, RES_SEND_SERVICE, command, response);
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/****************************************************************************************************/
int protocol_send_application_broadcast_message( const void* protocol, const RadioExchangeCommand* command, CommandResponse* response ) {
	ProtocolInfo* waveProtocol = get_protocol( protocol );
	if ( waveProtocol ) {
		return OK_PROTOCOL_SUCCESS;
	}
	return KO_PROTOCOL_NO_VALID_PARAMETER;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/****************************************************************************************************/
int protocol_send_application_request( const void* protocol, const RadioExchangeCommand* command, RequestResponse* response ) {

	LOG_DEBUG(g_Logger, ">>> protocol_send_application_request !" ,1);
	return protocol_send_request( protocol, REQ_SEND_FRAME, RES_SEND_FRAME, command, response );
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/****************************************************************************************************/
int protocol_send_service_request( const void* protocol, const NodeServiceId serviceId, const RadioExchangeCommand* command, RequestResponse* response ) {
	// TODO test on service ID
	return protocol_send_request( protocol, REQ_SEND_SERVICE, RES_SEND_SERVICE, command, response );
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/****************************************************************************************************/
int protocol_send_application_broadcast_request( const void* protocol, const RadioExchangeCommand* command,
						MulticastRequestResponse* response ) {
	ProtocolInfo* waveProtocol = get_protocol( protocol );
	LOG_DEBUG(g_Logger, "send_application_broadcast_request: Entered", 0);
	if ( waveProtocol && response && command ) {
		memset( response, 0,  sizeof(*response) );
		// 0. Is protocol opened ?
		if ( !protocol_is_open( waveProtocol ) ) {
			LOG_ERROR(g_Logger, "send_application_broadcast_request: Protocol not open", 0);
			return (response->response.status = KO_PROTOCOL_NOT_OPEN);
		}

		// 1. A command session is already in progress?
		if ( waveProtocol->status != PS_PROCESS_IN_PROGRESS ) {
			waveProtocol->status = PS_PROCESS_IN_PROGRESS;

			 LOG_DEBUG(g_Logger, "send_application_broadcast_request: Sending broadcast request ...", 0);
			//Point to Point mode  Max = 152 bytes of data
			//� Relay mode Max = 152 � (2 + 6 x Number of repeaters)
			//=> 1 repeater: 144 bytes of data
			//=> 2 repeaters: 138 bytes of data
			//=> 3 repeaters: 132 bytes of data

			 if (((Payload*)command->command.payload)->len <= 0) {
				 response->response.status = KO_PAYLOAD_REQUIRED;
				 waveProtocol->status = PS_IDLE;
				 LOG_ERROR( g_Logger, "protocol_send_application_broadcast_request Payload required ", 0);
				 return response->response.status;
			 }

			 if (((Payload*)&command->command.payload)->len > MAX_REQ_SEND_FRAME_BYTES_LEN) {
				 response->response.status = KO_PAYLOAD_NOT_VALID;
				 waveProtocol->status = PS_IDLE;
				 LOG_ERROR( g_Logger, "protocol_send_application_broadcast_request Payload is invalid ", 0);
				 return response->response.status;
			 }

			RadioSettingsCommand radioSettingsCommand;
			memset( &radioSettingsCommand, 0,  sizeof(radioSettingsCommand) );
			memset( &response->response, 0,  sizeof(response->response) );
			memcpy( &radioSettingsCommand.dynamicExpertParams, &command->dynamicExpertParams, sizeof(DynamicExpertParams) );

			// put protocol status in idle mode
			waveProtocol->status = PS_IDLE;
			protocol_write_radio_params( protocol, &radioSettingsCommand, &response->response );
			waveProtocol->status = PS_PROCESS_IN_PROGRESS;

			if (response->response.status  ==  KO_PROTOCOL_NO_VALID_PARAMETER) {
				LOG_DEBUG (g_Logger, "protocol_send_application_broadcast_request: There was an error writing parameters, but it is OK when no parameter are specified on port open", 0);
				response->response.status  = OK_PROTOCOL_SUCCESS;
			}

			LOG_DEBUG( g_Logger, "protocol_send_application_broadcast_request Radio params were written with status %d ", response->response.status);
			if ( response->response.status == OK_PROTOCOL_SUCCESS ) {
				//                   command            [Group number to query in broadcast mode ]
				SerialFrame sf = { { REQ_SEND_BROADCAST, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,       0 }, 7 };

				memcpy( sf.frame + sf.len, ((Payload*)&command->command.payload)->frame, ((Payload*)&command->command.payload)->len );
				sf.len += ((Payload*)&command->command.payload)->len;
				SerialFrame responseFrame;
				send_frame( waveProtocol, &sf, RES_SEND_FRAME, &response->response, &responseFrame);
				LOG_DEBUG( g_Logger, "protocol_send_application_broadcast_request Frame sent with status %d ", response->response.status);

				if ( response->response.status == OK_PROTOCOL_SUCCESS ) {
					notify_spies( waveProtocol, &sf, _true );
					_bool allFramesReceived = _false;
					int totalNumberOfFramesReceived = 0;
					int index = 0;
					int i = 0;
					char frameString[MAX_SERIALFRAME_LEN*2+1] = "";

					do {
						WavePortResponses wprCode = -1;
						// prepare receive buffer for requested data
						waveProtocol->receivedFrame.len = 0;
						const DynamicExpertParams* dynamicExpertParams = &waveProtocol->params.dynamicExpertParams;
						// if ( command->dynamicExpertParams.isSet & FLAG_WAKEUP_LENGTH )
							dynamicExpertParams = &command->dynamicExpertParams;

						wait_for_frame( waveProtocol, dynamicExpertParams, _true, _true, &response->response, &wprCode);
						LOG_DEBUG( g_Logger, "protocol_send_application_broadcast_request Wait for frame ended with status %d ", response->response.status);

						if ( response->response.status == OK_PROTOCOL_SUCCESS ) {
							switch( wprCode ) {
							case RECEIVED_BROADCAST_RESPONSE:
								totalNumberOfFramesReceived = waveProtocol->receivedFrame.frame[2];
								if ( totalNumberOfFramesReceived == 0 ) {
									allFramesReceived = _true;
									break;
								}
								// verify if we receive last frame ( last frame will have index 1 )
								if ( waveProtocol->receivedFrame.frame[3] == 1 )
									allFramesReceived = _true;
								for ( i = 4; i < ((MAX_RADIOADDRESS_LEN-1)/2+4); ++i )
									sprintf( ((RadioAddress*)&response->commands[index].dynamicStandardParams.nodeAddress)->addressValue +
											strlen(((RadioAddress*)&response->commands[index].dynamicStandardParams.nodeAddress)->addressValue), "%02X",
											waveProtocol->receivedFrame.frame[i] );
								((Payload*)&response->commands[index].command.payload)->len = waveProtocol->receivedFrame.len - ((MAX_RADIOADDRESS_LEN-1)/2 + 4);
								memcpy( ((Payload*)&response->commands[index].command.payload)->frame,
										waveProtocol->receivedFrame.frame + (waveProtocol->receivedFrame.len - ((Payload*)&response->commands[index].command.payload)->len),
										((Payload*)&response->commands[index].command.payload)->len );
								memcpy( &response->commands[index].dynamicExpertParams, &command->dynamicExpertParams, sizeof(DynamicExpertParams) );

								// set reception mode to broadcast for any received message
								LOG_DEBUG(g_Logger, "protocol_send_application_broadcast_request:  Setup reception mode", 0);
								response->commands[index].dynamicExpertParams.receptionMode = RECEPTION_MODE_BROADCAST;
								response->commands[index].dynamicExpertParams.isSet |= FLAG_RECEPTION_MODE;

								notify_spies_command( waveProtocol, &response->commands[index], _false );
								response->commandsLen = ++index;
								LOG_DEBUG( g_Logger, "TN: %02d index: %02d address: %s ",
										totalNumberOfFramesReceived,
										waveProtocol->receivedFrame.frame[3],
										((RadioAddress*)&response->commands[index-1].dynamicStandardParams.nodeAddress)->addressValue );


								break;
							default:

								sprintf( response->response.statusDescription, "unexpected received frame during waiting for waveport END_MESSAGE_EXCHANGE %s",
										get_hex_serialframe(frameString, sizeof(frameString), &waveProtocol->receivedFrame) );
								response->response.status = KO_RADIO_TRANSMISSION_FAILURE;
								break;
							}
						}
					} while ( response->response.status == OK_PROTOCOL_SUCCESS && !allFramesReceived && response->commandsLen < MAX_MULTICAST_REQUEST_COMMANDS_LEN );
				}
			} else {
				sprintf( response->response.statusDescription, "signals param in error %s", get_response_status_name( response->response.status ) );
			}

			// put protocol status in idle mode
			waveProtocol->status = PS_IDLE;

			LOG_DEBUG( g_Logger, "protocol_send_application_broadcast_request status %d ", response->response.status );
			return response->response.status;
		} else // 1. A command session is already in progress? -> NO
			response->response.status = KO_PROTOCOL_ALREADY_IN_USE;
	} else
		LOG_ERROR(g_Logger, "send_application_broadcast_request: Invalid parameters", 0);

	return KO_PROTOCOL_NO_VALID_PARAMETER;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


/****************************************************************************************************/
int  protocol_send_application_response( const void* protocol, const Command *command, CommandResponse* response ) {


	ProtocolInfo* waveProtocol = get_protocol( protocol );
	if (response) memset( response, 0,  sizeof(*response) );


	//--------------------------------------------------------------------------
	// Check passed parameters
	if (!waveProtocol || !response || !command) {
		if (response) response->status = KO_PROTOCOL_NO_VALID_PARAMETER;
		return KO_PROTOCOL_NO_VALID_PARAMETER;
	}

	if ( !protocol_is_open( waveProtocol ) )
		return (response->status = KO_PROTOCOL_NOT_OPEN);

	if (waveProtocol->status == PS_PROCESS_IN_PROGRESS) {
		return response->status = KO_PROTOCOL_ALREADY_IN_USE;
	}

	Payload* payload = (Payload *) command->payload;

	waveProtocol->status = PS_PROCESS_IN_PROGRESS;


	if (waveProtocol->isResponseSessionActive == 0) {
		LOG_DEBUG( g_Logger, " This is not a response session, exiting.", 0);
		waveProtocol->status = PS_IDLE;
		return response->status = KO_PROTOCOL_NO_RESPONSE_SESSION;
	}

	if (payload == NULL || payload->len == 0) {
		waveProtocol->status = PS_IDLE;
		return response->status = KO_PAYLOAD_REQUIRED;
	}

	if (payload->len > MAX_REQ_SEND_FRAME_BYTES_LEN) {
		waveProtocol->status = PS_IDLE;
		return response->status = KO_PAYLOAD_NOT_VALID;
	}

	if (strlen(((RadioAddress*)&waveProtocol->senderNodeAddress)->addressValue) == 0) {
		LOG_DEBUG( g_Logger, "protocol_send_application_response senderNodeAddress not defined %s ",
				((RadioAddress*)&waveProtocol->senderNodeAddress)->addressValue);
		waveProtocol->status = PS_IDLE;
		return response->status = KO_PROTOCOL_NO_RESPONSE_SESSION;
	}

	//--------------------------------------------------------------------------
	// Prepare frame to be sent
	SerialFrame serialFrame;
	memset(&serialFrame, 0, sizeof(serialFrame));

	int i;
	char *stopstring;

	LOG_DEBUG(g_Logger, "protocol_send_application_response: Compose response having flag reception mode %s and value %d",
			((waveProtocol->params.dynamicExpertParams.isSet & FLAG_RECEPTION_MODE) ? "true" : "false"),
			waveProtocol->params.dynamicExpertParams.receptionMode);

	_bool isBroadcastResponse = (waveProtocol->receptionSessionMode == RECEPTION_MODE_BROADCAST) ||
			((waveProtocol->params.dynamicExpertParams.isSet & FLAG_RECEPTION_MODE) &&
			(waveProtocol->params.dynamicExpertParams.receptionMode == RECEPTION_MODE_BROADCAST));
	if (isBroadcastResponse) {
		LOG_DEBUG(g_Logger, "protocol_send_application_response: Sending BROADCAST response", 0);
		serialFrame.frame[serialFrame.len ++] = REQ_SEND_BROADCAST_RESPONSE;
	}
	else {

		LOG_DEBUG(g_Logger, "protocol_send_application_response: Sending NORMAL response", 0);
		serialFrame.frame[serialFrame.len ++] = REQ_SEND_RESPONSE;
	}

	char *abuff = ((RadioAddress*)&waveProtocol->senderNodeAddress)->addressValue;
	for ( i = 0; i < MAX_RADIOADDRESS_LEN-1 ; i += 2 ) {
		char buff [3]=  {abuff[i], abuff[i + 1], 0};
		serialFrame.frame[serialFrame.len ++] = (byte)strtol( buff, &stopstring, 16 );
	}

	memcpy(&serialFrame.frame[serialFrame.len], payload->frame, payload->len);
	serialFrame.len += payload->len;

	//--------------------------------------------------------------------------
	// sending frame and wait RES_SEND_FRAME confirmation
	memset(response, 0, sizeof(response));
	send_frame (waveProtocol, &serialFrame, RES_SEND_FRAME, response, &serialFrame);

	if (response->status != OK_PROTOCOL_SUCCESS) {
		waveProtocol->status =  PS_IDLE;
		response->status = KO_PROTOCOL_PROCESS_FAILURE;
		return response->status;
	}

	//--------------------------------------------------------------------------
	// Close the response session and consume received confirmation frame
	waveProtocol->isResponseSessionActive = 0;
	waveProtocol->receivedFrame.len = 0;

	//--------------------------------------------------------------------------
	// wait for serial frame
	LOG_DEBUG( g_Logger, "protocol_send_application_response Waiting for END_MESSAGE_EXCHANGE ... ", 0);

	WavePortResponses responseId;
	wait_for_frame(waveProtocol, &waveProtocol->params.dynamicExpertParams,
			_false, isBroadcastResponse, response, &responseId);

	if ((response->status != OK_PROTOCOL_SUCCESS) ||
		(responseId != END_MESSAGE_EXCHANGE)) {

		LOG_ERROR( g_Logger, "protocol_send_application_response Failed to get END_MESSAGE_EXCHANGE", 0);
		waveProtocol->status =  PS_IDLE;
		return response->status = KO_PROTOCOL_PROCESS_FAILURE;
	}

	LOG_DEBUG( g_Logger, "protocol_send_application_response END_MESSAGE_EXCHANGE GOT!", 0);

	//--------------------------------------------------------------------------
	waveProtocol->status =  PS_IDLE;
	return response->status = OK_PROTOCOL_SUCCESS;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


/****************************************************************************************************/
/**
 * Function used to read dynamic expert parameters
 * @param response - response containing reading
 * @return 0 if success or positive number if unexpected error occurs during sending process
 */
int  protocol_read_dynamic_expert_params( const void* protocol, RadioSettingsResponse* response) {
	ProtocolInfo* waveProtocol = get_protocol( protocol );
	if ( waveProtocol && response ) {
		memset( response, 0,  sizeof(*response) );
		response->response.status = KO_PROTOCOL_NOT_OPEN;

		// 0. Is protocol opened ?
		int iRet = verify_serial_driver_port( waveProtocol );
		if ( iRet == OK_PROTOCOL_SUCCESS ) {
			response->response.status = KO_PROTOCOL_ALREADY_IN_USE;
			// 1. A command session is already in progress?
			if ( waveProtocol->status != PS_PROCESS_IN_PROGRESS ) {
				waveProtocol->status = PS_PROCESS_IN_PROGRESS;

				memset( &response->command.dynamicExpertParams, 0, sizeof(response->command.dynamicExpertParams) );

				SerialFrame command = { { REQ_READ_RADIO_PARAM, 0}, 2 };
				SerialFrame responseFrame;

				// read wakeup length parameter from port
				command.frame[1] = WAKEUP_LENGTH;
				iRet = send_frame( waveProtocol, &command, RES_READ_RADIO_PARAM, &response->response, &responseFrame);
				if ( iRet == OK_PROTOCOL_SUCCESS && ( responseFrame.len == 2 ) ) {
					response->command.dynamicExpertParams.wakeupLength = responseFrame.frame[1];
					response->command.dynamicExpertParams.wakeupLength <<= 8;
					response->command.dynamicExpertParams.wakeupLength += responseFrame.frame[0];
					response->command.dynamicExpertParams.isSet |= FLAG_WAKEUP_LENGTH;
					//printf("protocol_read_dynamic_expert_params:: wakeupLength %d\n", response->command.dynamicExpertParams.wakeupLength);
				}

				command.frame[1] = RADIO_USER_TIMEOUT;
				iRet = send_frame( waveProtocol, &command, RES_READ_RADIO_PARAM, &response->response, &responseFrame);
				if ( iRet == OK_PROTOCOL_SUCCESS ) {
					response->command.dynamicExpertParams.responseTimeout = ((unsigned int)responseFrame.frame[0])*100;
					response->command.dynamicExpertParams.isSet |= FLAG_RESPONSE_TIMEOUT;
					//printf("protocol_read_dynamic_expert_params:: responseTimeout %d\n", response->command.dynamicExpertParams.responseTimeout);

				}

				command.frame[1] = BCST_RECEPTION_TIMEOUT;
				iRet = send_frame( waveProtocol, &command, RES_READ_RADIO_PARAM, &response->response, &responseFrame);
				if ( iRet == OK_PROTOCOL_SUCCESS ) {
					response->command.dynamicExpertParams.broadcastSessionTimeout = ((unsigned int)responseFrame.frame[0])*100;
					response->command.dynamicExpertParams.isSet |= FLAG_BROADCAST_SESSION_TIMEOUT;
					//printf("protocol_read_dynamic_expert_params:: broadcastSessionTimeout %d\n", response->command.dynamicExpertParams.broadcastSessionTimeout);
				}

				command.frame[1] = BCST_MAX_RESPONSES_NB;
				iRet = send_frame( waveProtocol, &command, RES_READ_RADIO_PARAM, &response->response, &responseFrame);
				if ( iRet == OK_PROTOCOL_SUCCESS ) {
					response->command.dynamicExpertParams.broadcastSessionMaxResponsesNb = responseFrame.frame[0];
					response->command.dynamicExpertParams.isSet |= FLAG_BROADCAST_SESSION_MAX_RESPONSES_NB;
					//printf("protocol_read_dynamic_expert_params:: broadcastSessionMaxResponsesNb %d\n", response->command.dynamicExpertParams.broadcastSessionMaxResponsesNb);
				}

				command.frame[0] = REQ_READ_CHANNEL;
				command.len = 1;
				iRet = send_frame( waveProtocol, &command, RES_READ_CHANNEL, &response->response, &responseFrame);
				if ( iRet == OK_PROTOCOL_SUCCESS ) {
					response->command.dynamicExpertParams.transmissionRadioChannel = responseFrame.frame[0];
					response->command.dynamicExpertParams.isSet |= FLAG_TRANSMISSION_RADIO_CHANNEL;
					LOG_DEBUG(g_Logger, "protocol_read_dynamic_expert_params : Channel read %d", responseFrame.frame[0]);
					//printf("protocol_read_dynamic_expert_params:: transmissionRadioChannel %d\n", response->command.dynamicExpertParams.transmissionRadioChannel);
				} else {
					LOG_DEBUG(g_Logger, "protocol_read_dynamic_expert_params : Failed to read Channel ", 0);

				}

				LOG_DEBUG(g_Logger, "protocol_read_dynamic_expert_params : Reading of TX Power level", 0);

				command.frame[0] = REQ_READ_TX_POWER;
				iRet = send_frame( waveProtocol, &command, RES_READ_TX_POWER, &response->response, &responseFrame);
				if ( iRet == OK_PROTOCOL_SUCCESS ) {

					float transmissionPowerLevelRefValue = 0.f;
					if(payloadToTransmissionPowerLevel (responseFrame.frame[0],
							&transmissionPowerLevelRefValue)) {
						response->command.dynamicExpertParams.transmissionPowerLevel = transmissionPowerLevelRefValue;
						response->command.dynamicExpertParams.isSet |= FLAG_TRANSMISSION_POWER_LEVEL;
						LOG_DEBUG(g_Logger, "protocol_read_dynamic_expert_params : TX Power read %d", responseFrame.frame[0]);
						//printf("protocol_read_dynamic_expert_params:: transmissionPowerLevel %f\n",
						//		response->command.dynamicExpertParams.transmissionPowerLevel);

					} else {

						LOG_DEBUG(g_Logger, "protocol_read_dynamic_expert_params : TX Power cannot be converted %d", responseFrame.frame[0]);
					}
				} else {
					LOG_DEBUG(g_Logger, "protocol_read_dynamic_expert_params : Failed to read TX Power", 0);
				}

				response->response.status = OK_PROTOCOL_SUCCESS;
				waveProtocol->status = PS_IDLE;
			}
		}

		return iRet;
	}
	return KO_PROTOCOL_NO_VALID_PARAMETER;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/****************************************************************************************************/
/**
 * Function used to read static standard parameters
 * @param protocol - A valid handle to an open protocol.
 * @param response - response containing reading
 * @return 0 if success or positive number if unexpected error occurs during sending process
 */
int  protocol_read_static_standard_params( const void* protocol, RadioSettingsResponse* response) {
	ProtocolInfo* waveProtocol = get_protocol( protocol );
	if ( waveProtocol && response ) {
		memset( response, 0,  sizeof(*response) );
		response->response.status = KO_PROTOCOL_NOT_OPEN;
		// 0. Is protocol opened ?
		int iRet = verify_serial_driver_port( waveProtocol );
		if ( iRet == OK_PROTOCOL_SUCCESS ) {
			response->response.status = KO_PROTOCOL_ALREADY_IN_USE;
			// 1. A command session is already in progress?
			if ( waveProtocol->status != PS_PROCESS_IN_PROGRESS ) {
				waveProtocol->status = PS_PROCESS_IN_PROGRESS;

				memset( &response->command.staticStandardParams, 0, sizeof(response->command.staticStandardParams) );

				SerialFrame command = { { REQ_READ_RADIO_PARAM, 0}, 2 };
				SerialFrame responseFrame;

				// read radio ack activation parameter from port
				command.frame[1] = RADIO_ACKNOWLEDGE;
				iRet = send_frame( waveProtocol, &command, RES_READ_RADIO_PARAM, &response->response, &responseFrame);
				if ( iRet == OK_PROTOCOL_SUCCESS ) {
					response->command.staticStandardParams.radioAckActivation = responseFrame.frame[0];
					response->command.staticStandardParams.isSet |= FLAG_RADIO_ACK_ACTIVATION;
				}

				command.frame[1] = AWAKENING_PERIOD;
				iRet = send_frame( waveProtocol, &command, RES_READ_RADIO_PARAM, &response->response, &responseFrame);
				if ( iRet == OK_PROTOCOL_SUCCESS ) {
					response->command.staticStandardParams.firstScanActivityPeriod = ((unsigned int)responseFrame.frame[0])*100;
					response->command.staticStandardParams.isSet |= FLAG_ACTIVITY_PERIOD_ON_FIRST_SCAN;
				}

				command.frame[1] = SWITCH_MODE_STATUS;
				iRet = send_frame( waveProtocol, &command, RES_READ_RADIO_PARAM, &response->response, &responseFrame);
				if ( iRet == OK_PROTOCOL_SUCCESS ) {
					response->command.staticStandardParams.switchingModeOnRadioAddressActivation = responseFrame.frame[0];
					response->command.staticStandardParams.isSet |= FLAG_SWITCHING_MODE_ON_RADIO_ADDRESS_ACTIVATION;
				}

				command.frame[1] = WAVECARD_MULTICAST_GROUP;
				iRet = send_frame( waveProtocol, &command, RES_READ_RADIO_PARAM, &response->response, &responseFrame);
				if ( iRet == OK_PROTOCOL_SUCCESS ) {
					response->command.staticStandardParams.receptionGroupId = responseFrame.frame[0];
					response->command.staticStandardParams.isSet |= FLAG_RECEPTION_GROUPID;
				}

				response->command.staticStandardParams.firstScanEnableActivation = _true;
				response->command.staticStandardParams.isSet |= FLAG_FIRST_SCAN_ENABLE_ACTIVATION;
				response->response.status = OK_PROTOCOL_SUCCESS;
				waveProtocol->status = PS_IDLE;
			}
		}
		return iRet;
	}
	return KO_PROTOCOL_NO_VALID_PARAMETER;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


/****************************************************************************************************/
/**
 * Function used to read static expert parameters
 * @param protocol - A valid handle to an open protocol.
 * @param response - response containing reading
 * @return 0 if success or positive number if unexpected error occurs during sending process
 */
int  protocol_read_static_expert_params( const void* protocol, RadioSettingsResponse* response) {
	ProtocolInfo* waveProtocol = get_protocol( protocol );
	if ( waveProtocol && response ) {
		memset( response, 0,  sizeof(*response) );
		response->response.status = KO_PROTOCOL_NOT_OPEN;
		// 0. Is protocol opened ?
		int iRet = verify_serial_driver_port( waveProtocol );
		if ( iRet == OK_PROTOCOL_SUCCESS ) {
			response->response.status = KO_PROTOCOL_ALREADY_IN_USE;
			// 1. A command session is already in progress?
			if ( waveProtocol->status != PS_PROCESS_IN_PROGRESS ) {
				waveProtocol->status = PS_PROCESS_IN_PROGRESS;

				memset( &response->command.staticExpertParams, 0, sizeof(response->command.staticExpertParams) );

				SerialFrame command = { { REQ_READ_RADIO_PARAM, 0}, 2 };
				SerialFrame responseFrame;

				// read second scan enable activation parameter from port
				command.frame[1] = SECOND_SCAN_ENABLE_ACTIVATION;
				iRet = send_frame( waveProtocol, &command, RES_READ_RADIO_PARAM, &response->response, &responseFrame);
				if ( iRet == OK_PROTOCOL_SUCCESS ) {
					response->command.staticExpertParams.secondScanEnableActivation = responseFrame.frame[0];
					response->command.staticExpertParams.isSet |= FLAG_SECOND_SCAN_ENABLE_ACTIVATION;
					//printf("ReadStaticExpertParams:: secondScanEnableActivation %d\n",
					//		response->command.staticExpertParams.secondScanEnableActivation);
				}

				command.frame[1] = RADIO_CRC_ACTIVATION;
				iRet = send_frame( waveProtocol, &command, RES_READ_RADIO_PARAM, &response->response, &responseFrame);
				if ( iRet == OK_PROTOCOL_SUCCESS ) {
					response->command.staticExpertParams.radioCRCActivation = responseFrame.frame[0];
					response->command.staticExpertParams.isSet |= FLAG_RADIO_CRC_ACTIVATION;
					//printf("ReadStaticExpertParams:: secondScanEnableActivation %d\n",
					//		response->command.staticExpertParams.secondScanEnableActivation);
				}

				command.frame[1] = AWAKENING_PERIOD;
				iRet = send_frame( waveProtocol, &command, RES_READ_RADIO_PARAM, &response->response, &responseFrame);
				if ( iRet == OK_PROTOCOL_SUCCESS ) {
					response->command.staticExpertParams.secondScanActivityPeriod = ((unsigned int)responseFrame.frame[0])*100;
					response->command.staticExpertParams.isSet |= FLAG_ACTIVITY_PERIOD_ON_SECOND_SCAN;
					//printf("ReadStaticExpertParams:: secondScanActivityPeriod %d\n",
					//		response->command.staticExpertParams.secondScanActivityPeriod);
				}

				command.frame[0] = REQ_READ_CHANNEL;
				command.len = 1;
				iRet = send_frame( waveProtocol, &command, RES_READ_CHANNEL, &response->response, &responseFrame);
				if ( iRet == OK_PROTOCOL_SUCCESS ) {
					response->command.staticExpertParams.firstScanRadioChannel = responseFrame.frame[0];
					response->command.staticExpertParams.isSet |= FLAG_RADIO_CHANNEL_ON_FIRST_SCAN;
					//printf("ReadStaticExpertParams:: firstScanRadioChannel %d\n",
					//		response->command.staticExpertParams.firstScanRadioChannel);
				}

				command.frame[0] = REQ_READ_PHYCONFIG;
				iRet = send_frame( waveProtocol, &command, RES_READ_PHYCONFIG, &response->response, &responseFrame);
				if ( iRet == OK_PROTOCOL_SUCCESS && responseFrame.len == 2 ) {
					unsigned short firstScanPhysicalConfigPayload = responseFrame.frame[0];
					firstScanPhysicalConfigPayload <<= 8;
					firstScanPhysicalConfigPayload += responseFrame.frame[1];

					if (payloadToPhysicalLayer(firstScanPhysicalConfigPayload, &response->command.staticExpertParams.firstScanPhysicalConfig))
						response->command.staticExpertParams.isSet |= FLAG_PHYSICAL_CONFIG_ON_FIRST_SCAN;

					//printf("ReadStaticExpertParams:: firstScanPhysicalConfig %d\n",
					//		response->command.staticExpertParams.firstScanPhysicalConfig);

				}

				command.frame[0] = REQ_READ_SECOND_PHYCONFIG;
				iRet = send_frame( waveProtocol, &command, RES_READ_SECOND_PHYCONFIG, &response->response, &responseFrame);
				if ( iRet == OK_PROTOCOL_SUCCESS && responseFrame.len == 2 ) {
					unsigned short secondScanPhysicalConfigPayload = responseFrame.frame[0];
					secondScanPhysicalConfigPayload <<= 8;
					secondScanPhysicalConfigPayload += responseFrame.frame[1];

					if (payloadToPhysicalLayer(secondScanPhysicalConfigPayload, &response->command.staticExpertParams.secondScanPhysicalConfig))
						response->command.staticExpertParams.isSet |= FLAG_PHYSICAL_CONFIG_ON_SECOND_SCAN;

					//printf("ReadStaticExpertParams:: secondScanPhysicalConfig %d\n",
					//		response->command.staticExpertParams.secondScanPhysicalConfig);

				}

				response->command.staticExpertParams.complexScanEnableActivation = _false;
				response->command.staticExpertParams.isSet |= FLAG_COMPLEX_SCAN_ENABLE_ACTIVATION;
				response->command.staticExpertParams.firstScanSensitivityReductionFactor = 0;
				response->command.staticExpertParams.isSet |= FLAG_SENSITIVITY_REDUCTION_FACTOR_ON_FIRST_SCAN;
				response->command.staticExpertParams.secondScanRadioChannel = 0;
				response->command.staticExpertParams.isSet |= FLAG_RADIO_CHANNEL_ON_SECOND_SCAN;
				response->command.staticExpertParams.secondScanSensitivityReductionFactor = 0;
				response->command.staticExpertParams.isSet |= FLAG_SENSITIVITY_REDUCTION_FACTOR_ON_SECOND_SCAN;

				// Note: following parameters do not exist on waveport:
				// response->command.staticExpertParams.cwupTransmissionActivation;
				// response->command.staticExpertParams.cwupAutoSelectionActivation;
				// response->command.staticExpertParams.cwupRequestOnlyActivation;
				// response->command.staticExpertParams.cwupOnlyActivation;
				// response->command.staticExpertParams.nbScanOn;
				// response->command.staticExpertParams.nbScanOff;
				// response->command.staticExpertParams.frequencyHoppingChannelTable;

				response->response.status = OK_PROTOCOL_SUCCESS;

				waveProtocol->status = PS_IDLE;
			}
		}
		return iRet;
	}
	return KO_PROTOCOL_NO_VALID_PARAMETER;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


/****************************************************************************************************/
int  protocol_request_local_node_service( const void* protocol, const LocalNodeServiceCommand* command, LocalNodeServiceResponse* response ) {
	ProtocolInfo* waveProtocol = get_protocol( protocol );
	if ( waveProtocol && response && command ) {
		memset( response, 0,  sizeof(*response) );
		// 0. Is protocol opened ?
		if ( !protocol_is_open( waveProtocol ) )
			return (response->response.status = KO_PROTOCOL_NOT_OPEN);

		// 1. A command session is already in progress?
		if ( waveProtocol->status != PS_PROCESS_IN_PROGRESS ) {
			waveProtocol->status = PS_PROCESS_IN_PROGRESS;

			byte responseId = -1;
			_bool isSupported = _false;

			response->command.serviceId = command->serviceId;

			// 3. serviceId filled in?
			switch( command->serviceId ) {
			case STACK_SERVICE:
				LOG_DEBUG(g_Logger, "protocol_request_local_node_service: requesting stack service ... ", 0);
				isSupported = _true;
				break;
			case SDP_SERVICE:
				response->response.status = KO_PROTOCOL_NODE_SERVICE_NOT_SUPPORTED;
				break;
			case FIRMWARE_UPGRADE_SERVICE:
				response->response.status = KO_PROTOCOL_NODE_SERVICE_NOT_SUPPORTED;
				break;
			case NODE_SERVICE_ID_UNDEFINED:
			default:
				// 3. wrong/unknown serviceId!
				response->response.status = KO_PROTOCOL_NODE_SERVICE_ID_REQUIRED;
				break;
			}

			if ( isSupported ) {	// read parameter from port

				// 2. Payload filled in?
				if ( ((Payload*)&command->command.payload)->len > 0 && ((Payload*)&command->command.payload)->len <= MAX_REQ_SEND_FRAME_BYTES_LEN ) {

					SerialFrame sf;
					response->response.status = send_frame_from_payload( waveProtocol,
							(Payload*)&command->command.payload, responseId, &response->response, &sf);

					char sf_str [260];
					get_hex_serialframe(sf_str, sizeof(sf_str), &sf);

					LOG_DEBUG(g_Logger, "protocol_request_local_node_service: Response received len = %d, copy it to response->command.command.payload: %s",
							sf.len,
							sf_str);
					// response->command.command.payload
					memcpy(&response->command.command.payload, &sf,
							sizeof(response->command.command.payload));

					((Payload*) &response->command.command.payload)->len = sf.len;

					if ( response->response.status == KO_PROTOCOL_PARAM_READ_ERROR )
						response->response.status = KO_PROTOCOL_NODE_LOCAL_SERVICE_FAILURE;
				} else {// 3. Payload filled in? -> NO
					response->response.status = KO_PAYLOAD_REQUIRED;
					LOG_DEBUG( g_Logger, "protocol_request_local_node_service Payload not filled in %d ", response->response.status );
				}
			}

			// put protocol status in idle mode
			waveProtocol->status = PS_IDLE;

			LOG_DEBUG( g_Logger, "protocol_request_local_node_service status %d ", response->response.status );
			return response->response.status;
		} else // 1. A command session is already in progress? -> NO
			response->response.status = KO_PROTOCOL_ALREADY_IN_USE;
	}
	return KO_PROTOCOL_NO_VALID_PARAMETER;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/****************************************************************************************************/
_bool is_unexpected_command( byte commandId ) {
	_bool isUnexpectedCommand = _true;
	switch( commandId ) {
	case RECEIVED_FRAME:
	case RECEIVED_FRAME_POLLIN:
	case RECEIVED_BROADCAST_RESPONSE:
	case RECEIVED_FRAME_RELAYED:
	case RECEIVED_MULTIFRAME:
	case RECEIVED_BROADCAST_FRAME:
	case SERVICE_RESPONSE:
		isUnexpectedCommand = _false;
		break;
	default:
		isUnexpectedCommand = _true;
		break;
	}
	return isUnexpectedCommand;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/****************************************************************************************************/
int receive_frame( void* param, const pSerialExchangeData pFrame ) {

	SerialFrame *frame = (SerialFrame *)pFrame->serialFrame;
	return receive_internal_frame(param, frame);
}

int receive_internal_frame( void* param, const SerialFrame* frame ) {
	ProtocolInfo* waveProtocol = get_protocol( param );
	if ( waveProtocol && frame ) {
		int index = 10;
		// try to write frame after previous one has been handled
		while ( waveProtocol->receivedFrame.len != 0 && index < 10 ) {
			++index;
			SLEEP(10);
		}

		memcpy( &waveProtocol->receivedFrame, frame, sizeof(*frame) );

		waveProtocol->isResponseSessionActive = 1;

		// verify if received frame is spontaneous frame ( in case if protocol status is IDLE all received frames are spontaneous! )
		if ( waveProtocol->status == PS_IDLE ) {
			if ( is_unexpected_command( frame->frame[0]) ) {
				receive_event( param, PE_IGNORED_COMMAND_RECEIVED);
			} else {
				RadioExchangeCommand command;
				parse_frame(waveProtocol, (SerialFrame*)frame, &command, NULL, NULL, NULL, NULL );
				if ( waveProtocol->receiveCommand )
					waveProtocol->receiveCommand( waveProtocol->paramProtocolListener, &command );

				LOG_DEBUG(g_Logger,  "receive_frame: Notify spies on frame reception", 0);
				notify_spies_command( waveProtocol, &command, _false );
			}
		}

		//----------------------------------------------------------------------
		// wait for 2 seconds and close response session.
		if (waveProtocol->status == PS_IDLE || !frame ||
				frame->frame[0] == RECEIVED_BROADCAST_RESPONSE) {
			protocol_close_response_session (waveProtocol);
		}else {
			// just close session
			waveProtocol->isResponseSessionActive = 0;
		}
	}

	return 0;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


/****************************************************************************************************/
int receive_event( void* param, int eventCode  ) {
	ProtocolInfo* waveProtocol = get_protocol( param );
	if ( waveProtocol ) {
		switch( eventCode ) {
		case START_LISTENING:          eventCode = PE_START_LISTENING; break;
		case STOP_LISTENING:           eventCode = PE_STOP_LISTENING; break;
		case SERIALPORT_DISCONNECTION: eventCode = PE_SERIALDRIVER_DISCONNECTION; break;
		case IGNORED_FRAME_RECEIVED:
		default: eventCode = -1; break;
		}
		if ( eventCode != -1 ) {
			if ( waveProtocol->receiveEvent )
				waveProtocol->receiveEvent( waveProtocol->paramProtocolListener , eventCode );
#ifdef _JAVA_WRAPPER
			else // no C listener try to notify in Java way
				jni_notify_event_native( waveProtocol, eventCode );
#endif //_JAVA_WRAPPER
		}
	}
	return 0;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/****************************************************************************************************
 *                        HELPER SECTION - PROTOCOL SPECIFIC FUNCTIONS DEFINITION
 ****************************************************************************************************/

/*==================================================================================================*/
ProtocolInfo* get_protocol( const void* protocol ) {
	if ( protocol)  {
		 if (((ProtocolInfo*)protocol)->integrity == PROTOCOL_INTEGRITY ) {
			 return (ProtocolInfo*)protocol;
		 }
	}
	return NULL;
}
/*__________________________________________________________________________________________________*/

/*==================================================================================================*/
int verify_serial_driver_port( const ProtocolInfo* waveProtocol ) {
	int iRet = OK_PROTOCOL_SUCCESS;
	if ( !waveProtocol->serialDriverHandler.serialDriverHandle )
		iRet = KO_SERIALDRIVER_NOT_INITIALIZED;
	else
		if ( !waveProtocol->serialDriver )
			iRet = KO_SERIALDRIVER_PORT_NOT_OPENED;
	return iRet;
}
/*__________________________________________________________________________________________________*/


/**
 * Function used to waiting for response after send command to port
 * @param waveProtocol - handle to waveport protocol instance
 * @param responseId - response id for command
 * @return OK_PROTOCOL_SUCCESS if success, KO_PROTOCOL_CONFIGURATION_TIMEOUT if timeout expired,
 * KO_PROTOCOL_PARAM_UPDATE_ERROR if update failed and
 * KO_PROTOCOL_UNEXPECTED_FRAME if unexpected frame received.
 */
int wait_for_response( const ProtocolInfo* waveProtocol, CommandResponse* response, WavePortResponses responseId ) {
	response->status = KO_PROTOCOL_PROCESS_FAILURE;
	strcpy( response->statusDescription, "no receive waveport applicative acknowledge" );

	struct timeval startTime;
	struct timeval endTime;
	gettimeofday( &startTime, 0 );

	LOG_DEBUG(g_Logger, "wait_for_response: wait for %d ms", waveProtocol->params.configurationTimeout);
	// waiting for frame RES_WRITE_RADIO_PARAM
	while ( waveProtocol->receivedFrame.len == 0 ) {
		// wait for message body
		gettimeofday( &endTime, 0 );
		if ( millisecondtimediff( startTime, endTime ) > waveProtocol->params.configurationTimeout ) {
			LOG_DEBUG(g_Logger, "wait_for_response: Timeout occurred [%d ms]!", waveProtocol->params.configurationTimeout);
			break;
		}
		SLEEP( 10 );
	}

	gettimeofday( &endTime, 0 );
	LOG_DEBUG(g_Logger, "wait_for_response: Quit after %d ms", millisecondtimediff( startTime, endTime ));
	if ( waveProtocol->receivedFrame.len > 0 ) {
		LOG_DEBUG(g_Logger, "wait_for_response: Received some data 0x%2X, expected 0x%2X",
				waveProtocol->receivedFrame.frame[0], responseId);

		if ( responseId == (byte)-1 ) {
			// No expected responseId waited for --> all response are considered as OK
			response->status = OK_PROTOCOL_SUCCESS;
			response->statusDescription[0] = 0;
		} else if ( waveProtocol->receivedFrame.frame[0] == responseId) {
			// A specific responseId is waited for
			LOG_DEBUG(g_Logger, "wait_for_response: Response accepted, check it's status code %d, expected %d",
					waveProtocol->receivedFrame.frame[1], CMD_STATUS_OK);

			if ( waveProtocol->receivedFrame.frame[1] == CMD_STATUS_OK ) {
				response->status = OK_PROTOCOL_SUCCESS;
				response->statusDescription[0] = 0;
				LOG_DEBUG(g_Logger, "wait_for_response: Response and return code accepted!", 0);

			} else {
				char frameString[MAX_SERIALFRAME_LEN*2+1] = "";
				sprintf( response->statusDescription, "waveport applicative acknowledge is %d, expected %d. Frame is : %s",
						waveProtocol->receivedFrame.frame[1], responseId,
						get_hex_serialframe(frameString, sizeof(frameString), &waveProtocol->receivedFrame));

				LOG_ERROR(g_Logger, "Failed to receive response: %d", response->statusDescription);

			}
		} else
			strcpy( response->statusDescription, "unexpected received frame during waiting for waveport applicative acknowledge" );
	}

	return response->status;
}
/*__________________________________________________________________________________________________*/

/**
 * Function used to waiting for frame after send command to port
 * @param waveProtocol - handle to waveport protocol instance
 * @param dynamicExpertParams - handle to dynamic expert params that contain wakeup length that used as timeout counter
 * @param isRequest - if true it is request command, otherwise is message command.
 * @param response - response code and message
 * @param responseId - response id for command
 * @return OK_PROTOCOL_SUCCESS if success,
 * KO_PROTOCOL_PROCESS_FAILURE if send failed;
 */
int wait_for_frame( ProtocolInfo* waveProtocol,
		const DynamicExpertParams* dynamicExpertParams,
		_bool isRequest,
		_bool isBroadcast,
		CommandResponse* response,
		WavePortResponses* responseId ) {

	response->status = KO_RADIO_TRANSMISSION_FAILURE;
	strcpy( response->statusDescription, "no waveport END_MESSAGE_EXCHANGE received" );

	struct timeval startTime;
	struct timeval endTime;

	unsigned int wakeupLength = 1100 ;

	if (dynamicExpertParams->isSet & FLAG_WAKEUP_LENGTH) {
		wakeupLength = dynamicExpertParams->wakeupLength;
		LOG_DEBUG( g_Logger, "wakeupLength: use dynamicExpertParams", 0);
	}
	else if (waveProtocol->params.writeDefaultRadioSettings) {
		wakeupLength = waveProtocol->params.dynamicExpertParams.wakeupLength;
		LOG_DEBUG( g_Logger, "wakeupLength: use value from properties file", 0);
	}else  {
		LOG_DEBUG( g_Logger, "wakeupLength: use default value =  %d", wakeupLength);
	}

	// for boradcast messages waveProtocol->relayNb is always 0
	long txTo = ( wakeupLength * 2 + 500 ) +
			waveProtocol->relayNb * ( wakeupLength + 500 );

	LOG_DEBUG( g_Logger, "wait_for_frame Input data is :", 0);
	LOG_DEBUG( g_Logger, "wait_for_frame Broadcast dynamicExpertParams = %d", dynamicExpertParams->broadcastSessionTimeout);
	LOG_DEBUG( g_Logger, "wait_for_frame Broadcast waveProtocol->params.dynamicExpertParams = %d",
			waveProtocol->params.dynamicExpertParams.broadcastSessionTimeout);

	LOG_DEBUG( g_Logger, "wait_for_frame : dynamicExpertParams = %d", dynamicExpertParams->responseTimeout);
	LOG_DEBUG( g_Logger, "wait_for_frame : waveProtocol->params.dynamicExpertParams = %d",
			waveProtocol->params.dynamicExpertParams.responseTimeout);

	if ( isRequest ) {
		if (isBroadcast) {
			LOG_DEBUG( g_Logger, "wait_for_frame : This is a broadcast request.", 0);
			if (dynamicExpertParams->isSet & FLAG_BROADCAST_SESSION_TIMEOUT) {
				LOG_DEBUG( g_Logger, "wait_for_frame : Broadcast session timeout is set to %d",
						dynamicExpertParams->broadcastSessionTimeout);
				txTo += dynamicExpertParams->broadcastSessionTimeout;
			} else if (waveProtocol->params.dynamicExpertParams.isSet & FLAG_BROADCAST_SESSION_TIMEOUT) {
				// use the last configured timeout value (static value in waveport radio modem)
				LOG_DEBUG( g_Logger, "wait_for_frame : Broadcast session timeout is not set, use the last configured value %d",
						waveProtocol->params.dynamicExpertParams.broadcastSessionTimeout);
				txTo += waveProtocol->params.dynamicExpertParams.broadcastSessionTimeout;
			} else if ((waveProtocol->params.isSet & FLAG_BROADCAST_SESSION_TIMEOUT) && waveProtocol->params.writeDefaultRadioSettings) {
				// use configured timeout value
				LOG_DEBUG( g_Logger, "wait_for_frame : Broadcast session timeout is not set, use the property value %d",
						waveProtocol->params.dynamicExpertParams.broadcastSessionTimeout);
				txTo += waveProtocol->params.dynamicExpertParams.broadcastSessionTimeout;
			} else {
				// use default timeout value
				LOG_DEBUG( g_Logger, "wait_for_frame : Broadcast session timeout is not set and configured, use the default value 6000.",0);
				txTo += 6000;
			}
		} else {
			LOG_DEBUG( g_Logger, "wait_for_frame : This is a unicast request.", 0);
			if (dynamicExpertParams->isSet & FLAG_RESPONSE_TIMEOUT) {
				LOG_DEBUG( g_Logger, "wait_for_frame : Response timeout is set to %d",
						dynamicExpertParams->responseTimeout);
				txTo += dynamicExpertParams->responseTimeout;
			} else if (waveProtocol->params.dynamicExpertParams.isSet & FLAG_RESPONSE_TIMEOUT) {
				// use the last configured timeout value (static value in waveport radio modem)
				LOG_DEBUG( g_Logger, "wait_for_frame : Response timeout is not set, use the last configured value %d",
						waveProtocol->params.dynamicExpertParams.responseTimeout);
				txTo += waveProtocol->params.dynamicExpertParams.responseTimeout;
			} else if ((waveProtocol->params.dynamicExpertParams.isSet & FLAG_RESPONSE_TIMEOUT) && waveProtocol->params.writeDefaultRadioSettings) {
				// use configured timeout value
				LOG_DEBUG( g_Logger, "wait_for_frame : Response timeout is not set, use the property value %d",
						waveProtocol->params.dynamicExpertParams.responseTimeout);
				txTo += waveProtocol->params.dynamicExpertParams.responseTimeout;
			} else {
				// use default timeout value
				LOG_DEBUG( g_Logger, "wait_for_frame : Response timeout is not set and configured, use the default value 2000.",0);
				txTo += 2000;
			}
		}
	}

	LOG_DEBUG( g_Logger, "wait_for_frame : %d", txTo );
	gettimeofday( &startTime, 0 );

	// waiting for frame
	while ( waveProtocol->receivedFrame.len == 0 ) {
		// wait for message body
		gettimeofday( &endTime, 0 );
		if ( millisecondtimediff( startTime, endTime ) > txTo ) {
			LOG_ERROR( g_Logger, "wait_for_frame : Timeout occurred! ", 0);
			break;
		}
		SLEEP( 10 );
	}

	LOG_DEBUG( g_Logger, "wait_for_frame : Exit from wait loop with status %d and received frame length %d",
			response->status, waveProtocol->receivedFrame.len);

	if ( waveProtocol->receivedFrame.len > 0 ) {

		WavePortResponses cmd = waveProtocol->receivedFrame.frame[0];

		if ( responseId )
			*responseId = cmd;
		response->status = OK_PROTOCOL_SUCCESS;
		response->statusDescription[0] = 0;

		LOG_DEBUG( g_Logger, "wait_for_frame : Received frame with command 0x%02x", cmd);
	}

	return response->status ;
}

/*__________________________________________________________________________________________________*/

/**
 * Function used to send read command frames
 * @param waveProtocol - handle to waveport protocol instance
 * @param sf - serial frame that contain read command, requested parameter id
 * @param wprCode - response id for read command
 * @param response - response code and message
 * @param responseFrame - frame that contain requested value
 * @return OK_PROTOCOL_SUCCESS if success,
 * KO_PROTOCOL_PROCESS_FAILURE if send failed;
 */
int send_frame( ProtocolInfo* waveProtocol,
		SerialFrame* sf,
		WavePortResponses wprCode,
		CommandResponse* response,
		SerialFrame* responseFrame ) {
	int iRepeatIfError = 2;

	/*
	 * Before we send a new command we should be certain that previous command already handled ( let Waveport to finalize work with prev command)
	 */
	SLEEP( 10 );

	// prepare seriad exchange data
	SerialExchangeData sed = create_serial_exchange_data(*sf);
	char frameString[MAX_SERIALFRAME_LEN*2+1] = "";

	do {
		waveProtocol->receivedFrame.len = 0;
		response->status = waveProtocol->serialDriverHandler.send( waveProtocol->serialDriver, &sed );
		LOG_DEBUG( g_Logger, "send_frame AFTER  error:%s %s", get_sending_status_name(response->status),
				get_hex_serialframe(frameString, sizeof(frameString), sf)  );
		if ( OK_SERIALDRIVER_SUCCESS == (int)response->status ) {
			wait_for_response( waveProtocol, response, wprCode );
			iRepeatIfError = 0;
		} else {
			if ( KO_SENDING_FAILURE == (int)response->status ) {
				--iRepeatIfError;
				SLEEP( 20 );
			} else
				iRepeatIfError = 0; // other error occur (KO_SERIALFRAME_REQUIRED:KO_SERIALDRIVER_NOT_OPEN)!
			sprintf( response->statusDescription, "fails to send command on serial driver error code:%s", get_sending_status_name( response->status ) );
			LOG_ERROR( g_Logger, "send_frame : Failed to send data$ %s ", response->statusDescription);

			response->status = KO_PROTOCOL_PROCESS_FAILURE;
		}
	} while ( ((int)response->status != OK_SERIALDRIVER_SUCCESS) && (iRepeatIfError > 0) );

	if ( response->status == OK_PROTOCOL_SUCCESS ) {
		if ( responseFrame ) {
			// if no specific response is waited for, return all received frame
			if (wprCode == (byte)-1) {
				responseFrame->len = waveProtocol->receivedFrame.len;
				memcpy( responseFrame->frame, waveProtocol->receivedFrame.frame, responseFrame->len );
			// else extract the 2 first byte (command id and status)
			} else if ((waveProtocol->receivedFrame.len-2) > 0 ) {
				responseFrame->len = waveProtocol->receivedFrame.len-2;
				memcpy( responseFrame->frame, waveProtocol->receivedFrame.frame+2, responseFrame->len );
			}
		}
	} else if ( responseFrame && waveProtocol->receivedFrame.len > 0 ) {
		memcpy( responseFrame, &waveProtocol->receivedFrame, sizeof(waveProtocol->receivedFrame) );
		// TODO notify spontaneous message. actual process must be exited in KO_PROTOCOL_PROCESS_FAILURE
	}
	waveProtocol->receivedFrame.len = 0;

	LOG_DEBUG( g_Logger, "send_frame : Exiting with code %d ", response->status);
	return response->status;
}
/*__________________________________________________________________________________________________*/

int send_frame_from_payload( ProtocolInfo* waveProtocol,
					Payload* payload,
					WavePortResponses wprCode,
					CommandResponse* response,
					SerialFrame* responseFrame ) {

	SerialFrame serialFrame;

	memcpy(&serialFrame, payload, sizeof(SerialFrame));
	serialFrame.len = payload->len > MAX_SERIALFRAME_LEN ?
			MAX_SERIALFRAME_LEN : payload->len;

	LOG_DEBUG(g_Logger, "Sendig frame of length %d with first byt 0x%02x", serialFrame.len, serialFrame.frame[0]);
	return send_frame(waveProtocol, &serialFrame, wprCode, response, responseFrame);
}


/*==================================================================================================*/
int write_dynamic_standard_params( ProtocolInfo* waveProtocol, DynamicStandardParams* dynamicStandardParams, CommandResponse* response ) {
	response->status = KO_PROTOCOL_NO_VALID_PARAMETER;

	waveProtocol->relayNb = 0;
	int i = 0;

	// is node address valid?
	int len = strlen( ((RadioAddress*)&dynamicStandardParams->nodeAddress)->addressValue );
	if ( len == (MAX_RADIOADDRESS_LEN-1) ) {
		if ( !is_hex( ((RadioAddress*)&dynamicStandardParams->nodeAddress)->addressValue ) )
			return response->status = KO_PARAMETER_NODE_ADDRESS_NOT_VALID;
		response->status = OK_PROTOCOL_SUCCESS;
	} else if ( len != 0 )
		return response->status = KO_PARAMETER_NODE_ADDRESS_NOT_VALID;
	/* (len == 0) node address is not filled, it is OK */

	// is relay route valid?
	LOG_DEBUG(g_Logger, "Check if this is a relayed message [%d] ...", ARRAYSIZE(dynamicStandardParams->routerPath));
	for ( i = 0; i < ARRAYSIZE(dynamicStandardParams->routerPath); ++i ) {

		int len = strlen( ((RadioAddress*)&dynamicStandardParams->routerPath[i])->addressValue );
		LOG_DEBUG(g_Logger, "The length of path [%d]  is: %d", i, len);
		if ( len == (MAX_RADIOADDRESS_LEN-1) ) {
			if ( !is_hex( ((RadioAddress*)&dynamicStandardParams->routerPath[i])->addressValue ) )
				return response->status = KO_PARAMETER_RELAY_ROUTE_NOT_VALID;
			++waveProtocol->relayNb;
		} else if ( len != 0 ) {
			return response->status = KO_PARAMETER_RELAY_ROUTE_NOT_VALID;
		} else /* (len == 0) no additional path available */
			break;
	}

	// is group id valid?
	if ( dynamicStandardParams->groupId != -1 &&
			(dynamicStandardParams->groupId < 0 || dynamicStandardParams->groupId > 0xFF) )
		return response->status = KO_PARAMETER_GROUP_NUMBER_NOT_VALID;

	// set command
	SerialFrame sf;
	int index = 0;
	memset( &sf, 0, sizeof(sf) );
	sf.frame[index++] = REQ_WRITE_RADIO_PARAM;

	// router path filled in?
	if ( waveProtocol->relayNb > 0 ) {
		sf.frame[index++] = RELAY_ROUTE_STATUS;
		sf.frame[index++] = 1;// enable relay route !
		sf.len = index;
		if ( OK_PROTOCOL_SUCCESS == send_frame( waveProtocol, &sf, RES_WRITE_RADIO_PARAM, response, 0) ) {
			index = 1;
			// set command parameter
			sf.frame[index++] = RELAY_ROUTE;
			sf.frame[index++] = waveProtocol->relayNb;
			char *stopstring;
			for ( i = 0; i < waveProtocol->relayNb; ++i ) {
				int j = 0;
				for ( j = 0; j < (MAX_RADIOADDRESS_LEN - 1) ; j += 2 ) {
					char buff = ((RadioAddress*)&dynamicStandardParams->routerPath[i])->addressValue[j+2];
					((RadioAddress*)&dynamicStandardParams->routerPath[i])->addressValue[j+2] = 0;
					sf.frame[index++] = (byte)strtol( ((RadioAddress*)&dynamicStandardParams->routerPath[i])->addressValue + j, &stopstring, 16 );
					((RadioAddress*)&dynamicStandardParams->routerPath[i])->addressValue[j+2] = buff;
				}
			}
			sf.len = index;
			send_frame( waveProtocol, &sf, RES_WRITE_RADIO_PARAM, response, 0);
			if ( response->status == KO_PROTOCOL_PARAM_UPDATE_ERROR || response->status == KO_PROTOCOL_PARAM_READ_ERROR )
				response->status = KO_PARAMETER_RELAY_ROUTE_VALUE_NOT_SUPPORTED;
			else if ( response->status != OK_PROTOCOL_SUCCESS ) {
				response->status = KO_PARAM_WRITING_FAILURE;
				strcpy( response->statusDescription, "RELAY_ROUTE" );
			}
		} else {
			response->status = KO_PARAM_WRITING_FAILURE;
			strcpy( response->statusDescription, "RELAY_ROUTE_STATUS" );
		}
		index = 1;
	} else {
		sf.frame[index++] = RELAY_ROUTE_STATUS;
		sf.frame[index++] = 0;// disable relay route !
		sf.len = index;
		if ( OK_PROTOCOL_SUCCESS == send_frame( waveProtocol, &sf, RES_WRITE_RADIO_PARAM, response, 0) ) {
			index = 1;
			sf.frame[index++] = RELAY_ROUTE;
			sf.frame[index++] = 0;// disable relay route !
			sf.len = index;
			if ( OK_PROTOCOL_SUCCESS != send_frame( waveProtocol, &sf, RES_WRITE_RADIO_PARAM, response, 0) ) {
				response->status = KO_PARAM_WRITING_FAILURE;
				strcpy( response->statusDescription, "RELAY_ROUTE" );
			}
		} else {
			response->status = KO_PARAM_WRITING_FAILURE;
			strcpy( response->statusDescription, "RELAY_ROUTE_STATUS" );
		}
		index = 1;
	}

	// group id filled in?
	if ( (response->status == OK_PROTOCOL_SUCCESS || response->status == KO_PROTOCOL_NO_VALID_PARAMETER) &&
			dynamicStandardParams->groupId != -1 ) {
		// set command parameter
		sf.frame[index++] = GROUP_NUMBER;
		sf.frame[index++] = (byte)dynamicStandardParams->groupId;
		sf.len = index;
		send_frame( waveProtocol, &sf, RES_WRITE_RADIO_PARAM, response, 0);
		if ( response->status == KO_PROTOCOL_PARAM_UPDATE_ERROR || response->status == KO_PROTOCOL_PARAM_READ_ERROR )
			response->status = KO_PARAMETER_GROUP_NUMBER_VALUE_NOT_SUPPORTED;
		else if ( response->status != OK_PROTOCOL_SUCCESS ) {
			response->status = KO_PARAM_WRITING_FAILURE;
			strcpy( response->statusDescription, "GROUP_NUMBER" );
		}
	}

	if ( (response->status == OK_PROTOCOL_SUCCESS || response->status == KO_PROTOCOL_NO_VALID_PARAMETER) &&
		 strlen( ((RadioAddress*)&dynamicStandardParams->nodeAddress)->addressValue ) == (MAX_RADIOADDRESS_LEN-1) )
		response->status = OK_PROTOCOL_SUCCESS;

	return response->status;
}
/*__________________________________________________________________________________________________*/

/*==================================================================================================*/
int write_dynamic_expert_params( ProtocolInfo* waveProtocol, DynamicExpertParams* dynamicExpertParams, CommandResponse* response ) {

	if ( dynamicExpertParams->isSet == 0 )
		return response->status = OK_PROTOCOL_SUCCESS;

	LOG_DEBUG(g_Logger, "write_dynamic_expert_params : Entered, test parameter ranges", 0);
	// 1. wakeup length filled in?
	if ( dynamicExpertParams->isSet & FLAG_WAKEUP_LENGTH ) {
		// wakeup length is valid?
		if ( dynamicExpertParams->wakeupLength < WAKEUP_LENGTH_MIN ||
				dynamicExpertParams->wakeupLength > WAKEUP_LENGTH_MAX )
			return response->status = KO_PARAMETER_WAKEUP_LENGTH_NOT_VALID;
	}

	// 2. response timeout filled in?
	if ( dynamicExpertParams->isSet & FLAG_RESPONSE_TIMEOUT ) {
		// response timeout is valid?
		// Value in multiples of 100ms, so valid value must devide by 100 without remainder
		if ( dynamicExpertParams->responseTimeout < RADIO_USER_TIMEOUT_MIN ||
				dynamicExpertParams->responseTimeout > RADIO_USER_TIMEOUT_MAX ||
				((dynamicExpertParams->responseTimeout % 100) != 0) )
			return response->status = KO_PARAMETER_RESPONSE_TIMEOUT_NOT_VALID;
	}
	// 3. broadcast session timeout filled in?
	if ( dynamicExpertParams->isSet & FLAG_BROADCAST_SESSION_TIMEOUT ) {
		// broadcast session timeout is valid?
		// Value in multiples of 100ms, so valid value must divide by 100 without remainder
		if ( dynamicExpertParams->broadcastSessionTimeout < BCST_RECEPTION_TIMEOUT_MIN ||
				dynamicExpertParams->broadcastSessionTimeout > BCST_RECEPTION_TIMEOUT_MAX ||
				((dynamicExpertParams->broadcastSessionTimeout % 100) != 0) )
			return response->status = KO_PARAMETER_BROADCAST_SESSION_TIMEOUT_NOT_VALID;
	}
	// 4. number max of response waiting for broadcast filled in?
	if ( dynamicExpertParams->isSet & FLAG_BROADCAST_SESSION_MAX_RESPONSES_NB ) {
		// number max of response waiting for broadcast is valid?
		if ( dynamicExpertParams->broadcastSessionMaxResponsesNb <= 0 ||
				dynamicExpertParams->broadcastSessionMaxResponsesNb >= 0xFF )
			return response->status = KO_PARAMETER_BROADCAST_SESSION_MAX_RESPONSES_NB_NOT_VALID;
	}
	// 5. transmission radio channel filled in?
	if ( dynamicExpertParams->isSet & FLAG_TRANSMISSION_RADIO_CHANNEL ) {
		// transmission radio channel is valid?
		if ( dynamicExpertParams->transmissionRadioChannel < TRANSMISSION_RADIO_CHANNEL_MIN &&
				dynamicExpertParams->transmissionRadioChannel > TRANSMISSION_RADIO_CHANNEL_MAX )
			return response->status = KO_PARAMETER_TRANSMISSION_RADIO_CHANNEL_NOT_VALID;
	}

	// transmission power level is normalized by conversion routine

	response->status = KO_PROTOCOL_NO_VALID_PARAMETER;

	SerialFrame sf;
	int index = 0;
	memset( &sf, 0, sizeof(sf) );
	sf.frame[index++] = REQ_WRITE_RADIO_PARAM;

	/**************************************************************************/
	LOG_DEBUG(g_Logger, "write_dynamic_expert_params : Start writing of parameters", 0);
	// 1. wakeup length filled in?
	if ( dynamicExpertParams->isSet & FLAG_WAKEUP_LENGTH ) {
		sf.frame[index++] = WAKEUP_LENGTH;
		sf.frame[index++] = (byte)(dynamicExpertParams->wakeupLength & 0x00FF);
		sf.frame[index++] = (byte)(dynamicExpertParams->wakeupLength >> 8);
		sf.len = index;
		send_frame( waveProtocol, &sf, RES_WRITE_RADIO_PARAM, response, 0);
		if ( response->status == KO_PROTOCOL_PARAM_UPDATE_ERROR || response->status == KO_PROTOCOL_PARAM_READ_ERROR
				|| response->status == KO_PROTOCOL_PROCESS_FAILURE)
			response->status = KO_PARAMETER_WAKEUP_LENGTH_VALUE_NOT_SUPPORTED;
		else if ( response->status != OK_PROTOCOL_SUCCESS ) {
			response->status = KO_PARAM_WRITING_FAILURE;
			strcpy( response->statusDescription, "WAKEUP_LENGTH" );
		}
		index = 1;
	}

	LOG_DEBUG( g_Logger, "write_dynamic_expert_params: after FLAG_WAKEUP_LENGTH with status: %d %s", response->status, get_response_status_name( response->status ));

	// 2. response timeout filled in?
	if ( (response->status == OK_PROTOCOL_SUCCESS || response->status == KO_PROTOCOL_NO_VALID_PARAMETER) &&
			( dynamicExpertParams->isSet & FLAG_RESPONSE_TIMEOUT ) ) {
		sf.frame[index++] = RADIO_USER_TIMEOUT;
		sf.frame[index++] = (byte)(dynamicExpertParams->responseTimeout / 100);
		sf.len = index;
		send_frame( waveProtocol, &sf, RES_WRITE_RADIO_PARAM, response, 0);
		if ( response->status == KO_PROTOCOL_PARAM_UPDATE_ERROR || response->status == KO_PROTOCOL_PARAM_READ_ERROR
				|| response->status == KO_PROTOCOL_PROCESS_FAILURE)
			response->status = KO_PARAMETER_RESPONSE_TIMEOUT_VALUE_NOT_SUPPORTED;
		else if ( response->status != OK_PROTOCOL_SUCCESS ) {
			response->status = KO_PARAM_WRITING_FAILURE;
			strcpy( response->statusDescription, "RADIO_USER_TIMEOUT" );
		}
		index = 1;
	}
	LOG_DEBUG( g_Logger, "write_dynamic_expert_params: after FLAG_RESPONSE_TIMEOUT with status: %d %s", response->status, get_response_status_name( response->status ));

	// 3. broadcast session timeout filled in?
	if ( (response->status == OK_PROTOCOL_SUCCESS || response->status == KO_PROTOCOL_NO_VALID_PARAMETER) &&
			( dynamicExpertParams->isSet & FLAG_BROADCAST_SESSION_TIMEOUT ) ) {
		sf.frame[index++] = BCST_RECEPTION_TIMEOUT;
		sf.frame[index++] = (byte)(dynamicExpertParams->broadcastSessionTimeout / 100);
		sf.len = index;
		send_frame( waveProtocol, &sf, RES_WRITE_RADIO_PARAM, response, 0);
		if ( response->status == KO_PROTOCOL_PARAM_UPDATE_ERROR || response->status == KO_PROTOCOL_PARAM_READ_ERROR
				|| response->status == KO_PROTOCOL_PROCESS_FAILURE)
			response->status = KO_PARAMETER_BROADCAST_SESSION_TIMEOUT_VALUE_NOT_SUPPORTED;
		else if ( response->status != OK_PROTOCOL_SUCCESS ) {
			response->status = KO_PARAM_WRITING_FAILURE;
			strcpy( response->statusDescription, "BCST_RECEPTION_TIMEOUT" );
		}
		index = 1;
	}
	LOG_DEBUG( g_Logger, "write_dynamic_expert_params: after FLAG_BROADCAST_SESSION_TIMEOUT with status: %d %s", response->status, get_response_status_name( response->status ));

	// 4. number max of response waiting for broadcast filled in?
	if ( (response->status == OK_PROTOCOL_SUCCESS || response->status == KO_PROTOCOL_NO_VALID_PARAMETER) &&
			( dynamicExpertParams->isSet & FLAG_BROADCAST_SESSION_MAX_RESPONSES_NB ) ) {
		sf.frame[index++] = BCST_MAX_RESPONSES_NB;
		sf.frame[index++] = (byte)dynamicExpertParams->broadcastSessionMaxResponsesNb;
		sf.len = index;
		send_frame( waveProtocol, &sf, RES_WRITE_RADIO_PARAM, response, 0);
		if ( response->status == KO_PROTOCOL_PARAM_UPDATE_ERROR || response->status == KO_PROTOCOL_PARAM_READ_ERROR
				|| response->status == KO_PROTOCOL_PROCESS_FAILURE)
			response->status = KO_PARAMETER_BROADCAST_SESSION_MAX_RESPONSES_NB_VALUE_NOT_SUPPORTED;
		else if ( response->status != OK_PROTOCOL_SUCCESS ) {
			response->status = KO_PARAM_WRITING_FAILURE;
			strcpy( response->statusDescription, "BCST_MAX_RESPONSES_NB" );
		}
	}
	LOG_DEBUG( g_Logger, "write_dynamic_expert_params: after FLAG_BROADCAST_SESSION_MAX_RESPONSES_NB with status: %d %s", response->status, get_response_status_name( response->status ));

	index = 0;
	// 5. transmission radio channel filled in?
	if ( (response->status == OK_PROTOCOL_SUCCESS || response->status == KO_PROTOCOL_NO_VALID_PARAMETER) &&
			( dynamicExpertParams->isSet & FLAG_TRANSMISSION_RADIO_CHANNEL ) ) {
		sf.frame[index++] = REQ_SELECT_CHANNEL;
		sf.frame[index++] = (byte)dynamicExpertParams->transmissionRadioChannel;
		sf.len = index;
		send_frame( waveProtocol, &sf, RES_SELECT_CHANNEL, response, 0);
		if ( response->status == KO_PROTOCOL_PARAM_UPDATE_ERROR || response->status == KO_PROTOCOL_PARAM_READ_ERROR
				|| response->status == KO_PROTOCOL_PROCESS_FAILURE)
			response->status = KO_PARAMETER_TRANSMISSION_RADIO_CHANNEL_VALUE_NOT_SUPPORTED;
		else if ( response->status != OK_PROTOCOL_SUCCESS ) {
			response->status = KO_PARAM_WRITING_FAILURE;
			strcpy( response->statusDescription, "REQ_SELECT_CHANNEL" );
		}
		index = 0;
	}
	LOG_DEBUG( g_Logger, "write_dynamic_expert_params: after FLAG_TRANSMISSION_RADIO_CHANNEL with status: %d %s", response->status, get_response_status_name( response->status ));

	// 6. transmission power level filled in?
	if ( (response->status == OK_PROTOCOL_SUCCESS || response->status == KO_PROTOCOL_NO_VALID_PARAMETER) &&
			( dynamicExpertParams->isSet & FLAG_TRANSMISSION_POWER_LEVEL ) ) {


		byte transmissionPowerLevelPayload = 0;

		transmissionPowerLevelToPayload (dynamicExpertParams->transmissionPowerLevel,
				&transmissionPowerLevelPayload);

		sf.frame[index++] = REQ_CHANGE_TX_POWER;
		sf.frame[index++] = transmissionPowerLevelPayload;
		sf.len = index;
		send_frame( waveProtocol, &sf, RES_CHANGE_TX_POWER, response, 0);
		if ( response->status == KO_PROTOCOL_PARAM_UPDATE_ERROR || response->status == KO_PROTOCOL_PARAM_READ_ERROR
				|| response->status == KO_PROTOCOL_PROCESS_FAILURE)
			response->status = KO_PARAMETER_TRANSMISSION_POWER_LEVEL_VALUE_NOT_SUPPORTED;
		else if ( response->status != OK_PROTOCOL_SUCCESS ) {
			response->status = KO_PARAM_WRITING_FAILURE;
			strcpy( response->statusDescription, "REQ_CHANGE_TX_POWER" );
		}
	}
	LOG_DEBUG( g_Logger, "write_dynamic_expert_params: after FLAG_TRANSMISSION_POWER_LEVEL with status: %d %s", response->status, get_response_status_name( response->status ));

	return response->status;
}
/*__________________________________________________________________________________________________*/

/*==================================================================================================*/
int write_static_standard_params( ProtocolInfo* waveProtocol, StaticStandardParams* staticStandardParams, CommandResponse* response ) {
	response->status = OK_PROTOCOL_SUCCESS;

	LOG_DEBUG( g_Logger, "write_static_standard_params: Entered", 0);
	if ( staticStandardParams->isSet == 0 ) {
		return response->status;
	}

	LOG_DEBUG( g_Logger, "write_static_standard_params: there are some params, try to write them", 0);

	response->status = KO_PROTOCOL_NO_VALID_PARAMETER;

	// RadioAckActivation filled in ?
	if ( staticStandardParams->isSet & FLAG_RADIO_ACK_ACTIVATION ) {
		LOG_DEBUG( g_Logger, "write_static_standard_params: FLAG_RADIO_ACK_ACTIVATION is filled in", 0);
		// is RadioAckActivation valid ?
		if ( staticStandardParams->radioAckActivation != _false &&
				staticStandardParams->radioAckActivation != _true ){
			LOG_DEBUG( g_Logger, "write_static_standard_params: FLAG_RADIO_ACK_ACTIVATION not valid", 0);
			return response->status = KO_PARAMETER_RADIO_ACK_ACTIVATION_NOT_VALID;
		}
		LOG_WARN( g_Logger, "Radio acknowledge activation is not accessible in Waveport!", 0 );
	}

	// firstScanActivityPeriod filled in ?
	if ( staticStandardParams->isSet & FLAG_ACTIVITY_PERIOD_ON_FIRST_SCAN ) {
		// is firstScanActivityPeriod valid ?
		if ( staticStandardParams->firstScanActivityPeriod < ACTIVITY_PERIOD_MIN ||
				staticStandardParams->firstScanActivityPeriod > ACTIVITY_PERIOD_MAX ||
				( staticStandardParams->firstScanActivityPeriod % 100 ) != 0 )
			return response->status = KO_PARAMETER_FIRST_SCAN_ACTIVITY_PERIOD_NOT_VALID;
	}
	// switchingModeOnRadioAddressActivation filled in ?
	if ( staticStandardParams->isSet & FLAG_SWITCHING_MODE_ON_RADIO_ADDRESS_ACTIVATION ) {
		// is switchingModeOnRadioAddressActivation valid ?
		if ( staticStandardParams->switchingModeOnRadioAddressActivation != _false &&
				staticStandardParams->switchingModeOnRadioAddressActivation != _true )
			return response->status = KO_PARAMETER_SWITCHING_MODE_ON_RADIO_ADDRESS_ACTIVATION_NOT_VALID;
	}
	// receptionGroupId filled in ?
	if ( staticStandardParams->isSet & FLAG_RECEPTION_GROUPID ) {
		// is receptionGroupId valid ?
		if ( staticStandardParams->receptionGroupId > 0xFF )
			return response->status = KO_PARAMETER_RECEPTION_GROUP_ID_NOT_VALID;
	}

	// set command
	SerialFrame sf;
	int index = 0;
	memset( &sf, 0, sizeof(sf) );
	sf.frame[index++] = REQ_WRITE_RADIO_PARAM;

	// Note: RadioAckActivation is not accessible from Waveport (always true)
	LOG_DEBUG( g_Logger, "write_static_standard_params: Try to write FLAG_RADIO_ACK_ACTIVATION in port", 0);
	if ( staticStandardParams->isSet & FLAG_RADIO_ACK_ACTIVATION ) {
		LOG_DEBUG( g_Logger, "write_static_standard_params: Flag FLAG_RADIO_ACK_ACTIVATION is valid, compose frame", 0);
		sf.frame[index++] = RADIO_ACKNOWLEDGE;
		sf.frame[index++] = ( staticStandardParams->radioAckActivation == _true ) ? 1 : 0;
		sf.len = index;

		LOG_DEBUG( g_Logger, "write_static_standard_params: Sending frame for FLAG_RADIO_ACK_ACTIVATION value = %s",
				staticStandardParams->radioAckActivation == _true ? "true" : "false");

		send_frame( waveProtocol, &sf, RES_WRITE_RADIO_PARAM, response, 0);
		if ( response->status == KO_PROTOCOL_PARAM_UPDATE_ERROR || response->status == KO_PROTOCOL_PARAM_READ_ERROR ) {
			LOG_DEBUG( g_Logger, "write_static_standard_params: Param FLAG_RADIO_ACK_ACTIVATION not suppoerted", 0);
			response->status = KO_PARAMETER_RADIO_ACK_ACTIVATION_VALUE_NOT_SUPPORTED;
		}
		else if ( response->status != OK_PROTOCOL_SUCCESS ) {
			LOG_DEBUG( g_Logger, "write_static_standard_params: Failed to write param FLAG_RADIO_ACK_ACTIVATION ", 0);
			response->status = KO_PARAM_WRITING_FAILURE;
			strcpy( response->statusDescription, "RADIO_ACKNOWLEDGE" );
		}

		index = 1;
	}

	response->status = OK_PROTOCOL_SUCCESS;

	if ( staticStandardParams->isSet & FLAG_ACTIVITY_PERIOD_ON_FIRST_SCAN ) {
		sf.frame[index++] = AWAKENING_PERIOD;
		sf.frame[index++] = (byte)(staticStandardParams->firstScanActivityPeriod / 100);
		sf.len = index;
		send_frame( waveProtocol, &sf, RES_WRITE_RADIO_PARAM, response, 0);
		if ( response->status == KO_PROTOCOL_PARAM_UPDATE_ERROR || response->status == KO_PROTOCOL_PARAM_READ_ERROR )
			response->status = KO_PARAMETER_FIRST_SCAN_ACTIVITY_PERIOD_VALUE_NOT_SUPPORTED;
		else if ( response->status != OK_PROTOCOL_SUCCESS ) {
			response->status = KO_PARAM_WRITING_FAILURE;
			strcpy( response->statusDescription, "AWAKENING_PERIOD" );
		}
		index = 1;
	}

	if ( response->status == OK_PROTOCOL_SUCCESS &&
			staticStandardParams->isSet & FLAG_SWITCHING_MODE_ON_RADIO_ADDRESS_ACTIVATION ) {
		sf.frame[index++] = SWITCH_MODE_STATUS;
		sf.frame[index++] = ( staticStandardParams->switchingModeOnRadioAddressActivation == _true ) ? 1 : 0;
		sf.len = index;
		send_frame( waveProtocol, &sf, RES_WRITE_RADIO_PARAM, response, 0);
		if ( response->status == KO_PROTOCOL_PARAM_UPDATE_ERROR || response->status == KO_PROTOCOL_PARAM_READ_ERROR )
			response->status = KO_PARAMETER_SWITCHING_MODE_ON_RADIO_ADDRESS_ACTIVATION_VALUE_NOT_SUPPORTED;
		else if ( response->status != OK_PROTOCOL_SUCCESS ) {
			response->status = KO_PARAM_WRITING_FAILURE;
			strcpy( response->statusDescription, "SWITCH_MODE_STATUS" );
		}
		index = 1;
	}

	if ( response->status == OK_PROTOCOL_SUCCESS &&
			staticStandardParams->isSet & FLAG_RECEPTION_GROUPID ) {
		sf.frame[index++] = WAVECARD_MULTICAST_GROUP;
		sf.frame[index++] = staticStandardParams->receptionGroupId;
		sf.len = index;
		send_frame( waveProtocol, &sf, RES_WRITE_RADIO_PARAM, response, 0);
		if ( response->status == KO_PROTOCOL_PARAM_UPDATE_ERROR || response->status == KO_PROTOCOL_PARAM_READ_ERROR )
			response->status = KO_PARAMETER_RECEPTION_GROUP_ID_VALUE_NOT_SUPPORTED;
		else if ( response->status != OK_PROTOCOL_SUCCESS ) {
			response->status = KO_PARAM_WRITING_FAILURE;
			strcpy( response->statusDescription, "WAVECARD_MULTICAST_GROUP" );
		}
		index = 1;
	}

	return response->status;
}
/*__________________________________________________________________________________________________*/

/*==================================================================================================*/
int write_static_expert_params( ProtocolInfo* waveProtocol, StaticExpertParams* staticExpertParams, CommandResponse* response ) {
	response->status = OK_PROTOCOL_SUCCESS;

	unsigned short firstScanPhysicalConfigPayload = 0;
	unsigned short secondScanPhysicalConfigPayload = 0;

	LOG_DEBUG( g_Logger, "write_static_expert_params: Entered", 0);

	if ( staticExpertParams->isSet == 0 ) {
		return response->status;
	}

	response->status = KO_PROTOCOL_NO_VALID_PARAMETER;

	// SecondScanEnableActivation filled in ?
	if ( staticExpertParams->isSet & FLAG_SECOND_SCAN_ENABLE_ACTIVATION ) {
		// is SecondScanEnableActivation valid ?
		if ( staticExpertParams->secondScanEnableActivation != _false &&
				staticExpertParams->secondScanEnableActivation != _true )
			return response->status = KO_PARAMETER_SECOND_SCAN_ENABLE_ACTIVATION_NOT_VALID;
	}


	// RadioCRCActivation filled in ?
	if ( staticExpertParams->isSet & FLAG_RADIO_CRC_ACTIVATION ) {
		// is RadioCRCActivation valid ?
		if ( staticExpertParams->radioCRCActivation != _false &&
				staticExpertParams->radioCRCActivation != _true )
			return response->status = KO_PARAMETER_RADIO_CRC_ACTIVATION_NOT_VALID;
	}

	// FirstScanPhysicalConfig filled in ?
	if ( staticExpertParams->isSet & FLAG_PHYSICAL_CONFIG_ON_FIRST_SCAN ) {
		if (! physicalLayerToPayload (staticExpertParams->firstScanPhysicalConfig,
				&firstScanPhysicalConfigPayload)){
			return response->status = KO_PARAMETER_FIRST_SCAN_PHYSICAL_CONFIG_NOT_VALID;
		}
	}

	// SecondScanPhysicalConfig filled in ?
	if ( staticExpertParams->isSet & FLAG_PHYSICAL_CONFIG_ON_SECOND_SCAN ) {
		if (! physicalLayerToPayload (staticExpertParams->secondScanPhysicalConfig,
				&secondScanPhysicalConfigPayload)){
			return response->status = KO_PARAMETER_SECOND_SCAN_PHYSICAL_CONFIG_NOT_VALID;
		}

	}

	LOG_DEBUG( g_Logger, "write_static_expert_params: Start writting params", 0);

	// set command
	SerialFrame sf;
	int index = 0;
	memset( &sf, 0, sizeof(sf) );
	sf.frame[index++] = REQ_WRITE_RADIO_PARAM;

	response->status = OK_PROTOCOL_SUCCESS;
	if ( staticExpertParams->isSet & FLAG_SECOND_SCAN_ENABLE_ACTIVATION ) {
		sf.frame[index++] = SECOND_SCAN_ENABLE_ACTIVATION;
		sf.frame[index++] = ( staticExpertParams->secondScanEnableActivation == _true ) ? 1 : 0;
		sf.len = index;
		send_frame( waveProtocol, &sf, RES_WRITE_RADIO_PARAM, response, 0);
		if ( response->status == KO_PROTOCOL_PARAM_UPDATE_ERROR || response->status == KO_PROTOCOL_PARAM_READ_ERROR )
			response->status = KO_PARAMETER_SECOND_SCAN_ENABLE_ACTIVATION_VALUE_NOT_SUPPORTED;
		else if ( response->status != OK_PROTOCOL_SUCCESS ) {
			response->status = KO_PARAM_WRITING_FAILURE;
			strcpy( response->statusDescription, "SECOND_SCAN_ENABLE_ACTIVATION" );
		}
		index = 1;
	}

	LOG_DEBUG( g_Logger, "write_static_expert_params: after SECOND_SCAN_ENABLE_ACTIVATION status: %d %s", response->status, get_response_status_name( response->status ));

	if ( response->status == OK_PROTOCOL_SUCCESS &&
			staticExpertParams->isSet & FLAG_RADIO_CRC_ACTIVATION ) {
		index = 1;
		sf.frame[index++] = RADIO_CRC_ACTIVATION;
		sf.frame[index++] = ( staticExpertParams->radioCRCActivation == _true ) ? 1 : 0;
		sf.len = index;
		send_frame( waveProtocol, &sf, RES_WRITE_RADIO_PARAM, response, 0);
		index = 1;
	}
	LOG_DEBUG( g_Logger, "write_static_expert_params: after RADIO_CRC_ACTIVATION status: %d %s", response->status, get_response_status_name( response->status ));

	index = 0;
	if ( response->status == OK_PROTOCOL_SUCCESS &&
			staticExpertParams->isSet & FLAG_PHYSICAL_CONFIG_ON_SECOND_SCAN ) {
		sf.frame[index++] = REQ_SELECT_SECOND_PHYCONFIG;
		sf.frame[index++] = (byte)(secondScanPhysicalConfigPayload >> 8);
		sf.frame[index++] = (byte)(secondScanPhysicalConfigPayload & 0x00FF);
		sf.len = index;
		send_frame( waveProtocol, &sf, RES_SELECT_SECOND_PHYCONFIG, response, 0);
		if ( response->status == KO_PROTOCOL_PARAM_UPDATE_ERROR || response->status == KO_PROTOCOL_PARAM_READ_ERROR )
			response->status = KO_PARAMETER_SECOND_SCAN_PHYSICAL_CONFIG_VALUE_NOT_SUPPORTED;
		else if ( response->status != OK_PROTOCOL_SUCCESS ) {
			response->status = KO_PARAM_WRITING_FAILURE;
			strcpy( response->statusDescription, "REQ_SELECT_SECOND_PHYCONFIG" );
		}
		index = 0;
	}
	LOG_DEBUG( g_Logger, "write_static_expert_params: after REQ_SELECT_SECOND_PHYCONFIG status: %d %s", response->status, get_response_status_name( response->status ));

	if ( response->status == OK_PROTOCOL_SUCCESS &&
			staticExpertParams->isSet & FLAG_PHYSICAL_CONFIG_ON_FIRST_SCAN ) {
		sf.frame[index++] = REQ_SELECT_PHYCONFIG;
		sf.frame[index++] = (byte)(firstScanPhysicalConfigPayload >> 8);
		sf.frame[index++] = (byte)(firstScanPhysicalConfigPayload & 0x00FF);
		sf.len = index;
		send_frame( waveProtocol, &sf, RES_SELECT_PHYCONFIG, response, 0);
		if ( response->status == KO_PROTOCOL_PARAM_UPDATE_ERROR || response->status == KO_PROTOCOL_PARAM_READ_ERROR )
			response->status = KO_PARAMETER_FIRST_SCAN_PHYSICAL_CONFIG_VALUE_NOT_SUPPORTED;
		else if ( response->status != OK_PROTOCOL_SUCCESS ) {
			response->status = KO_PARAM_WRITING_FAILURE;
			strcpy( response->statusDescription, "REQ_SELECT_PHYCONFIG" );
		}
	}
	LOG_DEBUG( g_Logger, "write_static_expert_params: after REQ_SELECT_PHYCONFIG status: %d %s", response->status, get_response_status_name( response->status ));

	return response->status;
}
/*__________________________________________________________________________________________________*/

/*==================================================================================================*/

int parse_frame(ProtocolInfo* waveProtocol, const SerialFrame* frame, RadioExchangeCommand* command,
		int* status,
		int* totalNumber,
		int* index,
		int* types ) {
	int i = 0;
	int j = 0;
	byte cmd = 0;

	int ret = KO_UNEXPECTED_ERROR;
	if ( frame && frame->len && command ) {
		cmd = frame->frame[i];
		LOG_DEBUG(g_Logger, " +++++++++++++++++++++++ PARSE_FRAME 0x%02X", cmd);

		ret = OK_PROTOCOL_SUCCESS;
		memset( command, 0, sizeof(RadioExchangeCommand));
		switch ( cmd ) {
		case REQ_SEND_FRAME:
		case REQ_SEND_MESSAGE:
		case REQ_SEND_BROADCAST:
		case REQ_SEND_SERVICE:
		case REQ_SEND_SERVICE_MESSAGE:
		case RECEIVED_FRAME:
		case RECEIVED_BROADCAST_FRAME:

			// CMD [1 byte] Radio address of transmitter [6 bytes] data from received frame [len-7 bytes]
			if ( frame->len >= 7 ) {
				for ( ; j < 6; ++j )
					sprintf( ((RadioAddress*)&command->dynamicStandardParams.nodeAddress)->addressValue+j*2, "%02X", frame->frame[++i] );
			}

			LOG_DEBUG(g_Logger, "parse_frame: Saved radio address: %s", ((RadioAddress*)&command->dynamicStandardParams.nodeAddress)->addressValue);

			break;
		case RECEIVED_FRAME_POLLIN:
			// CMD [1 byte] STATUS_RECEPTION [1 byte] Radio address of queried module [6 bytes] data from received frame [len-8 bytes]
			if ( frame->len >= 8 ) {
				if ( status ) *status = frame->frame[++i]; // if 0 response OK, 1 no response from queried module
				for ( ; j < 6; ++j )
					sprintf( ((RadioAddress*)&command->dynamicStandardParams.nodeAddress)->addressValue+j*2, "%02X", frame->frame[++i] );
			}
			break;

		case RECEIVED_BROADCAST_RESPONSE:	// *
		case RECEIVED_MULTIFRAME:

			// CMD [1 byte] Status [1 byte] Total number [1 byte] Frame index [1 byte] Radio address of queried module [6 bytes] data from received frame [len-10 bytes]

			if ( frame->len >= 10 ) {

				if ( status ) *status = frame->frame[i + 1]; // if 0 reception OK, 1 : indicates that the number of received responses is higher than 255.
				                							// In this case, only the first 255 responses are forwarded to the host
				if ( totalNumber ) *totalNumber = frame->frame[i + 2];// Total number of frames received
				if ( index ) *index = frame->frame[i + 3];// Frame index

				i += 3;
				for ( ; j < 6; ++j )
					sprintf( ((RadioAddress*)&command->dynamicStandardParams.nodeAddress)->addressValue+j*2, "%02X", frame->frame[++i] );

			}
			break;
		case RECEIVED_FRAME_RELAYED:

			LOG_DEBUG(g_Logger, " +++++++++++++++++++++++ RELAYED FRAME RECEIVED", 0);

			// CMD [1 byte] Radio address of transmitter [6 bytes] Number of repeaters used [1 byte]
			// Radio address of repeaters [Number of repeaters used*6 bytes] data from received frame [len-X bytes]
			if ( frame->len >= 14 ) {
				for ( ; j < 6; ++j )
					sprintf( ((RadioAddress*)&command->dynamicStandardParams.nodeAddress)->addressValue+j*2, "%02X", frame->frame[++i] );

				int numRepeaters = (int)frame->frame[++i];
				int k = 0;
				int cnt = min( MAX_ROUTER_PATH, numRepeaters);

				for ( ; k < cnt; ++k ) {
					for ( j = 0; j < 6; ++j )
						sprintf( ((RadioAddress*)&command->dynamicStandardParams.routerPath[cnt - 1 - k])->addressValue+j*2, "%02X", frame->frame[++i] );

					LOG_DEBUG(g_Logger, "parse_frame: Saved relay %d address : %s", k,
							((RadioAddress*)&command->dynamicStandardParams.routerPath[k])->addressValue);
				}
			}

			break;
		case SERVICE_RESPONSE:
			// CMD [1 byte] Radio address of queried module [6 bytes] Service response type [1 byte] data from received frame [len-8 bytes]
			if ( frame->len >= 8 ) {

				for ( ; j < 6; ++j )
					sprintf( ((RadioAddress*)&command->dynamicStandardParams.nodeAddress)->addressValue+j*2,
							"%02X", frame->frame[++i] );
				if ( types ) *types = frame->frame[++i]; // response types
			}
			break;

		case RECEPTION_ERROR:
			LOG_ERROR(g_Logger, "parse_frame: RECEPTION_ERROR", 0);
			i = -1;
			ret = parse_reception_error( frame );
			break;

		default:
			i = -1;
			ret = KO_PROTOCOL_UNEXPECTED_FRAME;
			break;
		}
		//----------------------------------------------------------------------
		// *Save payload
		if ( ret == OK_PROTOCOL_SUCCESS ) {

			// save serial frame to response command
			int len = frame->len - (++i);
			memcpy( ((Payload*)&command->command.payload)->frame, frame->frame+i, len);
			((Payload*)&command->command.payload)->len = len;

			LOG_DEBUG(g_Logger, "parse_frame: Copied frame of len :%d", frame->len);

			char payload_text[2*256];
			memset(payload_text,0, sizeof(payload_text));

			int k = 0;
			for (; (k < sizeof(payload_text)) && (k < ((Payload*)&command->command.payload)->len); k ++)
			{
				sprintf(payload_text + k*2, "%02X", ((Payload*)&command->command.payload)->frame[k]);
			}
			LOG_DEBUG(g_Logger, "parse_frame: Received payload :%s", payload_text);
		}
	}

	// save node address
	if (waveProtocol) {
		LOG_DEBUG(g_Logger, " +++++++++++++++++++++++ SAVE NODE ADDRESS: %s",
				((RadioAddress*)&command->dynamicStandardParams.nodeAddress)->addressValue);

		if (strlen(((RadioAddress*)&command->dynamicStandardParams.nodeAddress)->addressValue)){
			memcpy((void*)&waveProtocol->senderNodeAddress,
					(const void*)&command->dynamicStandardParams.nodeAddress,
					sizeof(waveProtocol->senderNodeAddress));
		}
	}

	// setup unicast/broadcast modes
	if (cmd == RECEIVED_FRAME || cmd == RECEIVED_FRAME_RELAYED || cmd == SERVICE_RESPONSE) {
		LOG_DEBUG(g_Logger, "parse_frame: Set reception mode to unicast", 0);
		command->dynamicExpertParams.receptionMode = RECEPTION_MODE_UNICAST;
		command->dynamicExpertParams.isSet |= FLAG_RECEPTION_MODE;
		if (waveProtocol) waveProtocol->receptionSessionMode = RECEPTION_MODE_UNICAST;
	}
	else if (cmd == RECEIVED_BROADCAST_RESPONSE || cmd == RECEIVED_BROADCAST_FRAME) {
		command->dynamicExpertParams.receptionMode = RECEPTION_MODE_BROADCAST;
		command->dynamicExpertParams.isSet |= FLAG_RECEPTION_MODE;
		if (waveProtocol) waveProtocol->receptionSessionMode = RECEPTION_MODE_BROADCAST;
	} else {
		command->dynamicExpertParams.isSet &= ~FLAG_RECEPTION_MODE;	// reset flag
		if (waveProtocol) waveProtocol->receptionSessionMode = RECEPTION_MODE_UNDEFINED;
	}

	return ret;
}
/*__________________________________________________________________________________________________*/

/*==================================================================================================*/
int parse_reception_error( const SerialFrame* frame ) {
	int ret = KO_PROTOCOL_UNEXPECTED_FRAME;
	if ( frame && frame->len ) {
		if ( ( frame->frame[1] == 1 ) &&
				( frame->frame[2] == 1 || frame->frame[2] == 2 ) )
			ret = KO_RADIO_TRANSMISSION_FAILURE;
		else if ( frame->frame[1] == 2 && frame->frame[2] == 2 ) {
			switch ( frame->frame[3] ) {
			case 0:	ret = KO_RADIO_TRANSMISSION_FAILURE_WITH_DESTINATION_NODE; break;
			case 1:	ret = KO_RADIO_TRANSMISSION_FAILURE_WITH_FIRST_RELAY; break;
			case 2: ret = KO_RADIO_TRANSMISSION_FAILURE_WITH_SECOND_RELAY; break;
			case 3: ret = KO_RADIO_TRANSMISSION_FAILURE_WITH_THIRD_RELAY; break;
			default: break;
			}
		}
	}
	return ret;
}
/*__________________________________________________________________________________________________*/

/*==================================================================================================*/
int notify_spies( const ProtocolInfo* waveProtocol, const SerialFrame* frame, _bool sendEvent ) {
	// send spy event (Frame has been send with success! )
	RadioExchangeCommand command;
	parse_frame((ProtocolInfo*)waveProtocol, frame, &command, NULL, NULL, NULL, NULL );
	LOG_DEBUG(g_Logger, "notify_spies: Frame parsed, notify spies", 0);
	notify_spies_command( waveProtocol, &command, sendEvent );
	return OK_PROTOCOL_SUCCESS;
}
/*__________________________________________________________________________________________________*/

/*==================================================================================================*/
int notify_spies_command( const ProtocolInfo* waveProtocol, const RadioExchangeCommand* command, _bool sendEvent ) {
	int i = 0;
	_bool noListeners = _true;
	for ( ; i < MAX_SPY_NUM; ++i ) {
		protocol_spy_listener_t spy_listener = sendEvent ? waveProtocol->spiesSend[i] : waveProtocol->spiesRecv[i];
		if ( spy_listener ) {
			spy_listener( waveProtocol->paramSpies[i], i, command );
			noListeners = _false;
		}
	}

	// so we have c style Serial Driver and verify if we have JNI style listeners
#ifdef _JAVA_WRAPPER
	if ( noListeners && sendEvent ) {
		LOG_DEBUG(g_Logger, "Notify JNI style listeners", 0);
		jni_notify_spies_native_command( waveProtocol, command, sendEvent );
	}
#endif //_JAVA_WRAPPER

	return OK_PROTOCOL_SUCCESS;
}

/*__________________________________________________________________________________________________*/
int protocol_send_message( const void* protocol,
		const byte commandId,
		const byte responseId,
		const RadioExchangeCommand* command,
		CommandResponse* response )
{
	ProtocolInfo* waveProtocol = get_protocol( protocol );
//	_bool notWaitedFrameReceived = _false;
	if ( waveProtocol && response && command ) {


		memset( response, 0,  sizeof(*response) );
		// 0. Is protocol opened ?
		if ( !protocol_is_open( waveProtocol ) )
			return (response->status = KO_PROTOCOL_NOT_OPEN);

		// 1. A command session is already in progress?
		if ( waveProtocol->status != PS_PROCESS_IN_PROGRESS ) {
			waveProtocol->status = PS_PROCESS_IN_PROGRESS;

			//Point to Point mode  Max = 152 bytes of data
			//� Relay mode Max = 152 � (2 + 6 x Number of repeaters)
			//=> 1 repeater: 144 bytes of data
			//=> 2 repeaters: 138 bytes of data
			//=> 3 repeaters: 132 bytes of data


			// 2. Payload filled in?
			if ( ((Payload*)&command->command.payload)->len > 0 && ((Payload*)&command->command.payload)->len <= MAX_REQ_SEND_FRAME_BYTES_LEN ) {

				// 3. Node address filled in?
				if ( strlen( ((RadioAddress*)&command->dynamicStandardParams.nodeAddress)->addressValue ) == (MAX_RADIOADDRESS_LEN-1) ) {

					if (is_node_address_valid(((RadioAddress*)command->dynamicStandardParams.nodeAddress)->addressValue)) {

						RadioSettingsCommand radioSettingsCommand;
						memset( &radioSettingsCommand, 0,  sizeof(radioSettingsCommand) );
						memcpy( &radioSettingsCommand.dynamicStandardParams, &command->dynamicStandardParams, sizeof(DynamicStandardParams) );
						memcpy( &radioSettingsCommand.dynamicExpertParams, &command->dynamicExpertParams, sizeof(DynamicExpertParams) );

						// put protocol status in idle mode
						waveProtocol->status = PS_IDLE;
						// printf("protocol_send_message:: Writing radio params ... \n");
						protocol_write_radio_params( protocol, &radioSettingsCommand, response );
						waveProtocol->status = PS_PROCESS_IN_PROGRESS;

						if (response->status  ==  KO_PROTOCOL_NO_VALID_PARAMETER) {
							// printf("protocol_send_message:: No valid params ... \n");
							LOG_DEBUG (g_Logger, "protocol_send_message: There was an error writing parameters, but it is OK when no parameter are specified on port open", 0);
							response->status  = OK_PROTOCOL_SUCCESS;
						}

						if ( response->status == OK_PROTOCOL_SUCCESS ) {

							// printf("protocol_send_message:: Params written \n");
							SerialFrame sf = { { commandId, 0}, 1 };
							int index = 0;

							char *stopstring;
							size_t i = 0;
							for ( i = 0; i < MAX_RADIOADDRESS_LEN-1; i+=2 ) {
								char buff[3] = { ((RadioAddress*)command->dynamicStandardParams.nodeAddress)->addressValue[i],
										((RadioAddress*)command->dynamicStandardParams.nodeAddress)->addressValue[i+1], 0 };
								sf.frame[++index] = (byte)strtol( buff, &stopstring, 16 );
							}

							memcpy( sf.frame + 1 + index, ((Payload*)&command->command.payload)->frame, ((Payload*)&command->command.payload)->len );
							sf.len += index + ((Payload*)&command->command.payload)->len;

							send_frame( waveProtocol, &sf, responseId, response, NULL);

							if ( response->status == OK_PROTOCOL_SUCCESS ) {
								// printf("protocol_send_message:: Frame sent \n");
								WavePortResponses wprCode = -1;
								const DynamicExpertParams* dynamicExpertParams = &waveProtocol->params.dynamicExpertParams;
								if ( command->dynamicExpertParams.isSet & FLAG_WAKEUP_LENGTH )
									dynamicExpertParams = &command->dynamicExpertParams;
								wait_for_frame( waveProtocol, dynamicExpertParams, _false, _false, response, &wprCode);
								if ( response->status == OK_PROTOCOL_SUCCESS ) {

									char frameString[MAX_SERIALFRAME_LEN*2+1] = "";
									// printf("protocol_send_message:: Check code ...\n");
									switch( wprCode ) {
									case END_MESSAGE_EXCHANGE:
										// printf("protocol_send_message:: End message exchange\n");
										notify_spies( waveProtocol, &sf, _true );
										response->status = OK_PROTOCOL_SUCCESS;
										response->statusDescription[0] = 0;
										break;
									case RECEPTION_ERROR:
										// printf("protocol_send_message:: Reception error\n");
										response->status = parse_reception_error( &waveProtocol->receivedFrame );
										if ( response->status == KO_PROTOCOL_UNEXPECTED_FRAME )
											sprintf( response->statusDescription, "unexpected received frame during waiting for waveport END_MESSAGE_EXCHANGE %s",
													get_hex_serialframe(frameString, sizeof(frameString), &waveProtocol->receivedFrame) );
										else
											sprintf( response->statusDescription, "waveport RECEPTION_ERROR received %s",
													get_hex_serialframe(frameString, sizeof(frameString), &waveProtocol->receivedFrame) );
										response->status = KO_RADIO_TRANSMISSION_FAILURE;
										break;
									default:
										// printf("protocol_send_message:: Unexpected error \n");
										sprintf( response->statusDescription, "unexpected received frame during waiting for waveport END_MESSAGE_EXCHANGE %s",
												get_hex_serialframe(frameString, sizeof(frameString), &waveProtocol->receivedFrame) );
										response->status = KO_RADIO_TRANSMISSION_FAILURE;
										// TODO TCD launch spontaneous message transmission ?
//										notWaitedFrameReceived = _true;
										break;
									}
								}
							} else {
								// all errors from send frame are mapped to KO_PROTOCOL_PROCESS_FAILURE
								response->status = KO_PROTOCOL_PROCESS_FAILURE;
								// printf("protocol_send_message:: Process failure \n");
							}
						} else {
							//printf("protocol_send_message:: Parameter error: %s\n", get_response_status_name( response->status ) );
							sprintf( response->statusDescription, "signals param in error %s", get_response_status_name( response->status ) );
						}

					} else // Node address is valid ? -> NO
						response->status = KO_NODE_ADDRESS_NOT_VALID;
				} else // 3. Node address filled in? -> NO
					response->status = KO_NODE_ADDRESS_REQUIRED;

			} else // 2. Payload filled in? -> NO
				response->status = KO_PAYLOAD_REQUIRED;

			// put protocol status in idle mode
			waveProtocol->status = PS_IDLE;
			// TODO TCD
//			if (notWaitedFrameReceived) {
//				LOG_DEBUG( g_Logger, "protocol_send_message not waited frame has been received and must be process", 0);
//				receive_internal_frame((void*)protocol,  &waveProtocol->receivedFrame);
//			}
			LOG_DEBUG( g_Logger, "protocol_send_message status %d ", response->status );
			return response->status;
		} else // 1. A command session is already in progress? -> NO
			response->status = KO_PROTOCOL_ALREADY_IN_USE;
	}
	return KO_PROTOCOL_NO_VALID_PARAMETER;
}
/*__________________________________________________________________________________________________*/

/****************************************************************************************************/
int protocol_send_request( const void* protocol,
		const byte commandId,
		const byte responseId,
		const RadioExchangeCommand* command,
		RequestResponse* response ) {
	ProtocolInfo* waveProtocol = get_protocol( protocol );
	if ( waveProtocol && response && command ) {
		memset( response, 0,  sizeof(*response) );
		// 0. Is protocol opened ?
		if ( !protocol_is_open( waveProtocol ) )
			return (response->response.status = KO_PROTOCOL_NOT_OPEN);

		LOG_DEBUG(g_Logger, "protocol_send_request: Sending in C way", 0);
		// 1. A command session is already in progress?
		if ( waveProtocol->status != PS_PROCESS_IN_PROGRESS ) {
			waveProtocol->status = PS_PROCESS_IN_PROGRESS;

			//Point to Point mode  Max = 152 bytes of data
			//� Relay mode Max = 152 � (2 + 6 x Number of repeaters)
			//=> 1 repeater: 144 bytes of data
			//=> 2 repeaters: 138 bytes of data
			//=> 3 repeaters: 132 bytes of data

			// 2. Payload filled in?
			if ( ((Payload*)&command->command.payload)->len > 0 && ((Payload*)&command->command.payload)->len <= MAX_REQ_SEND_FRAME_BYTES_LEN ) {

				// 3. Node address filled in?
				if ( strlen( ((RadioAddress*)&command->dynamicStandardParams.nodeAddress)->addressValue ) == (MAX_RADIOADDRESS_LEN-1) ) {

					 if (is_node_address_valid(((RadioAddress*)&command->dynamicStandardParams.nodeAddress)->addressValue)) {

						RadioSettingsCommand radioSettingsCommand;
						memset( &radioSettingsCommand, 0,  sizeof(radioSettingsCommand) );
						memcpy( &radioSettingsCommand.dynamicStandardParams, &command->dynamicStandardParams, sizeof(DynamicStandardParams) );
						memcpy( &radioSettingsCommand.dynamicExpertParams, &command->dynamicExpertParams, sizeof(DynamicExpertParams) );

						// put protocol status in idle mode
						waveProtocol->status = PS_IDLE;
						protocol_write_radio_params( protocol, &radioSettingsCommand, &response->response );
						waveProtocol->status = PS_PROCESS_IN_PROGRESS;

						if (response->response.status  ==  KO_PROTOCOL_NO_VALID_PARAMETER) {
							response->response.status  = OK_PROTOCOL_SUCCESS;
						}

						if ( response->response.status == OK_PROTOCOL_SUCCESS ) {

							SerialFrame sf = { { REQ_SEND_FRAME, 0}, 1 };
							int index = 0;

							char *stopstring;
							size_t i = 0;
							for ( i = 0; i < MAX_RADIOADDRESS_LEN-1; i+=2 ) {
								char buff[3] = { ((RadioAddress*)&command->dynamicStandardParams.nodeAddress)->addressValue[i],
										((RadioAddress*)&command->dynamicStandardParams.nodeAddress)->addressValue[i+1], 0 };
								sf.frame[++index] = (byte)strtol( buff, &stopstring, 16 );
							}

							memcpy( sf.frame + 1 + index, ((Payload*)&command->command.payload)->frame, ((Payload*)&command->command.payload)->len );
							sf.len += index + ((Payload*)&command->command.payload)->len;

							SerialFrame respSerialFrame;
							memset(&respSerialFrame, 0, sizeof(respSerialFrame));
							send_frame( waveProtocol, &sf, RES_SEND_FRAME, &response->response, &respSerialFrame);

							//memcpy(response->command.command.payload, &respSerialFrame, sizeof (response->command.command.payload));
							//((Payload*)&response->command.command.payload)->len = respSerialFrame.len;

							//memcpy(response->response.serialFrame, &respSerialFrame, sizeof (response->response.serialFrame));
							//((SerialFrame*)&response->response.serialFrame)->len = respSerialFrame.len;

							if ( response->response.status == OK_PROTOCOL_SUCCESS ) {
								WavePortResponses wprCode = -1;
								const DynamicExpertParams* dynamicExpertParams = &waveProtocol->params.dynamicExpertParams;
								if ( command->dynamicExpertParams.isSet & FLAG_WAKEUP_LENGTH )
									dynamicExpertParams = &command->dynamicExpertParams;
								wait_for_frame( waveProtocol, dynamicExpertParams, _true, _false, &response->response, &wprCode);
								LOG_DEBUG(g_Logger, "protocol_send_request: Wait for frame finished with code %d", response->response.status);

								if ( response->response.status == OK_PROTOCOL_SUCCESS ) {
									notify_spies( waveProtocol, &sf, _true );
									response->response.statusDescription[0] = 0;

									LOG_DEBUG( g_Logger, "protocol_send_request : Parsing frame 0x%02x", waveProtocol->receivedFrame.frame[0]);
									protocol_dump_frame (&waveProtocol->receivedFrame);

									response->response.status = parse_frame(waveProtocol, &waveProtocol->receivedFrame, &response->command, NULL, NULL, NULL, NULL );
									LOG_DEBUG( g_Logger, "protocol_send_request : What was parsed :", 0);
									protocol_dump_payload ((Payload*)&response->command.command.payload);

									//memcpy( response->command.command.payload, &waveProtocol->receivedFrame, sizeof(response->command.command.payload));
									//((Payload*)&response->command.command.payload)->len = waveProtocol->receivedFrame.len;

									memcpy( response->response.serialFrame, &waveProtocol->receivedFrame, sizeof(response->response.serialFrame));
									((SerialFrame*)&response->response.serialFrame)->len = waveProtocol->receivedFrame.len;

									//LOG_DEBUG( g_Logger, "protocol_send_request : What was copied :", 0);
									//protocol_dump_payload (&response->command.command.payload);
									char frameString[MAX_SERIALFRAME_LEN*2+1] = "";

									if ( response->response.status == KO_PROTOCOL_UNEXPECTED_FRAME ) {
										response->response.status = KO_RADIO_TRANSMISSION_FAILURE;
										sprintf( response->response.statusDescription, "unexpected received frame during waiting for waveport END_MESSAGE_EXCHANGE %s",
												get_hex_serialframe(frameString, sizeof(frameString), &waveProtocol->receivedFrame) );
									} else if ( response->response.status != OK_PROTOCOL_SUCCESS ) {
										sprintf( response->response.statusDescription, "waveport RECEPTION_ERROR received %s",
												get_hex_serialframe(frameString, sizeof(frameString), &waveProtocol->receivedFrame) );
									}
								}

								if ( response->response.status == OK_PROTOCOL_SUCCESS ) {

									_bool receptionModeIsSet 	= (response->command.dynamicExpertParams.isSet & FLAG_RECEPTION_MODE) != 0;
									ReceptionMode receptionMode = response->command.dynamicExpertParams.receptionMode;

									memcpy( &response->command.dynamicExpertParams, &command->dynamicExpertParams, sizeof(DynamicExpertParams) );
									if (receptionModeIsSet) {
										LOG_DEBUG(g_Logger, "protocol_send_request:: Reception mode is set", 0);
										response->command.dynamicExpertParams.isSet |= FLAG_RECEPTION_MODE;
										response->command.dynamicExpertParams.receptionMode = receptionMode;
									} else {
										response->command.dynamicExpertParams.isSet &= ~FLAG_RECEPTION_MODE;
									}

									notify_spies_command( waveProtocol, &response->command, _false );
								}
							}
						} else {
							sprintf( response->response.statusDescription, "signals param in error %s", get_response_status_name( response->response.status ) );
							response->response.status = KO_PARAM_WRITING_FAILURE;
						}

					}else
						response->response.status = KO_NODE_ADDRESS_NOT_VALID;
				}
				else // 3. Node address filled in? -> NO
					response->response.status = KO_NODE_ADDRESS_REQUIRED;

			} else // 2. Payload filled in? -> NO
				response->response.status = KO_PAYLOAD_REQUIRED;

			// put protocol status in idle mode
			waveProtocol->status = PS_IDLE;
			char payloadString[MAX_SERIALFRAME_LEN*2+1] = "";

			LOG_DEBUG( g_Logger, "protocol_send_request status %d %s",
					response->response.status,
					get_hex_payload(payloadString, sizeof(payloadString), (const Payload*)&response->command.command.payload));

			return response->response.status;
		} else // 1. A command session is already in progress? -> NO
			response->response.status = KO_PROTOCOL_ALREADY_IN_USE;
	}
	return KO_PROTOCOL_NO_VALID_PARAMETER;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int protocol_close_response_session(ProtocolInfo *waveProtocol) {

	if (waveProtocol == NULL) {
		return KO_PROTOCOL_NO_VALID_PARAMETER;
	}

	//----------------------------------------------------------------------
	// wait for 2 seconds and close response session
	struct timeval startTime;

	gettimeofday( &startTime, 0 );
	while ( waveProtocol->isResponseSessionActive  != 0) {
		struct timeval endTime;

		gettimeofday( &endTime, 0 );
		if ( millisecondtimediff( startTime, endTime ) > 2000 )	// 2 seconds
			break;
		SLEEP( 10 );
	}

	// close response session
	waveProtocol->isResponseSessionActive = 0;
	return OK_PROTOCOL_SUCCESS;
}


//*****************************************************************************
// prototype methods definition to solve problem with serial driver instance
int protocol_set_serial_driver_instance( const void* protocol, const void *serial_driver_instance) {
	int iRet = KO_PROTOCOL_NO_VALID_PARAMETER;

	// for the production (final) version we have to ensure that no other operations are in
	// progress (i.e. the state machine is idle).
	// serial_driver_instance
	if (protocol == 0 || serial_driver_instance == 0)
		return iRet;

	ProtocolInfo* waveProtocol = get_protocol( protocol );

	if (waveProtocol == 0)
		return iRet;

	if(((SerialDriverInstance*)serial_driver_instance)->driverInfo == 0)
		return iRet;

	// close previos serial driver instance if has been loaded
	if (waveProtocol->serialDriverHandler.serialDriverHandle )
	{
		// check if driver has been loaded by this protocol instance
		if ( waveProtocol->serialDriver &&
				waveProtocol->serialDriverHandler.serialDriverHandle)
		{
			LOG_DEBUG(g_Logger, "set_serial_driver_instance: free previous one: %s", waveProtocol->params.serialDriverId);
			if (waveProtocol->serialDriverHandler.is_open && waveProtocol->serialDriverHandler.is_open( waveProtocol->serialDriver )) {
				waveProtocol->serialDriverHandler.close( waveProtocol->serialDriver );
			}

			if (waveProtocol->serialDriverHandler.destroy)
				waveProtocol->serialDriverHandler.destroy( waveProtocol->serialDriver );
		}
		// remove allocated memory that holds the references to all API's
		free_serial_driver( &waveProtocol->serialDriverHandler );
	}

	iRet = OK_PROTOCOL_SUCCESS;
	memset( &waveProtocol->serialDriverHandler, 0, sizeof(SerialDriverLoader) );

	waveProtocol->serialDriverHandler = ((SerialDriverInstance*)serial_driver_instance)->loader;
	waveProtocol->serialDriver	     = ((SerialDriverInstance*)serial_driver_instance)->driverInfo;
	LOG_DEBUG(g_Logger, "set_serial_driver_instance: Ok Done!\n", 0);

	return iRet;
}

//==============================================================================
_bool transmissionPowerLevelToPayload (float transmissionPowerLevelRefValue,
		byte *transmissionPowerLevelPayload) {

	if (!transmissionPowerLevelPayload)
		return _false;

	if (transmissionPowerLevelRefValue 		> 12  )
		*transmissionPowerLevelPayload 				= 0x0A;
	else if (transmissionPowerLevelRefValue > 11  )
		*transmissionPowerLevelPayload 				= 0x09;
	else if (transmissionPowerLevelRefValue > 9.7 )
		*transmissionPowerLevelPayload 				= 0x08;
	else if (transmissionPowerLevelRefValue > 7.9 )
		*transmissionPowerLevelPayload 				= 0x07;
	else if (transmissionPowerLevelRefValue > 5.5 )
		*transmissionPowerLevelPayload 				= 0x06;
	else if (transmissionPowerLevelRefValue > 3.3 )
		*transmissionPowerLevelPayload 				= 0x05;
	else if (transmissionPowerLevelRefValue > 2.1 )
		*transmissionPowerLevelPayload 				= 0x04;
	else if (transmissionPowerLevelRefValue > -0.3)
		*transmissionPowerLevelPayload 				= 0x03;
	else if (transmissionPowerLevelRefValue > -4  )
		*transmissionPowerLevelPayload 				= 0x02;
	else if (transmissionPowerLevelRefValue > -16 )
		*transmissionPowerLevelPayload 				= 0x01;
	else if (transmissionPowerLevelRefValue <= -16)
		*transmissionPowerLevelPayload 				= 0x00;
	else
		return _false;

	return _true;
}

_bool payloadToTransmissionPowerLevel (byte transmissionPowerLevelPayload,
		float *transmissionPowerLevelRefValue) {

	if (!transmissionPowerLevelRefValue) return _false;

	if (transmissionPowerLevelPayload	   == 0x0A)
		*transmissionPowerLevelRefValue 		= 14;
	else if (transmissionPowerLevelPayload == 0x09)
		*transmissionPowerLevelRefValue 		= 12;
	else if (transmissionPowerLevelPayload == 0x08)
		*transmissionPowerLevelRefValue 		= 11;
	else if (transmissionPowerLevelPayload == 0x07)
		*transmissionPowerLevelRefValue 		= 9.7;
	else if (transmissionPowerLevelPayload == 0x06)
		*transmissionPowerLevelRefValue 		= 7.9;
	else if (transmissionPowerLevelPayload == 0x05)
		*transmissionPowerLevelRefValue 		= 5.5;
	else if (transmissionPowerLevelPayload == 0x04)
		*transmissionPowerLevelRefValue 		= 3.3;
	else if (transmissionPowerLevelPayload == 0x03)
		*transmissionPowerLevelRefValue 		= 2.1;
	else if (transmissionPowerLevelPayload == 0x02)
		*transmissionPowerLevelRefValue 		= -0.3;
	else if (transmissionPowerLevelPayload == 0x01)
		*transmissionPowerLevelRefValue 		= -4;
	else if (transmissionPowerLevelPayload == 0x01)
		*transmissionPowerLevelRefValue 		= -4;
	else if (transmissionPowerLevelPayload == 0x00)
		*transmissionPowerLevelRefValue 		= -16;
	else
		return _false;

	return _true;
}

_bool physicalLayerToPayload (PhysicalLayer physicalLayer, unsigned short *payload) {

	if (!payload)
		return _false;

	if (physicalLayer == MONO_CHANNEL_868MHZ_9600BD )
		*payload = 0x00A2;
	else if (physicalLayer == FHSS_868MHZ_9600BD )
		*payload = 0x00A3;
	else if (physicalLayer == MONO_CHANNEL_868MHZ_500mW_19200BD )
		*payload = 0x00B6;
	else
		return _false;

	return _true;
}

_bool payloadToPhysicalLayer (unsigned short payload, PhysicalLayer *physicalLayer) {

	if (!physicalLayer)
		return _false;

	if (payload == 0x00A2)
		*physicalLayer = MONO_CHANNEL_868MHZ_9600BD;
	else if (payload == 0x00A3)
		*physicalLayer =  FHSS_868MHZ_9600BD;
	else if (payload == 0x00B6)
		*physicalLayer =  MONO_CHANNEL_868MHZ_500mW_19200BD ;
	else {
		*physicalLayer = -1;
		return _false;
	}

	return _true;
}

