/*******************************************************************************
 * Copyright (c) 2012 Coronis
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Contributors:
 *     Coronis - initial API and implementation
 *******************************************************************************/

#ifndef PROTOCOL_H_
#define PROTOCOL_H_

// local includes
#include "protocol_properties.h"
#include "protocol_command_types.h"
// domain includes
#include "utils.h"
#include "serial_driver_types.h"
#include <loader_serial_driver.h>
// protocol api includes
#include <i_protocol_types.h>
#include <i_protocol_service.h>
#include <protocol_event.h>
// serialdriver api includes
#include <i_serial_frame.h>
#include <serial_driver_status.h>

typedef enum tagWavePortCommands {

	// Request to send a radio frame and wait for radio response.
	REQ_SEND_FRAME        = 0x20,
	// Request to send a radio frame and wait for radio response
	REQ_SEND_MESSAGE      = 0x22,

	// Request to send a radio frame in response to a broadcast frame
	REQ_SEND_BROADCAST_RESPONSE    = 0x24,

	// Request to send a radio frame in broadcast mode
	REQ_SEND_BROADCAST    = 0x28,

	// Request to send application response
	REQ_SEND_RESPONSE	= 0x2C,

	// Request to update radio parameters
	REQ_WRITE_RADIO_PARAM = 0x40,
	// Request to update radio board transmission power

	REQ_CHANGE_TX_POWER   = 0x44,
	// Request to read radio parameters
	REQ_READ_RADIO_PARAM  = 0x50,
	// Request to read radio board transmission power
	REQ_READ_TX_POWER     = 0x54,
	// Request to select operating radio channel when FHSS is deselected
	REQ_SELECT_CHANNEL    = 0x60,
	// Request to read the operating radio channel when  FHSS is deselected
	REQ_READ_CHANNEL      = 0x62,
	// Request to select RF communication mode
	REQ_SELECT_PHYCONFIG  = 0x64,
	// Request to read RF communication mode
	REQ_READ_PHYCONFIG    = 0x66,
	REQ_SELECT_SECOND_PHYCONFIG = 0x6C,
	// Request to read RF communication mode
	REQ_READ_SECOND_PHYCONFIG = 0x6E,
	// Request to send a service frame (and wait for response)
	REQ_SEND_SERVICE = 0x80,
	// Request to send a service frame
	REQ_SEND_SERVICE_MESSAGE = 0x84
} WavePortCommands;

typedef enum tagWavePortResponses {

	// 	Radio board response to frame transmission (response to requests 0x20, 0x22, 0x24, 0x26, 0x28, 0x2A
	RES_SEND_FRAME           = 0x21,
	// Frame received by radio board.
    RECEIVED_FRAME           = 0x30,
    // Frame indicating error type detected at the end of the last exchange in point-to-point or relay mode.
    RECEPTION_ERROR          = 0x31,
    // Radio frame received following a REQ_SEND_POLLING request
    RECEIVED_FRAME_POLLIN    = 0x32,
    // Received radio frame following a REQ_SEND_BROADCAST request
    RECEIVED_BROADCAST_RESPONSE = 0x34,
    // Relay frame received by the radio board. Reception of this command is possible only if the RELAY_ROUTE_STATUS (0x06) parameter is set.
    RECEIVED_FRAME_RELAYED   = 0x35,
    // Received radio frame in multi-frame mode. Indicates that subsequent frames are pending.
    RECEIVED_MULTIFRAME      = 0x36,
    // Frame indicating the end of message exchange. This frame is returned only after a 0x22 & 0x24, or 0x2A request command. Reception of this frame
    // depends on the value of EXCHANGE_STATUS.
    END_MESSAGE_EXCHANGE     = 0x37,
    // Received a radio frame transmitted in broadcast mode
    RECEIVED_BROADCAST_FRAME = 0x38,
	// Radio board response to radio parameter update
	RES_WRITE_RADIO_PARAM    = 0x41,
	// Radio board response to transmission power update
	RES_CHANGE_TX_POWER      = 0x45,
	// Radio board response to parameter reading
	RES_READ_RADIO_PARAM     = 0x51,
	// Radio board response to transmission power read
	RES_READ_TX_POWER        = 0x55,
	// Response to channel selection request
	RES_SELECT_CHANNEL       = 0x61,
	// Response to the read channel request
	RES_READ_CHANNEL         = 0x63,
	// Response to communication mode selection request
	RES_SELECT_PHYCONFIG     = 0x65,
	// Response to communication mode read request
	RES_READ_PHYCONFIG       = 0x67,
	RES_SELECT_SECOND_PHYCONFIG = 0x6D,
	// Response to communication mode read request
	RES_READ_SECOND_PHYCONFIG    = 0x6F,
	// REQ_SEND_SERVICE response
	RES_SEND_SERVICE         = 0x81,
	// Frame received following REQ_SEND_SERVICE transmission
	SERVICE_RESPONSE         = 0x82
} WavePortResponses;

#define WAKEUP_LENGTH_MIN 20
#define WAKEUP_LENGTH_MAX 10000
#define RADIO_USER_TIMEOUT_MIN 100
#define RADIO_USER_TIMEOUT_MAX 25500
#define BCST_RECEPTION_TIMEOUT_MIN 100
#define BCST_RECEPTION_TIMEOUT_MAX 25500
#define ACTIVITY_PERIOD_MIN 100
#define ACTIVITY_PERIOD_MAX 25500

#define TRANSMISSION_RADIO_CHANNEL_MIN 0
#define TRANSMISSION_RADIO_CHANNEL_MAX 21

typedef enum tagWavePortParameters {
	//
	// Parameters are accessible using commands REQ_READ_RADIO_PARAM and REQ_WRITE_RADIO_PARAM
	//
	// RF polling period, in multiples of 100 ms
	AWAKENING_PERIOD   = 0x00,
	//wake-up duration when long wake-up is set used This value must be higher than the RF polling period. Value in multiples of 1 ms, LSB defined first.
	//Default value: 1100 ms min. value = 20 ms (0x1400) max. value = 10 sec. (0x1027)
	WAKEUP_LENGTH      = 0x02,
	// indicates whether or not radio frames should be acknowledged by receiver.
	RADIO_ACKNOWLEDGE  = 0x04,
	// radio board address This value is set a the factory. Read-only. 6 bytes
	RADIO_ADDRESS      = 0x05,
	// Parameter related to relay route transmission in each relayed frame received.
	RELAY_ROUTE_STATUS = 0x06,
	// Table containing radio addresses of successive repeaters used to	reach the end module.
	RELAY_ROUTE        = 0x07,
	// Byte containing the group number of the remote modules to query in polling mode
	GROUP_NUMBER       = 0x09,
	// automatic selection of Radio communication mode used to address an 	equipment depending on radio address
	SWITCH_MODE_STATUS = 0x10,
	// time-out for receiving a response frame Value in multiples of 100ms default value = 0x14  (2 seconds)
	RADIO_USER_TIMEOUT = 0x0C,
	// parameter for activating or deactivating error or status frame management.
	EXCHANGE_STATUS    = 0x0E,
	//
	SECOND_SCAN_ENABLE_ACTIVATION = 0x15,
	// byte	containing the Wavecard multicast group (starting with version 2.00).
	WAVECARD_MULTICAST_GROUP = 0x16,
	// time-out for	receiving CSMA frame following a transmitted REQ_SEND_BROADCAST command (starting with firmware version 2.01)
	// Value in multiples of 100 ms. Default  = 0x3C (6 seconds)
	BCST_RECEPTION_TIMEOUT = 0x17,
	//
	BCST_MAX_RESPONSES_NB  = 0x19,
	//
	RADIO_CRC_ACTIVATION  = 0x1F

} WavePortParameters;

 // Command Status
typedef enum tagCommandStatus {
   CMD_STATUS_OK = 0x00,
   CMD_STATUS_KO = 0x01
} CommandStatus;


typedef enum tagProtocolStatus {
	PS_IDLE,
	PS_PROCESS_IN_PROGRESS
} ProtocolStatus;


typedef struct tagProtocolInfo {
	int   integrity;

	void* jniSerialDriver;
	SerialDriverLoader serialDriverHandler;
	void* serialDriver;

	// current status of protocol session, could be one of the following values
	// PS_IDLE - session is idle
	// PS_PROCESS_IN_PROGRESS - send command in progress
	int status;

	// this parameter is used as storage for incoming framse from SerialDriver instance
	SerialFrame receivedFrame;

	// protocol session parameters
	ProtocolParams params;

	protocol_listener_t receiveCommand;
	protocol_listener_event_t receiveEvent;
	void* paramProtocolListener;

	protocol_spy_listener_t spiesSend[MAX_SPY_NUM];
	protocol_spy_listener_t spiesRecv[MAX_SPY_NUM];
	void* paramSpies[MAX_SPY_NUM];
	//void* env;

	// this parameter is indicate how many relay route path was used to sena a command
	// it is filled in write_dynamic_standard_params
	int relayNb;

	// set to 1 during 2 seconds, when a response message is received
	byte    		isResponseSessionActive;
	ReceptionMode	receptionSessionMode;	// save reception session mode for send application response
	// the address of remote node, received in ReceiveMessage with 0x35 frame
	pIRadioAddress 	senderNodeAddress;

} ProtocolInfo;

extern ProtocolInfo g_WaveProtocol[MAX_PROTOCOL_NUM];
extern void* g_ModuleHandle;

/****************************************************************************************************
 *                        HELPER SECTION - PROTOCOL SPECIFIC FUNCTIONS DEFINITION
 ****************************************************************************************************/

// Verify and receive ProtocolInfo structure from given external parameter
ProtocolInfo* get_protocol( const void* protocol );

/*
 * This function verifies if serial driver and serial port are initialized/opened
 */
int verify_serial_driver_port( const ProtocolInfo* waveProtocol );

// internal serial driver listener callback
int receive_frame( void* param, const pSerialExchangeData i_frame );

// internal serial driver notify listener callback
int receive_event( void* param, int eventCode  );

/**
 * Function used to waiting for response after update command
 * @param waveProtocol - a valid handle to an open protocol.
 * @param response - response code and message
 * @param responseId - response id for update command
 * @return OK_PROTOCOL_SUCCESS if success,
 * KO_PROTOCOL_PROCESS_FAILURE if send failed;
 */
int wait_for_response( const ProtocolInfo* waveProtocol, CommandResponse* response, WavePortResponses responseId );

/**
 * Function used to waiting for frame after send command to port
 * @param waveProtocol - handle to waveport protocol instance
 * @param dynamicExpertParams - handle to dynamic expert params that contain wakeup length that used as timeout counter
 * @param isRequest - if true it is request command, otherwise is message command.
 * @param isBroadcast - if true, this is a broadcast message
 * @param response - response code and message
 * @param responseId - response id for command
 * @return OK_PROTOCOL_SUCCESS if success,
 * KO_PROTOCOL_PROCESS_FAILURE if send failed;
 */
int wait_for_frame( ProtocolInfo* waveProtocol,
		const DynamicExpertParams* dynamicExpertParams,
		_bool isRequest,
		_bool isBroadcast,
		CommandResponse* response,
		WavePortResponses* responseId );


/**
 * Function used to send command frames
 * @param waveProtocol - a valid handle to an open protocol.
 * @param sf - serial frame that contain command, requested parameter id
 * @param wprCode - response id for read command
 * @param response - response code and message
 * @param responseFrame - frame that contain requested value if any
 * @return OK_PROTOCOL_SUCCESS if success,
 * KO_PROTOCOL_PROCESS_FAILURE if send failed;
 */
int send_frame( ProtocolInfo* waveProtocol,
		SerialFrame* sf,
		WavePortResponses wprCode,
		CommandResponse* response,
		SerialFrame* responseFrame );

/**
 * Function used to send command frames
 * @param waveProtocol - a valid handle to an open protocol.
 * @param payload - the payload data, that already contains encoded serial frame.
 * @param wprCode - response id for read command
 * @param response - response code and message
 * @param responseFrame - frame that contain requested value if any
 * @return OK_PROTOCOL_SUCCESS if success,
 * KO_PROTOCOL_PROCESS_FAILURE if send failed;
 */
int send_frame_from_payload( ProtocolInfo* waveProtocol,
					Payload* payload,
					WavePortResponses wprCode,
					CommandResponse* response,
					SerialFrame* responseFrame );
int write_dynamic_standard_params( ProtocolInfo* waveProtocol, DynamicStandardParams* dynamicStandardParams, CommandResponse* response );
int write_dynamic_expert_params( ProtocolInfo* waveProtocol, DynamicExpertParams* dynamicExpertParams, CommandResponse* response );

/**
 * Function used to write static standard parameters to waveport
 * @param waveProtocol - a valid handle to an open protocol.
 * @param staticStandardParams - waveport static standard parameters
 * @return OK_PROTOCOL_SUCCESS if success,
 * KO_SERIAL_DRIVER_SENDING_FAILURE if send failed,
 * KO_PROTOCOL_CONFIGURATION_TIMEOUT if timeout expired,
 * KO_PROTOCOL_PARAM_UPDATE_ERROR if update failed and
 * KO_PROTOCOL_UNEXPECTED_FRAME if unexpected frame received
 * KO_PROTOCOL_NO_VALID_PARAMETER if no filled parameters in staticStandardParams
 * KO_PARAMETER_RADIO_ACK_ACTIVATION_NOT_VALID if RadioAckActivation parameter invalid
 * KO_PARAMETER_FIRST_SCAN_ACTIVITY_PERIOD_NOT_VALID if FirstScanActivityPeriod parameter invalid
 * KO_PARAMETER_SWITCHING_MODE_ON_RADIO_ADDRESS_ACTIVATION_NOT_VALID if SwitchingModeOnRadioAddressActivation parameter invalid
 * KO_PARAMETER_RECEPTION_GROUP_ID_NOT_VALID if ReceptionGroupId parameter invalid
 * KO_PARAMETER_FIRST_SCAN_ENABLE_ACTIVATION_NOT_VALID if FirstScanEnableActivation parameter invalid.
 */
int write_static_standard_params( ProtocolInfo* waveProtocol, StaticStandardParams* staticStandardParams, CommandResponse* response );

/**
 * Function used to write static expert parameters to waveport
 * @param waveProtocol - a valid handle to an open protocol.
 * @param staticExpertParams - waveport static expert parameters
 * @return OK_PROTOCOL_SUCCESS if success,
 * KO_SERIAL_DRIVER_SENDING_FAILURE if send failed,
 * KO_PROTOCOL_CONFIGURATION_TIMEOUT if timeout expired,
 * KO_PROTOCOL_PARAM_UPDATE_ERROR if update failed and
 * KO_PROTOCOL_UNEXPECTED_FRAME if unexpected frame received
 * KO_PROTOCOL_NO_VALID_PARAMETER if no filled parameters in staticStandardParams
 * if  parameter invalid
 */
int write_static_expert_params( ProtocolInfo* waveProtocol, StaticExpertParams* staticExpertParams, CommandResponse* response );

/**
 * Function used to parse frame, to extract command, address, payload.
 * @param frame - a radio frame to be parsed.
 * @param command - a special structure to store parsed data.
 * @param status - STATUS_RECEPTION for RECEIVED_FRAME_POLLIN or RECEIVED_BROADCAST_RESPONSE or RECEIVED_MULTIFRAME
 * @param totalNumber - Total number of frames received for RECEIVED_BROADCAST_RESPONSE or RECEIVED_MULTIFRAME
 * @param index - Frame index for RECEIVED_BROADCAST_RESPONSE or RECEIVED_MULTIFRAME
 * @param types - response types for SERVICE_RESPONSE
 * @return _true if success, or _false if frame format is unknown.
 */
int parse_frame(ProtocolInfo* waveProtocol, const SerialFrame* frame, RadioExchangeCommand* command,
		int* status,
		int* totalNumber,
		int* index,
		int* types );

/**
 * Function used to parse frame, to error code.
 * @param frame - a radio frame to be parsed.
 * @return error code obtained from frame.
 */
int parse_reception_error( const SerialFrame* frame );

/**
 * Function used to notify sent/receive spies about sent/received frames.
 * @param waveProtocol - a waveport protocol instance.
 * @param frame - a radio frame about witch spies will be notified.
 * @param sendEvent - a flag that manage witch of send or receive spies to be notified.
 * @return error code obtained from frame.
 */
int notify_spies( const ProtocolInfo* waveProtocol, const SerialFrame* frame, _bool sendEvent );

/**
 * Function used to notify sent/receive spies about sent/received frames.
 * @param waveProtocol - a waveport protocol instance.
 * @param command - a radio exchange frame about witch spies will be notified.
 * @param sendEvent - a flag that manage witch of send or receive spies to be notified.
 * @return error code obtained from frame.
 */
int notify_spies_command( const ProtocolInfo* waveProtocol, const RadioExchangeCommand* command, _bool sendEvent );

/**
 * Function used to notify sent/receive spies about sent/received frames.
 * @param env - a JNI environment instance.
 * @param waveProtocol - a waveport protocol instance.
 * @param command - a radio exchange frame about witch spies will be notified.
 * @param sendEvent - a flag that manage witch of send or receive spies to be notified.
 * @return error code obtained from frame.
 */
void jni_notify_spies_native_command( const ProtocolInfo* waveProtocol, const RadioExchangeCommand* command, _bool sendEvent );

/**
 * Function used to notify sent/receive spies about sent/received frames.
 * @param waveProtocol - a waveport protocol instance.
 * @param eventCode - a flag that indicate type of event START_LISTENING,STOP_LISTENING,...etc.
 * @return error code obtained from frame.
 */
void jni_notify_event_native( const ProtocolInfo* waveProtocol, int eventCode );

/**
 * Method used to send a application/service message to a device
 * without waiting for a device answer.
 * @param protocol - A valid handle to an open protocol.
 * @param commandId - A valid id of application/service message command.
 * @param responseId - A valid id of application/service message response.
 * @param command - contains the information of the message to send
 * @param response - the sending status of the command
 * @return 0 if success or positive number if unexpected error occurs during sending process
 */
int protocol_send_message( const void* protocol,
		const byte commandId,
		const byte responseId,
		const RadioExchangeCommand* command,
		CommandResponse* response );

/**
 * Method used to send a application/service request message to a device
 * and waiting for the device answer.
 * @param protocol - A valid handle to an open protocol.
 * @param commandId - A valid id of application/service request command.
 * @param responseId - A valid id of application/service request response.
 * @param command - contains the information of the request to send
 * @param response - the response to the command
 * @return 0 if success or positive number if unexpected error occurs during sending process
 */
int protocol_send_request( const void* protocol,
		const byte commandId,
		const byte responseId,
		const RadioExchangeCommand* command,
		RequestResponse* response );

/****************************************************************************************************
 *                        HELPER SECTION - OTHER FUNCTIONS DEFINITION
 ****************************************************************************************************/

_bool get_module_path( const void* moduleHandle, char* path, int pathLen );
_bool is_hex( const char* hexStr );

/**
 * Function used to wait 2 seconds and then close the response session
 * by setting isResponseSessionActive  to Zero.
 * @param waveProtocol - pointer to protocol descriptor.
 * @return error code .
 */
int protocol_close_response_session(ProtocolInfo *waveProtocol);

/**
 * Function determines if passed node address has a valid length and format.
 *
 * @param node_address - a hex string that represents a node address
 * @return _true for a valid node address
 */
_bool is_node_address_valid(const char *node_address);

void set_payload_data(pIPayload dest, const Payload *src);

/**
 *  Function takes the reference value of a power transmission level and
 *  converts it to payload equivalent
 * @param [IN]  transmissionPowerLevelRefValue - the reference(symbolic)  value
 * 				of transmission power level.
 * @param [OUT] transmissionPowerLevelPayload - value ready to be sent as a
 * 				payload and which will be understand by the waveport protocol.
 *
 * @return  	true if passed value for conversion is valid.
 */
_bool transmissionPowerLevelToPayload (float transmissionPowerLevelRefValue,
		byte *transmissionPowerLevelPayload);

/**
 *  Function takes the payload value of power transmission level and converts
 *  it to symbolic one (used internally to make an abstraction from used waveport
 *  protocol).
 * @param [IN]  transmissionPowerLevelPayload - the payload value received from
 * 				waveport.
 * @param [OUT] transmissionPowerLevelRefValue - a value used to represent
 * 				converted level as an abstract value by the SDK
 * @return  	true if passed value for conversion is valid.
 */
_bool payloadToTransmissionPowerLevel (byte transmissionPowerLevelPayload,
		float *transmissionPowerLevelRefValue);

/**
 *  Function takes the reference value of a physicla layer ID and
 *  converts it to payload equivalent
 * @param [IN]  physicalLayer - the reference(symbolic)  value phys. layer
 * @param [OUT] payload - value ready to be sent as a
 * 				payload and which will be understand by the waveport protocol.
 *
 * @return  	true if passed value for conversion is valid.
 */
_bool physicalLayerToPayload (PhysicalLayer physicalLayer, unsigned short *payload);

/**
 *  Function takes the payload value of physical layer and converts
 *  it to symbolic one (used internally to make an abstraction from used waveport
 *  protocol).
 * @param [IN]  payload - the payload value received from waveport.
 * @param [OUT] physicalLayer - a value used to represent
 * 				converted level as an abstract value by the SDK
 * @return  	true if passed value for conversion is valid.
 */
_bool payloadToPhysicalLayer (unsigned short payload,  PhysicalLayer *physicalLayer);

#endif /* PROTOCOL_H_ */
