/*******************************************************************************
 * Copyright (c) 2012 Coronis
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Contributors:
 *     Coronis - initial API and implementation
 *******************************************************************************/
#ifdef _JAVA_WRAPPER
#include "com_coronis_sdk_serialdriver_rs232driver_RS232Driver.h"
#include "loader_logger.h"
#include "../listeners.h"

// coronis domain includes
#include "utils.h"
#include "serial_driver_types_utils.h"

#include <JNI/domain_jni.h>

static jclass g_clazzSerialDriverEvent = NULL;
static jclass g_frameClass = NULL;

int java_event_handler( void* param, const pSerialExchangeData frame );
int java_event_notify_handler( void* param, int errorCode );
int java_event_spy_recv_handler( void* param, int index, const pSerialExchangeData frame );
int java_event_spy_send_handler( void* param, int index, const pSerialExchangeData frame );
jobject get_sending_jni_status (JNIEnv *env, int ret_code) ;

const char* get_serial_driver_event( SerialDriverEvent sde ) {
	static char serialDriverEventName[50] = "";

	switch( sde ) {
	case START_LISTENING:          strcpy( serialDriverEventName, "START_LISTENING" ); break;
	case STOP_LISTENING:           strcpy( serialDriverEventName, "STOP_LISTENING" ); break;
	case IGNORED_FRAME_RECEIVED:   strcpy( serialDriverEventName, "IGNORED_FRAME_RECEIVED" ); break;
	case SERIALPORT_DISCONNECTION: strcpy( serialDriverEventName, "SERIALPORT_DISCONNECTION" ); break;
	default: serialDriverEventName[0] = 0; break;
	}

	return serialDriverEventName;
}

void reset_driver_handle( JNIEnv *env, jobject this ) {
	jclass jcI = (*env)->GetObjectClass( env, this);
	if ( jcI != 0 ) {
		jfieldID jfV = (*env)->GetFieldID( env, jcI, "rs232handle", "I");
		if ( jfV != 0 )
			(*env)->SetIntField( env, this, jfV, 0 );
		(*env)->DeleteLocalRef( env, jcI );
	}
}

void* get_driver_handle( JNIEnv *env, jobject this ) {
	jclass jcI = (*env)->GetObjectClass( env, this);
	if ( jcI != 0 ) {
		jfieldID jfV = (*env)->GetFieldID( env, jcI, "rs232handle", "I");
		if ( jfV != 0 )
			return (void*)(*env)->GetIntField( env, this, jfV );
		(*env)->DeleteLocalRef( env, jcI );
	}

	return 0;
}

void throw_exception( JNIEnv *env, jobject this, int errorCode ) {

	if ( errorCode == OK_SERIALDRIVER_SUCCESS )	return;
	LOG_DEBUG( g_Logger, "throw_exception EXCEPTION: %d", errorCode );

	const char* fieldName = get_sending_status_name( errorCode );
	if ( strlen( fieldName ) == 0 ) return;

	jclass g_excCls = (*env)->FindClass(env, "com/coronis/sdk/serialdriver/api/SerialDriverException");
	if ( g_excCls == 0) return;

	// Search for identically named exception static filed in SerialDriverException
	// If found, consider it as an exception, otherwise, ignore it
	jfieldID id = (*env)->GetStaticFieldID( env, g_excCls, fieldName+3, "Ljava/lang/String;");
	if ( id != NULL ) {
		jstring jstr = (*env)->GetStaticObjectField( env, g_excCls, id);
		if ( jstr != NULL ) {

			const char* msg = (*env)->GetStringUTFChars(env, jstr, NULL);
			if ( msg != 0 ) {
				LOG_DEBUG( g_Logger, "throw_exception EXCEPTION: %s", fieldName );
				(*env)->ThrowNew(env, g_excCls, msg);
				(*env)->ReleaseStringUTFChars(env, jstr, msg);
			}
		}
	}

	(*env)->DeleteLocalRef( env, g_excCls );
}

jint JNICALL JNIFUNCTION(getDriverVersion)(JNIEnv * env, jobject this) {
	return (jint)driver_get_version();
}

jint JNICALL JNIFUNCTION(create)(JNIEnv * env, jobject this) {
	return (jint)driver_create( );
}

void JNICALL JNIFUNCTION( close )(JNIEnv *env, jobject this) {
	throw_exception( env, this, driver_close( get_driver_handle( env, this) ) );
}

void JNICALL JNIFUNCTION( destroy )(JNIEnv * env, jobject this) {
	JNIFUNCTION( removeSerialDriverListener )( env, this);
	JNIFUNCTION( removeSerialDriverSpyListener )( env, this, NULL );
	throw_exception( env, this, driver_destroy( get_driver_handle( env, this) ) );
	reset_driver_handle( env, this );
}

void JNICALL JNIFUNCTION( initNative )(JNIEnv *env, jobject this) {
	throw_exception( env, this, driver_init( get_driver_handle( env, this) ) );
}

void JNICALL JNIFUNCTION( init ) (JNIEnv *env, jobject this, jobject url) {
	DriverInfo* serialDriver = get_driver( get_driver_handle( env, this) );

	if ( serialDriver != NULL ) {
		if ( url != NULL ) {
			jclass jcI = (*env)->GetObjectClass( env, url);
			if ( jcI != NULL ) {
				jmethodID jmtd = (*env)->GetMethodID( env, jcI, "getPath", "()Ljava/lang/String;");
				if ( jmtd != NULL ) {
					jstring jstr = (jobject)(*env)->CallObjectMethod( env, url, jmtd );
					if ( jstr != NULL ) {
						const char* str = (*env)->GetStringUTFChars(env, jstr, NULL);
						if ( str != NULL) {
							const char* propFileName = str;
#ifdef _WINDOWS
							propFileName = str[0] == '/' ? str+1 : str;
#endif
							throw_exception( env, this, driver_init_from_file( get_driver_handle( env, this), propFileName ) );
							(*env)->ReleaseStringUTFChars(env, jstr, str);
						}
					}
					(*env)->DeleteLocalRef(env, jstr);
				}
				(*env)->DeleteLocalRef( env, jcI );
			}
		} else
			throw_exception( env, this, driver_init( get_driver_handle( env, this) ) );
	}
}

jboolean JNICALL JNIFUNCTION( isOpen )(JNIEnv *env, jobject this) {
	return driver_is_open( get_driver_handle( env, this) );
}

void JNICALL JNIFUNCTION( open )(JNIEnv *env, jobject this) {
	throw_exception( env, this, driver_open( get_driver_handle( env, this) ) );
}

void JNICALL JNIFUNCTION( addSerialDriverListener )(JNIEnv *env, jobject this, jobject listener) {
	JNIFUNCTION( removeSerialDriverListener )( env, this);
	void* driver = NULL;
	DriverInfo* serialDriver = get_driver( driver = get_driver_handle( env, this) );
	if ( serialDriver ) {
		serialDriver->java = _true;
		jobject grefListener = (*env)->NewGlobalRef( env, listener);

		driver_add_listener(
			driver,
			java_event_handler,
			java_event_notify_handler,
			(void*)grefListener);

	}
}

void JNICALL JNIFUNCTION( removeSerialDriverListener )(JNIEnv *env, jobject this) {
	DriverInfo* serialDriver = get_driver( get_driver_handle( env, this) );

	if ( serialDriver != NULL ) {
		jobject listener = (jobject)serialDriver->realListeners.paramNotify;
		driver_remove_listener(get_driver_handle( env, this));
		(*env)->DeleteGlobalRef( env, listener );
	}
}

void JNICALL JNIFUNCTION( addSerialDriverSpyListener )(JNIEnv *env, jobject this, jobject listener) {
	DriverInfo* serialDriver = get_driver( get_driver_handle( env, this) );

	if ( serialDriver != NULL ) {
		int spyIndex = 0;
		int i = 0;
		for ( ; i < MAX_SPY_NUM; ++i ) {
			if ( serialDriver->realListeners.paramSpies[i] != NULL )
				if ( (*env)->IsSameObject( env, listener, (jobject)serialDriver->realListeners.paramSpies[i] ) == JNI_TRUE )
					return ;
		}
		serialDriver->java = _true;
		driver_add_spy_listener( get_driver_handle( env, this),
				java_event_spy_send_handler,
				java_event_spy_recv_handler,
				(void*)(*env)->NewGlobalRef( env, listener),
				&spyIndex );
	}
}

void JNICALL JNIFUNCTION( removeSerialDriverSpyListener )(JNIEnv *env, jobject this, jobject listener) {
	DriverInfo* serialDriver = get_driver( get_driver_handle( env, this) );

	if ( serialDriver != NULL ) {
		int i = 0;
		for ( ; i < MAX_SPY_NUM; ++i ) {
			if ( listener == NULL ) {
				if ( serialDriver->realListeners.paramSpies[i] != NULL )
					(*env)->DeleteGlobalRef( env, (jobject)serialDriver->realListeners.paramSpies[i]);
				driver_remove_spy_listener( get_driver_handle( env, this), i);
			} else if ( (*env)->IsSameObject( env, listener, (jobject)serialDriver->realListeners.paramSpies[i] ) == JNI_TRUE ) {
				if ( serialDriver->realListeners.paramSpies[i] != NULL )
					(*env)->DeleteGlobalRef( env, (jobject)serialDriver->realListeners.paramSpies[i]);
				driver_remove_spy_listener( get_driver_handle( env, this), i);
				break;
			}
		}
	}
}

jobject JNICALL JNIFUNCTION( send )(JNIEnv *env, jobject this, jobject frame) {
	int iRet = KO_SERIALPORT_UNEXPECTED_ERROR;

	LOG_DEBUG( g_Logger, "JNI(send): start sending frame", 0);

	if (frame == 0) {
		LOG_ERROR( g_Logger, "JNI(send): null frame passed", 0);
		return get_sending_jni_status (env, KO_SERIALFRAME_REQUIRED);
	}

	jclass jcI = (*env)->GetObjectClass( env, frame);
	if ( jcI != 0 ) {
		jmethodID jmtd = (*env)->GetMethodID( env, jcI, "getBuffer", "()Ljava/nio/ByteBuffer;");
		if ( jmtd != 0 ) {
			jobject byteBuffer = (jobject)(*env)->CallObjectMethod( env, frame, jmtd );
			if ( byteBuffer != 0 ) {
				jclass jcBB = (*env)->GetObjectClass( env, byteBuffer);
				if ( jcBB != 0 ) {

					jmethodID jmtdBB = (*env)->GetMethodID( env, jcBB, "capacity", "()I");
					if ( jmtdBB != 0 ) {
						jint len = (*env)->CallIntMethod( env, byteBuffer, jmtdBB );
						iRet = KO_SERIALFRAME_REQUIRED;
						if ( len > 0 ) {
							jmethodID jmtdBBb = (*env)->GetMethodID( env, jcBB, "get", "(I)B");
							if ( jmtdBBb != 0 ) {
								byte data[MAX_SERIALFRAME_LEN] = {0};
								int i = 0;
								for ( ; i < len; ++i ) {
									data[i] = (*env)->CallByteMethod( env, byteBuffer, jmtdBBb, i );
								}

								SerialExchangeData frameData;
								memcpy( ((SerialFrame*)&frameData.serialFrame)->frame, data, len );
								((SerialFrame*)&frameData.serialFrame)->len = len;
								LOG_DEBUG( g_Logger, "JNI(send): call  driver_send", 0);

								iRet = driver_send( get_driver_handle( env, this), &frameData );
							}
						}
					}
					(*env)->DeleteLocalRef( env, jcBB );
				}
			}
		}
		(*env)->DeleteLocalRef( env, jcI );
	}

	LOG_DEBUG( g_Logger, "JNI(send): -1- %d", iRet );

	// according to SerialDriver_Send_RS232Normal state diagram, exception should be thrown only in one case
	if ( iRet == KO_SERIALDRIVER_ALREADY_IN_USE )
		throw_exception( env, this, iRet );

	const char* fieldName = get_sending_status_name( iRet );
	LOG_DEBUG( g_Logger, "JNI(send): -2- %s", fieldName );
	return get_sending_jni_status (env, iRet);
}

jobject get_sending_jni_status (JNIEnv *env, int ret_code) {
	jobject sendStatus = NULL;

	const char* fieldName = get_sending_status_name( ret_code );

	jclass ssClass = (*env)->FindClass( env, "com/coronis/sdk/serialdriver/api/SendingStatus");
	if ( ssClass != NULL ) {
		jfieldID jfid = (*env)->GetStaticFieldID( env, ssClass, fieldName, "Lcom/coronis/sdk/serialdriver/api/SendingStatus;");
		if ( jfid != NULL )
			sendStatus = (*env)->GetStaticObjectField( env, ssClass, jfid);
		(*env)->DeleteLocalRef( env, ssClass);
	}

	return sendStatus;
}

void JNICALL JNIFUNCTION( setFlowControlModeParam )(JNIEnv *env, jobject this, jint flowControlIn, jint flowControlOut) {
	throw_exception( env, this, driver_set_flow_control_mode_params( get_driver_handle( env, this), flowControlIn, flowControlOut ) );
}

void JNICALL JNIFUNCTION( setListeningPeriodParam )(JNIEnv *env, jobject this, jint listeningPeriod) {
	throw_exception( env, this, driver_set_listening_period_param( get_driver_handle( env, this), listeningPeriod) );
}

void JNICALL JNIFUNCTION( setSerialPortParams )(JNIEnv *env, jobject this, jint baudRate, jint dataBits, jint stopBits, jint parity) {
	throw_exception( env, this, driver_set_port_params( get_driver_handle( env, this), baudRate, dataBits, stopBits, parity ) );
}

void JNICALL JNIFUNCTION( setSerialPortId )(JNIEnv *env, jobject this, jstring serialPortId) {
	if ( serialPortId ) {
		const char *str = (*env)->GetStringUTFChars(env, serialPortId, NULL);
		if ( str != 0 ) {
			int iRet = driver_set_port_id( get_driver_handle( env, this), (char*)str );
			(*env)->ReleaseStringUTFChars(env, serialPortId, str);
			throw_exception( env, this, iRet );
		}
	}
}

jstring JNICALL JNIFUNCTION( getSerialPortId )(JNIEnv *env, jobject this) {
	char id[DRIVER_ID_LEN] = "";
	driver_get_port_id( get_driver_handle( env, this), id, sizeof(id)/sizeof(id[0]) );
	return (*env)->NewStringUTF(env, id);
}

jint JNICALL JNIFUNCTION(getSerialDriverHandler) (JNIEnv *env, jobject this) {

	return driver_get_serial_driver_instance(get_driver_handle( env, this));
}

int java_event_base_handler( void* param, int index, const pSerialExchangeData frame, const char* methodeName ) {

	jobject listener = (jobject)param;
	JavaVMAttachArgs args;
	args.version = JNI_VERSION_1_2;
	args.name = "user";
	args.group = NULL;
	JNIEnv *env;
	if ( g_jvm )
		(*g_jvm)->AttachCurrentThread( g_jvm, (void **)&env, &args);

	LOG_INFO(g_Logger, "java_event_base_handler: Start listener notification", 0);
	if ( listener && env &&
		( JNI_FALSE == (*env)->IsSameObject( env, listener, NULL) ) ) { //if listener still refers to a live object
		jclass jcls = NULL;
		jmethodID jmtd = NULL;

		if ( ( jcls = (*env)->GetObjectClass( env, listener) ) != NULL ) {
			if ( (jmtd = (*env)->GetMethodID( env, jcls, methodeName, "(Lcom/coronis/sdk/serialdriver/api/ISerialFrame;)V") ) != NULL ) {
				if ( g_frameClass ) {
					// Get the method ID for the String(char[]) constructor
					jmethodID cid = (*env)->GetMethodID( env, g_frameClass, "<init>", "(I)V");
					if ( cid != NULL ) {
						// Construct a SerialFrame object
						jobject recvFrame = (*env)->NewObject( env, g_frameClass, cid, ((SerialFrame*)frame)->len);
						if ( recvFrame != NULL ) {
							jmethodID jmtdGB = (*env)->GetMethodID( env, g_frameClass, "getBuffer", "()Ljava/nio/ByteBuffer;");
							if ( jmtdGB != 0 ) {
								jobject byteBuffer = (jobject)(*env)->CallObjectMethod( env, recvFrame, jmtdGB );
								if ( byteBuffer != 0 ) {
									jclass jcBB = (*env)->GetObjectClass( env, byteBuffer);
									if ( jcBB != 0 ) {
										jmethodID jmtdBB = (*env)->GetMethodID( env, jcBB, "put", "(IB)Ljava/nio/ByteBuffer;");
										if ( jmtdBB != 0 ) {
											int i = 0;
											for ( ; i < ((SerialFrame*)&frame->serialFrame)->len; ++i )
												(*env)->CallObjectMethod( env, byteBuffer, jmtdBB, i, ((SerialFrame*)&frame->serialFrame)->frame[i] );
										}
										(*env)->DeleteLocalRef( env, jcBB );
									}
									(*env)->DeleteLocalRef( env, byteBuffer );
								}
							}
							LOG_INFO(g_Logger, "java_event_base_handler: Notify listener", 0);
							(*env)->CallVoidMethod( env, listener, jmtd, recvFrame);
							(*env)->DeleteLocalRef( env, recvFrame );
							LOG_INFO(g_Logger, "java_event_base_handler: Listener notified", 0);
						}
					}
				}
			}// jmt != NULL
			(*env)->DeleteLocalRef( env, jcls );
		}// jcls != NULL
	}

	if ( env && g_jvm )
		(*g_jvm)->DetachCurrentThread( g_jvm );

	LOG_INFO(g_Logger, "java_event_base_handler: End listener notification", 0);
	return 0;
}

int java_event_handler( void* param, const pSerialExchangeData frame ) {
	return java_event_base_handler( param, 0, frame, "receiveFrame");
}

int java_event_notify_handler( void* param, int errorCode ) {
	jobject listener = (jobject)param;
	const char* fieldIdName = NULL;
	switch( errorCode ) {
	case START_LISTENING:          fieldIdName = "START_LISTENING"; break;
	case STOP_LISTENING:           fieldIdName = "STOP_LISTENING"; break;
	case IGNORED_FRAME_RECEIVED:   fieldIdName = "IGNORED_FRAME_RECEIVED"; break;
	case SERIALPORT_DISCONNECTION: fieldIdName = "SERIALPORT_DISCONNECTION"; break;
	default:                       fieldIdName = ""; break;
	}

	LOG_DEBUG( g_Logger, "EVENT: %d %s %X <-", errorCode, fieldIdName, listener );

	if ( strlen(fieldIdName) ) {
		JavaVMAttachArgs args;
		args.version = JNI_VERSION_1_2;
		args.name = "user";
		args.group = NULL;
		JNIEnv *env;
		if ( g_jvm )
			(*g_jvm)->AttachCurrentThread( g_jvm, (void **)&env, &args);
		jclass jcls = NULL;
		jmethodID jmtd = NULL;
		if ( listener && env &&
				( JNI_FALSE == (*env)->IsSameObject( env, listener, NULL) ) ) { //if listener still refers to a live object
			if ( ( jcls = (*env)->GetObjectClass( env, listener) ) != NULL ) {
				if ( (jmtd = (*env)->GetMethodID( env, jcls, "receiveEvent", "(Lcom/coronis/sdk/serialdriver/api/SerialDriverEvent;)V") ) != NULL ) {
					if ( g_clazzSerialDriverEvent != NULL ) {
						jfieldID id = (*env)->GetStaticFieldID( env, g_clazzSerialDriverEvent, fieldIdName, "Lcom/coronis/sdk/serialdriver/api/SerialDriverEvent;");
						if ( id != NULL ) {
							jobject jenum = (*env)->GetStaticObjectField( env, g_clazzSerialDriverEvent, id);
							if ( jenum != NULL ) {
								LOG_DEBUG( g_Logger, "java_event_notify_handler: %s", fieldIdName );
								(*env)->CallVoidMethod( env, listener, jmtd,  jenum);
							}
						}
					}
				}// jmt != NULL
				(*env)->DeleteLocalRef( env, jcls );
			}// jcls != NULL
		}

		if ( env && g_jvm )
			(*g_jvm)->DetachCurrentThread( g_jvm );
	}
	LOG_DEBUG( g_Logger, "EVENT Transmitted!!!", 0 );
	return 0;
}

int java_event_spy_recv_handler( void* param, int index, const pSerialExchangeData frame ) {
	return java_event_base_handler( param, index, frame, "receivedFrame");
}

int java_event_spy_send_handler( void* param, int index, const pSerialExchangeData frame ) {
	return java_event_base_handler( param, index, frame, "sentFrame");
}

void create_jni_global_ref_to_param( RealListeners* rl ) {
	JavaVMAttachArgs args;
	args.version = JNI_VERSION_1_2;
	args.name = "user";
	args.group = NULL;
	JNIEnv *env;
	if ( g_jvm )
		(*g_jvm)->AttachCurrentThread( g_jvm, (void **)&env, &args);
	switch ( rl->eventId ) {
	case NOTIFY_EVENT:
		rl->paramNotify = (*env)->NewGlobalRef( env, (jobject)rl->paramNotify );
		break;
	case FRAME_EVENT:
		rl->paramReceiveFrame =  (*env)->NewGlobalRef( env, (jobject)rl->paramReceiveFrame );
		break;
	case SPY_SEND_EVENT:
	case SPY_RECV_EVENT:
		rl->paramSpies[rl->id] =  (*env)->NewGlobalRef( env, (jobject)rl->paramSpies[rl->id] );
		break;
	default: break;
	}
	if ( env && g_jvm )
		(*g_jvm)->DetachCurrentThread( g_jvm );
}

/* caching the class references, or method and field IDs
 * definition of this function*/
void initialize_global_class_references( JavaVM *jvm, jint jniVersion ) {
	JNIEnv *env;
	if ( !(*jvm)->GetEnv(jvm, (void **)&env, jniVersion) ) {
		jclass clazz = (*env)->FindClass( env, "com/coronis/sdk/serialdriver/api/SerialDriverEvent");
		if ( clazz != NULL )
			g_clazzSerialDriverEvent = (*env)->NewWeakGlobalRef(env, clazz);
		clazz = (*env)->FindClass( env, "com/coronis/sdk/domain/SerialFrame");
		if ( clazz != NULL )
			g_frameClass = (*env)->NewWeakGlobalRef(env, clazz);
	}
}

/* uncaching the class references, or method and field IDs
* definition of this function*/
void uninitialize_global_class_references( JavaVM *jvm, jint jniVersion ) {
	JNIEnv *env;
	if ( !(*jvm)->GetEnv(jvm, (void **)&env, jniVersion) ) {
		if ( g_clazzSerialDriverEvent )
			(*env)->DeleteWeakGlobalRef(env, g_clazzSerialDriverEvent);
		g_clazzSerialDriverEvent = NULL;
		if ( g_frameClass )
			(*env)->DeleteWeakGlobalRef(env, g_frameClass);
		g_frameClass = NULL;
	}
}
#endif //_JAVA_WRAPPER
