/*******************************************************************************
 * Copyright (c) 2012 Coronis
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Contributors:
 *     Coronis - initial API and implementation
 *******************************************************************************/

#ifndef SERIAL_PROPERTIES_H_
#define SERIAL_PROPERTIES_H_

#include <pthread.h>
#include <property_reader.h>

#include "utils.h"
#include "serial.h"

#define SERIAL_DRIVER_INTEGRITY 0x1F2E3D4C

#define SYNC 0xFF
#define STX  0x02
#define ETX  0x03

#define SERIAL_DRIVER_PROPERTIES_FILE                 "csrs232driver.properties"

// Id of the serial port (String value)
#define SERIAL_DRIVER_PORT_ID                         "serialDriver.portId"

// baudrate used to write and read data on the serial port (int value)
#define SERIAL_DRIVER_BAUD_RATE                       "serialDriver.baudRate"

// databit used to write and read data on the serial port (int value)
#define SERIAL_DRIVER_DATA_BITS                       "serialDriver.dataBit"

// stopbit used to write and read data on the serial port (int value)
#define SERIAL_DRIVER_STOP_BITS                       "serialDriver.stopBit"

// parity bit used to write and read data on the serial port (int value)
#define SERIAL_DRIVER_PARITY                          "serialDriver.parity"

//
#define SERIAL_DRIVER_LISTENING_PERIOD                "serialDriver.listeningPeriod"

//
#define SERIAL_DRIVER_FLOW_CONTROL_MODE               "serialDriver.flowControlMode"

//
#define SERIAL_DRIVER_LISTENING_STATUS                "serialDriver.listeningStatus"

//
#define SERIAL_DRIVER_RADIO_BOARD_ACK_STATUS          "rs232Driver.radioBoardAckStatus"

//
#define SERIAL_DRIVER_RADIO_BOARD_ACK_TIMEOUT         "rs232Driver.radioBoardAckTimeout"

//
#define SERIAL_DRIVER_HOST_ACK_STATUS                 "rs232Driver.hostAckStatus"

//
#define SERIAL_DRIVER_RETRY_NUMBER                    "rs232Driver.retryNumber"

// default value "true"
#define SERIAL_DRIVER_SYNC_CHARACTER_ACTIVATION       "rs232Driver.syncCharacterActivation"

// default value "true"
#define SERIAL_DRIVER_START_END_CHARACTERS_ACTIVATION "rs232Driver.startEndCharactersActivation"

// default value "true"
#define SERIAL_DRIVER_CRC16_ACTIVATION                "rs232Driver.crc16Activation"

// default value "true"
#define SERIAL_DRIVER_LENGTH_CHARACTER_ACTIVATION     "rs232Driver.lengthCharacterActivation"

// default value "true"
#define SERIAL_DRIVER_LOG_PROPERTIES_FILE_NAME       "rs232Driver.loggerFileNameProperties"


#define	FLAG_SERIAL_PORT_ID                  0x00000001
#define	FLAG_BAUD_RATE                       0x00000002
#define	FLAG_DATA_BIT                        0x00000004
#define	FLAG_STOP_BIT                        0x00000008
#define	FLAG_PARITY                          0x00000010
#define	FLAG_LISTENING_PERIOD                0x00000020
#define	FLAG_FLOWCONTROL_MODE                0x00000040
#define	FLAG_SERIAL_LISTENING_STATUS         0x00000080
#define	FLAG_RADIOBOARD_ACK_STATUS           0x00000100
#define	FLAG_RADIO_ACK_TIMEOUT               0x00000200
#define	FLAG_HOST_ACK_STATUS                 0x00000400
#define	FLAG_RETRY_COUNT_DEF                 0x00000800
#define FLAG_SYNC_CHARACTER_ACTIVATION       0x00001000
#define FLAG_START_END_CHARACTERS_ACTIVATION 0x00002000
#define FLAG_CRC16_ACTIVATION                0x00004000
#define FLAG_LENGTH_ACTIVATION                0x00008000
#define FLAG_LOGGER_PROP_FILE_NAME           0x00010000

typedef struct tagSerialParams {
	unsigned int isSet;
	char  id[DRIVER_ID_LEN];
	int   baudRate;
	int   dataBit;
	int   stopBit;
	int   parity;
	int   listeningPeriod;
	int   flowControlMode;
	_bool listeningStatus;
	_bool radioBoardAckStatus;
	int   radioAckTimeout;
	_bool hostAckStatus;
	int   retryCountDef;
	_bool syncCharacterActivation;
	_bool startEndCharactersActivation;
	_bool crc16Activation;
	_bool lengthActivation;
} SerialParams;

typedef enum tagListenersIDS {
	NOTIFY_EVENT,
	FRAME_EVENT,
	SPY_SEND_EVENT,
	SPY_RECV_EVENT
} ListenersIDS ;

typedef struct tagRealListeners {
	serial_driver_listener_t receiveFrame;
	void* paramReceiveFrame;

	serial_driver_listener_notify_t notifyListener;
	void* paramNotify;

	serial_driver_spy_t spiesSend[MAX_SPY_NUM];
	serial_driver_spy_t spiesRecv[MAX_SPY_NUM];
	void* paramSpies[MAX_SPY_NUM];

	/*
	 * type of event to be raise
	 * */
	ListenersIDS eventId;
	/*
	 * could be spy index id or notify event id, this depends on eventId value;
	 * */
	int id;
} RealListeners;

typedef struct tagDriverInfo {
	unsigned int integrity;
	SerialParams params;
	int   		 retryCount;

	void* 		 port;

	SerialFrame frame;
	int   		status;
	int   		sendingStatus;
	struct timeval startTime;
	pthread_mutex_t mutexProcessing;

	serial_driver_listener_t receiveFrame;
	void* paramReceiveFrame;

	serial_driver_listener_notify_t notifyListener;
	void* paramNotify;
	_bool startNotifyListenerEvent;

	serial_driver_spy_t spiesSend[MAX_SPY_NUM];
	serial_driver_spy_t spiesRecv[MAX_SPY_NUM];
	void* paramSpies[MAX_SPY_NUM];

	RealListeners 	realListeners;
	SerialFrame 	sendFrame;

	_bool java;
} DriverInfo;

extern DriverInfo drivers[MAX_DRIVER_NUM];
extern char g_loger_rs232_properties_file_name[];

void get_all_properties( const char* propertyString, SerialParams* params );

int port_open( DriverInfo* driver );
int port_close( DriverInfo* driver );
int port_send( DriverInfo* driver, const byte* frame, int frameLen );

DriverInfo* get_driver( const void* driver );

int copy_frame_body( const DriverInfo* driver, const SerialFrame* frameSrc, SerialFrame* frameDest );


#endif /* PROTOCOL_PROPERTIES_H_ */
